#include <libintl.h>

#include <stddef.h>

#define PROGRESS_LIM 262144

#define NUM_UTYPES   2

typedef struct {
	unsigned int map;
	void *store[NUM_UTYPES];
} undo_data;

typedef unsigned char png_byte;

typedef struct png_color_struct
{
   png_byte red;
   png_byte green;
   png_byte blue;
} png_color;

#include <string.h>

#define NUM_CHANNELS 4

typedef unsigned char *chanlist[NUM_CHANNELS];

void mem_free_chanlist(chanlist img)
{
	int i;

	for (i = 0; i < NUM_CHANNELS; i++)
	{
		if (!img[i]) continue;
		if (img[i] != (void *)(-1)) ;
	}
}

typedef struct {
	chanlist img;
	png_color *pal_;
	unsigned char *tileptr;
	undo_data *dataptr;
	size_t size;
	int width, height, flags;
	short cols, bpp, trans;
} undo_item;

typedef struct {
	undo_item *items;	// Pointer to undo images + current image being edited
	int pointer;		// Index of currently used image on canvas/screen
	int done;		// Undo images that we have behind current image (i.e. possible UNDO)
	int redo;		// Undo images that we have ahead of current image (i.e. possible REDO)
	int max;		// Total number of undo slots
	size_t size;		// Total used memory (0 means count it anew)
} undo_stack;

typedef struct {
	chanlist img;		// Array of pointers to image channels
	png_color pal[256];	// RGB entries for all 256 palette colours
	int cols;	// Number of colours in the palette: 1..256 or 0 for no image
	int bpp;		// Bytes per pixel = 1 or 3
	int trans;		// Transparent colour index (-1 if none)
	int width, height;	// Image geometry
	undo_stack undo_;	// Image's undo stack
	char *filename;		// File name of file loaded/saved
	void *tempfiles;	// List of up-to-date temp files
	int changed;		// Changed since last load/save flag
} image_info;

#define FREE_UNDO  2

#define FREE_IMAGE 1

void mem_free_image(image_info *image, int mode)
{
	int i, j = image->undo_.max, p = image->undo_.pointer;

	/* Delete current image (don't rely on undo frame being up to date) */
	if (mode & FREE_IMAGE)
	{
		mem_free_chanlist(image->img);
		memset(image->img, 0, sizeof(chanlist));
		image->width = image->height = 0;

		;
		image->filename = image->tempfiles = NULL;
	}

	/* Delete undo frames if any */
	image->undo_.pointer = image->undo_.done = image->undo_.redo = 0;
	if (!image->undo_.items) return;
	; // Already freed
	for (i = 0; i < j; i++) ;

	/* Delete undo stack if finalizing */
	if (mode & FREE_UNDO)
	{
		;
		image->undo_.items = NULL;
		image->undo_.max = 0;
	}
}

#include <stdlib.h>

#define FREE_ALL   3

void mem_rotate( char *new, char *old, int old_w, int old_h, int dir, int bpp )
{
	unsigned char *src;
	int i, j, k, l, flag;

	flag = (old_w * old_h > PROGRESS_LIM * 4);
	j = old_w * bpp;
	l = dir ? -bpp : bpp;
	k = -old_w * l;
	old += dir ? j - bpp: (old_h - 1) * j;

	if (flag) ;
	for (i = 0; i < old_w; i++)
	{
		if (flag && ((i * 5) % old_w >= old_w - 5))
				;
		src = old;
		if (bpp == 1)
		{
			for (j = 0; j < old_h; j++)
			{
				*new++ = *src;
				src += k;
			}
		}
		else
		{
			for (j = 0; j < old_h; j++)
			{
				*new++ = src[0];
				*new++ = src[1];
				*new++ = src[2];
				src += k;
			}
		}
		old += l;
	}
	if (flag) ;
}

#define mem_clip_bpp		mem_clip.bpp

#define mem_clip_h		mem_clip.height

#define mem_clip_w		mem_clip.width

#define mem_clip (*mem_clip__cc_)

int mem_sel_rot( int dir , image_info* mem_clip__cc_)			// Rotate clipboard 90 degrees
{
	unsigned char *buf = NULL;
	int i, j = mem_clip_w * mem_clip_h, bpp = mem_clip_bpp;

	for (i = 0; i < NUM_CHANNELS; i++ , bpp = 1)
	{
		if (!mem_clip.img[i]) continue;
		buf = malloc(j * bpp);
		if (!buf) break;	// Not enough memory
		mem_rotate(buf, mem_clip.img[i], mem_clip_w, mem_clip_h, dir, bpp);
		;
		mem_clip.img[i] = buf;
	}

	/* Don't leave a mix of rotated and unrotated channels */
	if (!buf && i) mem_free_image(&mem_clip, FREE_ALL);
	if (!buf) return (1);

	i = mem_clip_w;
	mem_clip_w = mem_clip_h;		// Flip geometry
	mem_clip_h = i;

	return (0);
}

