#include <stddef.h>
#include <inttypes.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>




static inline uint64_t __cc_xgetbv(void) {
  const uint32_t ecx = 0;
  uint32_t eax, edx;

  __asm__ volatile (
    ".byte 0x0f, 0x01, 0xd0\n"
    : "=a"(eax), "=d"(edx) : "c" (ecx));
  return ((uint64_t)edx << 32) | eax;
}

static inline void __cc_GetCPUInfo(int cpu_info[4], int info_type) {
  __asm__ volatile (
    "cpuid\n"
    : "=a"(cpu_info[0]), "=b"(cpu_info[1]), "=c"(cpu_info[2]), "=d"(cpu_info[3])
    : "a"(info_type), "c"(0));
}

typedef enum {
  kSSE2,
  kSSE3,
  kAVX,
  kAVX2,
  kNEON,
  kMIPS32
} __cc_CPUFeature;

static int __cc_x86CPUInfo(__cc_CPUFeature feature) {
  int cpu_info[4];
  __cc_GetCPUInfo(cpu_info, 1);
  if (feature == kSSE2) {
    return 0 != (cpu_info[3] & 0x04000000);
  }
  if (feature == kSSE3) {
    return 0 != (cpu_info[2] & 0x00000001);
  }
  if (feature == kAVX) {

    if ((cpu_info[2] & 0x18000000) == 0x18000000) {

      return (__cc_xgetbv() & 0x6) == 0x6;
    }
  }
  if (feature == kAVX2) {
    if (__cc_x86CPUInfo(kAVX)) {
      __cc_GetCPUInfo(cpu_info, 7);
      return ((cpu_info[1] & 0x00000020) == 0x00000020);
    }
  }
  return 0;
}

typedef int (*__cc_VP8CPUInfo)(__cc_CPUFeature feature);

__cc_VP8CPUInfo VP8GetCPUInfo = __cc_x86CPUInfo;







static uint32_t __cc_Mult(uint8_t x, uint32_t mult) {
  const uint32_t v = (x * mult + ((1u << 24) >> 1)) >> 24;
  ((v <= 255) ? (void) (0) : __assert_fail ("v <= 255", "extraction.1.c", 81, __PRETTY_FUNCTION__));
  return v;
}



static inline uint32_t __cc_GetScale(uint32_t a, int inverse) {
  return inverse ? (255u << 24) / a : a * ((1u << 24) / 255u);
}

static void __cc_MultARGBRow(uint32_t* const ptr, int width, int inverse) {
  int x;
  for (x = 0; x < width; ++x) {
    const uint32_t argb = ptr[x];
    if (argb < 0xff000000u) {
      if (argb <= 0x00ffffffu) {
        ptr[x] = 0;
      } else {
        const uint32_t alpha = (argb >> 24) & 0xff;
        const uint32_t scale = __cc_GetScale(alpha, inverse);
        uint32_t out = argb & 0xff000000u;
        out |= __cc_Mult(argb >> 0, scale) << 0;
        out |= __cc_Mult(argb >> 8, scale) << 8;
        out |= __cc_Mult(argb >> 16, scale) << 16;
        ptr[x] = out;
      }
    }
  }
}





static void __cc_ApplyAlphaMultiply(uint8_t* rgba, int alpha_first,
                               int w, int h, int stride) {
  while (h-- > 0) {
    uint8_t* const rgb = rgba + (alpha_first ? 1 : 0);
    const uint8_t* const alpha = rgba + (alpha_first ? 0 : 3);
    int i;
    for (i = 0; i < w; ++i) {
      const uint32_t a = alpha[4 * i];
      if (a != 0xff) {
        const uint32_t mult = ((a) * 32897U);
        rgb[4 * i + 0] = (((rgb[4 * i + 0]) * (mult)) >> 23);
        rgb[4 * i + 1] = (((rgb[4 * i + 1]) * (mult)) >> 23);
        rgb[4 * i + 2] = (((rgb[4 * i + 2]) * (mult)) >> 23);
      }
    }
    rgba += stride;
  }
}



void (*WebPMultARGBRow)(uint32_t* const ptr, int width, int inverse);

void (*WebPApplyAlphaMultiply)(uint8_t*, int, int, int, int);

void (*WebPApplyAlphaMultiply4444)(uint8_t*, int, int, int);

int (*WebPExtractAlpha)(const uint8_t*, int, int, int, uint8_t*, int);

static inline uint8_t __cc_dither_lo(uint8_t x) {
  return (x & 0x0f) | (x << 4);
}

static inline uint8_t __cc_dither_hi(uint8_t x) {
  return (x & 0xf0) | (x >> 4);
}

static inline uint8_t __cc_multiply(uint8_t x, uint32_t m) {
  return (x * m) >> 16;
}

static inline void __cc_ApplyAlphaMultiply4444(uint8_t* rgba4444,
                                               int w, int h, int stride,
                                               int rg_byte_pos ) {
  while (h-- > 0) {
    int i;
    for (i = 0; i < w; ++i) {
      const uint32_t rg = rgba4444[2 * i + rg_byte_pos];
      const uint32_t ba = rgba4444[2 * i + (rg_byte_pos ^ 1)];
      const uint8_t a = ba & 0x0f;
      const uint32_t mult = ((a) * 0x1111);
      const uint8_t r = __cc_multiply(__cc_dither_hi(rg), mult);
      const uint8_t g = __cc_multiply(__cc_dither_lo(rg), mult);
      const uint8_t b = __cc_multiply(__cc_dither_hi(ba), mult);
      rgba4444[2 * i + rg_byte_pos] = (r & 0xf0) | ((g >> 4) & 0x0f);
      rgba4444[2 * i + (rg_byte_pos ^ 1)] = (b & 0xf0) | a;
    }
    rgba4444 += stride;
  }
}

static void __cc_ApplyAlphaMultiply_16b(uint8_t* rgba4444,
                                   int w, int h, int stride) {



  __cc_ApplyAlphaMultiply4444(rgba4444, w, h, stride, 0);

}

static void __cc_MultRow(uint8_t* const ptr, const uint8_t* const alpha,
                    int width, int inverse) {
  int x;
  for (x = 0; x < width; ++x) {
    const uint32_t a = alpha[x];
    if (a != 255) {
      if (a == 0) {
        ptr[x] = 0;
      } else {
        const uint32_t scale = __cc_GetScale(a, inverse);
        ptr[x] = __cc_Mult(ptr[x], scale);
      }
    }
  }
}

void (*WebPMultRow)(uint8_t* const ptr, const uint8_t* const alpha,
                    int width, int inverse);

static int __cc_ExtractAlpha(const uint8_t* argb, int argb_stride,
                        int width, int height,
                        uint8_t* alpha, int alpha_stride) {
  uint8_t alpha_mask = 0xff;
  int i, j;

  for (j = 0; j < height; ++j) {
    for (i = 0; i < width; ++i) {
      const uint8_t alpha_value = argb[4 * i];
      alpha[i] = alpha_value;
      alpha_mask &= alpha_value;
    }
    argb += argb_stride;
    alpha += alpha_stride;
  }
  return (alpha_mask == 0xff);
}

extern __cc_VP8CPUInfo VP8GetCPUInfo;

static volatile __cc_VP8CPUInfo alpha_processing_last_cpuinfo_used =
    (__cc_VP8CPUInfo)&alpha_processing_last_cpuinfo_used;

void __cc_WebPInitAlphaProcessing(void) {
  if (alpha_processing_last_cpuinfo_used == VP8GetCPUInfo) return;

  WebPMultARGBRow = __cc_MultARGBRow;
  WebPMultRow = __cc_MultRow;
  WebPApplyAlphaMultiply = __cc_ApplyAlphaMultiply;
  WebPApplyAlphaMultiply4444 = __cc_ApplyAlphaMultiply_16b;
  WebPExtractAlpha = __cc_ExtractAlpha;


  if (VP8GetCPUInfo != ((void *)0)) {





  }
  alpha_processing_last_cpuinfo_used = VP8GetCPUInfo;
}

void __cc_WebPInitAlphaProcessing(void);

typedef struct WebPDecParams __cc_WebPDecParams;

typedef enum WEBP_CSP_MODE {
  MODE_RGB = 0, MODE_RGBA = 1,
  MODE_BGR = 2, MODE_BGRA = 3,
  MODE_ARGB = 4, MODE_RGBA_4444 = 5,
  MODE_RGB_565 = 6,

  MODE_rgbA = 7,
  MODE_bgrA = 8,
  MODE_Argb = 9,
  MODE_rgbA_4444 = 10,

  MODE_YUV = 11, MODE_YUVA = 12,
  MODE_LAST = 13
} __cc_WEBP_CSP_MODE;

typedef struct VP8Io __cc_VP8Io;

typedef void (*__cc_VP8IoTeardownHook)(const __cc_VP8Io* io);

typedef int (*__cc_VP8IoSetupHook)(__cc_VP8Io* io);

typedef int (*__cc_VP8IoPutHook)(const __cc_VP8Io* io);

struct VP8Io {

  int width, height;





  int mb_y;
  int mb_w;
  int mb_h;
  const uint8_t* y, *u, *v;
  int y_stride;
  int uv_stride;

  void* opaque;






  __cc_VP8IoPutHook put;





  __cc_VP8IoSetupHook setup;



  __cc_VP8IoTeardownHook teardown;




  int fancy_upsampling;


  size_t data_size;
  const uint8_t* data;





  int bypass_filtering;


  int use_cropping;
  int crop_left, crop_right, crop_top, crop_bottom;


  int use_scaling;
  int scaled_width, scaled_height;




  const uint8_t* a;
};

typedef int (*__cc_OutputFunc)(const __cc_VP8Io* const io, __cc_WebPDecParams* const p);

typedef uint32_t __cc_rescaler_t;

typedef struct {
  int x_expand;
  int y_expand;
  int num_channels;
  uint32_t fx_scale;
  uint32_t fy_scale;
  uint32_t fxy_scale;
  int y_accum;
  int y_add, y_sub;
  int x_add, x_sub;
  int src_width, src_height;
  int dst_width, dst_height;
  int src_y, dst_y;
  uint8_t* dst;
  int dst_stride;
  __cc_rescaler_t* irow, *frow;
} __cc_WebPRescaler;

typedef struct WebPDecoderOptions __cc_WebPDecoderOptions;

typedef int (*__cc_OutputRowFunc)(__cc_WebPDecParams* const p, int y_pos);



struct WebPDecoderOptions {
  int bypass_filtering;
  int no_fancy_upsampling;
  int use_cropping;
  int crop_left, crop_top;

  int crop_width, crop_height;
  int use_scaling;
  int scaled_width, scaled_height;
  int use_threads;
  int dithering_strength;
  int force_rotation;
  int no_enhancement;



  uint32_t pad[4];



};

typedef struct WebPDecBuffer __cc_WebPDecBuffer;

typedef struct WebPYUVABuffer __cc_WebPYUVABuffer;

typedef struct WebPRGBABuffer __cc_WebPRGBABuffer;

struct WebPRGBABuffer {
  uint8_t* rgba;
  int stride;
  size_t size;
};

struct WebPYUVABuffer {
  uint8_t* y, *u, *v, *a;
  int y_stride;
  int u_stride, v_stride;
  int a_stride;
  size_t y_size;
  size_t u_size, v_size;
  size_t a_size;
};

struct WebPDecBuffer {
  __cc_WEBP_CSP_MODE colorspace;
  int width, height;
  int is_external_memory;
  union {
    __cc_WebPRGBABuffer RGBA;
    __cc_WebPYUVABuffer YUVA;
  } u;
  uint32_t pad[4];

  uint8_t* private_memory;


};

struct WebPDecParams {
  __cc_WebPDecBuffer* output;
  uint8_t* tmp_y, *tmp_u, *tmp_v;


  int last_y;
  const __cc_WebPDecoderOptions* options;

  __cc_WebPRescaler scaler_y, scaler_u, scaler_v, scaler_a;
  void* memory;

  __cc_OutputFunc emit;
  __cc_OutputFunc emit_alpha;
  __cc_OutputRowFunc emit_alpha_row;
};

int __cc_WebPRescalerImport(__cc_WebPRescaler* const rescaler, int num_rows,
                       const uint8_t* src, int src_stride);

static int __cc_EmitRescaledAlphaRGB(const __cc_VP8Io* const io, __cc_WebPDecParams* const p) {
  if (io->a != ((void *)0)) {
    __cc_WebPRescaler* const scaler = &p->scaler_a;
    int j = 0;
    int pos = 0;
    while (j < io->mb_h) {
      j += __cc_WebPRescalerImport(scaler, io->mb_h - j,
                              io->a + j * io->width, io->width);
      pos += p->emit_alpha_row(p, pos);
    }
  }
  return 0;
}

static inline
int __cc_WebPRescalerOutputDone(const __cc_WebPRescaler* const rescaler) {
  return (rescaler->dst_y >= rescaler->dst_height);
}

static inline
int __cc_WebPRescalerHasPendingOutput(const __cc_WebPRescaler* const rescaler) {
  return !__cc_WebPRescalerOutputDone(rescaler) && (rescaler->y_accum <= 0);
}

typedef void (*__cc_WebPYUV444Converter)(const uint8_t* y,
                                    const uint8_t* u, const uint8_t* v,
                                    uint8_t* dst, int len);

extern const __cc_WebPYUV444Converter WebPYUV444Converters[ ];

void __cc_WebPRescalerExportRow(__cc_WebPRescaler* const wrk);

static int __cc_ExportRGB(__cc_WebPDecParams* const p, int y_pos) {
  const __cc_WebPYUV444Converter convert =
      WebPYUV444Converters[p->output->colorspace];
  const __cc_WebPRGBABuffer* const buf = &p->output->u.RGBA;
  uint8_t* dst = buf->rgba + (p->last_y + y_pos) * buf->stride;
  int num_lines_out = 0;


  while (__cc_WebPRescalerHasPendingOutput(&p->scaler_y) &&
         __cc_WebPRescalerHasPendingOutput(&p->scaler_u)) {
    ((p->last_y + y_pos + num_lines_out < p->output->height) ? (void) (0) : __assert_fail ("p->last_y + y_pos + num_lines_out < p->output->height", "extraction.1.c", 493, __PRETTY_FUNCTION__));
    ((p->scaler_u.y_accum == p->scaler_v.y_accum) ? (void) (0) : __assert_fail ("p->scaler_u.y_accum == p->scaler_v.y_accum", "extraction.1.c", 494, __PRETTY_FUNCTION__));
    __cc_WebPRescalerExportRow(&p->scaler_y);
    __cc_WebPRescalerExportRow(&p->scaler_u);
    __cc_WebPRescalerExportRow(&p->scaler_v);
    convert(p->scaler_y.dst, p->scaler_u.dst, p->scaler_v.dst,
            dst, p->scaler_y.dst_width);
    dst += buf->stride;
    ++num_lines_out;
  }
  return num_lines_out;
}

static int __cc_EmitRescaledRGB(const __cc_VP8Io* const io, __cc_WebPDecParams* const p) {
  const int mb_h = io->mb_h;
  const int uv_mb_h = (mb_h + 1) >> 1;
  int j = 0, uv_j = 0;
  int num_lines_out = 0;
  while (j < mb_h) {
    const int y_lines_in =
        __cc_WebPRescalerImport(&p->scaler_y, mb_h - j,
                           io->y + j * io->y_stride, io->y_stride);
    const int u_lines_in =
        __cc_WebPRescalerImport(&p->scaler_u, uv_mb_h - uv_j,
                           io->u + uv_j * io->uv_stride, io->uv_stride);
    const int v_lines_in =
        __cc_WebPRescalerImport(&p->scaler_v, uv_mb_h - uv_j,
                           io->v + uv_j * io->uv_stride, io->uv_stride);
    (void)v_lines_in;
    ((u_lines_in == v_lines_in) ? (void) (0) : __assert_fail ("u_lines_in == v_lines_in", "extraction.1.c", 522, __PRETTY_FUNCTION__));
    j += y_lines_in;
    uv_j += u_lines_in;
    num_lines_out += __cc_ExportRGB(p, num_lines_out);
  }
  return num_lines_out;
}



extern void* __cc_WebPSafeMalloc(uint64_t nmemb, size_t size);

extern void (*WebPApplyAlphaMultiply4444)(
    uint8_t* rgba4444, int w, int h, int stride);

static inline int __cc_WebPIsPremultipliedMode(__cc_WEBP_CSP_MODE mode) {
  return (mode == MODE_rgbA || mode == MODE_bgrA || mode == MODE_Argb ||
          mode == MODE_rgbA_4444);
}

static int __cc_ExportAlphaRGBA4444(__cc_WebPDecParams* const p, int y_pos) {
  const __cc_WebPRGBABuffer* const buf = &p->output->u.RGBA;
  uint8_t* const base_rgba = buf->rgba + (p->last_y + y_pos) * buf->stride;



  uint8_t* alpha_dst = base_rgba + 1;

  int num_lines_out = 0;
  const __cc_WEBP_CSP_MODE colorspace = p->output->colorspace;
  const int width = p->scaler_a.dst_width;
  const int is_premult_alpha = __cc_WebPIsPremultipliedMode(colorspace);
  uint32_t alpha_mask = 0x0f;

  while (__cc_WebPRescalerHasPendingOutput(&p->scaler_a)) {
    int i;
    ((p->last_y + y_pos + num_lines_out < p->output->height) ? (void) (0) : __assert_fail ("p->last_y + y_pos + num_lines_out < p->output->height", "extraction.1.c", 558, __PRETTY_FUNCTION__));
    __cc_WebPRescalerExportRow(&p->scaler_a);
    for (i = 0; i < width; ++i) {

      const uint32_t alpha_value = p->scaler_a.dst[i] >> 4;
      alpha_dst[2 * i] = (alpha_dst[2 * i] & 0xf0) | alpha_value;
      alpha_mask &= alpha_value;
    }
    alpha_dst += buf->stride;
    ++num_lines_out;
  }
  if (is_premult_alpha && alpha_mask != 0x0f) {
    WebPApplyAlphaMultiply4444(base_rgba, width, num_lines_out, buf->stride);
  }
  return num_lines_out;
}

static inline int __cc_WebPIsAlphaMode(__cc_WEBP_CSP_MODE mode) {
  return (mode == MODE_RGBA || mode == MODE_BGRA || mode == MODE_ARGB ||
          mode == MODE_RGBA_4444 || mode == MODE_YUVA ||
          __cc_WebPIsPremultipliedMode(mode));
}

void __cc_WebPRescalerInit(__cc_WebPRescaler* const rescaler,
                      int src_width, int src_height,
                      uint8_t* const dst,
                      int dst_width, int dst_height, int dst_stride,
                      int num_channels,
                      __cc_rescaler_t* const work);

extern void (*WebPApplyAlphaMultiply)(
    uint8_t* rgba, int alpha_first, int w, int h, int stride);

static int __cc_ExportAlpha(__cc_WebPDecParams* const p, int y_pos) {
  const __cc_WebPRGBABuffer* const buf = &p->output->u.RGBA;
  uint8_t* const base_rgba = buf->rgba + (p->last_y + y_pos) * buf->stride;
  const __cc_WEBP_CSP_MODE colorspace = p->output->colorspace;
  const int alpha_first =
      (colorspace == MODE_ARGB || colorspace == MODE_Argb);
  uint8_t* dst = base_rgba + (alpha_first ? 0 : 3);
  int num_lines_out = 0;
  const int is_premult_alpha = __cc_WebPIsPremultipliedMode(colorspace);
  uint32_t alpha_mask = 0xff;
  const int width = p->scaler_a.dst_width;

  while (__cc_WebPRescalerHasPendingOutput(&p->scaler_a)) {
    int i;
    ((p->last_y + y_pos + num_lines_out < p->output->height) ? (void) (0) : __assert_fail ("p->last_y + y_pos + num_lines_out < p->output->height", "extraction.1.c", 605, __PRETTY_FUNCTION__));
    __cc_WebPRescalerExportRow(&p->scaler_a);
    for (i = 0; i < width; ++i) {
      const uint32_t alpha_value = p->scaler_a.dst[i];
      dst[4 * i] = alpha_value;
      alpha_mask &= alpha_value;
    }
    dst += buf->stride;
    ++num_lines_out;
  }
  if (is_premult_alpha && alpha_mask != 0xff) {
    WebPApplyAlphaMultiply(base_rgba, alpha_first,
                           width, num_lines_out, buf->stride);
  }
  return num_lines_out;
}

static int __cc_InitRGBRescaler(const __cc_VP8Io* const io, __cc_WebPDecParams* const p) {
  const int has_alpha = __cc_WebPIsAlphaMode(p->output->colorspace);
  const int out_width = io->scaled_width;
  const int out_height = io->scaled_height;
  const int uv_in_width = (io->mb_w + 1) >> 1;
  const int uv_in_height = (io->mb_h + 1) >> 1;
  const size_t work_size = 2 * out_width;
  __cc_rescaler_t* work;
  uint8_t* tmp;
  size_t tmp_size1, tmp_size2, total_size;

  tmp_size1 = 3 * work_size;
  tmp_size2 = 3 * out_width;
  if (has_alpha) {
    tmp_size1 += work_size;
    tmp_size2 += out_width;
  }
  total_size = tmp_size1 * sizeof(*work) + tmp_size2 * sizeof(*tmp);
  p->memory = __cc_WebPSafeMalloc(1ULL, total_size);
  if (p->memory == ((void *)0)) {
    return 0;
  }
  work = (__cc_rescaler_t*)p->memory;
  tmp = (uint8_t*)(work + tmp_size1);
  __cc_WebPRescalerInit(&p->scaler_y, io->mb_w, io->mb_h,
                   tmp + 0 * out_width, out_width, out_height, 0, 1,
                   work + 0 * work_size);
  __cc_WebPRescalerInit(&p->scaler_u, uv_in_width, uv_in_height,
                   tmp + 1 * out_width, out_width, out_height, 0, 1,
                   work + 1 * work_size);
  __cc_WebPRescalerInit(&p->scaler_v, uv_in_width, uv_in_height,
                   tmp + 2 * out_width, out_width, out_height, 0, 1,
                   work + 2 * work_size);
  p->emit = __cc_EmitRescaledRGB;

  if (has_alpha) {
    __cc_WebPRescalerInit(&p->scaler_a, io->mb_w, io->mb_h,
                     tmp + 3 * out_width, out_width, out_height, 0, 1,
                     work + 3 * work_size);
    p->emit_alpha = __cc_EmitRescaledAlphaRGB;
    if (p->output->colorspace == MODE_RGBA_4444 ||
        p->output->colorspace == MODE_rgbA_4444) {
      p->emit_alpha_row = __cc_ExportAlphaRGBA4444;
    } else {
      p->emit_alpha_row = __cc_ExportAlpha;
    }
    __cc_WebPInitAlphaProcessing();
  }
  return 1;
}

extern void __cc_WebPSafeFree(void* const ptr);

static void __cc_CustomTeardown(const __cc_VP8Io* io) {
  __cc_WebPDecParams* const p = (__cc_WebPDecParams*)io->opaque;
  __cc_WebPSafeFree(p->memory);
  p->memory = ((void *)0);
}

static int __cc_GetAlphaSourceRow(const __cc_VP8Io* const io,
                             const uint8_t** alpha, int* const num_rows) {
  int start_y = io->mb_y;
  *num_rows = io->mb_h;



  if (io->fancy_upsampling) {
    if (start_y == 0) {

      --*num_rows;
    } else {
      --start_y;



      *alpha -= io->width;
    }
    if (io->crop_top + io->mb_y + io->mb_h == io->crop_bottom) {

      *num_rows = io->crop_bottom - io->crop_top - start_y;
    }
  }
  return start_y;
}

static int __cc_EmitAlphaRGBA4444(const __cc_VP8Io* const io, __cc_WebPDecParams* const p) {
  const uint8_t* alpha = io->a;
  if (alpha != ((void *)0)) {
    const int mb_w = io->mb_w;
    const __cc_WEBP_CSP_MODE colorspace = p->output->colorspace;
    const __cc_WebPRGBABuffer* const buf = &p->output->u.RGBA;
    int num_rows;
    const int start_y = __cc_GetAlphaSourceRow(io, &alpha, &num_rows);
    uint8_t* const base_rgba = buf->rgba + start_y * buf->stride;



    uint8_t* alpha_dst = base_rgba + 1;

    uint32_t alpha_mask = 0x0f;
    int i, j;

    for (j = 0; j < num_rows; ++j) {
      for (i = 0; i < mb_w; ++i) {

        const uint32_t alpha_value = alpha[i] >> 4;
        alpha_dst[2 * i] = (alpha_dst[2 * i] & 0xf0) | alpha_value;
        alpha_mask &= alpha_value;
      }
      alpha += io->width;
      alpha_dst += buf->stride;
    }
    if (alpha_mask != 0x0f && __cc_WebPIsPremultipliedMode(colorspace)) {
      WebPApplyAlphaMultiply4444(base_rgba, mb_w, num_rows, buf->stride);
    }
  }
  return 0;
}

typedef void (*__cc_WebPSamplerRowFunc)(const uint8_t* y,
                                   const uint8_t* u, const uint8_t* v,
                                   uint8_t* dst, int len);

extern __cc_WebPSamplerRowFunc WebPSamplers[ ];

void __cc_WebPSamplerProcessPlane(const uint8_t* y, int y_stride,
                             const uint8_t* u, const uint8_t* v, int uv_stride,
                             uint8_t* dst, int dst_stride,
                             int width, int height, __cc_WebPSamplerRowFunc func);

static int __cc_EmitSampledRGB(const __cc_VP8Io* const io, __cc_WebPDecParams* const p) {
  __cc_WebPDecBuffer* const output = p->output;
  __cc_WebPRGBABuffer* const buf = &output->u.RGBA;
  uint8_t* const dst = buf->rgba + io->mb_y * buf->stride;
  __cc_WebPSamplerProcessPlane(io->y, io->y_stride,
                          io->u, io->v, io->uv_stride,
                          dst, buf->stride, io->mb_w, io->mb_h,
                          WebPSamplers[output->colorspace]);
  return io->mb_h;
}


static int __cc_EmitAlphaYUV(const __cc_VP8Io* const io, __cc_WebPDecParams* const p) {
  const uint8_t* alpha = io->a;
  const __cc_WebPYUVABuffer* const buf = &p->output->u.YUVA;
  const int mb_w = io->mb_w;
  const int mb_h = io->mb_h;
  uint8_t* dst = buf->a + io->mb_y * buf->a_stride;
  int j;

  if (alpha != ((void *)0)) {
    for (j = 0; j < mb_h; ++j) {
      memcpy(dst, alpha, mb_w * sizeof(*dst));
      alpha += io->width;
      dst += buf->a_stride;
    }
  } else if (buf->a != ((void *)0)) {

    for (j = 0; j < mb_h; ++j) {
      memset(dst, 0xff, mb_w * sizeof(*dst));
      dst += buf->a_stride;
    }
  }
  return 0;
}



static int __cc_EmitAlphaRGB(const __cc_VP8Io* const io, __cc_WebPDecParams* const p) {
  const uint8_t* alpha = io->a;
  if (alpha != ((void *)0)) {
    const int mb_w = io->mb_w;
    const __cc_WEBP_CSP_MODE colorspace = p->output->colorspace;
    const int alpha_first =
        (colorspace == MODE_ARGB || colorspace == MODE_Argb);
    const __cc_WebPRGBABuffer* const buf = &p->output->u.RGBA;
    int num_rows;
    const int start_y = __cc_GetAlphaSourceRow(io, &alpha, &num_rows);
    uint8_t* const base_rgba = buf->rgba + start_y * buf->stride;
    uint8_t* dst = base_rgba + (alpha_first ? 0 : 3);
    uint32_t alpha_mask = 0xff;
    int i, j;

    for (j = 0; j < num_rows; ++j) {
      for (i = 0; i < mb_w; ++i) {
        const uint32_t alpha_value = alpha[i];
        dst[4 * i] = alpha_value;
        alpha_mask &= alpha_value;
      }
      alpha += io->width;
      dst += buf->stride;
    }

    if (alpha_mask != 0xff && __cc_WebPIsPremultipliedMode(colorspace)) {
      WebPApplyAlphaMultiply(base_rgba, alpha_first,
                             mb_w, num_rows, buf->stride);
    }
  }
  return 0;
}

static int __cc_EmitYUV(const __cc_VP8Io* const io, __cc_WebPDecParams* const p) {
  __cc_WebPDecBuffer* output = p->output;
  const __cc_WebPYUVABuffer* const buf = &output->u.YUVA;
  uint8_t* const y_dst = buf->y + io->mb_y * buf->y_stride;
  uint8_t* const u_dst = buf->u + (io->mb_y >> 1) * buf->u_stride;
  uint8_t* const v_dst = buf->v + (io->mb_y >> 1) * buf->v_stride;
  const int mb_w = io->mb_w;
  const int mb_h = io->mb_h;
  const int uv_w = (mb_w + 1) / 2;
  const int uv_h = (mb_h + 1) / 2;
  int j;
  for (j = 0; j < mb_h; ++j) {
    memcpy(y_dst + j * buf->y_stride, io->y + j * io->y_stride, mb_w);
  }
  for (j = 0; j < uv_h; ++j) {
    memcpy(u_dst + j * buf->u_stride, io->u + j * io->uv_stride, uv_w);
    memcpy(v_dst + j * buf->v_stride, io->v + j * io->uv_stride, uv_w);
  }
  return io->mb_h;
}

void __cc_WebPInitSamplers(void);

typedef void (*__cc_WebPUpsampleLinePairFunc)(
    const uint8_t* top_y, const uint8_t* bottom_y,
    const uint8_t* top_u, const uint8_t* top_v,
    const uint8_t* cur_u, const uint8_t* cur_v,
    uint8_t* top_dst, uint8_t* bottom_dst, int len);

extern __cc_WebPUpsampleLinePairFunc WebPUpsamplers[ ];

static int __cc_EmitFancyRGB(const __cc_VP8Io* const io, __cc_WebPDecParams* const p) {
  int num_lines_out = io->mb_h;
  const __cc_WebPRGBABuffer* const buf = &p->output->u.RGBA;
  uint8_t* dst = buf->rgba + io->mb_y * buf->stride;
  __cc_WebPUpsampleLinePairFunc upsample = WebPUpsamplers[p->output->colorspace];
  const uint8_t* cur_y = io->y;
  const uint8_t* cur_u = io->u;
  const uint8_t* cur_v = io->v;
  const uint8_t* top_u = p->tmp_u;
  const uint8_t* top_v = p->tmp_v;
  int y = io->mb_y;
  const int y_end = io->mb_y + io->mb_h;
  const int mb_w = io->mb_w;
  const int uv_w = (mb_w + 1) / 2;

  if (y == 0) {

    upsample(cur_y, ((void *)0), cur_u, cur_v, cur_u, cur_v, dst, ((void *)0), mb_w);
  } else {

    upsample(p->tmp_y, cur_y, top_u, top_v, cur_u, cur_v,
             dst - buf->stride, dst, mb_w);
    ++num_lines_out;
  }

  for (; y + 2 < y_end; y += 2) {
    top_u = cur_u;
    top_v = cur_v;
    cur_u += io->uv_stride;
    cur_v += io->uv_stride;
    dst += 2 * buf->stride;
    cur_y += 2 * io->y_stride;
    upsample(cur_y - io->y_stride, cur_y,
             top_u, top_v, cur_u, cur_v,
             dst - buf->stride, dst, mb_w);
  }

  cur_y += io->y_stride;
  if (io->crop_top + y_end < io->crop_bottom) {

    memcpy(p->tmp_y, cur_y, mb_w * sizeof(*p->tmp_y));
    memcpy(p->tmp_u, cur_u, uv_w * sizeof(*p->tmp_u));
    memcpy(p->tmp_v, cur_v, uv_w * sizeof(*p->tmp_v));


    num_lines_out--;
  } else {

    if (!(y_end & 1)) {
      upsample(cur_y, ((void *)0), cur_u, cur_v, cur_u, cur_v,
               dst + buf->stride, ((void *)0), mb_w);
    }
  }
  return num_lines_out;
}

static inline int __cc_WebPIsRGBMode(__cc_WEBP_CSP_MODE mode) {
  return (mode < MODE_YUV);
}

void __cc_WebPInitUpsamplers(void);

int __cc_WebPIoInitFromOptions(const __cc_WebPDecoderOptions* const options,
                          __cc_VP8Io* const io, __cc_WEBP_CSP_MODE src_colorspace);

void __cc_VP8YUVInit(void);

void __cc_WebPMultRows(uint8_t* ptr, int stride,
                  const uint8_t* alpha, int alpha_stride,
                  int width, int num_rows, int inverse);

int __cc_WebPRescalerExport(__cc_WebPRescaler* const rescaler);

static int __cc_Rescale(const uint8_t* src, int src_stride,
                   int new_lines, __cc_WebPRescaler* const wrk) {
  int num_lines_out = 0;
  while (new_lines > 0) {
    const int lines_in = __cc_WebPRescalerImport(wrk, new_lines, src, src_stride);
    src += lines_in * src_stride;
    new_lines -= lines_in;
    num_lines_out += __cc_WebPRescalerExport(wrk);
  }
  return num_lines_out;
}

static int __cc_EmitRescaledAlphaYUV(const __cc_VP8Io* const io, __cc_WebPDecParams* const p) {
  if (io->a != ((void *)0)) {
    const __cc_WebPYUVABuffer* const buf = &p->output->u.YUVA;
    uint8_t* dst_y = buf->y + p->last_y * buf->y_stride;
    const uint8_t* src_a = buf->a + p->last_y * buf->a_stride;
    const int num_lines_out = __cc_Rescale(io->a, io->width, io->mb_h, &p->scaler_a);
    if (num_lines_out > 0) {
      __cc_WebPMultRows(dst_y, buf->y_stride, src_a, buf->a_stride,
                   p->scaler_a.dst_width, num_lines_out, 1);
    }
  }
  return 0;
}

static int __cc_EmitRescaledYUV(const __cc_VP8Io* const io, __cc_WebPDecParams* const p) {
  const int mb_h = io->mb_h;
  const int uv_mb_h = (mb_h + 1) >> 1;
  __cc_WebPRescaler* const scaler = &p->scaler_y;
  int num_lines_out = 0;
  if (__cc_WebPIsAlphaMode(p->output->colorspace) && io->a != ((void *)0)) {




    __cc_WebPMultRows((uint8_t*)io->y, io->y_stride,
                 io->a, io->width, io->mb_w, mb_h, 0);
  }
  num_lines_out = __cc_Rescale(io->y, io->y_stride, mb_h, scaler);
  __cc_Rescale(io->u, io->uv_stride, uv_mb_h, &p->scaler_u);
  __cc_Rescale(io->v, io->uv_stride, uv_mb_h, &p->scaler_v);
  return num_lines_out;
}

static int __cc_InitYUVRescaler(const __cc_VP8Io* const io, __cc_WebPDecParams* const p) {
  const int has_alpha = __cc_WebPIsAlphaMode(p->output->colorspace);
  const __cc_WebPYUVABuffer* const buf = &p->output->u.YUVA;
  const int out_width = io->scaled_width;
  const int out_height = io->scaled_height;
  const int uv_out_width = (out_width + 1) >> 1;
  const int uv_out_height = (out_height + 1) >> 1;
  const int uv_in_width = (io->mb_w + 1) >> 1;
  const int uv_in_height = (io->mb_h + 1) >> 1;
  const size_t work_size = 2 * out_width;
  const size_t uv_work_size = 2 * uv_out_width;
  size_t tmp_size;
  __cc_rescaler_t* work;

  tmp_size = (work_size + 2 * uv_work_size) * sizeof(*work);
  if (has_alpha) {
    tmp_size += work_size * sizeof(*work);
  }
  p->memory = __cc_WebPSafeMalloc(1ULL, tmp_size);
  if (p->memory == ((void *)0)) {
    return 0;
  }
  work = (__cc_rescaler_t*)p->memory;
  __cc_WebPRescalerInit(&p->scaler_y, io->mb_w, io->mb_h,
                   buf->y, out_width, out_height, buf->y_stride, 1,
                   work);
  __cc_WebPRescalerInit(&p->scaler_u, uv_in_width, uv_in_height,
                   buf->u, uv_out_width, uv_out_height, buf->u_stride, 1,
                   work + work_size);
  __cc_WebPRescalerInit(&p->scaler_v, uv_in_width, uv_in_height,
                   buf->v, uv_out_width, uv_out_height, buf->v_stride, 1,
                   work + work_size + uv_work_size);
  p->emit = __cc_EmitRescaledYUV;

  if (has_alpha) {
    __cc_WebPRescalerInit(&p->scaler_a, io->mb_w, io->mb_h,
                     buf->a, out_width, out_height, buf->a_stride, 1,
                     work + work_size + 2 * uv_work_size);
    p->emit_alpha = __cc_EmitRescaledAlphaYUV;
    __cc_WebPInitAlphaProcessing();
  }
  return 1;
}

static int __cc_CustomSetup(__cc_VP8Io* io) {
  __cc_WebPDecParams* const p = (__cc_WebPDecParams*)io->opaque;
  const __cc_WEBP_CSP_MODE colorspace = p->output->colorspace;
  const int is_rgb = __cc_WebPIsRGBMode(colorspace);
  const int is_alpha = __cc_WebPIsAlphaMode(colorspace);

  p->memory = ((void *)0);
  p->emit = ((void *)0);
  p->emit_alpha = ((void *)0);
  p->emit_alpha_row = ((void *)0);
  if (!__cc_WebPIoInitFromOptions(p->options, io, is_alpha ? MODE_YUV : MODE_YUVA)) {
    return 0;
  }
  if (is_alpha && __cc_WebPIsPremultipliedMode(colorspace)) {
    __cc_WebPInitUpsamplers();
  }
  if (io->use_scaling) {
    const int ok = is_rgb ? __cc_InitRGBRescaler(io, p) : __cc_InitYUVRescaler(io, p);
    if (!ok) {
      return 0;
    }
  } else {
    if (is_rgb) {
      p->emit = __cc_EmitSampledRGB;
      if (io->fancy_upsampling) {

        const int uv_width = (io->mb_w + 1) >> 1;
        p->memory = __cc_WebPSafeMalloc(1ULL, (size_t)(io->mb_w + 2 * uv_width));
        if (p->memory == ((void *)0)) {
          return 0;
        }
        p->tmp_y = (uint8_t*)p->memory;
        p->tmp_u = p->tmp_y + io->mb_w;
        p->tmp_v = p->tmp_u + uv_width;
        p->emit = __cc_EmitFancyRGB;
        __cc_WebPInitUpsamplers();

      } else {
        __cc_WebPInitSamplers();
      }
    } else {
      p->emit = __cc_EmitYUV;
    }
    if (is_alpha) {
      p->emit_alpha =
          (colorspace == MODE_RGBA_4444 || colorspace == MODE_rgbA_4444) ?
              __cc_EmitAlphaRGBA4444
          : is_rgb ? __cc_EmitAlphaRGB
          : __cc_EmitAlphaYUV;
      if (is_rgb) {
        __cc_WebPInitAlphaProcessing();
      }
    }
  }

  if (is_rgb) {
    __cc_VP8YUVInit();
  }
  return 1;
}

static int __cc_CustomPut(const __cc_VP8Io* io) {
  __cc_WebPDecParams* const p = (__cc_WebPDecParams*)io->opaque;
  const int mb_w = io->mb_w;
  const int mb_h = io->mb_h;
  int num_lines_out;
  ((!(io->mb_y & 1)) ? (void) (0) : __assert_fail ("!(io->mb_y & 1)", "extraction.1.c", 1083, __PRETTY_FUNCTION__));

  if (mb_w <= 0 || mb_h <= 0) {
    return 0;
  }
  num_lines_out = p->emit(io, p);
  if (p->emit_alpha != ((void *)0)) {
    p->emit_alpha(io, p);
  }
  p->last_y += num_lines_out;
  return 1;
}

void __cc_WebPInitCustomIo(__cc_WebPDecParams* const params, __cc_VP8Io* const io) {
  io->put = __cc_CustomPut;
  io->setup = __cc_CustomSetup;
  io->teardown = __cc_CustomTeardown;
  io->opaque = params;
}

typedef enum {
  NOT_OK = 0,
  OK,
  WORK
} __cc_WebPWorkerStatus;

typedef struct WebPWorkerImpl __cc_WebPWorkerImpl;

typedef int (*__cc_WebPWorkerHook)(void*, void*);

typedef struct {
  __cc_WebPWorkerImpl* impl_;
  __cc_WebPWorkerStatus status_;
  __cc_WebPWorkerHook hook;
  void* data1;
  void* data2;
  int had_error;
} __cc_WebPWorker;

static int __cc_Sync(__cc_WebPWorker* const worker) {



  ((worker->status_ <= OK) ? (void) (0) : __assert_fail ("worker->status_ <= OK", "extraction.1.c", 1126, __PRETTY_FUNCTION__));
  return !worker->had_error;
}

static void __cc_Init(__cc_WebPWorker* const worker) {
  memset(worker, 0, sizeof(*worker));
  worker->status_ = NOT_OK;
}

static void __cc_End(__cc_WebPWorker* const worker) {
  worker->status_ = NOT_OK;
  ((worker->impl_ == ((void *)0)) ? (void) (0) : __assert_fail ("worker->impl_ == ((void *)0)", "extraction.1.c", 1147, __PRETTY_FUNCTION__));

  ((worker->status_ == NOT_OK) ? (void) (0) : __assert_fail ("worker->status_ == NOT_OK", "extraction.1.c", 1149, __PRETTY_FUNCTION__));
}

static void __cc_Execute(__cc_WebPWorker* const worker) {
  if (worker->hook != ((void *)0)) {
    worker->had_error |= !worker->hook(worker->data1, worker->data2);
  }
}

static void __cc_Launch(__cc_WebPWorker* const worker) {



  __cc_Execute(worker);

}

static int __cc_Reset(__cc_WebPWorker* const worker) {
  int ok = 1;
  worker->had_error = 0;
  if (worker->status_ < OK) {
    worker->status_ = OK;

  } else if (worker->status_ > OK) {
    ok = __cc_Sync(worker);
  }
  ((!ok || (worker->status_ == OK)) ? (void) (0) : __assert_fail ("!ok || (worker->status_ == OK)", "extraction.1.c", 1200, __PRETTY_FUNCTION__));
  return ok;
}

typedef struct {

  void (*Init)(__cc_WebPWorker* const worker);


  int (*Reset)(__cc_WebPWorker* const worker);


  int (*Sync)(__cc_WebPWorker* const worker);



  void (*Launch)(__cc_WebPWorker* const worker);




  void (*Execute)(__cc_WebPWorker* const worker);


  void (*End)(__cc_WebPWorker* const worker);
} __cc_WebPWorkerInterface;

static __cc_WebPWorkerInterface g_worker_interface = {
  __cc_Init, __cc_Reset, __cc_Sync, __cc_Launch, __cc_Execute, __cc_End
};

const __cc_WebPWorkerInterface* __cc_WebPGetWorkerInterface(void) {
  return &g_worker_interface;
}



static const uint32_t kRandomTable[55] = {
  0x0de15230, 0x03b31886, 0x775faccb, 0x1c88626a, 0x68385c55, 0x14b3b828,
  0x4a85fef8, 0x49ddb84b, 0x64fcf397, 0x5c550289, 0x4a290000, 0x0d7ec1da,
  0x5940b7ab, 0x5492577d, 0x4e19ca72, 0x38d38c69, 0x0c01ee65, 0x32a1755f,
  0x5437f652, 0x5abb2c32, 0x0faa57b1, 0x73f533e7, 0x685feeda, 0x7563cce2,
  0x6e990e83, 0x4730a7ed, 0x4fc0d9c6, 0x496b153c, 0x4f1403fa, 0x541afb0c,
  0x73990b32, 0x26d7cb1c, 0x6fcc3706, 0x2cbb77d8, 0x75762f2a, 0x6425ccdd,
  0x24b35461, 0x0a7d8715, 0x220414a8, 0x141ebf67, 0x56b41583, 0x73e502e3,
  0x44cab16f, 0x28264d42, 0x73baaefb, 0x0a50ebed, 0x1d6ab6fb, 0x0d3ad40b,
  0x35db3b68, 0x2b081e83, 0x77ce6b95, 0x5181e5f0, 0x78853bbc, 0x009f9494,
  0x27e5ed3c
};

typedef struct {
  uint32_t *colors_;
  int hash_shift_;
} __cc_VP8LColorCache;

void __cc_VP8LColorCacheClear(__cc_VP8LColorCache* const cc) {
  if (cc != ((void *)0)) {
    __cc_WebPSafeFree(cc->colors_);
    cc->colors_ = ((void *)0);
  }
}

void __cc_WebPMultRows(uint8_t* ptr, int stride,
                  const uint8_t* alpha, int alpha_stride,
                  int width, int num_rows, int inverse) {
  int n;
  for (n = 0; n < num_rows; ++n) {
    WebPMultRow(ptr, alpha, width, inverse);
    ptr += stride;
    alpha += alpha_stride;
  }
}



static int __cc_CheckSizeArgumentsOverflow(uint64_t nmemb, size_t size) {
  const uint64_t total_size = nmemb * size;
  if (nmemb == 0) return 1;
  if ((uint64_t)size > (1ULL << 40) / nmemb) return 0;
  if (total_size != (size_t)total_size) return 0;
  return 1;
}






void* __cc_WebPSafeCalloc(uint64_t nmemb, size_t size) {
  void* ptr;
  do {} while (0);
  if (!__cc_CheckSizeArgumentsOverflow(nmemb, size)) return ((void *)0);
  ((nmemb * size > 0) ? (void) (0) : __assert_fail ("nmemb * size > 0", "extraction.1.c", 1304, __PRETTY_FUNCTION__));
  ptr = calloc((size_t)nmemb, size);
  do {} while (0);
  return ptr;
}
static inline int __cc_GradientPredictor(uint8_t a, uint8_t b, uint8_t c) {
  const int g = a + b - c;
  return ((g & ~0xff) == 0) ? g : (g < 0) ? 0 : 255;
}

static inline void __cc_PredictLine(const uint8_t* src, const uint8_t* pred,
                                    uint8_t* dst, int length, int inverse) {
  int i;
  if (inverse) {
    for (i = 0; i < length; ++i) dst[i] = src[i] + pred[i];
  } else {
    for (i = 0; i < length; ++i) dst[i] = src[i] - pred[i];
  }
}

static inline void __cc_DoGradientFilter(const uint8_t* in,
                                         int width, int height, int stride,
                                         int row, int num_rows,
                                         int inverse, uint8_t* out) {
  const uint8_t* preds;
  const size_t start_offset = row * stride;
  const int last_row = row + num_rows;
  ((in != ((void *)0)) ? (void) (0) : __assert_fail ("in != ((void *)0)", "extraction.1.c", 1341, __PRETTY_FUNCTION__)); ((out != ((void *)0)) ? (void) (0) : __assert_fail ("out != ((void *)0)", "extraction.1.c", 1341, __PRETTY_FUNCTION__)); ((width > 0) ? (void) (0) : __assert_fail ("width > 0", "extraction.1.c", 1341, __PRETTY_FUNCTION__)); ((height > 0) ? (void) (0) : __assert_fail ("height > 0", "extraction.1.c", 1341, __PRETTY_FUNCTION__)); ((stride >= width) ? (void) (0) : __assert_fail ("stride >= width", "extraction.1.c", 1341, __PRETTY_FUNCTION__)); ((row >= 0 && num_rows > 0 && row + num_rows <= height) ? (void) (0) : __assert_fail ("row >= 0 && num_rows > 0 && row + num_rows <= height", "extraction.1.c", 1341, __PRETTY_FUNCTION__)); (void)height;;
  in += start_offset;
  out += start_offset;
  preds = inverse ? out : in;


  if (row == 0) {
    out[0] = in[0];
    __cc_PredictLine(in + 1, preds, out + 1, width - 1, inverse);
    row = 1;
    preds += stride;
    in += stride;
    out += stride;
  }


  while (row < last_row) {
    int w;

    __cc_PredictLine(in, preds - stride, out, 1, inverse);
    for (w = 1; w < width; ++w) {
      const int pred = __cc_GradientPredictor(preds[w - 1],
                                         preds[w - stride],
                                         preds[w - stride - 1]);
      out[w] = in[w] + (inverse ? pred : -pred);
    }
    ++row;
    preds += stride;
    in += stride;
    out += stride;
  }
}

static inline void __cc_DoVerticalFilter(const uint8_t* in,
                                         int width, int height, int stride,
                                         int row, int num_rows,
                                         int inverse, uint8_t* out) {
  const uint8_t* preds;
  const size_t start_offset = row * stride;
  const int last_row = row + num_rows;
  ((in != ((void *)0)) ? (void) (0) : __assert_fail ("in != ((void *)0)", "extraction.1.c", 1381, __PRETTY_FUNCTION__)); ((out != ((void *)0)) ? (void) (0) : __assert_fail ("out != ((void *)0)", "extraction.1.c", 1381, __PRETTY_FUNCTION__)); ((width > 0) ? (void) (0) : __assert_fail ("width > 0", "extraction.1.c", 1381, __PRETTY_FUNCTION__)); ((height > 0) ? (void) (0) : __assert_fail ("height > 0", "extraction.1.c", 1381, __PRETTY_FUNCTION__)); ((stride >= width) ? (void) (0) : __assert_fail ("stride >= width", "extraction.1.c", 1381, __PRETTY_FUNCTION__)); ((row >= 0 && num_rows > 0 && row + num_rows <= height) ? (void) (0) : __assert_fail ("row >= 0 && num_rows > 0 && row + num_rows <= height", "extraction.1.c", 1381, __PRETTY_FUNCTION__)); (void)height;;
  in += start_offset;
  out += start_offset;
  preds = inverse ? out : in;

  if (row == 0) {

    out[0] = in[0];

    __cc_PredictLine(in + 1, preds, out + 1, width - 1, inverse);
    row = 1;
    in += stride;
    out += stride;
  } else {

    preds -= stride;
  }


  while (row < last_row) {
    __cc_PredictLine(in, preds, out, width, inverse);
    ++row;
    preds += stride;
    in += stride;
    out += stride;
  }
}

static inline void __cc_DoHorizontalFilter(const uint8_t* in,
                                           int width, int height, int stride,
                                           int row, int num_rows,
                                           int inverse, uint8_t* out) {
  const uint8_t* preds;
  const size_t start_offset = row * stride;
  const int last_row = row + num_rows;
  ((in != ((void *)0)) ? (void) (0) : __assert_fail ("in != ((void *)0)", "extraction.1.c", 1416, __PRETTY_FUNCTION__)); ((out != ((void *)0)) ? (void) (0) : __assert_fail ("out != ((void *)0)", "extraction.1.c", 1416, __PRETTY_FUNCTION__)); ((width > 0) ? (void) (0) : __assert_fail ("width > 0", "extraction.1.c", 1416, __PRETTY_FUNCTION__)); ((height > 0) ? (void) (0) : __assert_fail ("height > 0", "extraction.1.c", 1416, __PRETTY_FUNCTION__)); ((stride >= width) ? (void) (0) : __assert_fail ("stride >= width", "extraction.1.c", 1416, __PRETTY_FUNCTION__)); ((row >= 0 && num_rows > 0 && row + num_rows <= height) ? (void) (0) : __assert_fail ("row >= 0 && num_rows > 0 && row + num_rows <= height", "extraction.1.c", 1416, __PRETTY_FUNCTION__)); (void)height;;
  in += start_offset;
  out += start_offset;
  preds = inverse ? out : in;

  if (row == 0) {

    out[0] = in[0];
    __cc_PredictLine(in + 1, preds, out + 1, width - 1, inverse);
    row = 1;
    preds += stride;
    in += stride;
    out += stride;
  }


  while (row < last_row) {

    __cc_PredictLine(in, preds - stride, out, 1, inverse);
    __cc_PredictLine(in + 1, preds, out + 1, width - 1, inverse);
    ++row;
    preds += stride;
    in += stride;
    out += stride;
  }
}

typedef enum {
  WEBP_FILTER_NONE = 0,
  WEBP_FILTER_HORIZONTAL,
  WEBP_FILTER_VERTICAL,
  WEBP_FILTER_GRADIENT,
  WEBP_FILTER_LAST = WEBP_FILTER_GRADIENT + 1,
  WEBP_FILTER_BEST,
  WEBP_FILTER_FAST
} __cc_WEBP_FILTER_TYPE;

void __cc_WebPInitCustomIo(__cc_WebPDecParams* const params, __cc_VP8Io* const io);

int __cc_VP8InitIoInternal(__cc_VP8Io* const, int);

static inline int __cc_VP8InitIo(__cc_VP8Io* const io) {
  return __cc_VP8InitIoInternal(io, 0x0203);
}

extern void* __cc_WebPSafeCalloc(uint64_t nmemb, size_t size);

extern int __cc_WebPInitDecBufferInternal(__cc_WebPDecBuffer*, int);

static inline int __cc_WebPInitDecBuffer(__cc_WebPDecBuffer* buffer) {
  return __cc_WebPInitDecBufferInternal(buffer, 0x0203);
}

typedef struct WebPDecoderConfig __cc_WebPDecoderConfig;

typedef struct WebPBitstreamFeatures __cc_WebPBitstreamFeatures;

struct WebPBitstreamFeatures {
  int width;
  int height;
  int has_alpha;
  int has_animation;
  int format;


  int no_incremental_decoding;

  int rotate;
  int uv_sampling;
  uint32_t pad[2];
};

struct WebPDecoderConfig {
  __cc_WebPBitstreamFeatures input;
  __cc_WebPDecBuffer output;
  __cc_WebPDecoderOptions options;
};

typedef enum VP8StatusCode {
  VP8_STATUS_OK = 0,
  VP8_STATUS_OUT_OF_MEMORY,
  VP8_STATUS_INVALID_PARAM,
  VP8_STATUS_BITSTREAM_ERROR,
  VP8_STATUS_UNSUPPORTED_FEATURE,
  VP8_STATUS_SUSPENDED,
  VP8_STATUS_USER_ABORT,
  VP8_STATUS_NOT_ENOUGH_DATA
} __cc_VP8StatusCode;

extern __cc_VP8StatusCode __cc_WebPGetFeaturesInternal(
    const uint8_t*, size_t, __cc_WebPBitstreamFeatures*, int);

static inline __cc_VP8StatusCode __cc_WebPGetFeatures(
    const uint8_t* data, size_t data_size,
    __cc_WebPBitstreamFeatures* features) {
  return __cc_WebPGetFeaturesInternal(data, data_size, features,
                                 0x0203);
}

static void __cc_VerticalUnfilter(int width, int height, int stride, int row,
                             int num_rows, uint8_t* data) {
  __cc_DoVerticalFilter(data, width, height, stride, row, num_rows, 1, data);
}

typedef enum {
  PREDICTOR_TRANSFORM = 0,
  CROSS_COLOR_TRANSFORM = 1,
  SUBTRACT_GREEN = 2,
  COLOR_INDEXING_TRANSFORM = 3
} __cc_VP8LImageTransformType;

struct VP8LTransform {
  __cc_VP8LImageTransformType type_;
  int bits_;
  int xsize_;
  int ysize_;
  uint32_t *data_;
};

typedef struct HuffmanTree __cc_HuffmanTree;

typedef struct {
  int symbol_;
  int children_;
} __cc_HuffmanTreeNode;





struct HuffmanTree {

  uint8_t lut_bits_[(1U << 7)];
  int16_t lut_symbol_[(1U << 7)];
  int16_t lut_jump_[(1U << 7)];

  __cc_HuffmanTreeNode* root_;
  int max_nodes_;
  int num_nodes_;
};



struct HTreeGroup {
  __cc_HuffmanTree htrees_[5];
};

typedef struct HTreeGroup __cc_HTreeGroup;

typedef struct {
  int color_cache_size_;
  __cc_VP8LColorCache color_cache_;

  int huffman_mask_;
  int huffman_subsample_bits_;
  int huffman_xsize_;
  uint32_t *huffman_image_;
  int num_htree_groups_;
  __cc_HTreeGroup *htree_groups_;
} __cc_VP8LMetadata;

typedef uint64_t __cc_vp8l_val_t;

typedef struct {
  __cc_vp8l_val_t val_;
  const uint8_t* buf_;
  size_t len_;
  size_t pos_;
  int bit_pos_;
  int eos_;
  int error_;
} __cc_VP8LBitReader;

typedef struct VP8LTransform __cc_VP8LTransform;

typedef enum {
  READ_DATA = 0,
  READ_HDR = 1,
  READ_DIM = 2
} __cc_VP8LDecodeState;



struct VP8LDecoder {
  __cc_VP8StatusCode status_;
  __cc_VP8LDecodeState action_;
  __cc_VP8LDecodeState state_;
  __cc_VP8Io *io_;

  const __cc_WebPDecBuffer *output_;

  uint32_t *pixels_;

  uint32_t *argb_cache_;

  __cc_VP8LBitReader br_;

  int width_;
  int height_;
  int last_row_;
  int last_pixel_;


  int last_out_row_;

  __cc_VP8LMetadata hdr_;

  int next_transform_;
  __cc_VP8LTransform transforms_[4];

  uint32_t transforms_seen_;

  uint8_t *rescaler_memory;
  __cc_WebPRescaler *rescaler;
};

typedef struct ALPHDecoder __cc_ALPHDecoder;

struct ALPHDecoder;

typedef struct {
  int16_t coeffs_[384];
  uint8_t is_i4x4_;
  uint8_t imodes_[16];
  uint8_t uvmode_;







  uint32_t non_zero_y_;
  uint32_t non_zero_uv_;
  uint8_t dither_;
  uint8_t skip_;
  uint8_t segment_;
} __cc_VP8MBData;

typedef struct {
  uint8_t f_limit_;
  uint8_t f_ilevel_;
  uint8_t f_inner_;
  uint8_t hev_thresh_;
} __cc_VP8FInfo;

typedef struct {
  uint8_t nz_;
  uint8_t nz_dc_;
} __cc_VP8MB;

struct VP8LDecoder;

struct ALPHDecoder {
  int width_;
  int height_;
  int method_;
  __cc_WEBP_FILTER_TYPE filter_;
  int pre_processing_;
  struct VP8LDecoder* vp8l_dec_;
  __cc_VP8Io io_;
  int use_8b_decode;


};

typedef int __cc_quant_t[2];

typedef struct {
  __cc_quant_t y1_mat_, y2_mat_, uv_mat_;

  int uv_quant_;
  int dither_;
} __cc_VP8QuantMatrix;

enum { MB_FEATURE_TREE_PROBS = 3,
       NUM_MB_SEGMENTS = 4,
       NUM_REF_LF_DELTAS = 4,
       NUM_MODE_LF_DELTAS = 4,
       MAX_NUM_PARTITIONS = 8,

       NUM_TYPES = 4,
       NUM_BANDS = 8,
       NUM_CTX = 3,
       NUM_PROBAS = 11,
       NUM_MV_PROBAS = 19 };

typedef struct {
  int index1_, index2_;
  uint32_t tab_[55];
  int amp_;
} __cc_VP8Random;

typedef uint8_t __cc_VP8ProbaArray[NUM_PROBAS];

typedef struct {
  __cc_VP8ProbaArray probas_[NUM_CTX];
} __cc_VP8BandProbas;

typedef struct {
  uint8_t segments_[MB_FEATURE_TREE_PROBS];

  __cc_VP8BandProbas bands_[NUM_TYPES][NUM_BANDS];
} __cc_VP8Proba;

typedef struct {
  int id_;
  int mb_y_;
  int filter_row_;
  __cc_VP8FInfo* f_info_;
  __cc_VP8MBData* mb_data_;
  __cc_VP8Io io_;
} __cc_VP8ThreadContext;

typedef struct {
  int simple_;
  int level_;
  int sharpness_;
  int use_lf_delta_;
  int ref_lf_delta_[NUM_REF_LF_DELTAS];
  int mode_lf_delta_[NUM_MODE_LF_DELTAS];
} __cc_VP8FilterHeader;

typedef struct {
  uint8_t y[16], u[8], v[8];
} __cc_VP8TopSamples;

typedef struct {
  int use_segment_;
  int update_map_;
  int absolute_delta_;
  int8_t quantizer_[NUM_MB_SEGMENTS];
  int8_t filter_strength_[NUM_MB_SEGMENTS];
} __cc_VP8SegmentHeader;

typedef struct {
  uint16_t width_;
  uint16_t height_;
  uint8_t xscale_;
  uint8_t yscale_;
  uint8_t colorspace_;
  uint8_t clamp_type_;
} __cc_VP8PictureHeader;

typedef struct {
  uint8_t key_frame_;
  uint8_t profile_;
  uint8_t show_;
  uint32_t partition_length_;
} __cc_VP8FrameHeader;

typedef struct VP8BitReader __cc_VP8BitReader;

typedef uint32_t __cc_range_t;

typedef uint32_t __cc_bit_t;

struct VP8BitReader {

  __cc_bit_t value_;
  __cc_range_t range_;
  int bits_;

  const uint8_t* buf_;
  const uint8_t* buf_end_;
  int eof_;
};

struct VP8Decoder {
  __cc_VP8StatusCode status_;
  int ready_;
  const char* error_msg_;


  __cc_VP8BitReader br_;


  __cc_VP8FrameHeader frm_hdr_;
  __cc_VP8PictureHeader pic_hdr_;
  __cc_VP8FilterHeader filter_hdr_;
  __cc_VP8SegmentHeader segment_hdr_;


  __cc_WebPWorker worker_;
  int mt_method_;

  int cache_id_;
  int num_caches_;
  __cc_VP8ThreadContext thread_ctx_;


  int mb_w_, mb_h_;


  int tl_mb_x_, tl_mb_y_;
  int br_mb_x_, br_mb_y_;


  int num_parts_;

  __cc_VP8BitReader parts_[MAX_NUM_PARTITIONS];


  int dither_;
  __cc_VP8Random dithering_rg_;


  __cc_VP8QuantMatrix dqm_[NUM_MB_SEGMENTS];


  __cc_VP8Proba proba_;
  int use_skip_proba_;
  uint8_t skip_p_;


  uint8_t* intra_t_;
  uint8_t intra_l_[4];

  __cc_VP8TopSamples* yuv_t_;

  __cc_VP8MB* mb_info_;
  __cc_VP8FInfo* f_info_;
  uint8_t* yuv_b_;

  uint8_t* cache_y_;
  uint8_t* cache_u_;
  uint8_t* cache_v_;
  int cache_y_stride_;
  int cache_uv_stride_;


  void* mem_;
  size_t mem_size_;


  int mb_x_, mb_y_;
  __cc_VP8MBData* mb_data_;


  int filter_type_;
  __cc_VP8FInfo fstrengths_[NUM_MB_SEGMENTS][2];


  struct ALPHDecoder* alph_dec_;
  const uint8_t* alpha_data_;
  size_t alpha_data_size_;
  int is_alpha_decoded_;
  uint8_t* alpha_plane_;
  int alpha_dithering_;
};

typedef struct VP8Decoder __cc_VP8Decoder;

typedef struct VP8LDecoder __cc_VP8LDecoder;





struct WebPPicture;

struct WebPPicture;

struct WebPPicture;

struct WebPPicture;

typedef struct WebPPicture __cc_WebPPicture;

typedef int (*__cc_WebPProgressHook)(int percent, const __cc_WebPPicture* picture);

typedef enum WebPEncodingError {
  VP8_ENC_OK = 0,
  VP8_ENC_ERROR_OUT_OF_MEMORY,
  VP8_ENC_ERROR_BITSTREAM_OUT_OF_MEMORY,
  VP8_ENC_ERROR_NULL_PARAMETER,
  VP8_ENC_ERROR_INVALID_CONFIGURATION,
  VP8_ENC_ERROR_BAD_DIMENSION,
  VP8_ENC_ERROR_PARTITION0_OVERFLOW,
  VP8_ENC_ERROR_PARTITION_OVERFLOW,
  VP8_ENC_ERROR_BAD_WRITE,
  VP8_ENC_ERROR_FILE_TOO_BIG,
  VP8_ENC_ERROR_USER_ABORT,
  VP8_ENC_ERROR_LAST
} __cc_WebPEncodingError;

typedef struct WebPAuxStats __cc_WebPAuxStats;

struct WebPAuxStats {
  int coded_size;

  float PSNR[5];
  int block_count[3];
  int header_bytes[2];

  int residual_bytes[3][4];

  int segment_size[4];
  int segment_quant[4];
  int segment_level[4];

  int alpha_data_size;
  int layer_data_size;


  uint32_t lossless_features;

  int histogram_bits;
  int transform_bits;
  int cache_bits;
  int palette_size;
  int lossless_size;

  uint32_t pad[4];
};

typedef int (*__cc_WebPWriterFunction)(const uint8_t* data, size_t data_size,
                                  const __cc_WebPPicture* picture);

typedef enum WebPEncCSP {

  WEBP_YUV420 = 0,
  WEBP_YUV420A = 4,
  WEBP_CSP_UV_MASK = 3,
  WEBP_CSP_ALPHA_BIT = 4
} __cc_WebPEncCSP;

struct WebPPicture {






  int use_argb;


  __cc_WebPEncCSP colorspace;
  int width, height;
  uint8_t *y, *u, *v;
  int y_stride, uv_stride;
  uint8_t* a;
  int a_stride;
  uint32_t pad1[2];


  uint32_t* argb;
  int argb_stride;
  uint32_t pad2[3];




  __cc_WebPWriterFunction writer;
  void* custom_ptr;


  int extra_info_type;



  uint8_t* extra_info;







  __cc_WebPAuxStats* stats;


  __cc_WebPEncodingError error_code;


  __cc_WebPProgressHook progress_hook;

  void* user_data;


  uint32_t pad3[3];


  uint8_t *pad4, *pad5;
  uint32_t pad6[8];



  void* memory_;
  void* memory_argb_;
  void* pad7[2];
};




static const uint32_t kBitMask[24 + 1] = {
  0,
  0x000001, 0x000003, 0x000007, 0x00000f,
  0x00001f, 0x00003f, 0x00007f, 0x0000ff,
  0x0001ff, 0x0003ff, 0x0007ff, 0x000fff,
  0x001fff, 0x003fff, 0x007fff, 0x00ffff,
  0x01ffff, 0x03ffff, 0x07ffff, 0x0fffff,
  0x1fffff, 0x3fffff, 0x7fffff, 0xffffff
};

void __cc_VP8LInitBitReader(__cc_VP8LBitReader* const br, const uint8_t* const start,
                       size_t length) {
  size_t i;
  __cc_vp8l_val_t value = 0;
  ((br != ((void *)0)) ? (void) (0) : __assert_fail ("br != ((void *)0)", "extraction.1.c", 2027, __PRETTY_FUNCTION__));
  ((start != ((void *)0)) ? (void) (0) : __assert_fail ("start != ((void *)0)", "extraction.1.c", 2028, __PRETTY_FUNCTION__));
  ((length < 0xfffffff8u) ? (void) (0) : __assert_fail ("length < 0xfffffff8u", "extraction.1.c", 2029, __PRETTY_FUNCTION__));

  br->len_ = length;
  br->val_ = 0;
  br->bit_pos_ = 0;
  br->eos_ = 0;
  br->error_ = 0;

  if (length > sizeof(br->val_)) {
    length = sizeof(br->val_);
  }
  for (i = 0; i < length; ++i) {
    value |= (__cc_vp8l_val_t)start[i] << (8 * i);
  }
  br->val_ = value;
  br->pos_ = length;
  br->buf_ = start;
}



const __cc_range_t kVP8NewRange[128] = {
  127, 127, 191, 127, 159, 191, 223, 127,
  143, 159, 175, 191, 207, 223, 239, 127,
  135, 143, 151, 159, 167, 175, 183, 191,
  199, 207, 215, 223, 231, 239, 247, 127,
  131, 135, 139, 143, 147, 151, 155, 159,
  163, 167, 171, 175, 179, 183, 187, 191,
  195, 199, 203, 207, 211, 215, 219, 223,
  227, 231, 235, 239, 243, 247, 251, 127,
  129, 131, 133, 135, 137, 139, 141, 143,
  145, 147, 149, 151, 153, 155, 157, 159,
  161, 163, 165, 167, 169, 171, 173, 175,
  177, 179, 181, 183, 185, 187, 189, 191,
  193, 195, 197, 199, 201, 203, 205, 207,
  209, 211, 213, 215, 217, 219, 221, 223,
  225, 227, 229, 231, 233, 235, 237, 239,
  241, 243, 245, 247, 249, 251, 253, 127
};



void __cc_VP8LoadFinalBytes(__cc_VP8BitReader* const br) {
  ((br != ((void *)0) && br->buf_ != ((void *)0)) ? (void) (0) : __assert_fail ("br != ((void *)0) && br->buf_ != ((void *)0)", "extraction.1.c", 2072, __PRETTY_FUNCTION__));

  if (br->buf_ < br->buf_end_) {
    br->bits_ += 8;
    br->value_ = (__cc_bit_t)(*br->buf_++) | (br->value_ << 8);
  } else if (!br->eof_) {
    br->value_ <<= 8;
    br->bits_ += 8;
    br->eof_ = 1;
  }
}



static inline int __cc_VP8LIsEndOfStream(const __cc_VP8LBitReader* const br) {
  ((br->pos_ <= br->len_) ? (void) (0) : __assert_fail ("br->pos_ <= br->len_", "extraction.1.c", 2087, __PRETTY_FUNCTION__));
  return (br->pos_ == br->len_) && (br->bit_pos_ > 64);
}

extern const __cc_WebPWorkerInterface* __cc_WebPGetWorkerInterface(void);

extern int __cc_WebPPictureAlloc(__cc_WebPPicture* picture);

extern void __cc_WebPPictureFree(__cc_WebPPicture* picture);





static const uint8_t
  CoeffsProba0[NUM_TYPES][NUM_BANDS][NUM_CTX][NUM_PROBAS] = {
  { { { 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128 },
      { 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128 },
      { 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128 }
    },
    { { 253, 136, 254, 255, 228, 219, 128, 128, 128, 128, 128 },
      { 189, 129, 242, 255, 227, 213, 255, 219, 128, 128, 128 },
      { 106, 126, 227, 252, 214, 209, 255, 255, 128, 128, 128 }
    },
    { { 1, 98, 248, 255, 236, 226, 255, 255, 128, 128, 128 },
      { 181, 133, 238, 254, 221, 234, 255, 154, 128, 128, 128 },
      { 78, 134, 202, 247, 198, 180, 255, 219, 128, 128, 128 },
    },
    { { 1, 185, 249, 255, 243, 255, 128, 128, 128, 128, 128 },
      { 184, 150, 247, 255, 236, 224, 128, 128, 128, 128, 128 },
      { 77, 110, 216, 255, 236, 230, 128, 128, 128, 128, 128 },
    },
    { { 1, 101, 251, 255, 241, 255, 128, 128, 128, 128, 128 },
      { 170, 139, 241, 252, 236, 209, 255, 255, 128, 128, 128 },
      { 37, 116, 196, 243, 228, 255, 255, 255, 128, 128, 128 }
    },
    { { 1, 204, 254, 255, 245, 255, 128, 128, 128, 128, 128 },
      { 207, 160, 250, 255, 238, 128, 128, 128, 128, 128, 128 },
      { 102, 103, 231, 255, 211, 171, 128, 128, 128, 128, 128 }
    },
    { { 1, 152, 252, 255, 240, 255, 128, 128, 128, 128, 128 },
      { 177, 135, 243, 255, 234, 225, 128, 128, 128, 128, 128 },
      { 80, 129, 211, 255, 194, 224, 128, 128, 128, 128, 128 }
    },
    { { 1, 1, 255, 128, 128, 128, 128, 128, 128, 128, 128 },
      { 246, 1, 255, 128, 128, 128, 128, 128, 128, 128, 128 },
      { 255, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128 }
    }
  },
  { { { 198, 35, 237, 223, 193, 187, 162, 160, 145, 155, 62 },
      { 131, 45, 198, 221, 172, 176, 220, 157, 252, 221, 1 },
      { 68, 47, 146, 208, 149, 167, 221, 162, 255, 223, 128 }
    },
    { { 1, 149, 241, 255, 221, 224, 255, 255, 128, 128, 128 },
      { 184, 141, 234, 253, 222, 220, 255, 199, 128, 128, 128 },
      { 81, 99, 181, 242, 176, 190, 249, 202, 255, 255, 128 }
    },
    { { 1, 129, 232, 253, 214, 197, 242, 196, 255, 255, 128 },
      { 99, 121, 210, 250, 201, 198, 255, 202, 128, 128, 128 },
      { 23, 91, 163, 242, 170, 187, 247, 210, 255, 255, 128 }
    },
    { { 1, 200, 246, 255, 234, 255, 128, 128, 128, 128, 128 },
      { 109, 178, 241, 255, 231, 245, 255, 255, 128, 128, 128 },
      { 44, 130, 201, 253, 205, 192, 255, 255, 128, 128, 128 }
    },
    { { 1, 132, 239, 251, 219, 209, 255, 165, 128, 128, 128 },
      { 94, 136, 225, 251, 218, 190, 255, 255, 128, 128, 128 },
      { 22, 100, 174, 245, 186, 161, 255, 199, 128, 128, 128 }
    },
    { { 1, 182, 249, 255, 232, 235, 128, 128, 128, 128, 128 },
      { 124, 143, 241, 255, 227, 234, 128, 128, 128, 128, 128 },
      { 35, 77, 181, 251, 193, 211, 255, 205, 128, 128, 128 }
    },
    { { 1, 157, 247, 255, 236, 231, 255, 255, 128, 128, 128 },
      { 121, 141, 235, 255, 225, 227, 255, 255, 128, 128, 128 },
      { 45, 99, 188, 251, 195, 217, 255, 224, 128, 128, 128 }
    },
    { { 1, 1, 251, 255, 213, 255, 128, 128, 128, 128, 128 },
      { 203, 1, 248, 255, 255, 128, 128, 128, 128, 128, 128 },
      { 137, 1, 177, 255, 224, 255, 128, 128, 128, 128, 128 }
    }
  },
  { { { 253, 9, 248, 251, 207, 208, 255, 192, 128, 128, 128 },
      { 175, 13, 224, 243, 193, 185, 249, 198, 255, 255, 128 },
      { 73, 17, 171, 221, 161, 179, 236, 167, 255, 234, 128 }
    },
    { { 1, 95, 247, 253, 212, 183, 255, 255, 128, 128, 128 },
      { 239, 90, 244, 250, 211, 209, 255, 255, 128, 128, 128 },
      { 155, 77, 195, 248, 188, 195, 255, 255, 128, 128, 128 }
    },
    { { 1, 24, 239, 251, 218, 219, 255, 205, 128, 128, 128 },
      { 201, 51, 219, 255, 196, 186, 128, 128, 128, 128, 128 },
      { 69, 46, 190, 239, 201, 218, 255, 228, 128, 128, 128 }
    },
    { { 1, 191, 251, 255, 255, 128, 128, 128, 128, 128, 128 },
      { 223, 165, 249, 255, 213, 255, 128, 128, 128, 128, 128 },
      { 141, 124, 248, 255, 255, 128, 128, 128, 128, 128, 128 }
    },
    { { 1, 16, 248, 255, 255, 128, 128, 128, 128, 128, 128 },
      { 190, 36, 230, 255, 236, 255, 128, 128, 128, 128, 128 },
      { 149, 1, 255, 128, 128, 128, 128, 128, 128, 128, 128 }
    },
    { { 1, 226, 255, 128, 128, 128, 128, 128, 128, 128, 128 },
      { 247, 192, 255, 128, 128, 128, 128, 128, 128, 128, 128 },
      { 240, 128, 255, 128, 128, 128, 128, 128, 128, 128, 128 }
    },
    { { 1, 134, 252, 255, 255, 128, 128, 128, 128, 128, 128 },
      { 213, 62, 250, 255, 255, 128, 128, 128, 128, 128, 128 },
      { 55, 93, 255, 128, 128, 128, 128, 128, 128, 128, 128 }
    },
    { { 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128 },
      { 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128 },
      { 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128 }
    }
  },
  { { { 202, 24, 213, 235, 186, 191, 220, 160, 240, 175, 255 },
      { 126, 38, 182, 232, 169, 184, 228, 174, 255, 187, 128 },
      { 61, 46, 138, 219, 151, 178, 240, 170, 255, 216, 128 }
    },
    { { 1, 112, 230, 250, 199, 191, 247, 159, 255, 255, 128 },
      { 166, 109, 228, 252, 211, 215, 255, 174, 128, 128, 128 },
      { 39, 77, 162, 232, 172, 180, 245, 178, 255, 255, 128 }
    },
    { { 1, 52, 220, 246, 198, 199, 249, 220, 255, 255, 128 },
      { 124, 74, 191, 243, 183, 193, 250, 221, 255, 255, 128 },
      { 24, 71, 130, 219, 154, 170, 243, 182, 255, 255, 128 }
    },
    { { 1, 182, 225, 249, 219, 240, 255, 224, 128, 128, 128 },
      { 149, 150, 226, 252, 216, 205, 255, 171, 128, 128, 128 },
      { 28, 108, 170, 242, 183, 194, 254, 223, 255, 255, 128 }
    },
    { { 1, 81, 230, 252, 204, 203, 255, 192, 128, 128, 128 },
      { 123, 102, 209, 247, 188, 196, 255, 233, 128, 128, 128 },
      { 20, 95, 153, 243, 164, 173, 255, 203, 128, 128, 128 }
    },
    { { 1, 222, 248, 255, 216, 213, 128, 128, 128, 128, 128 },
      { 168, 175, 246, 252, 235, 205, 255, 255, 128, 128, 128 },
      { 47, 116, 215, 255, 211, 212, 255, 255, 128, 128, 128 }
    },
    { { 1, 121, 236, 253, 212, 214, 255, 255, 128, 128, 128 },
      { 141, 84, 213, 252, 201, 202, 255, 219, 128, 128, 128 },
      { 42, 80, 160, 240, 162, 185, 255, 205, 128, 128, 128 }
    },
    { { 1, 1, 255, 128, 128, 128, 128, 128, 128, 128, 128 },
      { 244, 1, 255, 128, 128, 128, 128, 128, 128, 128, 128 },
      { 238, 1, 255, 128, 128, 128, 128, 128, 128, 128, 128 }
    }
  }
};

enum { B_DC_PRED = 0,
       B_TM_PRED,
       B_VE_PRED,
       B_HE_PRED,
       B_RD_PRED,
       B_VR_PRED,
       B_LD_PRED,
       B_VL_PRED,
       B_HD_PRED,
       B_HU_PRED,
       NUM_BMODES = B_HU_PRED + 1 - B_DC_PRED,


       DC_PRED = B_DC_PRED, V_PRED = B_VE_PRED,
       H_PRED = B_HE_PRED, TM_PRED = B_TM_PRED,
       B_PRED = NUM_BMODES,


       B_DC_PRED_NOTOP = 4,
       B_DC_PRED_NOLEFT = 5,
       B_DC_PRED_NOTOPLEFT = 6,
       NUM_B_DC_MODES = 7 };

static const int8_t kYModesIntra4[18] = {
  -B_DC_PRED, 1,
    -B_TM_PRED, 2,
      -B_VE_PRED, 3,
        4, 6,
          -B_HE_PRED, 5,
            -B_RD_PRED, -B_VR_PRED,
        -B_LD_PRED, 7,
          -B_VL_PRED, 8,
            -B_HD_PRED, -B_HU_PRED
};



static const int kQuantToDitherAmp[12] = {

  8, 7, 6, 4, 4, 2, 2, 2, 1, 1, 1, 1
};

typedef void (*__cc_VP8DecIdct2)(const int16_t* coeffs, uint8_t* dst, int do_two);

__cc_VP8DecIdct2 VP8Transform;



static inline int __cc_VP8RandomBits2(__cc_VP8Random* const rg, int num_bits,
                                      int amp) {
  int diff;
  ((num_bits + 8 <= 31) ? (void) (0) : __assert_fail ("num_bits + 8 <= 31", "extraction.1.c", 2288, __PRETTY_FUNCTION__));
  diff = rg->tab_[rg->index1_] - rg->tab_[rg->index2_];
  if (diff < 0) diff += (1u << 31);
  rg->tab_[rg->index1_] = diff;
  if (++rg->index1_ == 55) rg->index1_ = 0;
  if (++rg->index2_ == 55) rg->index2_ = 0;

  diff = (int)((uint32_t)diff << 1) >> (32 - num_bits);
  diff = (diff * amp) >> 8;
  diff += 1 << (num_bits - 1);
  return diff;
}
static void __cc_Dither8x8(__cc_VP8Random* const rg, uint8_t* dst, int bps, int amp) {
  int i, j;
  for (j = 0; j < 8; ++j) {
    for (i = 0; i < 8; ++i) {

      const int bits =
          __cc_VP8RandomBits2(rg, 8 + 1, amp) - (1 << 8);

      const int delta = (bits + (1 << (4 - 1))) >> 4;
      const int v = (int)dst[i] + delta;
      dst[i] = (v < 0) ? 0 : (v > 255) ? 255u : (uint8_t)v;
    }
    dst += bps;
  }
}



static void __cc_DitherRow(__cc_VP8Decoder* const dec) {
  int mb_x;
  ((dec->dither_) ? (void) (0) : __assert_fail ("dec->dither_", "extraction.1.c", 2329, __PRETTY_FUNCTION__));
  for (mb_x = dec->tl_mb_x_; mb_x < dec->br_mb_x_; ++mb_x) {
    const __cc_VP8ThreadContext* const ctx = &dec->thread_ctx_;
    const __cc_VP8MBData* const data = ctx->mb_data_ + mb_x;
    const int cache_id = ctx->id_;
    const int uv_bps = dec->cache_uv_stride_;
    if (data->dither_ >= 4) {
      uint8_t* const u_dst = dec->cache_u_ + cache_id * 8 * uv_bps + mb_x * 8;
      uint8_t* const v_dst = dec->cache_v_ + cache_id * 8 * uv_bps + mb_x * 8;
      __cc_Dither8x8(&dec->dithering_rg_, u_dst, uv_bps, data->dither_);
      __cc_Dither8x8(&dec->dithering_rg_, v_dst, uv_bps, data->dither_);
    }
  }
}



static const int kScan[16] = {
  0 + 0 * 32, 4 + 0 * 32, 8 + 0 * 32, 12 + 0 * 32,
  0 + 4 * 32, 4 + 4 * 32, 8 + 4 * 32, 12 + 4 * 32,
  0 + 8 * 32, 4 + 8 * 32, 8 + 8 * 32, 12 + 8 * 32,
  0 + 12 * 32, 4 + 12 * 32, 8 + 12 * 32, 12 + 12 * 32
};

void __cc_VP8InitScanline(__cc_VP8Decoder* const dec);

int __cc_VP8SetError(__cc_VP8Decoder* const dec,
                __cc_VP8StatusCode error, const char* const msg);

static const uint8_t kFilterExtraRows[3] = { 0, 2, 8 };





static int __cc_AllocateMemory(__cc_VP8Decoder* const dec) {
  const int num_caches = dec->num_caches_;
  const int mb_w = dec->mb_w_;

  const size_t intra_pred_mode_size = 4 * mb_w * sizeof(uint8_t);
  const size_t top_size = sizeof(__cc_VP8TopSamples) * mb_w;
  const size_t mb_info_size = (mb_w + 1) * sizeof(__cc_VP8MB);
  const size_t f_info_size =
      (dec->filter_type_ > 0) ?
          mb_w * (dec->mt_method_ > 0 ? 2 : 1) * sizeof(__cc_VP8FInfo)
        : 0;
  const size_t yuv_size = (32 * 17 + 32 * 9) * sizeof(*dec->yuv_b_);
  const size_t mb_data_size =
      (dec->mt_method_ == 2 ? 2 : 1) * mb_w * sizeof(*dec->mb_data_);
  const size_t cache_height = (16 * num_caches
                            + kFilterExtraRows[dec->filter_type_]) * 3 / 2;
  const size_t cache_size = top_size * cache_height;

  const uint64_t alpha_size = (dec->alpha_data_ != ((void *)0)) ?
      (uint64_t)dec->pic_hdr_.width_ * dec->pic_hdr_.height_ : 0ULL;
  const uint64_t needed = (uint64_t)intra_pred_mode_size
                        + top_size + mb_info_size + f_info_size
                        + yuv_size + mb_data_size
                        + cache_size + alpha_size + (32 - 1);
  uint8_t* mem;

  if (needed != (size_t)needed) return 0;
  if (needed > dec->mem_size_) {
    __cc_WebPSafeFree(dec->mem_);
    dec->mem_size_ = 0;
    dec->mem_ = __cc_WebPSafeMalloc(needed, sizeof(uint8_t));
    if (dec->mem_ == ((void *)0)) {
      return __cc_VP8SetError(dec, VP8_STATUS_OUT_OF_MEMORY,
                         "no memory during frame initialization.");
    }

    dec->mem_size_ = (size_t)needed;
  }

  mem = (uint8_t*)dec->mem_;
  dec->intra_t_ = (uint8_t*)mem;
  mem += intra_pred_mode_size;

  dec->yuv_t_ = (__cc_VP8TopSamples*)mem;
  mem += top_size;

  dec->mb_info_ = ((__cc_VP8MB*)mem) + 1;
  mem += mb_info_size;

  dec->f_info_ = f_info_size ? (__cc_VP8FInfo*)mem : ((void *)0);
  mem += f_info_size;
  dec->thread_ctx_.id_ = 0;
  dec->thread_ctx_.f_info_ = dec->f_info_;
  if (dec->mt_method_ > 0) {



    dec->thread_ctx_.f_info_ += mb_w;
  }

  mem = (uint8_t*)((uintptr_t)(mem + (32 - 1)) & ~(32 - 1));
  (((yuv_size & (32 - 1)) == 0) ? (void) (0) : __assert_fail ("(yuv_size & (32 - 1)) == 0", "extraction.1.c", 2425, __PRETTY_FUNCTION__));
  dec->yuv_b_ = (uint8_t*)mem;
  mem += yuv_size;

  dec->mb_data_ = (__cc_VP8MBData*)mem;
  dec->thread_ctx_.mb_data_ = (__cc_VP8MBData*)mem;
  if (dec->mt_method_ == 2) {
    dec->thread_ctx_.mb_data_ += mb_w;
  }
  mem += mb_data_size;

  dec->cache_y_stride_ = 16 * mb_w;
  dec->cache_uv_stride_ = 8 * mb_w;
  {
    const int extra_rows = kFilterExtraRows[dec->filter_type_];
    const int extra_y = extra_rows * dec->cache_y_stride_;
    const int extra_uv = (extra_rows / 2) * dec->cache_uv_stride_;
    dec->cache_y_ = ((uint8_t*)mem) + extra_y;
    dec->cache_u_ = dec->cache_y_
                  + 16 * num_caches * dec->cache_y_stride_ + extra_uv;
    dec->cache_v_ = dec->cache_u_
                  + 8 * num_caches * dec->cache_uv_stride_ + extra_uv;
    dec->cache_id_ = 0;
  }
  mem += cache_size;


  dec->alpha_plane_ = alpha_size ? (uint8_t*)mem : ((void *)0);
  mem += alpha_size;
  ((mem <= (uint8_t*)dec->mem_ + dec->mem_size_) ? (void) (0) : __assert_fail ("mem <= (uint8_t*)dec->mem_ + dec->mem_size_", "extraction.1.c", 2454, __PRETTY_FUNCTION__));


  memset(dec->mb_info_ - 1, 0, mb_info_size);
  __cc_VP8InitScanline(dec);


  memset(dec->intra_t_, B_DC_PRED, intra_pred_mode_size);

  return 1;
}



static inline void __cc_Put8x8uv(uint8_t value, uint8_t* dst) {
  int j;
  for (j = 0; j < 8; ++j) {
    memset(dst + j * 32, value, 8);
  }
}

static void __cc_InitIo(__cc_VP8Decoder* const dec, __cc_VP8Io* io) {

  io->mb_y = 0;
  io->y = dec->cache_y_;
  io->u = dec->cache_u_;
  io->v = dec->cache_v_;
  io->y_stride = dec->cache_y_stride_;
  io->uv_stride = dec->cache_uv_stride_;
  io->a = ((void *)0);
}

static void __cc_ReconstructRow(const __cc_VP8Decoder* const dec,
                           const __cc_VP8ThreadContext* ctx);

typedef void (*__cc_VP8LumaFilterFunc)(uint8_t* luma, int stride,
                                  int thresh, int ithresh, int hev_t);

extern __cc_VP8LumaFilterFunc VP8VFilter16i;

typedef void (*__cc_VP8ChromaFilterFunc)(uint8_t* u, uint8_t* v, int stride,
                                    int thresh, int ithresh, int hev_t);

extern __cc_VP8ChromaFilterFunc VP8VFilter8;

extern __cc_VP8ChromaFilterFunc VP8HFilter8i;

extern __cc_VP8ChromaFilterFunc VP8HFilter8;

typedef void (*__cc_VP8SimpleFilterFunc)(uint8_t* p, int stride, int thresh);

extern __cc_VP8SimpleFilterFunc VP8SimpleHFilter16i;

extern __cc_VP8LumaFilterFunc VP8VFilter16;

extern __cc_VP8LumaFilterFunc VP8HFilter16i;

extern __cc_VP8SimpleFilterFunc VP8SimpleHFilter16;

extern __cc_VP8ChromaFilterFunc VP8VFilter8i;

extern __cc_VP8SimpleFilterFunc VP8SimpleVFilter16;

extern __cc_VP8SimpleFilterFunc VP8SimpleVFilter16i;

extern __cc_VP8LumaFilterFunc VP8HFilter16;

static void __cc_DoFilter(const __cc_VP8Decoder* const dec, int mb_x, int mb_y) {
  const __cc_VP8ThreadContext* const ctx = &dec->thread_ctx_;
  const int cache_id = ctx->id_;
  const int y_bps = dec->cache_y_stride_;
  const __cc_VP8FInfo* const f_info = ctx->f_info_ + mb_x;
  uint8_t* const y_dst = dec->cache_y_ + cache_id * 16 * y_bps + mb_x * 16;
  const int ilevel = f_info->f_ilevel_;
  const int limit = f_info->f_limit_;
  if (limit == 0) {
    return;
  }
  ((limit >= 3) ? (void) (0) : __assert_fail ("limit >= 3", "extraction.1.c", 2532, __PRETTY_FUNCTION__));
  if (dec->filter_type_ == 1) {
    if (mb_x > 0) {
      VP8SimpleHFilter16(y_dst, y_bps, limit + 4);
    }
    if (f_info->f_inner_) {
      VP8SimpleHFilter16i(y_dst, y_bps, limit);
    }
    if (mb_y > 0) {
      VP8SimpleVFilter16(y_dst, y_bps, limit + 4);
    }
    if (f_info->f_inner_) {
      VP8SimpleVFilter16i(y_dst, y_bps, limit);
    }
  } else {
    const int uv_bps = dec->cache_uv_stride_;
    uint8_t* const u_dst = dec->cache_u_ + cache_id * 8 * uv_bps + mb_x * 8;
    uint8_t* const v_dst = dec->cache_v_ + cache_id * 8 * uv_bps + mb_x * 8;
    const int hev_thresh = f_info->hev_thresh_;
    if (mb_x > 0) {
      VP8HFilter16(y_dst, y_bps, limit + 4, ilevel, hev_thresh);
      VP8HFilter8(u_dst, v_dst, uv_bps, limit + 4, ilevel, hev_thresh);
    }
    if (f_info->f_inner_) {
      VP8HFilter16i(y_dst, y_bps, limit, ilevel, hev_thresh);
      VP8HFilter8i(u_dst, v_dst, uv_bps, limit, ilevel, hev_thresh);
    }
    if (mb_y > 0) {
      VP8VFilter16(y_dst, y_bps, limit + 4, ilevel, hev_thresh);
      VP8VFilter8(u_dst, v_dst, uv_bps, limit + 4, ilevel, hev_thresh);
    }
    if (f_info->f_inner_) {
      VP8VFilter16i(y_dst, y_bps, limit, ilevel, hev_thresh);
      VP8VFilter8i(u_dst, v_dst, uv_bps, limit, ilevel, hev_thresh);
    }
  }
}

static void __cc_FilterRow(const __cc_VP8Decoder* const dec) {
  int mb_x;
  const int mb_y = dec->thread_ctx_.mb_y_;
  ((dec->thread_ctx_.filter_row_) ? (void) (0) : __assert_fail ("dec->thread_ctx_.filter_row_", "extraction.1.c", 2573, __PRETTY_FUNCTION__));
  for (mb_x = dec->tl_mb_x_; mb_x < dec->br_mb_x_; ++mb_x) {
    __cc_DoFilter(dec, mb_x, mb_y);
  }
}

const uint8_t* __cc_VP8DecompressAlphaRows(__cc_VP8Decoder* const dec,
                                      int row, int num_rows);



static int __cc_FinishRow(__cc_VP8Decoder* const dec, __cc_VP8Io* const io) {
  int ok = 1;
  const __cc_VP8ThreadContext* const ctx = &dec->thread_ctx_;
  const int cache_id = ctx->id_;
  const int extra_y_rows = kFilterExtraRows[dec->filter_type_];
  const int ysize = extra_y_rows * dec->cache_y_stride_;
  const int uvsize = (extra_y_rows / 2) * dec->cache_uv_stride_;
  const int y_offset = cache_id * 16 * dec->cache_y_stride_;
  const int uv_offset = cache_id * 8 * dec->cache_uv_stride_;
  uint8_t* const ydst = dec->cache_y_ - ysize + y_offset;
  uint8_t* const udst = dec->cache_u_ - uvsize + uv_offset;
  uint8_t* const vdst = dec->cache_v_ - uvsize + uv_offset;
  const int mb_y = ctx->mb_y_;
  const int is_first_row = (mb_y == 0);
  const int is_last_row = (mb_y >= dec->br_mb_y_ - 1);

  if (dec->mt_method_ == 2) {
    __cc_ReconstructRow(dec, ctx);
  }

  if (ctx->filter_row_) {
    __cc_FilterRow(dec);
  }

  if (dec->dither_) {
    __cc_DitherRow(dec);
  }

  if (io->put != ((void *)0)) {
    int y_start = ((mb_y) * 16);
    int y_end = ((mb_y + 1) * 16);
    if (!is_first_row) {
      y_start -= extra_y_rows;
      io->y = ydst;
      io->u = udst;
      io->v = vdst;
    } else {
      io->y = dec->cache_y_ + y_offset;
      io->u = dec->cache_u_ + uv_offset;
      io->v = dec->cache_v_ + uv_offset;
    }

    if (!is_last_row) {
      y_end -= extra_y_rows;
    }
    if (y_end > io->crop_bottom) {
      y_end = io->crop_bottom;
    }
    io->a = ((void *)0);
    if (dec->alpha_data_ != ((void *)0) && y_start < y_end) {


      io->a = __cc_VP8DecompressAlphaRows(dec, y_start, y_end - y_start);
      if (io->a == ((void *)0)) {
        return __cc_VP8SetError(dec, VP8_STATUS_BITSTREAM_ERROR,
                           "Could not decode alpha data.");
      }
    }
    if (y_start < io->crop_top) {
      const int delta_y = io->crop_top - y_start;
      y_start = io->crop_top;
      ((!(delta_y & 1)) ? (void) (0) : __assert_fail ("!(delta_y & 1)", "extraction.1.c", 2645, __PRETTY_FUNCTION__));
      io->y += dec->cache_y_stride_ * delta_y;
      io->u += dec->cache_uv_stride_ * (delta_y >> 1);
      io->v += dec->cache_uv_stride_ * (delta_y >> 1);
      if (io->a != ((void *)0)) {
        io->a += io->width * delta_y;
      }
    }
    if (y_start < y_end) {
      io->y += io->crop_left;
      io->u += io->crop_left >> 1;
      io->v += io->crop_left >> 1;
      if (io->a != ((void *)0)) {
        io->a += io->crop_left;
      }
      io->mb_y = y_start - io->crop_top;
      io->mb_w = io->crop_right - io->crop_left;
      io->mb_h = y_end - y_start;
      ok = io->put(io);
    }
  }

  if (cache_id + 1 == dec->num_caches_) {
    if (!is_last_row) {
      memcpy(dec->cache_y_ - ysize, ydst + 16 * dec->cache_y_stride_, ysize);
      memcpy(dec->cache_u_ - uvsize, udst + 8 * dec->cache_uv_stride_, uvsize);
      memcpy(dec->cache_v_ - uvsize, vdst + 8 * dec->cache_uv_stride_, uvsize);
    }
  }

  return ok;
}



static int __cc_InitThreadContext(__cc_VP8Decoder* const dec) {
  dec->cache_id_ = 0;
  if (dec->mt_method_ > 0) {
    __cc_WebPWorker* const worker = &dec->worker_;
    if (!__cc_WebPGetWorkerInterface()->Reset(worker)) {
      return __cc_VP8SetError(dec, VP8_STATUS_OUT_OF_MEMORY,
                         "thread initialization failed.");
    }
    worker->data1 = dec;
    worker->data2 = (void*)&dec->thread_ctx_.io_;
    worker->hook = (__cc_WebPWorkerHook)__cc_FinishRow;
    dec->num_caches_ =
      (dec->filter_type_ > 0) ? 3 : 3 - 1;
  } else {
    dec->num_caches_ = 1;
  }
  return 1;
}

void __cc_VP8DspInit(void);

int __cc_VP8InitFrame(__cc_VP8Decoder* const dec, __cc_VP8Io* io) {
  if (!__cc_InitThreadContext(dec)) return 0;
  if (!__cc_AllocateMemory(dec)) return 0;
  __cc_InitIo(dec, io);
  __cc_VP8DspInit();
  return 1;
}



static int __cc_CheckMode(int mb_x, int mb_y, int mode) {
  if (mode == B_DC_PRED) {
    if (mb_x == 0) {
      return (mb_y == 0) ? B_DC_PRED_NOTOPLEFT : B_DC_PRED_NOLEFT;
    } else {
      return (mb_y == 0) ? B_DC_PRED_NOTOP : B_DC_PRED;
    }
  }
  return mode;
}

int __cc_VP8ProcessRow(__cc_VP8Decoder* const dec, __cc_VP8Io* const io) {
  int ok = 1;
  __cc_VP8ThreadContext* const ctx = &dec->thread_ctx_;
  const int filter_row =
      (dec->filter_type_ > 0) &&
      (dec->mb_y_ >= dec->tl_mb_y_) && (dec->mb_y_ <= dec->br_mb_y_);
  if (dec->mt_method_ == 0) {

    ctx->mb_y_ = dec->mb_y_;
    ctx->filter_row_ = filter_row;
    __cc_ReconstructRow(dec, ctx);
    ok = __cc_FinishRow(dec, io);
  } else {
    __cc_WebPWorker* const worker = &dec->worker_;

    ok &= __cc_WebPGetWorkerInterface()->Sync(worker);
    ((worker->status_ == OK) ? (void) (0) : __assert_fail ("worker->status_ == OK", "extraction.1.c", 2738, __PRETTY_FUNCTION__));
    if (ok) {
      ctx->io_ = *io;
      ctx->id_ = dec->cache_id_;
      ctx->mb_y_ = dec->mb_y_;
      ctx->filter_row_ = filter_row;
      if (dec->mt_method_ == 2) {
        __cc_VP8MBData* const tmp = ctx->mb_data_;
        ctx->mb_data_ = dec->mb_data_;
        dec->mb_data_ = tmp;
      } else {

        __cc_ReconstructRow(dec, ctx);
      }
      if (filter_row) {
        __cc_VP8FInfo* const tmp = ctx->f_info_;
        ctx->f_info_ = dec->f_info_;
        dec->f_info_ = tmp;
      }

      __cc_WebPGetWorkerInterface()->Launch(worker);
      if (++dec->cache_id_ == dec->num_caches_) {
        dec->cache_id_ = 0;
      }
    }
  }
  return ok;
}

int __cc_VP8ExitCritical(__cc_VP8Decoder* const dec, __cc_VP8Io* const io) {
  int ok = 1;
  if (dec->mt_method_ > 0) {
    ok = __cc_WebPGetWorkerInterface()->Sync(&dec->worker_);
  }

  if (io->teardown != ((void *)0)) {
    io->teardown(io);
  }
  return ok;
}

static void __cc_PrecomputeFilterStrengths(__cc_VP8Decoder* const dec) {
  if (dec->filter_type_ > 0) {
    int s;
    const __cc_VP8FilterHeader* const hdr = &dec->filter_hdr_;
    for (s = 0; s < NUM_MB_SEGMENTS; ++s) {
      int i4x4;

      int base_level;
      if (dec->segment_hdr_.use_segment_) {
        base_level = dec->segment_hdr_.filter_strength_[s];
        if (!dec->segment_hdr_.absolute_delta_) {
          base_level += hdr->level_;
        }
      } else {
        base_level = hdr->level_;
      }
      for (i4x4 = 0; i4x4 <= 1; ++i4x4) {
        __cc_VP8FInfo* const info = &dec->fstrengths_[s][i4x4];
        int level = base_level;
        if (hdr->use_lf_delta_) {

          level += hdr->ref_lf_delta_[0];
          if (i4x4) {
            level += hdr->mode_lf_delta_[0];
          }
        }
        level = (level < 0) ? 0 : (level > 63) ? 63 : level;
        if (level > 0) {
          int ilevel = level;
          if (hdr->sharpness_ > 0) {
            if (hdr->sharpness_ > 4) {
              ilevel >>= 2;
            } else {
              ilevel >>= 1;
            }
            if (ilevel > 9 - hdr->sharpness_) {
              ilevel = 9 - hdr->sharpness_;
            }
          }
          if (ilevel < 1) ilevel = 1;
          info->f_ilevel_ = ilevel;
          info->f_limit_ = 2 * level + ilevel;
          info->hev_thresh_ = (level >= 40) ? 2 : (level >= 15) ? 1 : 0;
        } else {
          info->f_limit_ = 0;
        }
        info->f_inner_ = i4x4;
      }
    }
  }
}

typedef void (*__cc_VP8DecIdct)(const int16_t* coeffs, uint8_t* dst);

extern __cc_VP8DecIdct VP8TransformDCUV;

extern __cc_VP8DecIdct VP8TransformUV;

static void __cc_DoUVTransform(uint32_t bits, const int16_t* const src,
                          uint8_t* const dst) {
  if (bits & 0xff) {
    if (bits & 0xaa) {
      VP8TransformUV(src, dst);
    } else {
      VP8TransformDCUV(src, dst);
    }
  }
}

typedef struct MetadataPayload {
  uint8_t* bytes;
  size_t size;
} __cc_MetadataPayload;

struct VP8LTransform;

void __cc_VP8LColorIndexInverseTransformAlpha(
    const struct VP8LTransform* const transform, int y_start, int y_end,
    const uint8_t* src, uint8_t* dst);

static void __cc_ApplyInverseTransformsAlpha(__cc_VP8LDecoder* const dec, int num_rows,
                                        const uint8_t* const rows) {
  const int start_row = dec->last_row_;
  const int end_row = start_row + num_rows;
  const uint8_t* rows_in = rows;
  uint8_t* rows_out = (uint8_t*)dec->io_->opaque + dec->io_->width * start_row;
  __cc_VP8LTransform* const transform = &dec->transforms_[0];
  ((dec->next_transform_ == 1) ? (void) (0) : __assert_fail ("dec->next_transform_ == 1", "extraction.1.c", 2866, __PRETTY_FUNCTION__));
  ((transform->type_ == COLOR_INDEXING_TRANSFORM) ? (void) (0) : __assert_fail ("transform->type_ == COLOR_INDEXING_TRANSFORM", "extraction.1.c", 2867, __PRETTY_FUNCTION__));
  __cc_VP8LColorIndexInverseTransformAlpha(transform, start_row, end_row, rows_in,
                                      rows_out);
}

typedef enum {
  GREEN = 0,
  RED = 1,
  BLUE = 2,
  ALPHA = 3,
  DIST = 4
} __cc_HuffIndex;

void __cc_VP8LConvertFromBGRA(const uint32_t* const in_data, int num_pixels,
                         __cc_WEBP_CSP_MODE out_colorspace, uint8_t* const rgba);

static int __cc_EmitRows(__cc_WEBP_CSP_MODE colorspace,
                    const uint8_t* row_in, int in_stride,
                    int mb_w, int mb_h,
                    uint8_t* const out, int out_stride) {
  int lines = mb_h;
  uint8_t* row_out = out;
  while (lines-- > 0) {
    __cc_VP8LConvertFromBGRA((const uint32_t*)row_in, mb_w, colorspace, row_out);
    row_in += in_stride;
    row_out += out_stride;
  }
  return mb_h;
}



static const uint8_t kCodeLengthCodeOrder[19] = {
  17, 18, 0, 1, 2, 3, 4, 5, 16, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15
};







static const uint16_t kAlphabetSize[5] = {
  256 + 24,
  256, 256, 256,
  40
};



static inline uint32_t __cc_VP8LSubSampleSize(uint32_t size,
                                              uint32_t sampling_bits) {
  return (size + (1 << sampling_bits) - 1) >> sampling_bits;
}

static void __cc_UpdateDecoder(__cc_VP8LDecoder* const dec, int width, int height) {
  __cc_VP8LMetadata* const hdr = &dec->hdr_;
  const int num_bits = hdr->huffman_subsample_bits_;
  dec->width_ = width;
  dec->height_ = height;

  hdr->huffman_xsize_ = __cc_VP8LSubSampleSize(width, num_bits);
  hdr->huffman_mask_ = (num_bits == 0) ? ~0 : (1 << num_bits) - 1;
}

static inline int __cc_HuffmanTreeNodeIsNotLeaf(
    const __cc_HuffmanTreeNode* const node) {
  return node->children_;
}

static inline const __cc_HuffmanTreeNode* __cc_HuffmanTreeNextNode(
    const __cc_HuffmanTreeNode* node, int right_child) {
  return node + node->children_ + right_child;
}

static inline void __cc_VP8LSetBitPos(__cc_VP8LBitReader* const br, int val) {
  br->bit_pos_ = val;
  br->eos_ = __cc_VP8LIsEndOfStream(br);
}

static inline uint32_t __cc_VP8LPrefetchBits(__cc_VP8LBitReader* const br) {
  return (uint32_t)(br->val_ >> br->bit_pos_);
}

static inline int __cc_ReadSymbol(const __cc_HuffmanTree* tree,
                                  __cc_VP8LBitReader* const br) {
  const __cc_HuffmanTreeNode* node = tree->root_;
  uint32_t bits = __cc_VP8LPrefetchBits(br);
  int bitpos = br->bit_pos_;

  const int lut_ix = bits & ((1U << 7) - 1);
  const int lut_bits = tree->lut_bits_[lut_ix];
  if (lut_bits <= 7) {
    __cc_VP8LSetBitPos(br, bitpos + lut_bits);
    return tree->lut_symbol_[lut_ix];
  }
  node += tree->lut_jump_[lut_ix];
  bitpos += 7;
  bits >>= 7;


  ((node != ((void *)0)) ? (void) (0) : __assert_fail ("node != ((void *)0)", "extraction.1.c", 2968, __PRETTY_FUNCTION__));
  do {
    node = __cc_HuffmanTreeNextNode(node, bits & 1);
    bits >>= 1;
    ++bitpos;
  } while (__cc_HuffmanTreeNodeIsNotLeaf(node));
  __cc_VP8LSetBitPos(br, bitpos);
  return node->symbol_;
}

static void __cc_SetOk(__cc_VP8Decoder* const dec) {
  dec->status_ = VP8_STATUS_OK;
  dec->error_msg_ = "OK";
}

__cc_VP8Decoder* __cc_VP8New(void) {
  __cc_VP8Decoder* const dec = (__cc_VP8Decoder*)__cc_WebPSafeCalloc(1ULL, sizeof(*dec));
  if (dec != ((void *)0)) {
    __cc_SetOk(dec);
    __cc_WebPGetWorkerInterface()->Init(&dec->worker_);
    dec->ready_ = 0;
    dec->num_parts_ = 1;
  }
  return dec;
}

static int __cc_ExpandColorMap(int num_colors, __cc_VP8LTransform* const transform) {
  int i;
  const int final_num_colors = 1 << (8 >> transform->bits_);
  uint32_t* const new_color_map =
      (uint32_t*)__cc_WebPSafeMalloc((uint64_t)final_num_colors,
                                sizeof(*new_color_map));
  if (new_color_map == ((void *)0)) {
    return 0;
  } else {
    uint8_t* const data = (uint8_t*)transform->data_;
    uint8_t* const new_data = (uint8_t*)new_color_map;
    new_color_map[0] = transform->data_[0];
    for (i = 4; i < 4 * num_colors; ++i) {

      new_data[i] = (data[i] + new_data[i - 4]) & 0xff;
    }
    for (; i < 4 * final_num_colors; ++i)
      new_data[i] = 0;
    __cc_WebPSafeFree(transform->data_);
    transform->data_ = new_color_map;
  }
  return 1;
}

int __cc_VP8CheckSignature(const uint8_t* const data, size_t data_size) {
  return (data_size >= 3 &&
          data[0] == 0x9d && data[1] == 0x01 && data[2] == 0x2a);
}



int __cc_VP8GetInfo(const uint8_t* data, size_t data_size, size_t chunk_size,
               int* const width, int* const height) {
  if (data == ((void *)0) || data_size < 10) {
    return 0;
  }

  if (!__cc_VP8CheckSignature(data + 3, data_size - 3)) {
    return 0;
  } else {
    const uint32_t bits = data[0] | (data[1] << 8) | (data[2] << 16);
    const int key_frame = !(bits & 1);
    const int w = ((data[7] << 8) | data[6]) & 0x3fff;
    const int h = ((data[9] << 8) | data[8]) & 0x3fff;

    if (!key_frame) {
      return 0;
    }

    if (((bits >> 1) & 7) > 3) {
      return 0;
    }
    if (!((bits >> 4) & 1)) {
      return 0;
    }
    if (((bits >> 5)) >= chunk_size) {
      return 0;
    }
    if (w == 0 || h == 0) {
      return 0;
    }

    if (width) {
      *width = w;
    }
    if (height) {
      *height = h;
    }

    return 1;
  }
}

extern void (*WebPMultARGBRow)(uint32_t* const ptr, int width, int inverse);

static int __cc_Export(__cc_WebPRescaler* const rescaler, __cc_WEBP_CSP_MODE colorspace,
                  int rgba_stride, uint8_t* const rgba) {
  uint32_t* const src = (uint32_t*)rescaler->dst;
  const int dst_width = rescaler->dst_width;
  int num_lines_out = 0;
  while (__cc_WebPRescalerHasPendingOutput(rescaler)) {
    uint8_t* const dst = rgba + num_lines_out * rgba_stride;
    __cc_WebPRescalerExportRow(rescaler);
    WebPMultARGBRow(src, dst_width, 1);
    __cc_VP8LConvertFromBGRA(src, dst_width, colorspace, dst);
    ++num_lines_out;
  }
  return num_lines_out;
}

int __cc_WebPRescaleNeededLines(const __cc_WebPRescaler* const rescaler,
                           int max_num_lines);

void __cc_WebPMultARGBRows(uint8_t* ptr, int stride, int width, int num_rows,
                      int inverse);

static int __cc_EmitRescaledRowsRGBA(const __cc_VP8LDecoder* const dec,
                                uint8_t* in, int in_stride, int mb_h,
                                uint8_t* const out, int out_stride) {
  const __cc_WEBP_CSP_MODE colorspace = dec->output_->colorspace;
  int num_lines_in = 0;
  int num_lines_out = 0;
  while (num_lines_in < mb_h) {
    uint8_t* const row_in = in + num_lines_in * in_stride;
    uint8_t* const row_out = out + num_lines_out * out_stride;
    const int lines_left = mb_h - num_lines_in;
    const int needed_lines = __cc_WebPRescaleNeededLines(dec->rescaler, lines_left);
    ((needed_lines > 0 && needed_lines <= lines_left) ? (void) (0) : __assert_fail ("needed_lines > 0 && needed_lines <= lines_left", "extraction.1.c", 3101, __PRETTY_FUNCTION__));
    __cc_WebPMultARGBRows(row_in, in_stride,
                     dec->rescaler->src_width, needed_lines, 0);
    __cc_WebPRescalerImport(dec->rescaler, lines_left, row_in, in_stride);
    num_lines_in += needed_lines;
    num_lines_out += __cc_Export(dec->rescaler, colorspace, out_stride, row_out);
  }
  return num_lines_out;
}

__cc_VP8LDecoder* __cc_VP8LNew(void);

typedef struct {
  const uint8_t* data;
  size_t data_size;
  int have_all_data;
  size_t offset;
  const uint8_t* alpha_data;
  size_t alpha_data_size;
  size_t compressed_size;
  size_t riff_size;
  int is_lossless;
} __cc_WebPHeaderStructure;

int __cc_VP8GetThreadMethod(const __cc_WebPDecoderOptions* const options,
                       const __cc_WebPHeaderStructure* const headers,
                       int width, int height);

__cc_VP8StatusCode __cc_WebPAllocateDecBuffer(int width, int height,
                                    const __cc_WebPDecoderOptions* const options,
                                    __cc_WebPDecBuffer* const buffer);

int __cc_VP8GetHeaders(__cc_VP8Decoder* const dec, __cc_VP8Io* const io);

void __cc_VP8Delete(__cc_VP8Decoder* const dec);

void __cc_VP8LDelete(__cc_VP8LDecoder* const dec);

__cc_VP8Decoder* __cc_VP8New(void);

extern void __cc_WebPFreeDecBuffer(__cc_WebPDecBuffer* buffer);

extern int __cc_VP8LGetInfo(
    const uint8_t* data, size_t data_size,
    int* const width, int* const height, int* const has_alpha);





static inline uint32_t __cc_get_le24(const uint8_t* const data) {
  return data[0] | (data[1] << 8) | (data[2] << 16);
}

static inline uint32_t __cc_get_le32(const uint8_t* const data) {
  return (uint32_t)__cc_get_le24(data) | (data[3] << 24);
}





static __cc_VP8StatusCode __cc_ParseOptionalChunks(const uint8_t** const data,
                                         size_t* const data_size,
                                         size_t const riff_size,
                                         const uint8_t** const alpha_data,
                                         size_t* const alpha_size) {
  const uint8_t* buf;
  size_t buf_size;
  uint32_t total_size = 4 +
                        8 +
                        10;
  ((data != ((void *)0)) ? (void) (0) : __assert_fail ("data != ((void *)0)", "extraction.1.c", 3173, __PRETTY_FUNCTION__));
  ((data_size != ((void *)0)) ? (void) (0) : __assert_fail ("data_size != ((void *)0)", "extraction.1.c", 3174, __PRETTY_FUNCTION__));
  buf = *data;
  buf_size = *data_size;

  ((alpha_data != ((void *)0)) ? (void) (0) : __assert_fail ("alpha_data != ((void *)0)", "extraction.1.c", 3178, __PRETTY_FUNCTION__));
  ((alpha_size != ((void *)0)) ? (void) (0) : __assert_fail ("alpha_size != ((void *)0)", "extraction.1.c", 3179, __PRETTY_FUNCTION__));
  *alpha_data = ((void *)0);
  *alpha_size = 0;

  while (1) {
    uint32_t chunk_size;
    uint32_t disk_chunk_size;

    *data = buf;
    *data_size = buf_size;

    if (buf_size < 8) {
      return VP8_STATUS_NOT_ENOUGH_DATA;
    }

    chunk_size = __cc_get_le32(buf + 4);
    if (chunk_size > (~0U - 8 - 1)) {
      return VP8_STATUS_BITSTREAM_ERROR;
    }

    disk_chunk_size = (8 + chunk_size + 1) & ~1;
    total_size += disk_chunk_size;


    if (riff_size > 0 && (total_size > riff_size)) {
      return VP8_STATUS_BITSTREAM_ERROR;
    }





    if (!memcmp(buf, "VP8 ", 4) ||
        !memcmp(buf, "VP8L", 4)) {
      return VP8_STATUS_OK;
    }

    if (buf_size < disk_chunk_size) {
      return VP8_STATUS_NOT_ENOUGH_DATA;
    }

    if (!memcmp(buf, "ALPH", 4)) {
      *alpha_data = buf + 8;
      *alpha_size = chunk_size;
    }


    buf += disk_chunk_size;
    buf_size -= disk_chunk_size;
  }
}



static __cc_VP8StatusCode __cc_ParseVP8X(const uint8_t** const data,
                               size_t* const data_size,
                               int* const found_vp8x,
                               int* const width_ptr, int* const height_ptr,
                               uint32_t* const flags_ptr) {
  const uint32_t vp8x_size = 8 + 10;
  ((data != ((void *)0)) ? (void) (0) : __assert_fail ("data != ((void *)0)", "extraction.1.c", 3239, __PRETTY_FUNCTION__));
  ((data_size != ((void *)0)) ? (void) (0) : __assert_fail ("data_size != ((void *)0)", "extraction.1.c", 3240, __PRETTY_FUNCTION__));
  ((found_vp8x != ((void *)0)) ? (void) (0) : __assert_fail ("found_vp8x != ((void *)0)", "extraction.1.c", 3241, __PRETTY_FUNCTION__));

  *found_vp8x = 0;

  if (*data_size < 8) {
    return VP8_STATUS_NOT_ENOUGH_DATA;
  }

  if (!memcmp(*data, "VP8X", 4)) {
    int width, height;
    uint32_t flags;
    const uint32_t chunk_size = __cc_get_le32(*data + 4);
    if (chunk_size != 10) {
      return VP8_STATUS_BITSTREAM_ERROR;
    }


    if (*data_size < vp8x_size) {
      return VP8_STATUS_NOT_ENOUGH_DATA;
    }
    flags = __cc_get_le32(*data + 8);
    width = 1 + __cc_get_le24(*data + 12);
    height = 1 + __cc_get_le24(*data + 15);
    if (width * (uint64_t)height >= (1ULL << 32)) {
      return VP8_STATUS_BITSTREAM_ERROR;
    }

    if (flags_ptr != ((void *)0)) *flags_ptr = flags;
    if (width_ptr != ((void *)0)) *width_ptr = width;
    if (height_ptr != ((void *)0)) *height_ptr = height;

    *data += vp8x_size;
    *data_size -= vp8x_size;
    *found_vp8x = 1;
  }
  return VP8_STATUS_OK;
}

extern int __cc_VP8GetInfo(
    const uint8_t* data,
    size_t data_size,
    size_t chunk_size,
    int* const width, int* const height);

extern int __cc_VP8LCheckSignature(const uint8_t* const data, size_t size);

static __cc_VP8StatusCode __cc_ParseVP8Header(const uint8_t** const data_ptr,
                                    size_t* const data_size, int have_all_data,
                                    size_t riff_size, size_t* const chunk_size,
                                    int* const is_lossless) {
  const uint8_t* const data = *data_ptr;
  const int is_vp8 = !memcmp(data, "VP8 ", 4);
  const int is_vp8l = !memcmp(data, "VP8L", 4);
  const uint32_t minimal_size =
      4 + 8;

  ((data != ((void *)0)) ? (void) (0) : __assert_fail ("data != ((void *)0)", "extraction.1.c", 3297, __PRETTY_FUNCTION__));
  ((data_size != ((void *)0)) ? (void) (0) : __assert_fail ("data_size != ((void *)0)", "extraction.1.c", 3298, __PRETTY_FUNCTION__));
  ((chunk_size != ((void *)0)) ? (void) (0) : __assert_fail ("chunk_size != ((void *)0)", "extraction.1.c", 3299, __PRETTY_FUNCTION__));
  ((is_lossless != ((void *)0)) ? (void) (0) : __assert_fail ("is_lossless != ((void *)0)", "extraction.1.c", 3300, __PRETTY_FUNCTION__));

  if (*data_size < 8) {
    return VP8_STATUS_NOT_ENOUGH_DATA;
  }

  if (is_vp8 || is_vp8l) {

    const uint32_t size = __cc_get_le32(data + 4);
    if ((riff_size >= minimal_size) && (size > riff_size - minimal_size)) {
      return VP8_STATUS_BITSTREAM_ERROR;
    }
    if (have_all_data && (size > *data_size - 8)) {
      return VP8_STATUS_NOT_ENOUGH_DATA;
    }

    *chunk_size = size;
    *data_ptr += 8;
    *data_size -= 8;
    *is_lossless = is_vp8l;
  } else {

    *is_lossless = __cc_VP8LCheckSignature(data, *data_size);
    *chunk_size = *data_size;
  }

  return VP8_STATUS_OK;
}



typedef enum WebPFeatureFlags {
  FRAGMENTS_FLAG = 0x00000001,
  ANIMATION_FLAG = 0x00000002,
  XMP_FLAG = 0x00000004,
  EXIF_FLAG = 0x00000008,
  ALPHA_FLAG = 0x00000010,
  ICCP_FLAG = 0x00000020
} __cc_WebPFeatureFlags;



static __cc_VP8StatusCode __cc_ParseRIFF(const uint8_t** const data,
                               size_t* const data_size, int have_all_data,
                               size_t* const riff_size) {
  ((data != ((void *)0)) ? (void) (0) : __assert_fail ("data != ((void *)0)", "extraction.1.c", 3345, __PRETTY_FUNCTION__));
  ((data_size != ((void *)0)) ? (void) (0) : __assert_fail ("data_size != ((void *)0)", "extraction.1.c", 3346, __PRETTY_FUNCTION__));
  ((riff_size != ((void *)0)) ? (void) (0) : __assert_fail ("riff_size != ((void *)0)", "extraction.1.c", 3347, __PRETTY_FUNCTION__));

  *riff_size = 0;
  if (*data_size >= 12 && !memcmp(*data, "RIFF", 4)) {
    if (memcmp(*data + 8, "WEBP", 4)) {
      return VP8_STATUS_BITSTREAM_ERROR;
    } else {
      const uint32_t size = __cc_get_le32(*data + 4);

      if (size < 4 + 8) {
        return VP8_STATUS_BITSTREAM_ERROR;
      }
      if (size > (~0U - 8 - 1)) {
        return VP8_STATUS_BITSTREAM_ERROR;
      }
      if (have_all_data && (size > *data_size - 8)) {
        return VP8_STATUS_NOT_ENOUGH_DATA;
      }

      *riff_size = size;
      *data += 12;
      *data_size -= 12;
    }
  }
  return VP8_STATUS_OK;
}

static __cc_VP8StatusCode __cc_ParseHeadersInternal(const uint8_t* data,
                                          size_t data_size,
                                          int* const width,
                                          int* const height,
                                          int* const has_alpha,
                                          int* const has_animation,
                                          int* const format,
                                          __cc_WebPHeaderStructure* const headers) {
  int canvas_width = 0;
  int canvas_height = 0;
  int image_width = 0;
  int image_height = 0;
  int found_riff = 0;
  int found_vp8x = 0;
  int animation_present = 0;
  int fragments_present = 0;
  const int have_all_data = (headers != ((void *)0)) ? headers->have_all_data : 0;

  __cc_VP8StatusCode status;
  __cc_WebPHeaderStructure hdrs;

  if (data == ((void *)0) || data_size < 12) {
    return VP8_STATUS_NOT_ENOUGH_DATA;
  }
  memset(&hdrs, 0, sizeof(hdrs));
  hdrs.data = data;
  hdrs.data_size = data_size;


  status = __cc_ParseRIFF(&data, &data_size, have_all_data, &hdrs.riff_size);
  if (status != VP8_STATUS_OK) {
    return status;
  }
  found_riff = (hdrs.riff_size > 0);


  {
    uint32_t flags = 0;
    status = __cc_ParseVP8X(&data, &data_size, &found_vp8x,
                       &canvas_width, &canvas_height, &flags);
    if (status != VP8_STATUS_OK) {
      return status;
    }
    animation_present = !!(flags & ANIMATION_FLAG);
    fragments_present = !!(flags & FRAGMENTS_FLAG);
    if (!found_riff && found_vp8x) {


      return VP8_STATUS_BITSTREAM_ERROR;
    }
    if (has_alpha != ((void *)0)) *has_alpha = !!(flags & ALPHA_FLAG);
    if (has_animation != ((void *)0)) *has_animation = animation_present;
    if (format != ((void *)0)) *format = 0;

    image_width = canvas_width;
    image_height = canvas_height;
    if (found_vp8x && (animation_present || fragments_present) &&
        headers == ((void *)0)) {
      status = VP8_STATUS_OK;
      goto ReturnWidthHeight;
    }
  }

  if (data_size < 4) {
    status = VP8_STATUS_NOT_ENOUGH_DATA;
    goto ReturnWidthHeight;
  }


  if ((found_riff && found_vp8x) ||
      (!found_riff && !found_vp8x && !memcmp(data, "ALPH", 4))) {
    status = __cc_ParseOptionalChunks(&data, &data_size, hdrs.riff_size,
                                 &hdrs.alpha_data, &hdrs.alpha_data_size);
    if (status != VP8_STATUS_OK) {
      goto ReturnWidthHeight;
    }
  }


  status = __cc_ParseVP8Header(&data, &data_size, have_all_data, hdrs.riff_size,
                          &hdrs.compressed_size, &hdrs.is_lossless);
  if (status != VP8_STATUS_OK) {
    goto ReturnWidthHeight;
  }
  if (hdrs.compressed_size > (~0U - 8 - 1)) {
    return VP8_STATUS_BITSTREAM_ERROR;
  }

  if (format != ((void *)0) && !(animation_present || fragments_present)) {
    *format = hdrs.is_lossless ? 2 : 1;
  }

  if (!hdrs.is_lossless) {
    if (data_size < 10) {
      status = VP8_STATUS_NOT_ENOUGH_DATA;
      goto ReturnWidthHeight;
    }

    if (!__cc_VP8GetInfo(data, data_size, (uint32_t)hdrs.compressed_size,
                    &image_width, &image_height)) {
      return VP8_STATUS_BITSTREAM_ERROR;
    }
  } else {
    if (data_size < 5) {
      status = VP8_STATUS_NOT_ENOUGH_DATA;
      goto ReturnWidthHeight;
    }

    if (!__cc_VP8LGetInfo(data, data_size, &image_width, &image_height, has_alpha)) {
      return VP8_STATUS_BITSTREAM_ERROR;
    }
  }

  if (found_vp8x) {
    if (canvas_width != image_width || canvas_height != image_height) {
      return VP8_STATUS_BITSTREAM_ERROR;
    }
  }
  if (headers != ((void *)0)) {
    *headers = hdrs;
    headers->offset = data - headers->data;
    (((uint64_t)(data - headers->data) < (~0U - 8 - 1)) ? (void) (0) : __assert_fail ("(uint64_t)(data - headers->data) < (~0U - 8 - 1)", "extraction.1.c", 3495, __PRETTY_FUNCTION__));
    ((headers->offset == headers->data_size - data_size) ? (void) (0) : __assert_fail ("headers->offset == headers->data_size - data_size", "extraction.1.c", 3496, __PRETTY_FUNCTION__));
  }
 ReturnWidthHeight:
  if (status == VP8_STATUS_OK ||
      (status == VP8_STATUS_NOT_ENOUGH_DATA && found_vp8x && headers == ((void *)0))) {
    if (has_alpha != ((void *)0)) {


      *has_alpha |= (hdrs.alpha_data != ((void *)0));
    }
    if (width != ((void *)0)) *width = image_width;
    if (height != ((void *)0)) *height = image_height;
    return VP8_STATUS_OK;
  } else {
    return status;
  }
}

__cc_VP8StatusCode __cc_WebPParseHeaders(__cc_WebPHeaderStructure* const headers) {
  __cc_VP8StatusCode status;
  int has_animation = 0;
  ((headers != ((void *)0)) ? (void) (0) : __assert_fail ("headers != ((void *)0)", "extraction.1.c", 3517, __PRETTY_FUNCTION__));

  status = __cc_ParseHeadersInternal(headers->data, headers->data_size,
                                ((void *)0), ((void *)0), ((void *)0), &has_animation,
                                ((void *)0), headers);
  if (status == VP8_STATUS_OK || status == VP8_STATUS_NOT_ENOUGH_DATA) {

    if (has_animation) {
      status = VP8_STATUS_UNSUPPORTED_FEATURE;
    }
  }
  return status;
}

void __cc_VP8InitDithering(const __cc_WebPDecoderOptions* const options,
                      __cc_VP8Decoder* const dec);

int __cc_VP8Decode(__cc_VP8Decoder* const dec, __cc_VP8Io* const io);

int __cc_VP8LDecodeImage(__cc_VP8LDecoder* const dec);

int __cc_VP8LDecodeHeader(__cc_VP8LDecoder* const dec, __cc_VP8Io* const io);

static __cc_VP8StatusCode __cc_DecodeInto(const uint8_t* const data, size_t data_size,
                                __cc_WebPDecParams* const params) {
  __cc_VP8StatusCode status;
  __cc_VP8Io io;
  __cc_WebPHeaderStructure headers;

  headers.data = data;
  headers.data_size = data_size;
  headers.have_all_data = 1;
  status = __cc_WebPParseHeaders(&headers);
  if (status != VP8_STATUS_OK) {
    return status;
  }

  ((params != ((void *)0)) ? (void) (0) : __assert_fail ("params != ((void *)0)", "extraction.1.c", 3554, __PRETTY_FUNCTION__));
  __cc_VP8InitIo(&io);
  io.data = headers.data + headers.offset;
  io.data_size = headers.data_size - headers.offset;
  __cc_WebPInitCustomIo(params, &io);

  if (!headers.is_lossless) {
    __cc_VP8Decoder* const dec = __cc_VP8New();
    if (dec == ((void *)0)) {
      return VP8_STATUS_OUT_OF_MEMORY;
    }
    dec->alpha_data_ = headers.alpha_data;
    dec->alpha_data_size_ = headers.alpha_data_size;


    if (!__cc_VP8GetHeaders(dec, &io)) {
      status = dec->status_;
    } else {

      status = __cc_WebPAllocateDecBuffer(io.width, io.height, params->options,
                                     params->output);
      if (status == VP8_STATUS_OK) {

        dec->mt_method_ = __cc_VP8GetThreadMethod(params->options, &headers,
                                             io.width, io.height);
        __cc_VP8InitDithering(params->options, dec);
        if (!__cc_VP8Decode(dec, &io)) {
          status = dec->status_;
        }
      }
    }
    __cc_VP8Delete(dec);
  } else {
    __cc_VP8LDecoder* const dec = __cc_VP8LNew();
    if (dec == ((void *)0)) {
      return VP8_STATUS_OUT_OF_MEMORY;
    }
    if (!__cc_VP8LDecodeHeader(dec, &io)) {
      status = dec->status_;
    } else {

      status = __cc_WebPAllocateDecBuffer(io.width, io.height, params->options,
                                     params->output);
      if (status == VP8_STATUS_OK) {
        if (!__cc_VP8LDecodeImage(dec)) {
          status = dec->status_;
        }
      }
    }
    __cc_VP8LDelete(dec);
  }

  if (status != VP8_STATUS_OK) {
    __cc_WebPFreeDecBuffer(params->output);
  }






  return status;
}

void __cc_WebPResetDecParams(__cc_WebPDecParams* const params) {
  if (params != ((void *)0)) {
    memset(params, 0, sizeof(*params));
  }
}

static const int kCodeLengthRepeatOffsets[3] = { 3, 3, 11 };

static void __cc_ClearTransform(__cc_VP8LTransform* const transform) {
  __cc_WebPSafeFree(transform->data_);
  transform->data_ = ((void *)0);
}

static void __cc_InitMetadata(__cc_VP8LMetadata* const hdr) {
  ((hdr) ? (void) (0) : __assert_fail ("hdr", "extraction.1.c", 3632, __PRETTY_FUNCTION__));
  memset(hdr, 0, sizeof(*hdr));
}

void __cc_VP8LColorCacheClear(__cc_VP8LColorCache* const color_cache);

void __cc_VP8LHtreeGroupsFree(__cc_HTreeGroup* htree_groups, int num_htree_groups);

static void __cc_ClearMetadata(__cc_VP8LMetadata* const hdr) {
  ((hdr) ? (void) (0) : __assert_fail ("hdr", "extraction.1.c", 3641, __PRETTY_FUNCTION__));

  __cc_WebPSafeFree(hdr->huffman_image_);
  __cc_VP8LHtreeGroupsFree(hdr->htree_groups_, hdr->num_htree_groups_);
  __cc_VP8LColorCacheClear(&hdr->color_cache_);
  __cc_InitMetadata(hdr);
}

void __cc_VP8LClear(__cc_VP8LDecoder* const dec) {
  int i;
  if (dec == ((void *)0)) return;
  __cc_ClearMetadata(&dec->hdr_);

  __cc_WebPSafeFree(dec->pixels_);
  dec->pixels_ = ((void *)0);
  for (i = 0; i < dec->next_transform_; ++i) {
    __cc_ClearTransform(&dec->transforms_[i]);
  }
  dec->next_transform_ = 0;
  dec->transforms_seen_ = 0;

  __cc_WebPSafeFree(dec->rescaler_memory);
  dec->rescaler_memory = ((void *)0);

  dec->output_ = ((void *)0);
}

void __cc_VP8LDelete(__cc_VP8LDecoder* const dec) {
  if (dec != ((void *)0)) {
    __cc_VP8LClear(dec);
    __cc_WebPSafeFree(dec);
  }
}

enum {
  YUV_FIX = 16,
  YUV_HALF = 1 << (YUV_FIX - 1),
  YUV_MASK = (256 << YUV_FIX) - 1,
  YUV_RANGE_MIN = -227,
  YUV_RANGE_MAX = 256 + 226,

  YUV_FIX2 = 14,
  YUV_HALF2 = 1 << (YUV_FIX2 - 1),
  YUV_MASK2 = (256 << YUV_FIX2) - 1
};

static inline int __cc_VP8ClipUV(int uv, int rounding) {
  uv = (uv + rounding + (128 << (YUV_FIX + 2))) >> (YUV_FIX + 2);
  return ((uv & ~0xff) == 0) ? uv : (uv < 0) ? 0 : 255;
}

static inline int __cc_VP8RGBToV(int r, int g, int b, int rounding) {
  const int v = +28800 * r - 24116 * g - 4684 * b;
  return __cc_VP8ClipUV(v, rounding);
}

static inline int __cc_VP8RGBToU(int r, int g, int b, int rounding) {
  const int u = -9719 * r - 19081 * g + 28800 * b;
  return __cc_VP8ClipUV(u, rounding);
}

static inline int __cc_VP8RGBToY(int r, int g, int b, int rounding) {
  const int luma = 16839 * r + 33059 * g + 6420 * b;
  return (luma + rounding + (16 << YUV_FIX)) >> YUV_FIX;
}

static void __cc_ConvertToYUVA(const uint32_t* const src, int width, int y_pos,
                          const __cc_WebPDecBuffer* const output) {
  const __cc_WebPYUVABuffer* const buf = &output->u.YUVA;

  {
    int i;
    uint8_t* const y = buf->y + y_pos * buf->y_stride;
    for (i = 0; i < width; ++i) {
      const uint32_t p = src[i];
      y[i] = __cc_VP8RGBToY((p >> 16) & 0xff, (p >> 8) & 0xff, (p >> 0) & 0xff,
                       YUV_HALF);
    }
  }


  {
    uint8_t* const u = buf->u + (y_pos >> 1) * buf->u_stride;
    uint8_t* const v = buf->v + (y_pos >> 1) * buf->v_stride;
    const int uv_width = width >> 1;
    int i;
    for (i = 0; i < uv_width; ++i) {
      const uint32_t v0 = src[2 * i + 0];
      const uint32_t v1 = src[2 * i + 1];


      const int r = ((v0 >> 15) & 0x1fe) + ((v1 >> 15) & 0x1fe);
      const int g = ((v0 >> 7) & 0x1fe) + ((v1 >> 7) & 0x1fe);
      const int b = ((v0 << 1) & 0x1fe) + ((v1 << 1) & 0x1fe);
      if (!(y_pos & 1)) {
        u[i] = __cc_VP8RGBToU(r, g, b, YUV_HALF << 2);
        v[i] = __cc_VP8RGBToV(r, g, b, YUV_HALF << 2);
      } else {
        const int tmp_u = __cc_VP8RGBToU(r, g, b, YUV_HALF << 2);
        const int tmp_v = __cc_VP8RGBToV(r, g, b, YUV_HALF << 2);

        u[i] = (u[i] + tmp_u + 1) >> 1;
        v[i] = (v[i] + tmp_v + 1) >> 1;
      }
    }
    if (width & 1) {
      const uint32_t v0 = src[2 * i + 0];
      const int r = (v0 >> 14) & 0x3fc;
      const int g = (v0 >> 6) & 0x3fc;
      const int b = (v0 << 2) & 0x3fc;
      if (!(y_pos & 1)) {
        u[i] = __cc_VP8RGBToU(r, g, b, YUV_HALF << 2);
        v[i] = __cc_VP8RGBToV(r, g, b, YUV_HALF << 2);
      } else {
        const int tmp_u = __cc_VP8RGBToU(r, g, b, YUV_HALF << 2);
        const int tmp_v = __cc_VP8RGBToV(r, g, b, YUV_HALF << 2);
        u[i] = (u[i] + tmp_u + 1) >> 1;
        v[i] = (v[i] + tmp_v + 1) >> 1;
      }
    }
  }

  if (buf->a != ((void *)0)) {
    int i;
    uint8_t* const a = buf->a + y_pos * buf->a_stride;
    for (i = 0; i < width; ++i) a[i] = (src[i] >> 24);
  }
}

static int __cc_ExportYUVA(const __cc_VP8LDecoder* const dec, int y_pos) {
  __cc_WebPRescaler* const rescaler = dec->rescaler;
  uint32_t* const src = (uint32_t*)rescaler->dst;
  const int dst_width = rescaler->dst_width;
  int num_lines_out = 0;
  while (__cc_WebPRescalerHasPendingOutput(rescaler)) {
    __cc_WebPRescalerExportRow(rescaler);
    WebPMultARGBRow(src, dst_width, 1);
    __cc_ConvertToYUVA(src, dst_width, y_pos, dec->output_);
    ++y_pos;
    ++num_lines_out;
  }
  return num_lines_out;
}

static int __cc_SetCropWindow(__cc_VP8Io* const io, int y_start, int y_end,
                         uint8_t** const in_data, int pixel_stride) {
  ((y_start < y_end) ? (void) (0) : __assert_fail ("y_start < y_end", "extraction.1.c", 3787, __PRETTY_FUNCTION__));
  ((io->crop_left < io->crop_right) ? (void) (0) : __assert_fail ("io->crop_left < io->crop_right", "extraction.1.c", 3788, __PRETTY_FUNCTION__));
  if (y_end > io->crop_bottom) {
    y_end = io->crop_bottom;
  }
  if (y_start < io->crop_top) {
    const int delta = io->crop_top - y_start;
    y_start = io->crop_top;
    *in_data += delta * pixel_stride;
  }
  if (y_start >= y_end) return 0;

  *in_data += io->crop_left * sizeof(uint32_t);

  io->mb_y = y_start - io->crop_top;
  io->mb_w = io->crop_right - io->crop_left;
  io->mb_h = y_end - y_start;
  return 1;
}

void __cc_VP8LInverseTransform(const struct VP8LTransform* const transform,
                          int row_start, int row_end,
                          const uint32_t* const in, uint32_t* const out);

static void __cc_ApplyInverseTransforms(__cc_VP8LDecoder* const dec, int num_rows,
                                   const uint32_t* const rows) {
  int n = dec->next_transform_;
  const int cache_pixs = dec->width_ * num_rows;
  const int start_row = dec->last_row_;
  const int end_row = start_row + num_rows;
  const uint32_t* rows_in = rows;
  uint32_t* const rows_out = dec->argb_cache_;



  memcpy(rows_out, rows_in, cache_pixs * sizeof(*rows_out));
  while (n-- > 0) {
    __cc_VP8LTransform* const transform = &dec->transforms_[n];
    __cc_VP8LInverseTransform(transform, start_row, end_row, rows_in, rows_out);
    rows_in = rows_out;
  }
}

static void __cc_ExtractAlphaRows(__cc_VP8LDecoder* const dec, int row) {
  const int num_rows = row - dec->last_row_;
  const uint32_t* const in = dec->pixels_ + dec->width_ * dec->last_row_;

  if (num_rows <= 0) return;
  __cc_ApplyInverseTransforms(dec, num_rows, in);


  {
    const int width = dec->io_->width;
    const int cache_pixs = width * num_rows;
    uint8_t* const dst = (uint8_t*)dec->io_->opaque + width * dec->last_row_;
    const uint32_t* const src = dec->argb_cache_;
    int i;
    for (i = 0; i < cache_pixs; ++i) dst[i] = (src[i] >> 8) & 0xff;
  }
  dec->last_row_ = dec->last_out_row_ = row;
}

void __cc_VP8LDspInit(void);

__cc_VP8LDecoder* __cc_VP8LNew(void) {
  __cc_VP8LDecoder* const dec = (__cc_VP8LDecoder*)__cc_WebPSafeCalloc(1ULL, sizeof(*dec));
  if (dec == ((void *)0)) return ((void *)0);
  dec->status_ = VP8_STATUS_OK;
  dec->action_ = READ_DIM;
  dec->state_ = READ_DIM;

  __cc_VP8LDspInit();

  return dec;
}

static int __cc_AllocateInternalBuffers32b(__cc_VP8LDecoder* const dec, int final_width) {
  const uint64_t num_pixels = (uint64_t)dec->width_ * dec->height_;


  const uint64_t cache_top_pixels = (uint16_t)final_width;

  const uint64_t cache_pixels = (uint64_t)final_width * 16;
  const uint64_t total_num_pixels =
      num_pixels + cache_top_pixels + cache_pixels;

  ((dec->width_ <= final_width) ? (void) (0) : __assert_fail ("dec->width_ <= final_width", "extraction.1.c", 3873, __PRETTY_FUNCTION__));
  dec->pixels_ = (uint32_t*)__cc_WebPSafeMalloc(total_num_pixels, sizeof(uint32_t));
  if (dec->pixels_ == ((void *)0)) {
    dec->argb_cache_ = ((void *)0);
    dec->status_ = VP8_STATUS_OUT_OF_MEMORY;
    return 0;
  }
  dec->argb_cache_ = dec->pixels_ + num_pixels + cache_top_pixels;
  return 1;
}



static const uint8_t kCodeToPlane[120] = {
  0x18, 0x07, 0x17, 0x19, 0x28, 0x06, 0x27, 0x29, 0x16, 0x1a,
  0x26, 0x2a, 0x38, 0x05, 0x37, 0x39, 0x15, 0x1b, 0x36, 0x3a,
  0x25, 0x2b, 0x48, 0x04, 0x47, 0x49, 0x14, 0x1c, 0x35, 0x3b,
  0x46, 0x4a, 0x24, 0x2c, 0x58, 0x45, 0x4b, 0x34, 0x3c, 0x03,
  0x57, 0x59, 0x13, 0x1d, 0x56, 0x5a, 0x23, 0x2d, 0x44, 0x4c,
  0x55, 0x5b, 0x33, 0x3d, 0x68, 0x02, 0x67, 0x69, 0x12, 0x1e,
  0x66, 0x6a, 0x22, 0x2e, 0x54, 0x5c, 0x43, 0x4d, 0x65, 0x6b,
  0x32, 0x3e, 0x78, 0x01, 0x77, 0x79, 0x53, 0x5d, 0x11, 0x1f,
  0x64, 0x6c, 0x42, 0x4e, 0x76, 0x7a, 0x21, 0x2f, 0x75, 0x7b,
  0x31, 0x3f, 0x63, 0x6d, 0x52, 0x5e, 0x00, 0x74, 0x7c, 0x41,
  0x4f, 0x10, 0x20, 0x62, 0x6e, 0x30, 0x73, 0x7d, 0x51, 0x5f,
  0x40, 0x72, 0x7e, 0x61, 0x6f, 0x50, 0x71, 0x7f, 0x60, 0x70
};

static inline int __cc_PlaneCodeToDistance(int xsize, int plane_code) {
  if (plane_code > 120) {
    return plane_code - 120;
  } else {
    const int dist_code = kCodeToPlane[plane_code - 1];
    const int yoffset = dist_code >> 4;
    const int xoffset = 8 - (dist_code & 0xf);
    const int dist = yoffset * xsize + xoffset;
    return (dist >= 1) ? dist : 1;
  }
}

uint32_t __cc_VP8LReadBits(__cc_VP8LBitReader* const br, int n_bits);

static inline int __cc_GetCopyDistance(int distance_symbol,
                                       __cc_VP8LBitReader* const br) {
  int extra_bits, offset;
  if (distance_symbol < 4) {
    return distance_symbol + 1;
  }
  extra_bits = (distance_symbol - 2) >> 1;
  offset = (2 + (distance_symbol & 1)) << extra_bits;
  return offset + __cc_VP8LReadBits(br, extra_bits) + 1;
}

static inline int __cc_GetCopyLength(int length_symbol,
                                     __cc_VP8LBitReader* const br) {

  return __cc_GetCopyDistance(length_symbol, br);
}

static int __cc_Is8bOptimizable(const __cc_VP8LMetadata* const hdr) {
  int i;
  if (hdr->color_cache_size_ > 0) return 0;


  for (i = 0; i < hdr->num_htree_groups_; ++i) {
    const __cc_HuffmanTree* const htrees = hdr->htree_groups_[i].htrees_;
    if (htrees[RED].num_nodes_ > 1) return 0;
    if (htrees[BLUE].num_nodes_ > 1) return 0;
    if (htrees[ALPHA].num_nodes_ > 1) return 0;
  }
  return 1;
}

static inline int __cc_GetMetaIndex(
    const uint32_t* const image, int xsize, int bits, int x, int y) {
  if (bits == 0) return 0;
  return image[xsize * (y >> bits) + (x >> bits)];
}

static inline __cc_HTreeGroup* __cc_GetHtreeGroupForPos(__cc_VP8LMetadata* const hdr,
                                                   int x, int y) {
  const int meta_index = __cc_GetMetaIndex(hdr->huffman_image_, hdr->huffman_xsize_,
                                      hdr->huffman_subsample_bits_, x, y);
  ((meta_index < hdr->num_htree_groups_) ? (void) (0) : __assert_fail ("meta_index < hdr->num_htree_groups_", "extraction.1.c", 3956, __PRETTY_FUNCTION__));
  return hdr->htree_groups_ + meta_index;
}

extern void __cc_VP8LDoFillBitWindow(__cc_VP8LBitReader* const br);



static inline void __cc_VP8LFillBitWindow(__cc_VP8LBitReader* const br) {
  if (br->bit_pos_ >= 32) __cc_VP8LDoFillBitWindow(br);
}

static void __cc_ExtractPalettedAlphaRows(__cc_VP8LDecoder* const dec, int row) {
  const int num_rows = row - dec->last_row_;
  const uint8_t* const in =
      (uint8_t*)dec->pixels_ + dec->width_ * dec->last_row_;
  if (num_rows > 0) {
    __cc_ApplyInverseTransformsAlpha(dec, num_rows, in);
  }
  dec->last_row_ = dec->last_out_row_ = row;
}

static int __cc_DecodeAlphaData(__cc_VP8LDecoder* const dec, uint8_t* const data,
                           int width, int height, int last_row) {
  int ok = 1;
  int row = dec->last_pixel_ / width;
  int col = dec->last_pixel_ % width;
  __cc_VP8LBitReader* const br = &dec->br_;
  __cc_VP8LMetadata* const hdr = &dec->hdr_;
  const __cc_HTreeGroup* htree_group = __cc_GetHtreeGroupForPos(hdr, col, row);
  int pos = dec->last_pixel_;
  const int end = width * height;
  const int last = width * last_row;
  const int len_code_limit = 256 + 24;
  const int mask = hdr->huffman_mask_;
  ((htree_group != ((void *)0)) ? (void) (0) : __assert_fail ("htree_group != ((void *)0)", "extraction.1.c", 3991, __PRETTY_FUNCTION__));
  ((pos < end) ? (void) (0) : __assert_fail ("pos < end", "extraction.1.c", 3992, __PRETTY_FUNCTION__));
  ((last_row <= height) ? (void) (0) : __assert_fail ("last_row <= height", "extraction.1.c", 3993, __PRETTY_FUNCTION__));
  ((__cc_Is8bOptimizable(hdr)) ? (void) (0) : __assert_fail ("Is8bOptimizable(hdr)", "extraction.1.c", 3994, __PRETTY_FUNCTION__));

  while (!br->eos_ && pos < last) {
    int code;

    if ((col & mask) == 0) {
      htree_group = __cc_GetHtreeGroupForPos(hdr, col, row);
    }
    __cc_VP8LFillBitWindow(br);
    code = __cc_ReadSymbol(&htree_group->htrees_[GREEN], br);
    if (code < 256) {
      data[pos] = code;
      ++pos;
      ++col;
      if (col >= width) {
        col = 0;
        ++row;
        if (row % 16 == 0) {
          __cc_ExtractPalettedAlphaRows(dec, row);
        }
      }
    } else if (code < len_code_limit) {
      int dist_code, dist;
      const int length_sym = code - 256;
      const int length = __cc_GetCopyLength(length_sym, br);
      const int dist_symbol = __cc_ReadSymbol(&htree_group->htrees_[DIST], br);
      __cc_VP8LFillBitWindow(br);
      dist_code = __cc_GetCopyDistance(dist_symbol, br);
      dist = __cc_PlaneCodeToDistance(width, dist_code);
      if (pos >= dist && end - pos >= length) {
        int i;
        for (i = 0; i < length; ++i) data[pos + i] = data[pos + i - dist];
      } else {
        ok = 0;
        goto End;
      }
      pos += length;
      col += length;
      while (col >= width) {
        col -= width;
        ++row;
        if (row % 16 == 0) {
          __cc_ExtractPalettedAlphaRows(dec, row);
        }
      }
      if (pos < last && (col & mask)) {
        htree_group = __cc_GetHtreeGroupForPos(hdr, col, row);
      }
    } else {
      ok = 0;
      goto End;
    }
    ((br->eos_ == __cc_VP8LIsEndOfStream(br)) ? (void) (0) : __assert_fail ("br->eos_ == VP8LIsEndOfStream(br)", "extraction.1.c", 4046, __PRETTY_FUNCTION__));
    ok = !br->error_;
    if (!ok) goto End;
  }

  __cc_ExtractPalettedAlphaRows(dec, row);

 End:
  if (br->error_ || !ok || (br->eos_ && pos < end)) {
    ok = 0;
    dec->status_ = br->eos_ ? VP8_STATUS_SUSPENDED
                            : VP8_STATUS_BITSTREAM_ERROR;
  } else {
    dec->last_pixel_ = (int)pos;
    if (pos == end) dec->state_ = READ_DATA;
  }
  return ok;
}

static int __cc_EmitRowsYUVA(const __cc_VP8LDecoder* const dec,
                        const uint8_t* in, int in_stride,
                        int mb_w, int num_rows) {
  int y_pos = dec->last_out_row_;
  while (num_rows-- > 0) {
    __cc_ConvertToYUVA((const uint32_t*)in, mb_w, y_pos, dec->output_);
    in += in_stride;
    ++y_pos;
  }
  return y_pos;
}

static int __cc_EmitRescaledRowsYUVA(const __cc_VP8LDecoder* const dec,
                                uint8_t* in, int in_stride, int mb_h) {
  int num_lines_in = 0;
  int y_pos = dec->last_out_row_;
  while (num_lines_in < mb_h) {
    const int lines_left = mb_h - num_lines_in;
    const int needed_lines = __cc_WebPRescaleNeededLines(dec->rescaler, lines_left);
    __cc_WebPMultARGBRows(in, in_stride, dec->rescaler->src_width, needed_lines, 0);
    __cc_WebPRescalerImport(dec->rescaler, lines_left, in, in_stride);
    num_lines_in += needed_lines;
    in += needed_lines * in_stride;
    y_pos += __cc_ExportYUVA(dec, y_pos);
  }
  return y_pos;
}

static void __cc_ProcessRows(__cc_VP8LDecoder* const dec, int row) {
  const uint32_t* const rows = dec->pixels_ + dec->width_ * dec->last_row_;
  const int num_rows = row - dec->last_row_;

  if (num_rows <= 0) return;
  __cc_ApplyInverseTransforms(dec, num_rows, rows);


  {
    __cc_VP8Io* const io = dec->io_;
    uint8_t* rows_data = (uint8_t*)dec->argb_cache_;
    const int in_stride = io->width * sizeof(uint32_t);
    if (!__cc_SetCropWindow(io, dec->last_row_, row, &rows_data, in_stride)) {

    } else {
      const __cc_WebPDecBuffer* const output = dec->output_;
      if (output->colorspace < MODE_YUV) {
        const __cc_WebPRGBABuffer* const buf = &output->u.RGBA;
        uint8_t* const rgba = buf->rgba + dec->last_out_row_ * buf->stride;
        const int num_rows_out = io->use_scaling ?
            __cc_EmitRescaledRowsRGBA(dec, rows_data, in_stride, io->mb_h,
                                 rgba, buf->stride) :
            __cc_EmitRows(output->colorspace, rows_data, in_stride,
                     io->mb_w, io->mb_h, rgba, buf->stride);

        dec->last_out_row_ += num_rows_out;
      } else {
        dec->last_out_row_ = io->use_scaling ?
            __cc_EmitRescaledRowsYUVA(dec, rows_data, in_stride, io->mb_h) :
            __cc_EmitRowsYUVA(dec, rows_data, in_stride, io->mb_w, io->mb_h);
      }
      ((dec->last_out_row_ <= output->height) ? (void) (0) : __assert_fail ("dec->last_out_row_ <= output->height", "extraction.1.c", 4124, __PRETTY_FUNCTION__));
    }
  }


  dec->last_row_ = row;
  ((dec->last_row_ <= dec->height_) ? (void) (0) : __assert_fail ("dec->last_row_ <= dec->height_", "extraction.1.c", 4130, __PRETTY_FUNCTION__));
}

static int __cc_DecodeImageStream(int xsize, int ysize,
                             int is_level0,
                             __cc_VP8LDecoder* const dec,
                             uint32_t** const decoded_data);

static int __cc_ReadTransform(int* const xsize, int const* ysize,
                         __cc_VP8LDecoder* const dec) {
  int ok = 1;
  __cc_VP8LBitReader* const br = &dec->br_;
  __cc_VP8LTransform* transform = &dec->transforms_[dec->next_transform_];
  const __cc_VP8LImageTransformType type =
      (__cc_VP8LImageTransformType)__cc_VP8LReadBits(br, 2);


  if (dec->transforms_seen_ & (1U << type)) {
    return 0;
  }
  dec->transforms_seen_ |= (1U << type);

  transform->type_ = type;
  transform->xsize_ = *xsize;
  transform->ysize_ = *ysize;
  transform->data_ = ((void *)0);
  ++dec->next_transform_;
  ((dec->next_transform_ <= 4) ? (void) (0) : __assert_fail ("dec->next_transform_ <= 4", "extraction.1.c", 4157, __PRETTY_FUNCTION__));

  switch (type) {
    case PREDICTOR_TRANSFORM:
    case CROSS_COLOR_TRANSFORM:
      transform->bits_ = __cc_VP8LReadBits(br, 3) + 2;
      ok = __cc_DecodeImageStream(__cc_VP8LSubSampleSize(transform->xsize_,
                                               transform->bits_),
                             __cc_VP8LSubSampleSize(transform->ysize_,
                                               transform->bits_),
                             0, dec, &transform->data_);
      break;
    case COLOR_INDEXING_TRANSFORM: {
       const int num_colors = __cc_VP8LReadBits(br, 8) + 1;
       const int bits = (num_colors > 16) ? 0
                      : (num_colors > 4) ? 1
                      : (num_colors > 2) ? 2
                      : 3;
       *xsize = __cc_VP8LSubSampleSize(transform->xsize_, bits);
       transform->bits_ = bits;
       ok = __cc_DecodeImageStream(num_colors, 1, 0, dec, &transform->data_);
       ok = ok && __cc_ExpandColorMap(num_colors, transform);
      break;
    }
    case SUBTRACT_GREEN:
      break;
    default:
      ((0) ? (void) (0) : __assert_fail ("0", "extraction.1.c", 4184, __PRETTY_FUNCTION__));
      break;
  }

  return ok;
}



static inline int __cc_VP8RandomBits(__cc_VP8Random* const rg, int num_bits) {
  return __cc_VP8RandomBits2(rg, num_bits, rg->amp_);
}

static int __cc_RGBToY(int r, int g, int b, __cc_VP8Random* const rg) {
  return (rg == ((void *)0)) ? __cc_VP8RGBToY(r, g, b, YUV_HALF)
                      : __cc_VP8RGBToY(r, g, b, __cc_VP8RandomBits(rg, YUV_FIX));
}

static inline void __cc_ConvertRowToY(const uint8_t* const r_ptr,
                                      const uint8_t* const g_ptr,
                                      const uint8_t* const b_ptr,
                                      int step,
                                      uint8_t* const dst_y,
                                      int width,
                                      __cc_VP8Random* const rg) {
  int i, j;
  for (i = 0, j = 0; i < width; ++i, j += step) {
    dst_y[i] = __cc_RGBToY(r_ptr[j], g_ptr[j], b_ptr[j], rg);
  }
}







static int kLinearToGammaTab[(1 << (12 - 7)) + 1];



static uint16_t kGammaToLinearTab[256];



static int kGammaTablesOk = 0;

static void __cc_InitGammaTables(void) {
  if (!kGammaTablesOk) {
    int v;
    const double scale = (double)(1 << 7) / ((1 << 12) - 1);
    const double norm = 1. / 255.;
    for (v = 0; v <= 255; ++v) {
      kGammaToLinearTab[v] =
          (uint16_t)(pow(norm * v, 0.80) * ((1 << 12) - 1) + .5);
    }
    for (v = 0; v <= (1 << (12 - 7)); ++v) {
      kLinearToGammaTab[v] = (int)(255. * pow(scale * v, 1. / 0.80) + .5);
    }
    kGammaTablesOk = 1;
  }
}

extern int (*WebPExtractAlpha)(const uint8_t* argb, int argb_stride,
                               int width, int height,
                               uint8_t* alpha, int alpha_stride);

static int __cc_RGBToGray(int r, int g, int b) {
  const int luma = 19595 * r + 38470 * g + 7471 * b + YUV_HALF;
  return (luma >> YUV_FIX);
}

typedef uint16_t __cc_fixed_y_t;





static __cc_fixed_y_t __cc_clip_y(int y) {
  return (!(y & ~((256 << 2) - 1))) ? (__cc_fixed_y_t)y : (y < 0) ? 0 : ((256 << 2) - 1);
}



static inline int __cc_FixedYToW(int v) {



  return v << (6 - 2);



}

typedef int16_t __cc_fixed_t;

static inline int __cc_Filter(const __cc_fixed_t* const A, const __cc_fixed_t* const B,
                              int rightwise) {
  int v;
  if (!rightwise) {
    v = (A[0] * 9 + A[-3] * 3 + B[0] * 3 + B[-3]);
  } else {
    v = (A[0] * 9 + A[+3] * 3 + B[0] * 3 + B[+3]);
  }
  return (v + 8) >> 4;
}

static inline int __cc_FixedWToY(int v) {





  return v >> (6 - 2);

}

static __cc_fixed_y_t __cc_clip_fixed_t(__cc_fixed_t v) {
  const int y = __cc_FixedWToY(v);
  const __cc_fixed_y_t w = __cc_clip_y(y);
  return w;
}

static inline int __cc_Filter2(int A, int B) { return (A * 3 + B + 2) >> 2; }

static void __cc_InterpolateTwoRows(const __cc_fixed_y_t* const best_y,
                               const __cc_fixed_t* const prev_uv,
                               const __cc_fixed_t* const cur_uv,
                               const __cc_fixed_t* const next_uv,
                               int w,
                               __cc_fixed_y_t* const out1,
                               __cc_fixed_y_t* const out2) {
  int i, k;
  {
    const int W0 = __cc_FixedYToW(best_y[0]);
    const int W1 = __cc_FixedYToW(best_y[w]);
    for (k = 0; k <= 2; ++k) {
      out1[k] = __cc_clip_fixed_t(__cc_Filter2(cur_uv[k], prev_uv[k]) + W0);
      out2[k] = __cc_clip_fixed_t(__cc_Filter2(cur_uv[k], next_uv[k]) + W1);
    }
  }
  for (i = 1; i < w - 1; ++i) {
    const int W0 = __cc_FixedYToW(best_y[i + 0]);
    const int W1 = __cc_FixedYToW(best_y[i + w]);
    const int off = 3 * (i >> 1);
    for (k = 0; k <= 2; ++k) {
      const int tmp0 = __cc_Filter(cur_uv + off + k, prev_uv + off + k, i & 1);
      const int tmp1 = __cc_Filter(cur_uv + off + k, next_uv + off + k, i & 1);
      out1[3 * i + k] = __cc_clip_fixed_t(tmp0 + W0);
      out2[3 * i + k] = __cc_clip_fixed_t(tmp1 + W1);
    }
  }
  {
    const int W0 = __cc_FixedYToW(best_y[i + 0]);
    const int W1 = __cc_FixedYToW(best_y[i + w]);
    const int off = 3 * (i >> 1);
    for (k = 0; k <= 2; ++k) {
      out1[3 * i + k] =
          __cc_clip_fixed_t(__cc_Filter2(cur_uv[off + k], prev_uv[off + k]) + W0);
      out2[3 * i + k] =
          __cc_clip_fixed_t(__cc_Filter2(cur_uv[off + k], next_uv[off + k]) + W1);
    }
  }
}

static float __cc_RGBToGrayF(float r, float g, float b) {
  return 0.299f * r + 0.587f * g + 0.114f * b;
}

static float kLinearToGammaTabF[(1 << (12 - 7)) + 2];

static inline float __cc_LinearToGammaF(float value) {
  const float v = value * (1 << (12 - 7));
  const int tab_pos = (int)v;
  const float x = v - (float)tab_pos;
  const float v0 = kLinearToGammaTabF[tab_pos + 0];
  const float v1 = kLinearToGammaTabF[tab_pos + 1];
  const float y = v1 * x + v0 * (1.f - x);
  return y;
}

static float kGammaToLinearTabF[((256 << 2) - 1) + 1];

static inline float __cc_GammaToLinearF(int v) {
  return kGammaToLinearTabF[v];
}

static float __cc_ScaleDown(int a, int b, int c, int d) {
  const float A = __cc_GammaToLinearF(a);
  const float B = __cc_GammaToLinearF(b);
  const float C = __cc_GammaToLinearF(c);
  const float D = __cc_GammaToLinearF(d);
  return __cc_LinearToGammaF(0.25f * (A + B + C + D));
}

static inline void __cc_UpdateChroma(const __cc_fixed_y_t* src1,
                                     const __cc_fixed_y_t* src2,
                                     __cc_fixed_t* dst, __cc_fixed_y_t* tmp, int len) {
  while (len--> 0) {
    const float r = __cc_ScaleDown(src1[0], src1[3], src2[0], src2[3]);
    const float g = __cc_ScaleDown(src1[1], src1[4], src2[1], src2[4]);
    const float b = __cc_ScaleDown(src1[2], src1[5], src2[2], src2[5]);
    const float W = __cc_RGBToGrayF(r, g, b);
    dst[0] = (__cc_fixed_t)__cc_FixedYToW((int)(r - W));
    dst[1] = (__cc_fixed_t)__cc_FixedYToW((int)(g - W));
    dst[2] = (__cc_fixed_t)__cc_FixedYToW((int)(b - W));
    dst += 3;
    src1 += 6;
    src2 += 6;
    if (tmp != ((void *)0)) {
      tmp[0] = tmp[1] = __cc_clip_y((int)(W + .5));
      tmp += 2;
    }
  }
}

static inline void __cc_UpdateW(const __cc_fixed_y_t* src, __cc_fixed_y_t* dst, int len) {
  while (len-- > 0) {
    const float R = __cc_GammaToLinearF(src[0]);
    const float G = __cc_GammaToLinearF(src[1]);
    const float B = __cc_GammaToLinearF(src[2]);
    const float Y = __cc_RGBToGrayF(R, G, B);
    *dst++ = (__cc_fixed_y_t)(__cc_LinearToGammaF(Y) + .5);
    src += 3;
  }
}

static const int kNumIterations = 6;

int __cc_WebPEncodingSetError(const __cc_WebPPicture* const pic, __cc_WebPEncodingError error);



static uint8_t __cc_clip_8b(__cc_fixed_t v) {
  return (!(v & ~0xff)) ? (uint8_t)v : (v < 0) ? 0u : 255u;
}



static inline uint8_t __cc_ConvertRGBToU(int r, int g, int b) {
  const int u = -9719 * r - 19081 * g + 28800 * b + (1 << (YUV_FIX + 6 - 1));
  return __cc_clip_8b(128 + (u >> (YUV_FIX + 6)));
}

static inline uint8_t __cc_ConvertRGBToY(int r, int g, int b) {
  const int luma = 16839 * r + 33059 * g + 6420 * b + (1 << (YUV_FIX + 6 - 1));
  return __cc_clip_8b(16 + (luma >> (YUV_FIX + 6)));
}

static inline uint8_t __cc_ConvertRGBToV(int r, int g, int b) {
  const int v = +28800 * r - 24116 * g - 4684 * b + (1 << (YUV_FIX + 6 - 1));
  return __cc_clip_8b(128 + (v >> (YUV_FIX + 6)));
}

static int __cc_ConvertWRGBToYUV(const __cc_fixed_y_t* const best_y,
                            const __cc_fixed_t* const best_uv,
                            __cc_WebPPicture* const picture) {
  int i, j;
  const int w = (picture->width + 1) & ~1;
  const int h = (picture->height + 1) & ~1;
  const int uv_w = w >> 1;
  const int uv_h = h >> 1;
  for (j = 0; j < picture->height; ++j) {
    for (i = 0; i < picture->width; ++i) {
      const int off = 3 * ((i >> 1) + (j >> 1) * uv_w);
      const int off2 = i + j * picture->y_stride;
      const int W = __cc_FixedYToW(best_y[i + j * w]);
      const int r = best_uv[off + 0] + W;
      const int g = best_uv[off + 1] + W;
      const int b = best_uv[off + 2] + W;
      picture->y[off2] = __cc_ConvertRGBToY(r, g, b);
    }
  }
  for (j = 0; j < uv_h; ++j) {
    uint8_t* const dst_u = picture->u + j * picture->uv_stride;
    uint8_t* const dst_v = picture->v + j * picture->uv_stride;
    for (i = 0; i < uv_w; ++i) {
      const int off = 3 * (i + j * uv_w);
      const int r = best_uv[off + 0];
      const int g = best_uv[off + 1];
      const int b = best_uv[off + 2];
      dst_u[i] = __cc_ConvertRGBToU(r, g, b);
      dst_v[i] = __cc_ConvertRGBToV(r, g, b);
    }
  }
  return 1;
}

static const int kMinDimensionIterativeConversion = 4;

static inline __cc_fixed_y_t __cc_UpLift(uint8_t a) {
  return ((__cc_fixed_y_t)a << 2) | (1 << (2 - 1));
}

static void __cc_ImportOneRow(const uint8_t* const r_ptr,
                         const uint8_t* const g_ptr,
                         const uint8_t* const b_ptr,
                         int step,
                         int pic_width,
                         __cc_fixed_y_t* const dst) {
  int i;
  for (i = 0; i < pic_width; ++i) {
    const int off = i * step;
    dst[3 * i + 0] = __cc_UpLift(r_ptr[off]);
    dst[3 * i + 1] = __cc_UpLift(g_ptr[off]);
    dst[3 * i + 2] = __cc_UpLift(b_ptr[off]);
  }
  if (pic_width & 1) {
    memcpy(dst + 3 * pic_width, dst + 3 * (pic_width - 1), 3 * sizeof(*dst));
  }
}

static int __cc_PreprocessARGB(const uint8_t* const r_ptr,
                          const uint8_t* const g_ptr,
                          const uint8_t* const b_ptr,
                          int step, int rgb_stride,
                          __cc_WebPPicture* const picture) {

  const int w = (picture->width + 1) & ~1;
  const int h = (picture->height + 1) & ~1;
  const int uv_w = w >> 1;
  const int uv_h = h >> 1;
  int i, j, iter;



  __cc_fixed_y_t* const tmp_buffer = ((__cc_fixed_y_t*)__cc_WebPSafeMalloc((w * 3) * (2), sizeof(__cc_fixed_y_t)));
  __cc_fixed_y_t* const best_y = ((__cc_fixed_y_t*)__cc_WebPSafeMalloc((w) * (h), sizeof(__cc_fixed_y_t)));
  __cc_fixed_y_t* const target_y = ((__cc_fixed_y_t*)__cc_WebPSafeMalloc((w) * (h), sizeof(__cc_fixed_y_t)));
  __cc_fixed_y_t* const best_rgb_y = ((__cc_fixed_y_t*)__cc_WebPSafeMalloc((w) * (2), sizeof(__cc_fixed_y_t)));
  __cc_fixed_t* const best_uv = ((__cc_fixed_t*)__cc_WebPSafeMalloc((uv_w * 3) * (uv_h), sizeof(__cc_fixed_t)));
  __cc_fixed_t* const target_uv = ((__cc_fixed_t*)__cc_WebPSafeMalloc((uv_w * 3) * (uv_h), sizeof(__cc_fixed_t)));
  __cc_fixed_t* const best_rgb_uv = ((__cc_fixed_t*)__cc_WebPSafeMalloc((uv_w * 3) * (1), sizeof(__cc_fixed_t)));
  int ok;

  if (best_y == ((void *)0) || best_uv == ((void *)0) ||
      target_y == ((void *)0) || target_uv == ((void *)0) ||
      best_rgb_y == ((void *)0) || best_rgb_uv == ((void *)0) ||
      tmp_buffer == ((void *)0)) {
    ok = __cc_WebPEncodingSetError(picture, VP8_ENC_ERROR_OUT_OF_MEMORY);
    goto End;
  }
  ((picture->width >= kMinDimensionIterativeConversion) ? (void) (0) : __assert_fail ("picture->width >= kMinDimensionIterativeConversion", "extraction.1.c", 4526, __PRETTY_FUNCTION__));
  ((picture->height >= kMinDimensionIterativeConversion) ? (void) (0) : __assert_fail ("picture->height >= kMinDimensionIterativeConversion", "extraction.1.c", 4527, __PRETTY_FUNCTION__));


  for (j = 0; j < picture->height; j += 2) {
    const int is_last_row = (j == picture->height - 1);
    __cc_fixed_y_t* const src1 = tmp_buffer;
    __cc_fixed_y_t* const src2 = tmp_buffer + 3 * w;
    const int off1 = j * rgb_stride;
    const int off2 = off1 + rgb_stride;
    const int uv_off = (j >> 1) * 3 * uv_w;
    __cc_fixed_y_t* const dst_y = best_y + j * w;


    __cc_ImportOneRow(r_ptr + off1, g_ptr + off1, b_ptr + off1,
                 step, picture->width, src1);
    if (!is_last_row) {
      __cc_ImportOneRow(r_ptr + off2, g_ptr + off2, b_ptr + off2,
                   step, picture->width, src2);
    } else {
      memcpy(src2, src1, 3 * w * sizeof(*src2));
    }
    __cc_UpdateW(src1, target_y + (j + 0) * w, w);
    __cc_UpdateW(src2, target_y + (j + 1) * w, w);
    __cc_UpdateChroma(src1, src2, target_uv + uv_off, dst_y, uv_w);
    memcpy(best_uv + uv_off, target_uv + uv_off, 3 * uv_w * sizeof(*best_uv));
    memcpy(dst_y + w, dst_y, w * sizeof(*dst_y));
  }


  for (iter = 0; iter < kNumIterations; ++iter) {
    int k;
    const __cc_fixed_t* cur_uv = best_uv;
    const __cc_fixed_t* prev_uv = best_uv;
    for (j = 0; j < h; j += 2) {
      __cc_fixed_y_t* const src1 = tmp_buffer;
      __cc_fixed_y_t* const src2 = tmp_buffer + 3 * w;

      {
        const __cc_fixed_t* const next_uv = cur_uv + ((j < h - 2) ? 3 * uv_w : 0);
        __cc_InterpolateTwoRows(best_y + j * w, prev_uv, cur_uv, next_uv,
                           w, src1, src2);
        prev_uv = cur_uv;
        cur_uv = next_uv;
      }

      __cc_UpdateW(src1, best_rgb_y + 0 * w, w);
      __cc_UpdateW(src2, best_rgb_y + 1 * w, w);
      __cc_UpdateChroma(src1, src2, best_rgb_uv, ((void *)0), uv_w);


      for (i = 0; i < 2 * w; ++i) {
        const int off = i + j * w;
        const int diff_y = target_y[off] - best_rgb_y[i];
        const int new_y = (int)best_y[off] + diff_y;
        best_y[off] = __cc_clip_y(new_y);
      }
      for (i = 0; i < uv_w; ++i) {
        const int off = 3 * (i + (j >> 1) * uv_w);
        int W;
        for (k = 0; k <= 2; ++k) {
          const int diff_uv = (int)target_uv[off + k] - best_rgb_uv[3 * i + k];
          best_uv[off + k] += diff_uv;
        }
        W = __cc_RGBToGray(best_uv[off + 0], best_uv[off + 1], best_uv[off + 2]);
        for (k = 0; k <= 2; ++k) {
          best_uv[off + k] -= W;
        }
      }
    }

  }


  ok = __cc_ConvertWRGBToYUV(best_y, best_uv, picture);

 End:
  __cc_WebPSafeFree(best_y);
  __cc_WebPSafeFree(best_uv);
  __cc_WebPSafeFree(target_y);
  __cc_WebPSafeFree(target_uv);
  __cc_WebPSafeFree(best_rgb_y);
  __cc_WebPSafeFree(best_rgb_uv);
  __cc_WebPSafeFree(tmp_buffer);
  return ok;
}

int __cc_WebPPictureAllocYUVA(__cc_WebPPicture* const picture, int width, int height);

static int __cc_RGBToV(int r, int g, int b, __cc_VP8Random* const rg) {
  return (rg == ((void *)0)) ? __cc_VP8RGBToV(r, g, b, YUV_HALF << 2)
                      : __cc_VP8RGBToV(r, g, b, __cc_VP8RandomBits(rg, YUV_FIX + 2));
}

static inline uint32_t __cc_GammaToLinear(uint8_t v) {
  return kGammaToLinearTab[v];
}

static int __cc_RGBToU(int r, int g, int b, __cc_VP8Random* const rg) {
  return (rg == ((void *)0)) ? __cc_VP8RGBToU(r, g, b, YUV_HALF << 2)
                      : __cc_VP8RGBToU(r, g, b, __cc_VP8RandomBits(rg, YUV_FIX + 2));
}





static inline int __cc_Interpolate(int v) {
  const int tab_pos = v >> (7 + 2);
  const int x = v & (((1 << 7) << 2) - 1);
  const int v0 = kLinearToGammaTab[tab_pos];
  const int v1 = kLinearToGammaTab[tab_pos + 1];
  const int y = v1 * x + v0 * (((1 << 7) << 2) - x);
  ((tab_pos + 1 < (1 << (12 - 7)) + 1) ? (void) (0) : __assert_fail ("tab_pos + 1 < (1 << (12 - 7)) + 1", "extraction.1.c", 4639, __PRETTY_FUNCTION__));
  return y;
}

static inline int __cc_LinearToGamma(uint32_t base_value, int shift) {
  const int y = __cc_Interpolate(base_value << shift);
  return (y + ((1 << 7) >> 1)) >> 7;
}
static inline void __cc_ConvertRowsToUV(const uint8_t* const r_ptr,
                                        const uint8_t* const g_ptr,
                                        const uint8_t* const b_ptr,
                                        int step, int rgb_stride,
                                        uint8_t* const dst_u,
                                        uint8_t* const dst_v,
                                        int width,
                                        __cc_VP8Random* const rg) {
  int i, j;
  for (i = 0, j = 0; i < (width >> 1); ++i, j += 2 * step) {
    const int r = __cc_LinearToGamma( __cc_GammaToLinear((r_ptr + j)[0]) + __cc_GammaToLinear((r_ptr + j)[(step)]) + __cc_GammaToLinear((r_ptr + j)[rgb_stride]) + __cc_GammaToLinear((r_ptr + j)[rgb_stride + (step)]), 0);
    const int g = __cc_LinearToGamma( __cc_GammaToLinear((g_ptr + j)[0]) + __cc_GammaToLinear((g_ptr + j)[(step)]) + __cc_GammaToLinear((g_ptr + j)[rgb_stride]) + __cc_GammaToLinear((g_ptr + j)[rgb_stride + (step)]), 0);
    const int b = __cc_LinearToGamma( __cc_GammaToLinear((b_ptr + j)[0]) + __cc_GammaToLinear((b_ptr + j)[(step)]) + __cc_GammaToLinear((b_ptr + j)[rgb_stride]) + __cc_GammaToLinear((b_ptr + j)[rgb_stride + (step)]), 0);
    dst_u[i] = __cc_RGBToU(r, g, b, rg);
    dst_v[i] = __cc_RGBToV(r, g, b, rg);
  }
  if (width & 1) {
    const int r = __cc_LinearToGamma(__cc_GammaToLinear((r_ptr + j)[0]) + __cc_GammaToLinear((r_ptr + j)[rgb_stride]), 1);
    const int g = __cc_LinearToGamma(__cc_GammaToLinear((g_ptr + j)[0]) + __cc_GammaToLinear((g_ptr + j)[rgb_stride]), 1);
    const int b = __cc_LinearToGamma(__cc_GammaToLinear((b_ptr + j)[0]) + __cc_GammaToLinear((b_ptr + j)[rgb_stride]), 1);
    dst_u[i] = __cc_RGBToU(r, g, b, rg);
    dst_v[i] = __cc_RGBToV(r, g, b, rg);
  }
}

static int __cc_CheckNonOpaque(const uint8_t* alpha, int width, int height,
                          int x_step, int y_step) {
  if (alpha == ((void *)0)) return 0;
  while (height-- > 0) {
    int x;
    for (x = 0; x < width * x_step; x += x_step) {
      if (alpha[x] != 0xff) return 1;
    }
    alpha += y_step;
  }
  return 0;
}

static const int kAlphaFix = 19;

static int kGammaTablesFOk = 0;



static void __cc_InitGammaTablesF(void) {
  if (!kGammaTablesFOk) {
    int v;
    const double norm = 1. / ((256 << 2) - 1);
    const double scale = 1. / (1 << (12 - 7));
    for (v = 0; v <= ((256 << 2) - 1); ++v) {
      kGammaToLinearTabF[v] = (float)pow(norm * v, 2.2);
    }
    for (v = 0; v <= (1 << (12 - 7)); ++v) {
      kLinearToGammaTabF[v] = (float)(((256 << 2) - 1) * pow(scale * v, 1. / 2.2));
    }

    kLinearToGammaTabF[(1 << (12 - 7)) + 1] = kLinearToGammaTabF[(1 << (12 - 7))];
    kGammaTablesFOk = 1;
  }
}

static const uint32_t kInvAlpha[4 * 0xff + 1] = {
  0,
  524288, 262144, 174762, 131072, 104857, 87381, 74898, 65536,
  58254, 52428, 47662, 43690, 40329, 37449, 34952, 32768,
  30840, 29127, 27594, 26214, 24966, 23831, 22795, 21845,
  20971, 20164, 19418, 18724, 18078, 17476, 16912, 16384,
  15887, 15420, 14979, 14563, 14169, 13797, 13443, 13107,
  12787, 12483, 12192, 11915, 11650, 11397, 11155, 10922,
  10699, 10485, 10280, 10082, 9892, 9709, 9532, 9362,
  9198, 9039, 8886, 8738, 8594, 8456, 8322, 8192,
  8065, 7943, 7825, 7710, 7598, 7489, 7384, 7281,
  7182, 7084, 6990, 6898, 6808, 6721, 6636, 6553,
  6472, 6393, 6316, 6241, 6168, 6096, 6026, 5957,
  5890, 5825, 5761, 5698, 5637, 5577, 5518, 5461,
  5405, 5349, 5295, 5242, 5190, 5140, 5090, 5041,
  4993, 4946, 4899, 4854, 4809, 4766, 4723, 4681,
  4639, 4599, 4559, 4519, 4481, 4443, 4405, 4369,
  4332, 4297, 4262, 4228, 4194, 4161, 4128, 4096,
  4064, 4032, 4002, 3971, 3942, 3912, 3883, 3855,
  3826, 3799, 3771, 3744, 3718, 3692, 3666, 3640,
  3615, 3591, 3566, 3542, 3518, 3495, 3472, 3449,
  3426, 3404, 3382, 3360, 3339, 3318, 3297, 3276,
  3256, 3236, 3216, 3196, 3177, 3158, 3139, 3120,
  3102, 3084, 3066, 3048, 3030, 3013, 2995, 2978,
  2962, 2945, 2928, 2912, 2896, 2880, 2864, 2849,
  2833, 2818, 2803, 2788, 2774, 2759, 2744, 2730,
  2716, 2702, 2688, 2674, 2661, 2647, 2634, 2621,
  2608, 2595, 2582, 2570, 2557, 2545, 2532, 2520,
  2508, 2496, 2484, 2473, 2461, 2449, 2438, 2427,
  2416, 2404, 2394, 2383, 2372, 2361, 2351, 2340,
  2330, 2319, 2309, 2299, 2289, 2279, 2269, 2259,
  2250, 2240, 2231, 2221, 2212, 2202, 2193, 2184,
  2175, 2166, 2157, 2148, 2139, 2131, 2122, 2114,
  2105, 2097, 2088, 2080, 2072, 2064, 2056, 2048,
  2040, 2032, 2024, 2016, 2008, 2001, 1993, 1985,
  1978, 1971, 1963, 1956, 1949, 1941, 1934, 1927,
  1920, 1913, 1906, 1899, 1892, 1885, 1879, 1872,
  1865, 1859, 1852, 1846, 1839, 1833, 1826, 1820,
  1814, 1807, 1801, 1795, 1789, 1783, 1777, 1771,
  1765, 1759, 1753, 1747, 1741, 1736, 1730, 1724,
  1718, 1713, 1707, 1702, 1696, 1691, 1685, 1680,
  1675, 1669, 1664, 1659, 1653, 1648, 1643, 1638,
  1633, 1628, 1623, 1618, 1613, 1608, 1603, 1598,
  1593, 1588, 1583, 1579, 1574, 1569, 1565, 1560,
  1555, 1551, 1546, 1542, 1537, 1533, 1528, 1524,
  1519, 1515, 1510, 1506, 1502, 1497, 1493, 1489,
  1485, 1481, 1476, 1472, 1468, 1464, 1460, 1456,
  1452, 1448, 1444, 1440, 1436, 1432, 1428, 1424,
  1420, 1416, 1413, 1409, 1405, 1401, 1398, 1394,
  1390, 1387, 1383, 1379, 1376, 1372, 1368, 1365,
  1361, 1358, 1354, 1351, 1347, 1344, 1340, 1337,
  1334, 1330, 1327, 1323, 1320, 1317, 1314, 1310,
  1307, 1304, 1300, 1297, 1294, 1291, 1288, 1285,
  1281, 1278, 1275, 1272, 1269, 1266, 1263, 1260,
  1257, 1254, 1251, 1248, 1245, 1242, 1239, 1236,
  1233, 1230, 1227, 1224, 1222, 1219, 1216, 1213,
  1210, 1208, 1205, 1202, 1199, 1197, 1194, 1191,
  1188, 1186, 1183, 1180, 1178, 1175, 1172, 1170,
  1167, 1165, 1162, 1159, 1157, 1154, 1152, 1149,
  1147, 1144, 1142, 1139, 1137, 1134, 1132, 1129,
  1127, 1125, 1122, 1120, 1117, 1115, 1113, 1110,
  1108, 1106, 1103, 1101, 1099, 1096, 1094, 1092,
  1089, 1087, 1085, 1083, 1081, 1078, 1076, 1074,
  1072, 1069, 1067, 1065, 1063, 1061, 1059, 1057,
  1054, 1052, 1050, 1048, 1046, 1044, 1042, 1040,
  1038, 1036, 1034, 1032, 1030, 1028, 1026, 1024,
  1022, 1020, 1018, 1016, 1014, 1012, 1010, 1008,
  1006, 1004, 1002, 1000, 998, 996, 994, 992,
  991, 989, 987, 985, 983, 981, 979, 978,
  976, 974, 972, 970, 969, 967, 965, 963,
  961, 960, 958, 956, 954, 953, 951, 949,
  948, 946, 944, 942, 941, 939, 937, 936,
  934, 932, 931, 929, 927, 926, 924, 923,
  921, 919, 918, 916, 914, 913, 911, 910,
  908, 907, 905, 903, 902, 900, 899, 897,
  896, 894, 893, 891, 890, 888, 887, 885,
  884, 882, 881, 879, 878, 876, 875, 873,
  872, 870, 869, 868, 866, 865, 863, 862,
  860, 859, 858, 856, 855, 853, 852, 851,
  849, 848, 846, 845, 844, 842, 841, 840,
  838, 837, 836, 834, 833, 832, 830, 829,
  828, 826, 825, 824, 823, 821, 820, 819,
  817, 816, 815, 814, 812, 811, 810, 809,
  807, 806, 805, 804, 802, 801, 800, 799,
  798, 796, 795, 794, 793, 791, 790, 789,
  788, 787, 786, 784, 783, 782, 781, 780,
  779, 777, 776, 775, 774, 773, 772, 771,
  769, 768, 767, 766, 765, 764, 763, 762,
  760, 759, 758, 757, 756, 755, 754, 753,
  752, 751, 750, 748, 747, 746, 745, 744,
  743, 742, 741, 740, 739, 738, 737, 736,
  735, 734, 733, 732, 731, 730, 729, 728,
  727, 726, 725, 724, 723, 722, 721, 720,
  719, 718, 717, 716, 715, 714, 713, 712,
  711, 710, 709, 708, 707, 706, 705, 704,
  703, 702, 701, 700, 699, 699, 698, 697,
  696, 695, 694, 693, 692, 691, 690, 689,
  688, 688, 687, 686, 685, 684, 683, 682,
  681, 680, 680, 679, 678, 677, 676, 675,
  674, 673, 673, 672, 671, 670, 669, 668,
  667, 667, 666, 665, 664, 663, 662, 661,
  661, 660, 659, 658, 657, 657, 656, 655,
  654, 653, 652, 652, 651, 650, 649, 648,
  648, 647, 646, 645, 644, 644, 643, 642,
  641, 640, 640, 639, 638, 637, 637, 636,
  635, 634, 633, 633, 632, 631, 630, 630,
  629, 628, 627, 627, 626, 625, 624, 624,
  623, 622, 621, 621, 620, 619, 618, 618,
  617, 616, 616, 615, 614, 613, 613, 612,
  611, 611, 610, 609, 608, 608, 607, 606,
  606, 605, 604, 604, 603, 602, 601, 601,
  600, 599, 599, 598, 597, 597, 596, 595,
  595, 594, 593, 593, 592, 591, 591, 590,
  589, 589, 588, 587, 587, 586, 585, 585,
  584, 583, 583, 582, 581, 581, 580, 579,
  579, 578, 578, 577, 576, 576, 575, 574,
  574, 573, 572, 572, 571, 571, 570, 569,
  569, 568, 568, 567, 566, 566, 565, 564,
  564, 563, 563, 562, 561, 561, 560, 560,
  559, 558, 558, 557, 557, 556, 555, 555,
  554, 554, 553, 553, 552, 551, 551, 550,
  550, 549, 548, 548, 547, 547, 546, 546,
  545, 544, 544, 543, 543, 542, 542, 541,
  541, 540, 539, 539, 538, 538, 537, 537,
  536, 536, 535, 534, 534, 533, 533, 532,
  532, 531, 531, 530, 530, 529, 529, 528,
  527, 527, 526, 526, 525, 525, 524, 524,
  523, 523, 522, 522, 521, 521, 520, 520,
  519, 519, 518, 518, 517, 517, 516, 516,
  515, 515, 514, 514
};



static inline int __cc_LinearToGammaWeighted(const uint8_t* src,
                                             const uint8_t* a_ptr,
                                             uint32_t total_a, int step,
                                             int rgb_stride) {
  const uint32_t sum =
      a_ptr[0] * __cc_GammaToLinear(src[0]) +
      a_ptr[step] * __cc_GammaToLinear(src[step]) +
      a_ptr[rgb_stride] * __cc_GammaToLinear(src[rgb_stride]) +
      a_ptr[rgb_stride + step] * __cc_GammaToLinear(src[rgb_stride + step]);
  ((total_a > 0 && total_a <= 4 * 0xff) ? (void) (0) : __assert_fail ("total_a > 0 && total_a <= 4 * 0xff", "extraction.1.c", 4861, __PRETTY_FUNCTION__));

  (((uint64_t)sum * kInvAlpha[total_a] < ((uint64_t)1 << 32)) ? (void) (0) : __assert_fail ("(uint64_t)sum * kInvAlpha[total_a] < ((uint64_t)1 << 32)", "extraction.1.c", 4863, __PRETTY_FUNCTION__));

  return __cc_LinearToGamma((((sum) * kInvAlpha[(total_a)]) >> (kAlphaFix - 2)), 0);
}





static inline void __cc_ConvertRowsToUVWithAlpha(const uint8_t* const r_ptr,
                                                 const uint8_t* const g_ptr,
                                                 const uint8_t* const b_ptr,
                                                 const uint8_t* const a_ptr,
                                                 int rgb_stride,
                                                 uint8_t* const dst_u,
                                                 uint8_t* const dst_v,
                                                 int width,
                                                 __cc_VP8Random* const rg) {
  int i, j;

  for (i = 0, j = 0; i < (width >> 1); ++i, j += 2 * sizeof(uint32_t)) {
    const uint32_t a = (((a_ptr + j)[0] + (a_ptr + j)[rgb_stride]) + (((a_ptr + j) + 4)[0] + ((a_ptr + j) + 4)[rgb_stride]));
    int r, g, b;
    if (a == 4 * 0xff || a == 0) {
      r = __cc_LinearToGamma( __cc_GammaToLinear((r_ptr + j)[0]) + __cc_GammaToLinear((r_ptr + j)[(4)]) + __cc_GammaToLinear((r_ptr + j)[rgb_stride]) + __cc_GammaToLinear((r_ptr + j)[rgb_stride + (4)]), 0);
      g = __cc_LinearToGamma( __cc_GammaToLinear((g_ptr + j)[0]) + __cc_GammaToLinear((g_ptr + j)[(4)]) + __cc_GammaToLinear((g_ptr + j)[rgb_stride]) + __cc_GammaToLinear((g_ptr + j)[rgb_stride + (4)]), 0);
      b = __cc_LinearToGamma( __cc_GammaToLinear((b_ptr + j)[0]) + __cc_GammaToLinear((b_ptr + j)[(4)]) + __cc_GammaToLinear((b_ptr + j)[rgb_stride]) + __cc_GammaToLinear((b_ptr + j)[rgb_stride + (4)]), 0);
    } else {
      r = __cc_LinearToGammaWeighted(r_ptr + j, a_ptr + j, a, 4, rgb_stride);
      g = __cc_LinearToGammaWeighted(g_ptr + j, a_ptr + j, a, 4, rgb_stride);
      b = __cc_LinearToGammaWeighted(b_ptr + j, a_ptr + j, a, 4, rgb_stride);
    }
    dst_u[i] = __cc_RGBToU(r, g, b, rg);
    dst_v[i] = __cc_RGBToV(r, g, b, rg);
  }
  if (width & 1) {
    const uint32_t a = 2u * ((a_ptr + j)[0] + (a_ptr + j)[rgb_stride]);
    int r, g, b;
    if (a == 4 * 0xff || a == 0) {
      r = __cc_LinearToGamma(__cc_GammaToLinear((r_ptr + j)[0]) + __cc_GammaToLinear((r_ptr + j)[rgb_stride]), 1);
      g = __cc_LinearToGamma(__cc_GammaToLinear((g_ptr + j)[0]) + __cc_GammaToLinear((g_ptr + j)[rgb_stride]), 1);
      b = __cc_LinearToGamma(__cc_GammaToLinear((b_ptr + j)[0]) + __cc_GammaToLinear((b_ptr + j)[rgb_stride]), 1);
    } else {
      r = __cc_LinearToGammaWeighted(r_ptr + j, a_ptr + j, a, 0, rgb_stride);
      g = __cc_LinearToGammaWeighted(g_ptr + j, a_ptr + j, a, 0, rgb_stride);
      b = __cc_LinearToGammaWeighted(b_ptr + j, a_ptr + j, a, 0, rgb_stride);
    }
    dst_u[i] = __cc_RGBToU(r, g, b, rg);
    dst_v[i] = __cc_RGBToV(r, g, b, rg);
  }
}

void __cc_VP8InitRandom(__cc_VP8Random* const rg, float dithering);

static int __cc_ImportYUVAFromRGBA(const uint8_t* const r_ptr,
                              const uint8_t* const g_ptr,
                              const uint8_t* const b_ptr,
                              const uint8_t* const a_ptr,
                              int step,
                              int rgb_stride,
                              float dithering,
                              int use_iterative_conversion,
                              __cc_WebPPicture* const picture) {
  int y;
  const int width = picture->width;
  const int height = picture->height;
  const int has_alpha = __cc_CheckNonOpaque(a_ptr, width, height, step, rgb_stride);

  picture->colorspace = has_alpha ? WEBP_YUV420A : WEBP_YUV420;
  picture->use_argb = 0;


  if (width < kMinDimensionIterativeConversion ||
      height < kMinDimensionIterativeConversion) {
    use_iterative_conversion = 0;
  }

  if (!__cc_WebPPictureAllocYUVA(picture, width, height)) {
    return 0;
  }
  if (has_alpha) {
    __cc_WebPInitAlphaProcessing();
    ((step == 4) ? (void) (0) : __assert_fail ("step == 4", "extraction.1.c", 4945, __PRETTY_FUNCTION__));

    ((kAlphaFix + 12 <= 31) ? (void) (0) : __assert_fail ("kAlphaFix + 12 <= 31", "extraction.1.c", 4947, __PRETTY_FUNCTION__));

  }

  if (use_iterative_conversion) {
    __cc_InitGammaTablesF();
    if (!__cc_PreprocessARGB(r_ptr, g_ptr, b_ptr, step, rgb_stride, picture)) {
      return 0;
    }
    if (has_alpha) {
      WebPExtractAlpha(a_ptr, rgb_stride, width, height,
                       picture->a, picture->a_stride);
    }
  } else {
    uint8_t* dst_y = picture->y;
    uint8_t* dst_u = picture->u;
    uint8_t* dst_v = picture->v;
    uint8_t* dst_a = picture->a;

    __cc_VP8Random base_rg;
    __cc_VP8Random* rg = ((void *)0);
    if (dithering > 0.) {
      __cc_VP8InitRandom(&base_rg, dithering);
      rg = &base_rg;
    }

    __cc_InitGammaTables();


    for (y = 0; y < (height >> 1); ++y) {
      int rows_have_alpha = has_alpha;
      const int off1 = (2 * y + 0) * rgb_stride;
      const int off2 = (2 * y + 1) * rgb_stride;
      __cc_ConvertRowToY(r_ptr + off1, g_ptr + off1, b_ptr + off1, step,
                    dst_y, width, rg);
      __cc_ConvertRowToY(r_ptr + off2, g_ptr + off2, b_ptr + off2, step,
                    dst_y + picture->y_stride, width, rg);
      dst_y += 2 * picture->y_stride;
      if (has_alpha) {
        rows_have_alpha &= !WebPExtractAlpha(a_ptr + off1, rgb_stride,
                                             width, 2,
                                             dst_a, picture->a_stride);
        dst_a += 2 * picture->a_stride;
      }
      if (!rows_have_alpha) {
        __cc_ConvertRowsToUV(r_ptr + off1, g_ptr + off1, b_ptr + off1,
                        step, rgb_stride, dst_u, dst_v, width, rg);
      } else {
        __cc_ConvertRowsToUVWithAlpha(r_ptr + off1, g_ptr + off1, b_ptr + off1,
                                 a_ptr + off1, rgb_stride,
                                 dst_u, dst_v, width, rg);
      }
      dst_u += picture->uv_stride;
      dst_v += picture->uv_stride;
    }
    if (height & 1) {
      const int off = 2 * y * rgb_stride;
      int row_has_alpha = has_alpha;
      __cc_ConvertRowToY(r_ptr + off, g_ptr + off, b_ptr + off, step,
                    dst_y, width, rg);
      if (row_has_alpha) {
        row_has_alpha &= !WebPExtractAlpha(a_ptr + off, 0, width, 1, dst_a, 0);
      }
      if (!row_has_alpha) {
        __cc_ConvertRowsToUV(r_ptr + off, g_ptr + off, b_ptr + off,
                        step, 0, dst_u, dst_v, width, rg);
      } else {
        __cc_ConvertRowsToUVWithAlpha(r_ptr + off, g_ptr + off, b_ptr + off,
                                 a_ptr + off, 0,
                                 dst_u, dst_v, width, rg);
      }
    }
  }
  return 1;
}

static inline uint32_t __cc_MakeARGB32(int a, int r, int g, int b) {
  return (((uint32_t)a << 24) | (r << 16) | (g << 8) | b);
}

static int __cc_Import(__cc_WebPPicture* const picture,
                  const uint8_t* const rgb, int rgb_stride,
                  int step, int swap_rb, int import_alpha) {
  int y;
  const uint8_t* const r_ptr = rgb + (swap_rb ? 2 : 0);
  const uint8_t* const g_ptr = rgb + 1;
  const uint8_t* const b_ptr = rgb + (swap_rb ? 0 : 2);
  const uint8_t* const a_ptr = import_alpha ? rgb + 3 : ((void *)0);
  const int width = picture->width;
  const int height = picture->height;

  if (!picture->use_argb) {
    return __cc_ImportYUVAFromRGBA(r_ptr, g_ptr, b_ptr, a_ptr, step, rgb_stride,
                              0.f , 0, picture);
  }
  if (!__cc_WebPPictureAlloc(picture)) return 0;

  ((step >= (import_alpha ? 4 : 3)) ? (void) (0) : __assert_fail ("step >= (import_alpha ? 4 : 3)", "extraction.1.c", 5044, __PRETTY_FUNCTION__));
  for (y = 0; y < height; ++y) {
    uint32_t* const dst = &picture->argb[y * picture->argb_stride];
    int x;
    for (x = 0; x < width; ++x) {
      const int offset = step * x + y * rgb_stride;
      dst[x] = __cc_MakeARGB32(import_alpha ? a_ptr[offset] : 0xff,
                          r_ptr[offset], g_ptr[offset], b_ptr[offset]);
    }
  }
  return 1;
}

typedef void (*__cc_ProcessRowsFunc)(__cc_VP8LDecoder* const dec, int row);

static int __cc_AllocateAndInitRescaler(__cc_VP8LDecoder* const dec, __cc_VP8Io* const io) {
  const int num_channels = 4;
  const int in_width = io->mb_w;
  const int out_width = io->scaled_width;
  const int in_height = io->mb_h;
  const int out_height = io->scaled_height;
  const uint64_t work_size = 2 * num_channels * (uint64_t)out_width;
  __cc_rescaler_t* work;
  const uint64_t scaled_data_size = (uint64_t)out_width;
  uint32_t* scaled_data;
  const uint64_t memory_size = sizeof(*dec->rescaler) +
                               work_size * sizeof(*work) +
                               scaled_data_size * sizeof(*scaled_data);
  uint8_t* memory = (uint8_t*)__cc_WebPSafeMalloc(memory_size, sizeof(*memory));
  if (memory == ((void *)0)) {
    dec->status_ = VP8_STATUS_OUT_OF_MEMORY;
    return 0;
  }
  ((dec->rescaler_memory == ((void *)0)) ? (void) (0) : __assert_fail ("dec->rescaler_memory == ((void *)0)", "extraction.1.c", 5077, __PRETTY_FUNCTION__));
  dec->rescaler_memory = memory;

  dec->rescaler = (__cc_WebPRescaler*)memory;
  memory += sizeof(*dec->rescaler);
  work = (__cc_rescaler_t*)memory;
  memory += work_size * sizeof(*work);
  scaled_data = (uint32_t*)memory;

  __cc_WebPRescalerInit(dec->rescaler, in_width, in_height, (uint8_t*)scaled_data,
                   out_width, out_height, 0, num_channels, work);
  return 1;
}

static int __cc_HuffmanTreeAllocate(__cc_HuffmanTree* const tree, int num_nodes) {
  ((tree != ((void *)0)) ? (void) (0) : __assert_fail ("tree != ((void *)0)", "extraction.1.c", 5092, __PRETTY_FUNCTION__));
  tree->root_ =
      (__cc_HuffmanTreeNode*)__cc_WebPSafeMalloc(num_nodes, sizeof(*tree->root_));
  return (tree->root_ != ((void *)0));
}

static void __cc_TreeNodeInit(__cc_HuffmanTreeNode* const node) {
  node->children_ = -1;
}

static int __cc_HuffmanTreeMaxNodes(int num_leaves) {
  return (2 * num_leaves - 1);
}

static int __cc_TreeInit(__cc_HuffmanTree* const tree, int num_leaves) {
  ((tree != ((void *)0)) ? (void) (0) : __assert_fail ("tree != ((void *)0)", "extraction.1.c", 5107, __PRETTY_FUNCTION__));
  if (num_leaves == 0) return 0;
  tree->max_nodes_ = __cc_HuffmanTreeMaxNodes(num_leaves);
  ((tree->max_nodes_ < (1 << 16)) ? (void) (0) : __assert_fail ("tree->max_nodes_ < (1 << 16)", "extraction.1.c", 5110, __PRETTY_FUNCTION__));
  if (!__cc_HuffmanTreeAllocate(tree, tree->max_nodes_)) return 0;
  __cc_TreeNodeInit(tree->root_);
  tree->num_nodes_ = 1;
  memset(tree->lut_bits_, 255, sizeof(tree->lut_bits_));
  memset(tree->lut_jump_, 0, sizeof(tree->lut_jump_));
  return 1;
}

void __cc_VP8LHuffmanTreeFree(__cc_HuffmanTree* const tree) {
  if (tree != ((void *)0)) {
    __cc_WebPSafeFree(tree->root_);
    tree->root_ = ((void *)0);
    tree->max_nodes_ = 0;
    tree->num_nodes_ = 0;
  }
}

void __cc_VP8LHtreeGroupsFree(__cc_HTreeGroup* htree_groups, int num_htree_groups) {
  if (htree_groups != ((void *)0)) {
    int i, j;
    for (i = 0; i < num_htree_groups; ++i) {
      __cc_HuffmanTree* const htrees = htree_groups[i].htrees_;
      for (j = 0; j < 5; ++j) {
        __cc_VP8LHuffmanTreeFree(&htrees[j]);
      }
    }
    __cc_WebPSafeFree(htree_groups);
  }
}

static int __cc_IsFull(const __cc_HuffmanTree* const tree) {
  return (tree->num_nodes_ == tree->max_nodes_);
}

static void __cc_AssignChildren(__cc_HuffmanTree* const tree,
                           __cc_HuffmanTreeNode* const node) {
  __cc_HuffmanTreeNode* const children = tree->root_ + tree->num_nodes_;
  node->children_ = (int)(children - node);
  ((children - node == (int)(children - node)) ? (void) (0) : __assert_fail ("children - node == (int)(children - node)", "extraction.1.c", 5149, __PRETTY_FUNCTION__));
  tree->num_nodes_ += 2;
  __cc_TreeNodeInit(children + 0);
  __cc_TreeNodeInit(children + 1);
}

static int __cc_ReverseBitsShort(int bits, int num_bits) {
  int retval = 0;
  int i;
  ((num_bits <= 8) ? (void) (0) : __assert_fail ("num_bits <= 8", "extraction.1.c", 5158, __PRETTY_FUNCTION__));
  for (i = 0; i < num_bits; ++i) {
    retval <<= 1;
    retval |= bits & 1;
    bits >>= 1;
  }
  return retval;
}

static int __cc_NodeIsEmpty(const __cc_HuffmanTreeNode* const node) {
  return (node->children_ < 0);
}

static int __cc_TreeAddSymbol(__cc_HuffmanTree* const tree,
                         int symbol, int code, int code_length) {
  int step = 7;
  int base_code;
  __cc_HuffmanTreeNode* node = tree->root_;
  const __cc_HuffmanTreeNode* const max_node = tree->root_ + tree->max_nodes_;
  ((symbol == (int16_t)symbol) ? (void) (0) : __assert_fail ("symbol == (int16_t)symbol", "extraction.1.c", 5177, __PRETTY_FUNCTION__));
  if (code_length <= 7) {
    int i;
    base_code = __cc_ReverseBitsShort(code, code_length);
    for (i = 0; i < (1 << (7 - code_length)); ++i) {
      const int idx = base_code | (i << code_length);
      tree->lut_symbol_[idx] = (int16_t)symbol;
      tree->lut_bits_[idx] = code_length;
    }
  } else {
    base_code = __cc_ReverseBitsShort((code >> (code_length - 7)),
                                 7);
  }
  while (code_length-- > 0) {
    if (node >= max_node) {
      return 0;
    }
    if (__cc_NodeIsEmpty(node)) {
      if (__cc_IsFull(tree)) return 0;
      __cc_AssignChildren(tree, node);
    } else if (!__cc_HuffmanTreeNodeIsNotLeaf(node)) {
      return 0;
    }
    node += node->children_ + ((code >> code_length) & 1);
    if (--step == 0) {
      tree->lut_jump_[base_code] = (int16_t)(node - tree->root_);
    }
  }
  if (__cc_NodeIsEmpty(node)) {
    node->children_ = 0;
  } else if (__cc_HuffmanTreeNodeIsNotLeaf(node)) {
    return 0;
  }
  node->symbol_ = symbol;
  return 1;
}



int __cc_VP8LHuffmanTreeBuildExplicit(__cc_HuffmanTree* const tree,
                                 const int* const code_lengths,
                                 const int* const codes,
                                 const int* const symbols, int max_symbol,
                                 int num_symbols) {
  int ok = 0;
  int i;
  ((tree != ((void *)0)) ? (void) (0) : __assert_fail ("tree != ((void *)0)", "extraction.1.c", 5223, __PRETTY_FUNCTION__));
  ((code_lengths != ((void *)0)) ? (void) (0) : __assert_fail ("code_lengths != ((void *)0)", "extraction.1.c", 5224, __PRETTY_FUNCTION__));
  ((codes != ((void *)0)) ? (void) (0) : __assert_fail ("codes != ((void *)0)", "extraction.1.c", 5225, __PRETTY_FUNCTION__));
  ((symbols != ((void *)0)) ? (void) (0) : __assert_fail ("symbols != ((void *)0)", "extraction.1.c", 5226, __PRETTY_FUNCTION__));


  if (!__cc_TreeInit(tree, num_symbols)) return 0;


  for (i = 0; i < num_symbols; ++i) {
    if (codes[i] != (-1)) {
      if (symbols[i] < 0 || symbols[i] >= max_symbol) {
        goto End;
      }
      if (!__cc_TreeAddSymbol(tree, symbols[i], codes[i], code_lengths[i])) {
        goto End;
      }
    }
  }
  ok = 1;
 End:
  ok = ok && __cc_IsFull(tree);
  if (!ok) __cc_VP8LHuffmanTreeFree(tree);
  return ok;
}



__cc_HTreeGroup* __cc_VP8LHtreeGroupsNew(int num_htree_groups) {
  __cc_HTreeGroup* const htree_groups =
      (__cc_HTreeGroup*)__cc_WebPSafeCalloc(num_htree_groups, sizeof(*htree_groups));
  ((num_htree_groups <= 0x10000) ? (void) (0) : __assert_fail ("num_htree_groups <= 0x10000", "extraction.1.c", 5254, __PRETTY_FUNCTION__));
  if (htree_groups == ((void *)0)) {
    return ((void *)0);
  }
  return htree_groups;
}

static void __cc_WebPPictureResetBufferYUVA(__cc_WebPPicture* const picture) {
  picture->memory_ = ((void *)0);
  picture->y = picture->u = picture->v = picture->a = ((void *)0);
  picture->y_stride = picture->uv_stride = 0;
  picture->a_stride = 0;
}

static void __cc_WebPPictureResetBufferARGB(__cc_WebPPicture* const picture) {
  picture->memory_argb_ = ((void *)0);
  picture->argb = ((void *)0);
  picture->argb_stride = 0;
}

void __cc_WebPPictureResetBuffers(__cc_WebPPicture* const picture) {
  __cc_WebPPictureResetBufferARGB(picture);
  __cc_WebPPictureResetBufferYUVA(picture);
}

void __cc_WebPPictureFree(__cc_WebPPicture* picture) {
  if (picture != ((void *)0)) {
    __cc_WebPSafeFree(picture->memory_);
    __cc_WebPSafeFree(picture->memory_argb_);
    __cc_WebPPictureResetBuffers(picture);
  }
}

int __cc_WebPPictureAllocARGB(__cc_WebPPicture* const picture, int width, int height) {
  void* memory;
  const uint64_t argb_size = (uint64_t)width * height;

  ((picture != ((void *)0)) ? (void) (0) : __assert_fail ("picture != ((void *)0)", "extraction.1.c", 5291, __PRETTY_FUNCTION__));

  __cc_WebPSafeFree(picture->memory_argb_);
  __cc_WebPPictureResetBufferARGB(picture);

  if (width <= 0 || height <= 0) {
    return __cc_WebPEncodingSetError(picture, VP8_ENC_ERROR_BAD_DIMENSION);
  }

  memory = __cc_WebPSafeMalloc(argb_size, sizeof(*picture->argb));
  if (memory == ((void *)0)) {
    return __cc_WebPEncodingSetError(picture, VP8_ENC_ERROR_OUT_OF_MEMORY);
  }

  picture->memory_argb_ = memory;
  picture->argb = (uint32_t*)memory;
  picture->argb_stride = width;
  return 1;
}

int __cc_WebPPictureAllocYUVA(__cc_WebPPicture* const picture, int width, int height) {
  const __cc_WebPEncCSP uv_csp = picture->colorspace & WEBP_CSP_UV_MASK;
  const int has_alpha = picture->colorspace & WEBP_CSP_ALPHA_BIT;
  const int y_stride = width;
  const int uv_width = (width + 1) >> 1;
  const int uv_height = (height + 1) >> 1;
  const int uv_stride = uv_width;
  int a_width, a_stride;
  uint64_t y_size, uv_size, a_size, total_size;
  uint8_t* mem;

  ((picture != ((void *)0)) ? (void) (0) : __assert_fail ("picture != ((void *)0)", "extraction.1.c", 5322, __PRETTY_FUNCTION__));

  __cc_WebPSafeFree(picture->memory_);
  __cc_WebPPictureResetBufferYUVA(picture);

  if (uv_csp != WEBP_YUV420) {
    return __cc_WebPEncodingSetError(picture, VP8_ENC_ERROR_INVALID_CONFIGURATION);
  }


  a_width = has_alpha ? width : 0;
  a_stride = a_width;
  y_size = (uint64_t)y_stride * height;
  uv_size = (uint64_t)uv_stride * uv_height;
  a_size = (uint64_t)a_stride * height;

  total_size = y_size + a_size + 2 * uv_size;


  if (width <= 0 || height <= 0 ||
      uv_width < 0 || uv_height < 0) {
    return __cc_WebPEncodingSetError(picture, VP8_ENC_ERROR_BAD_DIMENSION);
  }

  mem = (uint8_t*)__cc_WebPSafeMalloc(total_size, sizeof(*mem));
  if (mem == ((void *)0)) {
    return __cc_WebPEncodingSetError(picture, VP8_ENC_ERROR_OUT_OF_MEMORY);
  }


  picture->memory_ = (void*)mem;
  picture->y_stride = y_stride;
  picture->uv_stride = uv_stride;
  picture->a_stride = a_stride;


  picture->y = mem;
  mem += y_size;

  picture->u = mem;
  mem += uv_size;
  picture->v = mem;
  mem += uv_size;

  if (a_size > 0) {
    picture->a = mem;
    mem += a_size;
  }
  (void)mem;
  return 1;
}

static const int kModeBpp[MODE_LAST] = {
  3, 4, 3, 4, 4, 2, 2,
  4, 4, 4, 2,
  1, 1 };

void __cc_VP8LoadFinalBytes(__cc_VP8BitReader* const br);





static inline uint32_t __cc_BSwap32(uint32_t x) {
  return __builtin_bswap32(x);
}

typedef uint32_t __cc_lbit_t;

static inline void __cc_VP8LoadNewBytes(__cc_VP8BitReader* const br) {
  ((br != ((void *)0) && br->buf_ != ((void *)0)) ? (void) (0) : __assert_fail ("br != ((void *)0) && br->buf_ != ((void *)0)", "extraction.1.c", 5411, __PRETTY_FUNCTION__));

  if (br->buf_ + sizeof(__cc_lbit_t) <= br->buf_end_) {

    __cc_bit_t bits;
    const __cc_lbit_t in_bits = *(const __cc_lbit_t*)br->buf_;

    br->buf_ += 24 >> 3;





    bits = __cc_BSwap32(in_bits);
    bits >>= (32 - 24);
    br->value_ = bits | (br->value_ << 24);
    br->bits_ += 24;
  } else {
    __cc_VP8LoadFinalBytes(br);
  }
}

void __cc_VP8InitBitReader(__cc_VP8BitReader* const br,
                      const uint8_t* const start, const uint8_t* const end) {
  ((br != ((void *)0)) ? (void) (0) : __assert_fail ("br != ((void *)0)", "extraction.1.c", 5462, __PRETTY_FUNCTION__));
  ((start != ((void *)0)) ? (void) (0) : __assert_fail ("start != ((void *)0)", "extraction.1.c", 5463, __PRETTY_FUNCTION__));
  ((start <= end) ? (void) (0) : __assert_fail ("start <= end", "extraction.1.c", 5464, __PRETTY_FUNCTION__));
  br->range_ = 255 - 1;
  br->buf_ = start;
  br->buf_end_ = end;
  br->value_ = 0;
  br->bits_ = -8;
  br->eof_ = 0;
  __cc_VP8LoadNewBytes(br);
}


typedef struct timeval __cc_Stopwatch;

static inline double __cc_StopwatchReadAndReset(__cc_Stopwatch* watch) {
  struct timeval old_value;
  double delta_sec, delta_usec;
  memcpy(&old_value, watch, sizeof(old_value));
  gettimeofday(watch, ((void *)0));
  delta_sec = (double)watch->tv_sec - old_value.tv_sec;
  delta_usec = (double)watch->tv_usec - old_value.tv_usec;
  return delta_sec + delta_usec / 1000000.0;
}

extern __cc_VP8StatusCode __cc_WebPDecode(const uint8_t* data, size_t data_size,
                                      __cc_WebPDecoderConfig* config);

static void __cc_PrintAnimationWarning(const __cc_WebPDecoderConfig* const config) {
  if (config->input.has_animation) {
    fprintf(stderr,
            "Error! Decoding of an animated WebP file is not supported.\n"
            "       Use webpmux to extract the individual frames or\n"
            "       vwebp to view this image.\n");
  }
}

static inline void __cc_StopwatchReset(__cc_Stopwatch* watch) {
  gettimeofday(watch, ((void *)0));
}

__cc_VP8StatusCode __cc_ExUtilDecodeWebP(const uint8_t* const data, size_t data_size,
                               int verbose, __cc_WebPDecoderConfig* const config) {
  __cc_Stopwatch stop_watch;
  __cc_VP8StatusCode status = VP8_STATUS_OK;
  if (config == ((void *)0)) return VP8_STATUS_INVALID_PARAM;

  __cc_PrintAnimationWarning(config);

  __cc_StopwatchReset(&stop_watch);


  status = __cc_WebPDecode(data, data_size, config);

  if (verbose) {
    const double decode_time = __cc_StopwatchReadAndReset(&stop_watch);
    fprintf(stderr, "Time to decode picture: %.3fs\n", decode_time);
  }
  return status;
}

static inline int __cc_VP8Clip8(int v) {
  return ((v & ~YUV_MASK2) == 0) ? (v >> YUV_FIX2) : (v < 0) ? 0 : 255;
}
static inline int __cc_VP8YUVToG(int y, int u, int v) {
  return __cc_VP8Clip8(19077 * y - 6419 * u - 13320 * v + (-19077 * 16 + 6419 * 128 + 13320 * 128 + YUV_HALF2));
}

typedef void (*__cc_VP8PredFunc)(uint8_t* dst);

extern const __cc_VP8PredFunc VP8PredChroma8[ ];

extern __cc_VP8DecIdct VP8TransformAC3;



static const uint8_t kBModesProba[NUM_BMODES][NUM_BMODES][NUM_BMODES - 1] = {
  { { 231, 120, 48, 89, 115, 113, 120, 152, 112 },
    { 152, 179, 64, 126, 170, 118, 46, 70, 95 },
    { 175, 69, 143, 80, 85, 82, 72, 155, 103 },
    { 56, 58, 10, 171, 218, 189, 17, 13, 152 },
    { 114, 26, 17, 163, 44, 195, 21, 10, 173 },
    { 121, 24, 80, 195, 26, 62, 44, 64, 85 },
    { 144, 71, 10, 38, 171, 213, 144, 34, 26 },
    { 170, 46, 55, 19, 136, 160, 33, 206, 71 },
    { 63, 20, 8, 114, 114, 208, 12, 9, 226 },
    { 81, 40, 11, 96, 182, 84, 29, 16, 36 } },
  { { 134, 183, 89, 137, 98, 101, 106, 165, 148 },
    { 72, 187, 100, 130, 157, 111, 32, 75, 80 },
    { 66, 102, 167, 99, 74, 62, 40, 234, 128 },
    { 41, 53, 9, 178, 241, 141, 26, 8, 107 },
    { 74, 43, 26, 146, 73, 166, 49, 23, 157 },
    { 65, 38, 105, 160, 51, 52, 31, 115, 128 },
    { 104, 79, 12, 27, 217, 255, 87, 17, 7 },
    { 87, 68, 71, 44, 114, 51, 15, 186, 23 },
    { 47, 41, 14, 110, 182, 183, 21, 17, 194 },
    { 66, 45, 25, 102, 197, 189, 23, 18, 22 } },
  { { 88, 88, 147, 150, 42, 46, 45, 196, 205 },
    { 43, 97, 183, 117, 85, 38, 35, 179, 61 },
    { 39, 53, 200, 87, 26, 21, 43, 232, 171 },
    { 56, 34, 51, 104, 114, 102, 29, 93, 77 },
    { 39, 28, 85, 171, 58, 165, 90, 98, 64 },
    { 34, 22, 116, 206, 23, 34, 43, 166, 73 },
    { 107, 54, 32, 26, 51, 1, 81, 43, 31 },
    { 68, 25, 106, 22, 64, 171, 36, 225, 114 },
    { 34, 19, 21, 102, 132, 188, 16, 76, 124 },
    { 62, 18, 78, 95, 85, 57, 50, 48, 51 } },
  { { 193, 101, 35, 159, 215, 111, 89, 46, 111 },
    { 60, 148, 31, 172, 219, 228, 21, 18, 111 },
    { 112, 113, 77, 85, 179, 255, 38, 120, 114 },
    { 40, 42, 1, 196, 245, 209, 10, 25, 109 },
    { 88, 43, 29, 140, 166, 213, 37, 43, 154 },
    { 61, 63, 30, 155, 67, 45, 68, 1, 209 },
    { 100, 80, 8, 43, 154, 1, 51, 26, 71 },
    { 142, 78, 78, 16, 255, 128, 34, 197, 171 },
    { 41, 40, 5, 102, 211, 183, 4, 1, 221 },
    { 51, 50, 17, 168, 209, 192, 23, 25, 82 } },
  { { 138, 31, 36, 171, 27, 166, 38, 44, 229 },
    { 67, 87, 58, 169, 82, 115, 26, 59, 179 },
    { 63, 59, 90, 180, 59, 166, 93, 73, 154 },
    { 40, 40, 21, 116, 143, 209, 34, 39, 175 },
    { 47, 15, 16, 183, 34, 223, 49, 45, 183 },
    { 46, 17, 33, 183, 6, 98, 15, 32, 183 },
    { 57, 46, 22, 24, 128, 1, 54, 17, 37 },
    { 65, 32, 73, 115, 28, 128, 23, 128, 205 },
    { 40, 3, 9, 115, 51, 192, 18, 6, 223 },
    { 87, 37, 9, 115, 59, 77, 64, 21, 47 } },
  { { 104, 55, 44, 218, 9, 54, 53, 130, 226 },
    { 64, 90, 70, 205, 40, 41, 23, 26, 57 },
    { 54, 57, 112, 184, 5, 41, 38, 166, 213 },
    { 30, 34, 26, 133, 152, 116, 10, 32, 134 },
    { 39, 19, 53, 221, 26, 114, 32, 73, 255 },
    { 31, 9, 65, 234, 2, 15, 1, 118, 73 },
    { 75, 32, 12, 51, 192, 255, 160, 43, 51 },
    { 88, 31, 35, 67, 102, 85, 55, 186, 85 },
    { 56, 21, 23, 111, 59, 205, 45, 37, 192 },
    { 55, 38, 70, 124, 73, 102, 1, 34, 98 } },
  { { 125, 98, 42, 88, 104, 85, 117, 175, 82 },
    { 95, 84, 53, 89, 128, 100, 113, 101, 45 },
    { 75, 79, 123, 47, 51, 128, 81, 171, 1 },
    { 57, 17, 5, 71, 102, 57, 53, 41, 49 },
    { 38, 33, 13, 121, 57, 73, 26, 1, 85 },
    { 41, 10, 67, 138, 77, 110, 90, 47, 114 },
    { 115, 21, 2, 10, 102, 255, 166, 23, 6 },
    { 101, 29, 16, 10, 85, 128, 101, 196, 26 },
    { 57, 18, 10, 102, 102, 213, 34, 20, 43 },
    { 117, 20, 15, 36, 163, 128, 68, 1, 26 } },
  { { 102, 61, 71, 37, 34, 53, 31, 243, 192 },
    { 69, 60, 71, 38, 73, 119, 28, 222, 37 },
    { 68, 45, 128, 34, 1, 47, 11, 245, 171 },
    { 62, 17, 19, 70, 146, 85, 55, 62, 70 },
    { 37, 43, 37, 154, 100, 163, 85, 160, 1 },
    { 63, 9, 92, 136, 28, 64, 32, 201, 85 },
    { 75, 15, 9, 9, 64, 255, 184, 119, 16 },
    { 86, 6, 28, 5, 64, 255, 25, 248, 1 },
    { 56, 8, 17, 132, 137, 255, 55, 116, 128 },
    { 58, 15, 20, 82, 135, 57, 26, 121, 40 } },
  { { 164, 50, 31, 137, 154, 133, 25, 35, 218 },
    { 51, 103, 44, 131, 131, 123, 31, 6, 158 },
    { 86, 40, 64, 135, 148, 224, 45, 183, 128 },
    { 22, 26, 17, 131, 240, 154, 14, 1, 209 },
    { 45, 16, 21, 91, 64, 222, 7, 1, 197 },
    { 56, 21, 39, 155, 60, 138, 23, 102, 213 },
    { 83, 12, 13, 54, 192, 255, 68, 47, 28 },
    { 85, 26, 85, 85, 128, 128, 32, 146, 171 },
    { 18, 11, 7, 63, 144, 171, 4, 4, 246 },
    { 35, 27, 10, 146, 174, 171, 12, 26, 128 } },
  { { 190, 80, 35, 99, 180, 80, 126, 54, 45 },
    { 85, 126, 47, 87, 176, 51, 41, 20, 32 },
    { 101, 75, 128, 139, 118, 146, 116, 128, 85 },
    { 56, 41, 15, 176, 236, 85, 37, 9, 62 },
    { 71, 30, 17, 119, 118, 255, 17, 18, 138 },
    { 101, 38, 60, 138, 55, 70, 43, 26, 142 },
    { 146, 36, 19, 30, 171, 255, 97, 27, 20 },
    { 138, 45, 61, 62, 219, 1, 81, 188, 64 },
    { 32, 41, 20, 117, 151, 142, 20, 21, 163 },
    { 112, 19, 12, 61, 195, 128, 48, 4, 24 } }
};

extern const __cc_range_t kVP8NewRange[128];

extern const uint8_t kVP8Log2Range[128];

static inline int __cc_VP8GetBit(__cc_VP8BitReader* const br, int prob) {



  __cc_range_t range = br->range_;
  if (br->bits_ < 0) {
    __cc_VP8LoadNewBytes(br);
  }
  {
    const int pos = br->bits_;
    const __cc_range_t split = (range * prob) >> 8;
    const __cc_range_t value = (__cc_range_t)(br->value_ >> pos);
    int bit;
    if (value > split) {
      range -= split + 1;
      br->value_ -= (__cc_bit_t)(split + 1) << pos;
      bit = 1;
    } else {
      range = split;
      bit = 0;
    }

    if (range <= (__cc_range_t)0x7e) {
      const int shift = kVP8Log2Range[range];
      range = kVP8NewRange[range];
      br->bits_ -= shift;
    }
    br->range_ = range;
    return bit;
  }
}

static void __cc_ParseIntraMode(__cc_VP8BitReader* const br,
                           __cc_VP8Decoder* const dec, int mb_x) {
  uint8_t* const top = dec->intra_t_ + 4 * mb_x;
  uint8_t* const left = dec->intra_l_;
  __cc_VP8MBData* const block = dec->mb_data_ + mb_x;



  if (dec->segment_hdr_.update_map_) {

    block->segment_ = !__cc_VP8GetBit(br, dec->proba_.segments_[0])
                    ? __cc_VP8GetBit(br, dec->proba_.segments_[1])
                    : 2 + __cc_VP8GetBit(br, dec->proba_.segments_[2]);
  } else {
    block->segment_ = 0;
  }
  if (dec->use_skip_proba_) block->skip_ = __cc_VP8GetBit(br, dec->skip_p_);

  block->is_i4x4_ = !__cc_VP8GetBit(br, 145);
  if (!block->is_i4x4_) {

    const int ymode =
        __cc_VP8GetBit(br, 156) ? (__cc_VP8GetBit(br, 128) ? TM_PRED : H_PRED)
                           : (__cc_VP8GetBit(br, 163) ? V_PRED : DC_PRED);
    block->imodes_[0] = ymode;
    memset(top, ymode, 4 * sizeof(*top));
    memset(left, ymode, 4 * sizeof(*left));
  } else {
    uint8_t* modes = block->imodes_;
    int y;
    for (y = 0; y < 4; ++y) {
      int ymode = left[y];
      int x;
      for (x = 0; x < 4; ++x) {
        const uint8_t* const prob = kBModesProba[top[x]][ymode];


        int i = kYModesIntra4[__cc_VP8GetBit(br, prob[0])];
        while (i > 0) {
          i = kYModesIntra4[2 * i + __cc_VP8GetBit(br, prob[i])];
        }
        ymode = -i;
        top[x] = ymode;
      }
      memcpy(modes, top, 4 * sizeof(*top));
      modes += 4;
      left[y] = ymode;
    }
  }

  block->uvmode_ = !__cc_VP8GetBit(br, 142) ? DC_PRED
                 : !__cc_VP8GetBit(br, 114) ? V_PRED
                 : __cc_VP8GetBit(br, 183) ? TM_PRED : H_PRED;
}

int __cc_VP8ParseIntraModeRow(__cc_VP8BitReader* const br, __cc_VP8Decoder* const dec) {
  int mb_x;
  for (mb_x = 0; mb_x < dec->mb_w_; ++mb_x) {
    __cc_ParseIntraMode(br, dec, mb_x);
  }
  return !dec->br_.eof_;
}

static void __cc_DefaultFeatures(__cc_WebPBitstreamFeatures* const features) {
  ((features != ((void *)0)) ? (void) (0) : __assert_fail ("features != ((void *)0)", "extraction.1.c", 5772, __PRETTY_FUNCTION__));
  memset(features, 0, sizeof(*features));
}

static __cc_VP8StatusCode __cc_GetFeatures(const uint8_t* const data, size_t data_size,
                                 __cc_WebPBitstreamFeatures* const features) {
  if (features == ((void *)0) || data == ((void *)0)) {
    return VP8_STATUS_INVALID_PARAM;
  }
  __cc_DefaultFeatures(features);


  return __cc_ParseHeadersInternal(data, data_size,
                              &features->width, &features->height,
                              &features->has_alpha, &features->has_animation,
                              &features->format, ((void *)0));
}

int __cc_VP8ProcessRow(__cc_VP8Decoder* const dec, __cc_VP8Io* const io);

int __cc_VP8ExitCritical(__cc_VP8Decoder* const dec, __cc_VP8Io* const io);

int __cc_VP8ParseIntraModeRow(__cc_VP8BitReader* const br, __cc_VP8Decoder* const dec);

int __cc_VP8InitFrame(__cc_VP8Decoder* const dec, __cc_VP8Io* io);

__cc_VP8StatusCode __cc_VP8EnterCritical(__cc_VP8Decoder* const dec, __cc_VP8Io* const io);

typedef struct Metadata {
  __cc_MetadataPayload exif;
  __cc_MetadataPayload iccp;
  __cc_MetadataPayload xmp;
} __cc_Metadata;

extern int __cc_WebPInitDecoderConfigInternal(__cc_WebPDecoderConfig*, int);

__cc_VP8SimpleFilterFunc VP8SimpleHFilter16;

static inline void __cc_Put16(int v, uint8_t* dst) {
  int j;
  for (j = 0; j < 16; ++j) {
    memset(dst + j * 32, v, 16);
  }
}

static void __cc_DC16NoLeft(uint8_t *dst) {
  int DC = 8;
  int i;
  for (i = 0; i < 16; ++i) {
    DC += dst[i - 32];
  }
  __cc_Put16(DC >> 4, dst);
}

static inline uint32_t __cc_NzCodeBits(uint32_t nz_coeffs, int nz, int dc_nz) {
  nz_coeffs <<= 2;
  nz_coeffs |= (nz > 3) ? 3 : (nz > 1) ? 2 : dc_nz;
  return nz_coeffs;
}

static const int kCodeLengthRepeatCode = 16;

void __cc_VP8YUVInit(void) {}

typedef void (*__cc_VP8LConvertFunc)(const uint32_t* src, int num_pixels,
                                uint8_t* dst);

extern __cc_VP8LConvertFunc VP8LConvertBGRAToRGBA4444;

extern const uint8_t* const VP8kabs0;



int __cc_WebPInitDecBufferInternal(__cc_WebPDecBuffer* buffer, int version) {
  if ((((version) >> 8) != ((0x0203) >> 8))) {
    return 0;
  }
  if (buffer == ((void *)0)) return 0;
  memset(buffer, 0, sizeof(*buffer));
  return 1;
}

static int __cc_IsValidColorspace(int webp_csp_mode) {
  return (webp_csp_mode >= MODE_RGB && webp_csp_mode < MODE_LAST);
}




static __cc_VP8StatusCode __cc_CheckDecBuffer(const __cc_WebPDecBuffer* const buffer) {
  int ok = 1;
  const __cc_WEBP_CSP_MODE mode = buffer->colorspace;
  const int width = buffer->width;
  const int height = buffer->height;
  if (!__cc_IsValidColorspace(mode)) {
    ok = 0;
  } else if (!__cc_WebPIsRGBMode(mode)) {
    const __cc_WebPYUVABuffer* const buf = &buffer->u.YUVA;
    const int uv_width = (width + 1) / 2;
    const int uv_height = (height + 1) / 2;
    const int y_stride = abs(buf->y_stride);
    const int u_stride = abs(buf->u_stride);
    const int v_stride = abs(buf->v_stride);
    const int a_stride = abs(buf->a_stride);
    const uint64_t y_size = (uint64_t)(y_stride) * ((height) - 1) + (width);
    const uint64_t u_size = (uint64_t)(u_stride) * ((uv_height) - 1) + (uv_width);
    const uint64_t v_size = (uint64_t)(v_stride) * ((uv_height) - 1) + (uv_width);
    const uint64_t a_size = (uint64_t)(a_stride) * ((height) - 1) + (width);
    ok &= (y_size <= buf->y_size);
    ok &= (u_size <= buf->u_size);
    ok &= (v_size <= buf->v_size);
    ok &= (y_stride >= width);
    ok &= (u_stride >= uv_width);
    ok &= (v_stride >= uv_width);
    ok &= (buf->y != ((void *)0));
    ok &= (buf->u != ((void *)0));
    ok &= (buf->v != ((void *)0));
    if (mode == MODE_YUVA) {
      ok &= (a_stride >= width);
      ok &= (a_size <= buf->a_size);
      ok &= (buf->a != ((void *)0));
    }
  } else {
    const __cc_WebPRGBABuffer* const buf = &buffer->u.RGBA;
    const int stride = abs(buf->stride);
    const uint64_t size = (uint64_t)(stride) * ((height) - 1) + (width);
    ok &= (size <= buf->size);
    ok &= (stride >= width * kModeBpp[mode]);
    ok &= (buf->rgba != ((void *)0));
  }
  return ok ? VP8_STATUS_OK : VP8_STATUS_INVALID_PARAM;
}
typedef struct {


  uint32_t* literal_;
  uint32_t red_[256];
  uint32_t blue_[256];
  uint32_t alpha_[256];

  uint32_t distance_[40];
  int palette_code_bits_;
  double bit_cost_;
  double literal_cost_;
  double red_cost_;
  double blue_cost_;
} __cc_VP8LHistogram;







static void __cc_InitCorrectionLUT(int16_t* const lut, int min_dist) {






  const int threshold1 = min_dist << 2;
  const int threshold2 = (3 * threshold1) >> 2;
  const int max_threshold = threshold2 << 0;
  const int delta = threshold1 - threshold2;
  int i;
  for (i = 1; i <= ((1 << (8 + 2)) - 1); ++i) {
    int c = (i <= threshold2) ? (i << 0)
          : (i < threshold1) ? max_threshold * (threshold1 - i) / delta
          : 0;
    c >>= 2;
    lut[+i] = +c;
    lut[-i] = -c;
  }
  lut[0] = 0;
}



typedef struct {
  int width_, height_;
  int row_;
  uint8_t* src_;
  uint8_t* dst_;

  int radius_;
  int scale_;

  void* mem_;


  uint16_t* start_;
  uint16_t* cur_;
  uint16_t* end_;
  uint16_t* top_;
  uint16_t* average_;


  int num_levels_;
  int min_, max_;
  int min_level_dist_;

  int16_t* correction_;
} __cc_SmoothParams;

static void __cc_ApplyFilter(__cc_SmoothParams* const p) {
  const uint16_t* const average = p->average_;
  const int w = p->width_;
  const int16_t* const correction = p->correction_;



  uint8_t* const dst = p->dst_;
  int x;
  for (x = 0; x < w; ++x) {
    const int v = dst[x];
    if (v < p->max_ && v > p->min_) {
      const int c = (v << 0) + correction[average[x] - (v << 2)];



      dst[x] = __cc_clip_8b(c);

    }
  }
  p->dst_ += w;
}



static void __cc_HFilter(__cc_SmoothParams* const p) {
  const uint16_t* const in = p->end_;
  uint16_t* const out = p->average_;
  const uint32_t scale = p->scale_;
  const int w = p->width_;
  const int r = p->radius_;

  int x;
  for (x = 0; x <= r; ++x) {
    const uint16_t delta = in[x + r - 1] + in[r - x];
    out[x] = (delta * scale) >> 16;
  }
  for (; x < w - r; ++x) {
    const uint16_t delta = in[x + r] - in[x - r - 1];
    out[x] = (delta * scale) >> 16;
  }
  for (; x < w; ++x) {
    const uint16_t delta =
        2 * in[w - 1] - in[2 * w - 2 - r - x] - in[x - r - 1];
    out[x] = (delta * scale) >> 16;
  }
}

static void __cc_CleanupParams(__cc_SmoothParams* const p) {
  __cc_WebPSafeFree(p->mem_);
}

static void __cc_VFilter(__cc_SmoothParams* const p) {
  const uint8_t* src = p->src_;
  const int w = p->width_;
  uint16_t* const cur = p->cur_;
  const uint16_t* const top = p->top_;
  uint16_t* const out = p->end_;
  uint16_t sum = 0;
  int x;

  for (x = 0; x < w; ++x) {
    uint16_t new_value;
    sum += src[x];
    new_value = top[x] + sum;
    out[x] = new_value - cur[x];
    cur[x] = new_value;
  }

  p->top_ = p->cur_;
  p->cur_ += w;
  if (p->cur_ == p->end_) p->cur_ = p->start_;


  if (p->row_ >= 0 && p->row_ < p->height_ - 1) {
    p->src_ += p->width_;
  }
}

static void __cc_CountLevels(const uint8_t* const data, int size,
                        __cc_SmoothParams* const p) {
  int i, last_level;
  uint8_t used_levels[256] = { 0 };
  p->min_ = 255;
  p->max_ = 0;
  for (i = 0; i < size; ++i) {
    const int v = data[i];
    if (v < p->min_) p->min_ = v;
    if (v > p->max_) p->max_ = v;
    used_levels[v] = 1;
  }

  p->min_level_dist_ = p->max_ - p->min_;
  last_level = -1;
  for (i = 0; i < 256; ++i) {
    if (used_levels[i]) {
      ++p->num_levels_;
      if (last_level >= 0) {
        const int level_dist = i - last_level;
        if (level_dist < p->min_level_dist_) {
          p->min_level_dist_ = level_dist;
        }
      }
      last_level = i;
    }
  }
}

static int __cc_InitParams(uint8_t* const data, int width, int height,
                      int radius, __cc_SmoothParams* const p) {
  const int R = 2 * radius + 1;

  const size_t size_scratch_m = (R + 1) * width * sizeof(*p->start_);
  const size_t size_m = width * sizeof(*p->average_);
  const size_t size_lut = (1 + 2 * ((1 << (8 + 2)) - 1)) * sizeof(*p->correction_);
  const size_t total_size = size_scratch_m + size_m + size_lut;
  uint8_t* mem = (uint8_t*)__cc_WebPSafeMalloc(1U, total_size);

  if (mem == ((void *)0)) return 0;
  p->mem_ = (void*)mem;

  p->start_ = (uint16_t*)mem;
  p->cur_ = p->start_;
  p->end_ = p->start_ + R * width;
  p->top_ = p->end_ - width;
  memset(p->top_, 0, width * sizeof(*p->top_));
  mem += size_scratch_m;

  p->average_ = (uint16_t*)mem;
  mem += size_m;

  p->width_ = width;
  p->height_ = height;
  p->src_ = data;
  p->dst_ = data;
  p->radius_ = radius;
  p->scale_ = (1 << (16 + 2)) / (R * R);
  p->row_ = -radius;


  __cc_CountLevels(data, width * height, p);


  p->correction_ = ((int16_t*)mem) + ((1 << (8 + 2)) - 1);
  __cc_InitCorrectionLUT(p->correction_, p->min_level_dist_);

  return 1;
}

int __cc_WebPDequantizeLevels(uint8_t* const data, int width, int height,
                         int strength) {
  const int radius = 4 * strength / 100;
  if (strength < 0 || strength > 100) return 0;
  if (data == ((void *)0) || width <= 0 || height <= 0) return 0;
  if (radius > 0) {
    __cc_SmoothParams p;
    memset(&p, 0, sizeof(p));
    if (!__cc_InitParams(data, width, height, radius, &p)) return 0;
    if (p.num_levels_ > 2) {
      for (; p.row_ < p.height_; ++p.row_) {
        __cc_VFilter(&p);


        if (p.row_ >= p.radius_) {
          __cc_HFilter(&p);
          __cc_ApplyFilter(&p);
        }
      }
    }
    __cc_CleanupParams(&p);
  }
  return 1;
}

static const uint32_t kHashMul = 0x1e35a7bd;

static inline void __cc_VP8LColorCacheInsert(const __cc_VP8LColorCache* const cc,
                                             uint32_t argb) {
  const uint32_t key = (kHashMul * argb) >> cc->hash_shift_;
  cc->colors_[key] = argb;
}

int __cc_VP8LColorCacheInit(__cc_VP8LColorCache* const color_cache, int hash_bits);

typedef float (*__cc_VP8LFastLog2SlowFunc)(uint32_t v);



static inline int __cc_VP8LHistogramNumCodes(int palette_code_bits) {
  return 256 + 24 +
      ((palette_code_bits > 0) ? (1 << palette_code_bits) : 0);
}

void __cc_ALPHDelete(__cc_ALPHDecoder* const dec) {
  if (dec != ((void *)0)) {
    __cc_VP8LDelete(dec->vp8l_dec_);
    dec->vp8l_dec_ = ((void *)0);
    __cc_WebPSafeFree(dec);
  }
}

int __cc_WebPDequantizeLevels(uint8_t* const data, int width, int height,
                         int strength);





typedef void (*__cc_VP8LProcessBlueAndRedFunc)(uint32_t* argb_data, int num_pixels);

void __cc_ExUtilPrintWebPError(const char* const in_file, int status);

void __cc_VP8InitScanline(__cc_VP8Decoder* const dec) {
  __cc_VP8MB* const left = dec->mb_info_ - 1;
  left->nz_ = 0;
  left->nz_dc_ = 0;
  memset(dec->intra_l_, B_DC_PRED, sizeof(dec->intra_l_));
  dec->mb_x_ = 0;
}

void __cc_VP8Clear(__cc_VP8Decoder* const dec);

void __cc_VP8Delete(__cc_VP8Decoder* const dec) {
  if (dec != ((void *)0)) {
    __cc_VP8Clear(dec);
    __cc_WebPSafeFree(dec);
  }
}

static inline uint32_t __cc_Clip255(uint32_t a) {
  if (a < 256) {
    return a;
  }


  return ~a >> 24;
}

static inline int __cc_AddSubtractComponentHalf(int a, int b) {
  return __cc_Clip255(a + (a - b) / 2);
}

static const uint8_t kCat6[] =
  { 254, 254, 243, 230, 196, 177, 153, 140, 133, 130, 129, 0 };

static const uint8_t kCat5[] = { 180, 157, 141, 134, 130, 0 };

static const uint8_t kCat4[] = { 176, 155, 140, 135, 0 };

static const uint8_t kCat3[] = { 173, 148, 140, 0 };

static const uint8_t* const kCat3456[] = { kCat3, kCat4, kCat5, kCat6 };

void __cc_VP8InitBitReader(__cc_VP8BitReader* const br,
                      const uint8_t* const start, const uint8_t* const end);

uint32_t __cc_VP8GetValue(__cc_VP8BitReader* const br, int num_bits);

static __cc_VP8StatusCode __cc_ParsePartitions(__cc_VP8Decoder* const dec,
                                     const uint8_t* buf, size_t size) {
  __cc_VP8BitReader* const br = &dec->br_;
  const uint8_t* sz = buf;
  const uint8_t* buf_end = buf + size;
  const uint8_t* part_start;
  int last_part;
  int p;

  dec->num_parts_ = 1 << __cc_VP8GetValue(br, 2);
  last_part = dec->num_parts_ - 1;
  part_start = buf + last_part * 3;
  if (buf_end < part_start) {

    return VP8_STATUS_NOT_ENOUGH_DATA;
  }
  for (p = 0; p < last_part; ++p) {
    const uint32_t psize = sz[0] | (sz[1] << 8) | (sz[2] << 16);
    const uint8_t* part_end = part_start + psize;
    if (part_end > buf_end) part_end = buf_end;
    __cc_VP8InitBitReader(dec->parts_ + p, part_start, part_end);
    part_start = part_end;
    sz += 3;
  }
  __cc_VP8InitBitReader(dec->parts_ + last_part, part_start, buf_end);
  return (part_start < buf_end) ? VP8_STATUS_OK :
           VP8_STATUS_SUSPENDED;
}

int32_t __cc_VP8GetSignedValue(__cc_VP8BitReader* const br, int num_bits);

static inline uint32_t __cc_VP8Get(__cc_VP8BitReader* const br) {
  return __cc_VP8GetValue(br, 1);
}

static int __cc_ParseFilterHeader(__cc_VP8BitReader* br, __cc_VP8Decoder* const dec) {
  __cc_VP8FilterHeader* const hdr = &dec->filter_hdr_;
  hdr->simple_ = __cc_VP8Get(br);
  hdr->level_ = __cc_VP8GetValue(br, 6);
  hdr->sharpness_ = __cc_VP8GetValue(br, 3);
  hdr->use_lf_delta_ = __cc_VP8Get(br);
  if (hdr->use_lf_delta_) {
    if (__cc_VP8Get(br)) {
      int i;
      for (i = 0; i < NUM_REF_LF_DELTAS; ++i) {
        if (__cc_VP8Get(br)) {
          hdr->ref_lf_delta_[i] = __cc_VP8GetSignedValue(br, 6);
        }
      }
      for (i = 0; i < NUM_MODE_LF_DELTAS; ++i) {
        if (__cc_VP8Get(br)) {
          hdr->mode_lf_delta_[i] = __cc_VP8GetSignedValue(br, 6);
        }
      }
    }
  }
  dec->filter_type_ = (hdr->level_ == 0) ? 0 : hdr->simple_ ? 1 : 2;
  return !br->eof_;
}

void __cc_VP8ParseQuant(__cc_VP8Decoder* const dec);

static int __cc_ParseSegmentHeader(__cc_VP8BitReader* br,
                              __cc_VP8SegmentHeader* hdr, __cc_VP8Proba* proba) {
  ((br != ((void *)0)) ? (void) (0) : __assert_fail ("br != ((void *)0)", "extraction.1.c", 6308, __PRETTY_FUNCTION__));
  ((hdr != ((void *)0)) ? (void) (0) : __assert_fail ("hdr != ((void *)0)", "extraction.1.c", 6309, __PRETTY_FUNCTION__));
  hdr->use_segment_ = __cc_VP8Get(br);
  if (hdr->use_segment_) {
    hdr->update_map_ = __cc_VP8Get(br);
    if (__cc_VP8Get(br)) {
      int s;
      hdr->absolute_delta_ = __cc_VP8Get(br);
      for (s = 0; s < NUM_MB_SEGMENTS; ++s) {
        hdr->quantizer_[s] = __cc_VP8Get(br) ? __cc_VP8GetSignedValue(br, 7) : 0;
      }
      for (s = 0; s < NUM_MB_SEGMENTS; ++s) {
        hdr->filter_strength_[s] = __cc_VP8Get(br) ? __cc_VP8GetSignedValue(br, 6) : 0;
      }
    }
    if (hdr->update_map_) {
      int s;
      for (s = 0; s < MB_FEATURE_TREE_PROBS; ++s) {
        proba->segments_[s] = __cc_VP8Get(br) ? __cc_VP8GetValue(br, 8) : 255u;
      }
    }
  } else {
    hdr->update_map_ = 0;
  }
  return !br->eof_;
}

static void __cc_ResetSegmentHeader(__cc_VP8SegmentHeader* const hdr) {
  ((hdr != ((void *)0)) ? (void) (0) : __assert_fail ("hdr != ((void *)0)", "extraction.1.c", 6336, __PRETTY_FUNCTION__));
  hdr->use_segment_ = 0;
  hdr->update_map_ = 0;
  hdr->absolute_delta_ = 1;
  memset(hdr->quantizer_, 0, sizeof(hdr->quantizer_));
  memset(hdr->filter_strength_, 0, sizeof(hdr->filter_strength_));
}

int __cc_VP8SetError(__cc_VP8Decoder* const dec,
                __cc_VP8StatusCode error, const char* const msg) {




  if (dec->status_ == VP8_STATUS_OK) {
    dec->status_ = error;
    dec->error_msg_ = msg;
    dec->ready_ = 0;
  }
  return 0;
}

void __cc_VP8ParseProba(__cc_VP8BitReader* const br, __cc_VP8Decoder* const dec);

void __cc_VP8ResetProba(__cc_VP8Proba* const proba);

int __cc_VP8GetHeaders(__cc_VP8Decoder* const dec, __cc_VP8Io* const io) {
  const uint8_t* buf;
  size_t buf_size;
  __cc_VP8FrameHeader* frm_hdr;
  __cc_VP8PictureHeader* pic_hdr;
  __cc_VP8BitReader* br;
  __cc_VP8StatusCode status;

  if (dec == ((void *)0)) {
    return 0;
  }
  __cc_SetOk(dec);
  if (io == ((void *)0)) {
    return __cc_VP8SetError(dec, VP8_STATUS_INVALID_PARAM,
                       "null VP8Io passed to VP8GetHeaders()");
  }
  buf = io->data;
  buf_size = io->data_size;
  if (buf_size < 4) {
    return __cc_VP8SetError(dec, VP8_STATUS_NOT_ENOUGH_DATA,
                       "Truncated header.");
  }


  {
    const uint32_t bits = buf[0] | (buf[1] << 8) | (buf[2] << 16);
    frm_hdr = &dec->frm_hdr_;
    frm_hdr->key_frame_ = !(bits & 1);
    frm_hdr->profile_ = (bits >> 1) & 7;
    frm_hdr->show_ = (bits >> 4) & 1;
    frm_hdr->partition_length_ = (bits >> 5);
    if (frm_hdr->profile_ > 3)
      return __cc_VP8SetError(dec, VP8_STATUS_BITSTREAM_ERROR,
                         "Incorrect keyframe parameters.");
    if (!frm_hdr->show_)
      return __cc_VP8SetError(dec, VP8_STATUS_UNSUPPORTED_FEATURE,
                         "Frame not displayable.");
    buf += 3;
    buf_size -= 3;
  }

  pic_hdr = &dec->pic_hdr_;
  if (frm_hdr->key_frame_) {

    if (buf_size < 7) {
      return __cc_VP8SetError(dec, VP8_STATUS_NOT_ENOUGH_DATA,
                         "cannot parse picture header");
    }
    if (!__cc_VP8CheckSignature(buf, buf_size)) {
      return __cc_VP8SetError(dec, VP8_STATUS_BITSTREAM_ERROR,
                         "Bad code word");
    }
    pic_hdr->width_ = ((buf[4] << 8) | buf[3]) & 0x3fff;
    pic_hdr->xscale_ = buf[4] >> 6;
    pic_hdr->height_ = ((buf[6] << 8) | buf[5]) & 0x3fff;
    pic_hdr->yscale_ = buf[6] >> 6;
    buf += 7;
    buf_size -= 7;

    dec->mb_w_ = (pic_hdr->width_ + 15) >> 4;
    dec->mb_h_ = (pic_hdr->height_ + 15) >> 4;

    io->width = pic_hdr->width_;
    io->height = pic_hdr->height_;
    io->use_scaling = 0;
    io->use_cropping = 0;
    io->crop_top = 0;
    io->crop_left = 0;
    io->crop_right = io->width;
    io->crop_bottom = io->height;
    io->mb_w = io->width;
    io->mb_h = io->height;

    __cc_VP8ResetProba(&dec->proba_);
    __cc_ResetSegmentHeader(&dec->segment_hdr_);
  }



  if (frm_hdr->partition_length_ > buf_size) {
    return __cc_VP8SetError(dec, VP8_STATUS_NOT_ENOUGH_DATA,
                       "bad partition length");
  }

  br = &dec->br_;
  __cc_VP8InitBitReader(br, buf, buf + frm_hdr->partition_length_);
  buf += frm_hdr->partition_length_;
  buf_size -= frm_hdr->partition_length_;

  if (frm_hdr->key_frame_) {
    pic_hdr->colorspace_ = __cc_VP8Get(br);
    pic_hdr->clamp_type_ = __cc_VP8Get(br);
  }
  if (!__cc_ParseSegmentHeader(br, &dec->segment_hdr_, &dec->proba_)) {
    return __cc_VP8SetError(dec, VP8_STATUS_BITSTREAM_ERROR,
                       "cannot parse segment header");
  }

  if (!__cc_ParseFilterHeader(br, dec)) {
    return __cc_VP8SetError(dec, VP8_STATUS_BITSTREAM_ERROR,
                       "cannot parse filter header");
  }
  status = __cc_ParsePartitions(dec, buf, buf_size);
  if (status != VP8_STATUS_OK) {
    return __cc_VP8SetError(dec, status, "cannot parse partitions");
  }


  __cc_VP8ParseQuant(dec);


  if (!frm_hdr->key_frame_) {
    return __cc_VP8SetError(dec, VP8_STATUS_UNSUPPORTED_FEATURE,
                       "Not a key frame.");
  }

  __cc_VP8Get(br);

  __cc_VP8ParseProba(br, dec);


  dec->ready_ = 1;
  return 1;
}

static void __cc_GradientUnfilter(int width, int height, int stride, int row,
                             int num_rows, uint8_t* data) {
  __cc_DoGradientFilter(data, width, height, stride, row, num_rows, 1, data);
}

static inline int __cc_VP8GetSigned(__cc_VP8BitReader* const br, int v) {
  if (br->bits_ < 0) {
    __cc_VP8LoadNewBytes(br);
  }
  {
    const int pos = br->bits_;
    const __cc_range_t split = br->range_ >> 1;
    const __cc_range_t value = (__cc_range_t)(br->value_ >> pos);
    const int32_t mask = (int32_t)(split - value) >> 31;
    br->bits_ -= 1;
    br->range_ += mask;
    br->range_ |= 1;
    br->value_ -= (__cc_bit_t)((split + 1) & mask) << pos;
    return (v ^ mask) - mask;
  }
}

static const uint8_t kZigzag[16] = {
  0, 1, 4, 8, 5, 2, 3, 6, 9, 12, 13, 10, 7, 11, 14, 15
};

static const int kBands[16 + 1] = {
  0, 1, 2, 3, 6, 4, 5, 6, 6, 6, 6, 6, 6, 6, 6, 7,
  0
};

static int __cc_GetLargeValue(__cc_VP8BitReader* const br, const uint8_t* const p) {
  int v;
  if (!__cc_VP8GetBit(br, p[3])) {
    if (!__cc_VP8GetBit(br, p[4])) {
      v = 2;
    } else {
      v = 3 + __cc_VP8GetBit(br, p[5]);
    }
  } else {
    if (!__cc_VP8GetBit(br, p[6])) {
      if (!__cc_VP8GetBit(br, p[7])) {
        v = 5 + __cc_VP8GetBit(br, 159);
      } else {
        v = 7 + 2 * __cc_VP8GetBit(br, 165);
        v += __cc_VP8GetBit(br, 145);
      }
    } else {
      const uint8_t* tab;
      const int bit1 = __cc_VP8GetBit(br, p[8]);
      const int bit0 = __cc_VP8GetBit(br, p[9 + bit1]);
      const int cat = 2 * bit1 + bit0;
      v = 0;
      for (tab = kCat3456[cat]; *tab; ++tab) {
        v += v + __cc_VP8GetBit(br, *tab);
      }
      v += 3 + (8 << cat);
    }
  }
  return v;
}

static int __cc_GetCoeffs(__cc_VP8BitReader* const br, const __cc_VP8BandProbas* const prob,
                     int ctx, const __cc_quant_t dq, int n, int16_t* out) {

  const uint8_t* p = prob[n].probas_[ctx];
  for (; n < 16; ++n) {
    if (!__cc_VP8GetBit(br, p[0])) {
      return n;
    }
    while (!__cc_VP8GetBit(br, p[1])) {
      p = prob[kBands[++n]].probas_[0];
      if (n == 16) return 16;
    }
    {
      const __cc_VP8ProbaArray* const p_ctx = &prob[kBands[n + 1]].probas_[0];
      int v;
      if (!__cc_VP8GetBit(br, p[2])) {
        v = 1;
        p = p_ctx[1];
      } else {
        v = __cc_GetLargeValue(br, p);
        p = p_ctx[2];
      }
      out[kZigzag[n]] = __cc_VP8GetSigned(br, v) * dq[n > 0];
    }
  }
  return 16;
}

typedef void (*__cc_VP8WHT)(const int16_t* in, int16_t* out);

extern __cc_VP8WHT VP8TransformWHT;

static int __cc_ParseResiduals(__cc_VP8Decoder* const dec,
                          __cc_VP8MB* const mb, __cc_VP8BitReader* const token_br) {
  __cc_VP8BandProbas (* const bands)[NUM_BANDS] = dec->proba_.bands_;
  const __cc_VP8BandProbas* ac_proba;
  __cc_VP8MBData* const block = dec->mb_data_ + dec->mb_x_;
  const __cc_VP8QuantMatrix* const q = &dec->dqm_[block->segment_];
  int16_t* dst = block->coeffs_;
  __cc_VP8MB* const left_mb = dec->mb_info_ - 1;
  uint8_t tnz, lnz;
  uint32_t non_zero_y = 0;
  uint32_t non_zero_uv = 0;
  int x, y, ch;
  uint32_t out_t_nz, out_l_nz;
  int first;

  memset(dst, 0, 384 * sizeof(*dst));
  if (!block->is_i4x4_) {
    int16_t dc[16] = { 0 };
    const int ctx = mb->nz_dc_ + left_mb->nz_dc_;
    const int nz = __cc_GetCoeffs(token_br, bands[1], ctx, q->y2_mat_, 0, dc);
    mb->nz_dc_ = left_mb->nz_dc_ = (nz > 0);
    if (nz > 1) {
      VP8TransformWHT(dc, dst);
    } else {
      int i;
      const int dc0 = (dc[0] + 3) >> 3;
      for (i = 0; i < 16 * 16; i += 16) dst[i] = dc0;
    }
    first = 1;
    ac_proba = bands[0];
  } else {
    first = 0;
    ac_proba = bands[3];
  }

  tnz = mb->nz_ & 0x0f;
  lnz = left_mb->nz_ & 0x0f;
  for (y = 0; y < 4; ++y) {
    int l = lnz & 1;
    uint32_t nz_coeffs = 0;
    for (x = 0; x < 4; ++x) {
      const int ctx = l + (tnz & 1);
      const int nz = __cc_GetCoeffs(token_br, ac_proba, ctx, q->y1_mat_, first, dst);
      l = (nz > first);
      tnz = (tnz >> 1) | (l << 7);
      nz_coeffs = __cc_NzCodeBits(nz_coeffs, nz, dst[0] != 0);
      dst += 16;
    }
    tnz >>= 4;
    lnz = (lnz >> 1) | (l << 7);
    non_zero_y = (non_zero_y << 8) | nz_coeffs;
  }
  out_t_nz = tnz;
  out_l_nz = lnz >> 4;

  for (ch = 0; ch < 4; ch += 2) {
    uint32_t nz_coeffs = 0;
    tnz = mb->nz_ >> (4 + ch);
    lnz = left_mb->nz_ >> (4 + ch);
    for (y = 0; y < 2; ++y) {
      int l = lnz & 1;
      for (x = 0; x < 2; ++x) {
        const int ctx = l + (tnz & 1);
        const int nz = __cc_GetCoeffs(token_br, bands[2], ctx, q->uv_mat_, 0, dst);
        l = (nz > 0);
        tnz = (tnz >> 1) | (l << 3);
        nz_coeffs = __cc_NzCodeBits(nz_coeffs, nz, dst[0] != 0);
        dst += 16;
      }
      tnz >>= 2;
      lnz = (lnz >> 1) | (l << 5);
    }

    non_zero_uv |= nz_coeffs << (4 * ch);
    out_t_nz |= (tnz << 4) << ch;
    out_l_nz |= (lnz & 0xf0) << ch;
  }
  mb->nz_ = out_t_nz;
  left_mb->nz_ = out_l_nz;

  block->non_zero_y_ = non_zero_y;
  block->non_zero_uv_ = non_zero_uv;




  block->dither_ = (non_zero_uv & 0xaaaa) ? 0 : q->dither_;

  return !(non_zero_y | non_zero_uv);
}

int __cc_VP8DecodeMB(__cc_VP8Decoder* const dec, __cc_VP8BitReader* const token_br) {
  __cc_VP8MB* const left = dec->mb_info_ - 1;
  __cc_VP8MB* const mb = dec->mb_info_ + dec->mb_x_;
  __cc_VP8MBData* const block = dec->mb_data_ + dec->mb_x_;
  int skip = dec->use_skip_proba_ ? block->skip_ : 0;

  if (!skip) {
    skip = __cc_ParseResiduals(dec, mb, token_br);
  } else {
    left->nz_ = mb->nz_ = 0;
    if (!block->is_i4x4_) {
      left->nz_dc_ = mb->nz_dc_ = 0;
    }
    block->non_zero_y_ = 0;
    block->non_zero_uv_ = 0;
    block->dither_ = 0;
  }

  if (dec->filter_type_ > 0) {
    __cc_VP8FInfo* const finfo = dec->f_info_ + dec->mb_x_;
    *finfo = dec->fstrengths_[block->segment_][block->is_i4x4_];
    finfo->f_inner_ |= !skip;
  }

  return !token_br->eof_;
}

static int __cc_ParseFrame(__cc_VP8Decoder* const dec, __cc_VP8Io* io) {
  for (dec->mb_y_ = 0; dec->mb_y_ < dec->br_mb_y_; ++dec->mb_y_) {

    __cc_VP8BitReader* const token_br =
        &dec->parts_[dec->mb_y_ & (dec->num_parts_ - 1)];
    if (!__cc_VP8ParseIntraModeRow(&dec->br_, dec)) {
      return __cc_VP8SetError(dec, VP8_STATUS_NOT_ENOUGH_DATA,
                         "Premature end-of-partition0 encountered.");
    }
    for (; dec->mb_x_ < dec->mb_w_; ++dec->mb_x_) {
      if (!__cc_VP8DecodeMB(dec, token_br)) {
        return __cc_VP8SetError(dec, VP8_STATUS_NOT_ENOUGH_DATA,
                           "Premature end-of-file encountered.");
      }
    }
    __cc_VP8InitScanline(dec);


    if (!__cc_VP8ProcessRow(dec, io)) {
      return __cc_VP8SetError(dec, VP8_STATUS_USER_ABORT, "Output aborted.");
    }
  }
  if (dec->mt_method_ > 0) {
    if (!__cc_WebPGetWorkerInterface()->Sync(&dec->worker_)) return 0;
  }

  return 1;
}

int __cc_VP8Decode(__cc_VP8Decoder* const dec, __cc_VP8Io* const io) {
  int ok = 0;
  if (dec == ((void *)0)) {
    return 0;
  }
  if (io == ((void *)0)) {
    return __cc_VP8SetError(dec, VP8_STATUS_INVALID_PARAM,
                       "NULL VP8Io parameter in VP8Decode().");
  }

  if (!dec->ready_) {
    if (!__cc_VP8GetHeaders(dec, io)) {
      return 0;
    }
  }
  ((dec->ready_) ? (void) (0) : __assert_fail ("dec->ready_", "extraction.1.c", 6743, __PRETTY_FUNCTION__));


  ok = (__cc_VP8EnterCritical(dec, io) == VP8_STATUS_OK);
  if (ok) {

    if (ok) ok = __cc_VP8InitFrame(dec, io);


    if (ok) ok = __cc_ParseFrame(dec, io);


    ok &= __cc_VP8ExitCritical(dec, io);
  }

  if (!ok) {
    __cc_VP8Clear(dec);
    return 0;
  }

  dec->ready_ = 0;
  return ok;
}

static inline uint32_t __cc_GetARGBValue(uint32_t val) {
  return val;
}

static inline uint32_t __cc_GetARGBIndex(uint32_t idx) {
  return (idx >> 8) & 0xff;
}
static void __cc_ColorIndexInverseTransform(const __cc_VP8LTransform* const transform, int y_start, int y_end, const uint32_t* src, uint32_t* dst) { int y; const int bits_per_pixel = 8 >> transform->bits_; const int width = transform->xsize_; const uint32_t* const color_map = transform->data_; if (bits_per_pixel < 8) { const int pixels_per_byte = 1 << transform->bits_; const int count_mask = pixels_per_byte - 1; const uint32_t bit_mask = (1 << bits_per_pixel) - 1; for (y = y_start; y < y_end; ++y) { uint32_t packed_pixels = 0; int x; for (x = 0; x < width; ++x) { if ((x & count_mask) == 0) packed_pixels = __cc_GetARGBIndex(*src++); *dst++ = __cc_GetARGBValue(color_map[packed_pixels & bit_mask]); packed_pixels >>= bits_per_pixel; } } } else { for (y = y_start; y < y_end; ++y) { int x; for (x = 0; x < width; ++x) { *dst++ = __cc_GetARGBValue(color_map[__cc_GetARGBIndex(*src++)]); } } } }


typedef struct {


  uint8_t green_to_red_;
  uint8_t green_to_blue_;
  uint8_t red_to_blue_;
} __cc_VP8LMultipliers;

typedef void (*__cc_VP8LTransformColorFunc)(const __cc_VP8LMultipliers* const m,
                                       uint32_t* argb_data, int num_pixels);

extern __cc_VP8LTransformColorFunc VP8LTransformColorInverse;

static inline void __cc_ColorCodeToMultipliers(uint32_t color_code,
                                               __cc_VP8LMultipliers* const m) {
  m->green_to_red_ = (color_code >> 0) & 0xff;
  m->green_to_blue_ = (color_code >> 8) & 0xff;
  m->red_to_blue_ = (color_code >> 16) & 0xff;
}

static void __cc_ColorSpaceInverseTransform(const __cc_VP8LTransform* const transform,
                                       int y_start, int y_end, uint32_t* data) {
  const int width = transform->xsize_;
  const int tile_width = 1 << transform->bits_;
  const int mask = tile_width - 1;
  const int safe_width = width & ~mask;
  const int remaining_width = width - safe_width;
  const int tiles_per_row = __cc_VP8LSubSampleSize(width, transform->bits_);
  int y = y_start;
  const uint32_t* pred_row =
      transform->data_ + (y >> transform->bits_) * tiles_per_row;

  while (y < y_end) {
    const uint32_t* pred = pred_row;
    __cc_VP8LMultipliers m = { 0, 0, 0 };
    const uint32_t* const data_safe_end = data + safe_width;
    const uint32_t* const data_end = data + width;
    while (data < data_safe_end) {
      __cc_ColorCodeToMultipliers(*pred++, &m);
      VP8LTransformColorInverse(&m, data, tile_width);
      data += tile_width;
    }
    if (data < data_end) {
      __cc_ColorCodeToMultipliers(*pred++, &m);
      VP8LTransformColorInverse(&m, data, remaining_width);
      data += remaining_width;
    }
    ++y;
    if ((y & mask) == 0) pred_row += tiles_per_row;
  }
}

extern __cc_VP8LProcessBlueAndRedFunc VP8LAddGreenToBlueAndRed;

typedef uint32_t (*__cc_VP8LPredictorFunc)(uint32_t left, const uint32_t* const top);

extern __cc_VP8LPredictorFunc VP8LPredictors[16];

static uint32_t __cc_Predictor2(uint32_t left, const uint32_t* const top) {
  (void)left;
  return top[0];
}

static inline void __cc_AddPixelsEq(uint32_t* a, uint32_t b) {
  const uint32_t alpha_and_green = (*a & 0xff00ff00u) + (b & 0xff00ff00u);
  const uint32_t red_and_blue = (*a & 0x00ff00ffu) + (b & 0x00ff00ffu);
  *a = (alpha_and_green & 0xff00ff00u) | (red_and_blue & 0x00ff00ffu);
}



static uint32_t __cc_Predictor0(uint32_t left, const uint32_t* const top) {
  (void)top;
  (void)left;
  return 0xff000000;
}

static uint32_t __cc_Predictor1(uint32_t left, const uint32_t* const top) {
  (void)top;
  return left;
}

static void __cc_PredictorInverseTransform(const __cc_VP8LTransform* const transform,
                                      int y_start, int y_end, uint32_t* data) {
  const int width = transform->xsize_;
  if (y_start == 0) {
    int x;
    const uint32_t pred0 = __cc_Predictor0(data[-1], ((void *)0));
    __cc_AddPixelsEq(data, pred0);
    for (x = 1; x < width; ++x) {
      const uint32_t pred1 = __cc_Predictor1(data[x - 1], ((void *)0));
      __cc_AddPixelsEq(data + x, pred1);
    }
    data += width;
    ++y_start;
  }

  {
    int y = y_start;
    const int tile_width = 1 << transform->bits_;
    const int mask = tile_width - 1;
    const int safe_width = width & ~mask;
    const int tiles_per_row = __cc_VP8LSubSampleSize(width, transform->bits_);
    const uint32_t* pred_mode_base =
        transform->data_ + (y >> transform->bits_) * tiles_per_row;

    while (y < y_end) {
      const uint32_t pred2 = __cc_Predictor2(data[-1], data - width);
      const uint32_t* pred_mode_src = pred_mode_base;
      __cc_VP8LPredictorFunc pred_func;
      int x = 1;
      int t = 1;

      __cc_AddPixelsEq(data, pred2);

      while (x < safe_width) {
        pred_func = VP8LPredictors[((*pred_mode_src++) >> 8) & 0xf];
        for (; t < tile_width; ++t, ++x) {
          const uint32_t pred = pred_func(data[x - 1], data + x - width);
          __cc_AddPixelsEq(data + x, pred);
        }
        t = 0;
      }
      if (x < width) {
        pred_func = VP8LPredictors[((*pred_mode_src++) >> 8) & 0xf];
        for (; x < width; ++x) {
          const uint32_t pred = pred_func(data[x - 1], data + x - width);
          __cc_AddPixelsEq(data + x, pred);
        }
      }
      data += width;
      ++y;
      if ((y & mask) == 0) {
        pred_mode_base += tiles_per_row;
      }
    }
  }
}

void __cc_VP8LInverseTransform(const __cc_VP8LTransform* const transform,
                          int row_start, int row_end,
                          const uint32_t* const in, uint32_t* const out) {
  const int width = transform->xsize_;
  ((row_start < row_end) ? (void) (0) : __assert_fail ("row_start < row_end", "extraction.1.c", 6955, __PRETTY_FUNCTION__));
  ((row_end <= transform->ysize_) ? (void) (0) : __assert_fail ("row_end <= transform->ysize_", "extraction.1.c", 6956, __PRETTY_FUNCTION__));
  switch (transform->type_) {
    case SUBTRACT_GREEN:
      VP8LAddGreenToBlueAndRed(out, (row_end - row_start) * width);
      break;
    case PREDICTOR_TRANSFORM:
      __cc_PredictorInverseTransform(transform, row_start, row_end, out);
      if (row_end != transform->ysize_) {


        memcpy(out - width, out + (row_end - row_start - 1) * width,
               width * sizeof(*out));
      }
      break;
    case CROSS_COLOR_TRANSFORM:
      __cc_ColorSpaceInverseTransform(transform, row_start, row_end, out);
      break;
    case COLOR_INDEXING_TRANSFORM:
      if (in == out && transform->bits_ > 0) {





        const int out_stride = (row_end - row_start) * width;
        const int in_stride = (row_end - row_start) *
            __cc_VP8LSubSampleSize(transform->xsize_, transform->bits_);
        uint32_t* const src = out + out_stride - in_stride;
        memmove(src, out, in_stride * sizeof(*src));
        __cc_ColorIndexInverseTransform(transform, row_start, row_end, src, out);
      } else {
        __cc_ColorIndexInverseTransform(transform, row_start, row_end, in, out);
      }
      break;
  }
}

void __cc_ALPHDelete(__cc_ALPHDecoder* const dec);

void __cc_VP8Clear(__cc_VP8Decoder* const dec) {
  if (dec == ((void *)0)) {
    return;
  }
  __cc_WebPGetWorkerInterface()->End(&dec->worker_);
  __cc_ALPHDelete(dec->alph_dec_);
  dec->alph_dec_ = ((void *)0);
  __cc_WebPSafeFree(dec->mem_);
  dec->mem_ = ((void *)0);
  dec->mem_size_ = 0;
  memset(&dec->br_, 0, sizeof(dec->br_));
  dec->ready_ = 0;
}

typedef double (*__cc_VP8LCostFunc)(const uint32_t* population, int length);

typedef struct {
  int counts[2];
  int streaks[2][2];
} __cc_VP8LStreaks;

typedef __cc_VP8LStreaks (*__cc_VP8LCostCountFunc)(const uint32_t* population,
                                         int length);

static const uint8_t
    CoeffsUpdateProba[NUM_TYPES][NUM_BANDS][NUM_CTX][NUM_PROBAS] = {
  { { { 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 }
    },
    { { 176, 246, 255, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 223, 241, 252, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 249, 253, 253, 255, 255, 255, 255, 255, 255, 255, 255 }
    },
    { { 255, 244, 252, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 234, 254, 254, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 253, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 }
    },
    { { 255, 246, 254, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 239, 253, 254, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 254, 255, 254, 255, 255, 255, 255, 255, 255, 255, 255 }
    },
    { { 255, 248, 254, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 251, 255, 254, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 }
    },
    { { 255, 253, 254, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 251, 254, 254, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 254, 255, 254, 255, 255, 255, 255, 255, 255, 255, 255 }
    },
    { { 255, 254, 253, 255, 254, 255, 255, 255, 255, 255, 255 },
      { 250, 255, 254, 255, 254, 255, 255, 255, 255, 255, 255 },
      { 254, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 }
    },
    { { 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 }
    }
  },
  { { { 217, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 225, 252, 241, 253, 255, 255, 254, 255, 255, 255, 255 },
      { 234, 250, 241, 250, 253, 255, 253, 254, 255, 255, 255 }
    },
    { { 255, 254, 255, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 223, 254, 254, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 238, 253, 254, 254, 255, 255, 255, 255, 255, 255, 255 }
    },
    { { 255, 248, 254, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 249, 254, 255, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 }
    },
    { { 255, 253, 255, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 247, 254, 255, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 }
    },
    { { 255, 253, 254, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 252, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 }
    },
    { { 255, 254, 254, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 253, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 }
    },
    { { 255, 254, 253, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 250, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 254, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 }
    },
    { { 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 }
    }
  },
  { { { 186, 251, 250, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 234, 251, 244, 254, 255, 255, 255, 255, 255, 255, 255 },
      { 251, 251, 243, 253, 254, 255, 254, 255, 255, 255, 255 }
    },
    { { 255, 253, 254, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 236, 253, 254, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 251, 253, 253, 254, 254, 255, 255, 255, 255, 255, 255 }
    },
    { { 255, 254, 254, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 254, 254, 254, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 }
    },
    { { 255, 254, 255, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 254, 254, 255, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 254, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 }
    },
    { { 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 254, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 }
    },
    { { 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 }
    },
    { { 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 }
    },
    { { 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 }
    }
  },
  { { { 248, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 250, 254, 252, 254, 255, 255, 255, 255, 255, 255, 255 },
      { 248, 254, 249, 253, 255, 255, 255, 255, 255, 255, 255 }
    },
    { { 255, 253, 253, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 246, 253, 253, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 252, 254, 251, 254, 254, 255, 255, 255, 255, 255, 255 }
    },
    { { 255, 254, 252, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 248, 254, 253, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 253, 255, 254, 254, 255, 255, 255, 255, 255, 255, 255 }
    },
    { { 255, 251, 254, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 245, 251, 254, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 253, 253, 254, 255, 255, 255, 255, 255, 255, 255, 255 }
    },
    { { 255, 251, 253, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 252, 253, 254, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 255, 254, 255, 255, 255, 255, 255, 255, 255, 255, 255 }
    },
    { { 255, 252, 255, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 249, 255, 254, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 255, 255, 254, 255, 255, 255, 255, 255, 255, 255, 255 }
    },
    { { 255, 255, 253, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 250, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 }
    },
    { { 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 254, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 },
      { 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255 }
    }
  }
};

void __cc_VP8ParseProba(__cc_VP8BitReader* const br, __cc_VP8Decoder* const dec) {
  __cc_VP8Proba* const proba = &dec->proba_;
  int t, b, c, p;
  for (t = 0; t < NUM_TYPES; ++t) {
    for (b = 0; b < NUM_BANDS; ++b) {
      for (c = 0; c < NUM_CTX; ++c) {
        for (p = 0; p < NUM_PROBAS; ++p) {
          const int v = __cc_VP8GetBit(br, CoeffsUpdateProba[t][b][c][p]) ?
                        __cc_VP8GetValue(br, 8) : CoeffsProba0[t][b][c][p];
          proba->bands_[t][b].probas_[c][p] = v;
        }
      }
    }
  }
  dec->use_skip_proba_ = __cc_VP8Get(br);
  if (dec->use_skip_proba_) {
    dec->skip_p_ = __cc_VP8GetValue(br, 8);
  }
}

typedef void (*__cc_VP8LHistogramAddFunc)(const __cc_VP8LHistogram* const a,
                                     const __cc_VP8LHistogram* const b,
                                     __cc_VP8LHistogram* const out);

typedef double (*__cc_VP8LCostCombinedFunc)(const uint32_t* X, const uint32_t* Y,
                                       int length);

typedef __cc_VP8LStreaks (*__cc_VP8LCostCombinedCountFunc)(const uint32_t* X,
                                                 const uint32_t* Y, int length);

static const int kC2 = 35468;



FILE* __cc_ExUtilSetBinaryMode(FILE* file) {






  return file;
}

int __cc_ExUtilReadFromStdin(const uint8_t** data, size_t* data_size) {
  static const size_t kBlockSize = 16384;
  size_t max_size = 0;
  size_t size = 0;
  uint8_t* input = ((void *)0);

  if (data == ((void *)0) || data_size == ((void *)0)) return 0;
  *data = ((void *)0);
  *data_size = 0;

  if (!__cc_ExUtilSetBinaryMode(stdin)) return 0;

  while (!feof(stdin)) {

    const size_t extra_size = (max_size == 0) ? kBlockSize : max_size;
    void* const new_data = realloc(input, max_size + extra_size);
    if (new_data == ((void *)0)) goto Error;
    input = (uint8_t*)new_data;
    max_size += extra_size;
    size += fread(input + size, 1, extra_size, stdin);
    if (size < max_size) break;
  }
  if (ferror(stdin)) goto Error;
  *data = input;
  *data_size = size;
  return 1;

 Error:
  free(input);
  fprintf(stderr, "Could not read from stdin\n");
  return 0;
}

int __cc_ExUtilReadFile(const char* const file_name,
                   const uint8_t** data, size_t* data_size) {
  int ok;
  void* file_data;
  size_t file_size;
  FILE* in;
  const int from_stdin = (file_name == ((void *)0)) || !strcmp(file_name, "-");

  if (from_stdin) return __cc_ExUtilReadFromStdin(data, data_size);

  if (data == ((void *)0) || data_size == ((void *)0)) return 0;
  *data = ((void *)0);
  *data_size = 0;

  in = fopen(file_name, "rb");
  if (in == ((void *)0)) {
    fprintf(stderr, "cannot open input file '%s'\n", file_name);
    return 0;
  }
  fseek(in, 0, 2);
  file_size = ftell(in);
  fseek(in, 0, 0);
  file_data = malloc(file_size);
  if (file_data == ((void *)0)) return 0;
  ok = (fread(file_data, file_size, 1, in) == 1);
  fclose(in);

  if (!ok) {
    fprintf(stderr, "Could not read %d bytes of data from file %s\n",
            (int)file_size, file_name);
    free(file_data);
    return 0;
  }
  *data = (uint8_t*)file_data;
  *data_size = file_size;
  return 1;
}

static const char* const kStatusMessages[VP8_STATUS_NOT_ENOUGH_DATA + 1] = {
  "OK", "OUT_OF_MEMORY", "INVALID_PARAM", "BITSTREAM_ERROR",
  "UNSUPPORTED_FEATURE", "SUSPENDED", "USER_ABORT", "NOT_ENOUGH_DATA"
};

void __cc_ExUtilPrintWebPError(const char* const in_file, int status) {
  fprintf(stderr, "Decoding of %s failed.\n", in_file);
  fprintf(stderr, "Status: %d", status);
  if (status >= VP8_STATUS_OK && status <= VP8_STATUS_NOT_ENOUGH_DATA) {
    fprintf(stderr, "(%s)", kStatusMessages[status]);
  }
  fprintf(stderr, "\n");
}

int __cc_WebPPictureImportRGBA(__cc_WebPPicture* picture,
                          const uint8_t* rgba, int rgba_stride) {
  return (picture != ((void *)0)) ? __cc_Import(picture, rgba, rgba_stride, 4, 0, 1) : 0;
}

void __cc_VP8LInitBitReader(__cc_VP8LBitReader* const br,
                       const uint8_t* const start,
                       size_t length);

static int __cc_AllocateInternalBuffers8b(__cc_VP8LDecoder* const dec) {
  const uint64_t total_num_pixels = (uint64_t)dec->width_ * dec->height_;
  dec->argb_cache_ = ((void *)0);
  dec->pixels_ = (uint32_t*)__cc_WebPSafeMalloc(total_num_pixels, sizeof(uint8_t));
  if (dec->pixels_ == ((void *)0)) {
    dec->status_ = VP8_STATUS_OUT_OF_MEMORY;
    return 0;
  }
  return 1;
}

static inline uint32_t __cc_VP8LColorCacheLookup(
    const __cc_VP8LColorCache* const cc, uint32_t key) {
  ((key <= (~0U >> cc->hash_shift_)) ? (void) (0) : __assert_fail ("key <= (~0U >> cc->hash_shift_)", "extraction.1.c", 7306, __PRETTY_FUNCTION__));
  return cc->colors_[key];
}

static int __cc_DecodeImageData(__cc_VP8LDecoder* const dec, uint32_t* const data,
                           int width, int height, int last_row,
                           __cc_ProcessRowsFunc process_func) {
  int ok = 1;
  int row = dec->last_pixel_ / width;
  int col = dec->last_pixel_ % width;
  __cc_VP8LBitReader* const br = &dec->br_;
  __cc_VP8LMetadata* const hdr = &dec->hdr_;
  __cc_HTreeGroup* htree_group = __cc_GetHtreeGroupForPos(hdr, col, row);
  uint32_t* src = data + dec->last_pixel_;
  uint32_t* last_cached = src;
  uint32_t* const src_end = data + width * height;
  uint32_t* const src_last = data + width * last_row;
  const int len_code_limit = 256 + 24;
  const int color_cache_limit = len_code_limit + hdr->color_cache_size_;
  __cc_VP8LColorCache* const color_cache =
      (hdr->color_cache_size_ > 0) ? &hdr->color_cache_ : ((void *)0);
  const int mask = hdr->huffman_mask_;
  ((htree_group != ((void *)0)) ? (void) (0) : __assert_fail ("htree_group != ((void *)0)", "extraction.1.c", 7328, __PRETTY_FUNCTION__));
  ((src < src_end) ? (void) (0) : __assert_fail ("src < src_end", "extraction.1.c", 7329, __PRETTY_FUNCTION__));
  ((src_last <= src_end) ? (void) (0) : __assert_fail ("src_last <= src_end", "extraction.1.c", 7330, __PRETTY_FUNCTION__));

  while (!br->eos_ && src < src_last) {
    int code;



    if ((col & mask) == 0) {
      htree_group = __cc_GetHtreeGroupForPos(hdr, col, row);
    }
    __cc_VP8LFillBitWindow(br);
    code = __cc_ReadSymbol(&htree_group->htrees_[GREEN], br);
    if (code < 256) {
      int red, green, blue, alpha;
      red = __cc_ReadSymbol(&htree_group->htrees_[RED], br);
      green = code;
      __cc_VP8LFillBitWindow(br);
      blue = __cc_ReadSymbol(&htree_group->htrees_[BLUE], br);
      alpha = __cc_ReadSymbol(&htree_group->htrees_[ALPHA], br);
      *src = ((uint32_t)alpha << 24) | (red << 16) | (green << 8) | blue;
    AdvanceByOne:
      ++src;
      ++col;
      if (col >= width) {
        col = 0;
        ++row;
        if ((row % 16 == 0) && (process_func != ((void *)0))) {
          process_func(dec, row);
        }
        if (color_cache != ((void *)0)) {
          while (last_cached < src) {
            __cc_VP8LColorCacheInsert(color_cache, *last_cached++);
          }
        }
      }
    } else if (code < len_code_limit) {
      int dist_code, dist;
      const int length_sym = code - 256;
      const int length = __cc_GetCopyLength(length_sym, br);
      const int dist_symbol = __cc_ReadSymbol(&htree_group->htrees_[DIST], br);
      __cc_VP8LFillBitWindow(br);
      dist_code = __cc_GetCopyDistance(dist_symbol, br);
      dist = __cc_PlaneCodeToDistance(width, dist_code);
      if (src - data < (ptrdiff_t)dist || src_end - src < (ptrdiff_t)length) {
        ok = 0;
        goto End;
      } else {
        int i;
        for (i = 0; i < length; ++i) src[i] = src[i - dist];
        src += length;
      }
      col += length;
      while (col >= width) {
        col -= width;
        ++row;
        if ((row % 16 == 0) && (process_func != ((void *)0))) {
          process_func(dec, row);
        }
      }
      if (src < src_end) {
        if (col & mask) htree_group = __cc_GetHtreeGroupForPos(hdr, col, row);
        if (color_cache != ((void *)0)) {
          while (last_cached < src) {
            __cc_VP8LColorCacheInsert(color_cache, *last_cached++);
          }
        }
      }
    } else if (code < color_cache_limit) {
      const int key = code - len_code_limit;
      ((color_cache != ((void *)0)) ? (void) (0) : __assert_fail ("color_cache != ((void *)0)", "extraction.1.c", 7399, __PRETTY_FUNCTION__));
      while (last_cached < src) {
        __cc_VP8LColorCacheInsert(color_cache, *last_cached++);
      }
      *src = __cc_VP8LColorCacheLookup(color_cache, key);
      goto AdvanceByOne;
    } else {
      ok = 0;
      goto End;
    }
    ((br->eos_ == __cc_VP8LIsEndOfStream(br)) ? (void) (0) : __assert_fail ("br->eos_ == VP8LIsEndOfStream(br)", "extraction.1.c", 7409, __PRETTY_FUNCTION__));
    ok = !br->error_;
    if (!ok) goto End;
  }

  if (process_func != ((void *)0)) process_func(dec, row);

 End:
  if (br->error_ || !ok || (br->eos_ && src < src_end)) {
    ok = 0;
    dec->status_ = br->eos_ ? VP8_STATUS_SUSPENDED
                            : VP8_STATUS_BITSTREAM_ERROR;
  } else {
    dec->last_pixel_ = (int)(src - data);
    if (src == src_end) dec->state_ = READ_DATA;
  }
  return ok;
}

int __cc_VP8LHuffmanTreeBuildImplicit(__cc_HuffmanTree* const tree,
                                 const int* const code_lengths,
                                 int* const huff_codes,
                                 int code_lengths_size);

int __cc_VP8LHuffmanTreeBuildExplicit(__cc_HuffmanTree* const tree,
                                 const int* const code_lengths,
                                 const int* const codes,
                                 const int* const symbols, int max_symbol,
                                 int num_symbols);

void __cc_VP8LHuffmanTreeFree(__cc_HuffmanTree* const tree);

static const int kCodeLengthLiterals = 16;

static const int kCodeLengthExtraBits[3] = { 2, 3, 7 };



static int __cc_ReadHuffmanCodeLengths(
    __cc_VP8LDecoder* const dec, const int* const code_length_code_lengths,
    int num_symbols, int* const code_lengths) {
  int ok = 0;
  __cc_VP8LBitReader* const br = &dec->br_;
  int symbol;
  int max_symbol;
  int prev_code_len = 8;
  __cc_HuffmanTree tree;
  int huff_codes[19] = { 0 };

  if (!__cc_VP8LHuffmanTreeBuildImplicit(&tree, code_length_code_lengths,
                                    huff_codes, 19)) {
    dec->status_ = VP8_STATUS_BITSTREAM_ERROR;
    return 0;
  }

  if (__cc_VP8LReadBits(br, 1)) {
    const int length_nbits = 2 + 2 * __cc_VP8LReadBits(br, 3);
    max_symbol = 2 + __cc_VP8LReadBits(br, length_nbits);
    if (max_symbol > num_symbols) {
      dec->status_ = VP8_STATUS_BITSTREAM_ERROR;
      goto End;
    }
  } else {
    max_symbol = num_symbols;
  }

  symbol = 0;
  while (symbol < num_symbols) {
    int code_len;
    if (max_symbol-- == 0) break;
    __cc_VP8LFillBitWindow(br);
    code_len = __cc_ReadSymbol(&tree, br);
    if (code_len < kCodeLengthLiterals) {
      code_lengths[symbol++] = code_len;
      if (code_len != 0) prev_code_len = code_len;
    } else {
      const int use_prev = (code_len == kCodeLengthRepeatCode);
      const int slot = code_len - kCodeLengthLiterals;
      const int extra_bits = kCodeLengthExtraBits[slot];
      const int repeat_offset = kCodeLengthRepeatOffsets[slot];
      int repeat = __cc_VP8LReadBits(br, extra_bits) + repeat_offset;
      if (symbol + repeat > num_symbols) {
        dec->status_ = VP8_STATUS_BITSTREAM_ERROR;
        goto End;
      } else {
        const int length = use_prev ? prev_code_len : 0;
        while (repeat-- > 0) code_lengths[symbol++] = length;
      }
    }
  }
  ok = 1;

 End:
  __cc_VP8LHuffmanTreeFree(&tree);
  if (!ok) dec->status_ = VP8_STATUS_BITSTREAM_ERROR;
  return ok;
}

static int __cc_ReadHuffmanCode(int alphabet_size, __cc_VP8LDecoder* const dec,
                           int* const code_lengths, int* const huff_codes,
                           __cc_HuffmanTree* const tree) {
  int ok = 0;
  __cc_VP8LBitReader* const br = &dec->br_;
  const int simple_code = __cc_VP8LReadBits(br, 1);

  if (simple_code) {
    int symbols[2];
    int codes[2];
    const int num_symbols = __cc_VP8LReadBits(br, 1) + 1;
    const int first_symbol_len_code = __cc_VP8LReadBits(br, 1);

    symbols[0] = __cc_VP8LReadBits(br, (first_symbol_len_code == 0) ? 1 : 8);
    codes[0] = 0;
    code_lengths[0] = num_symbols - 1;

    if (num_symbols == 2) {
      symbols[1] = __cc_VP8LReadBits(br, 8);
      codes[1] = 1;
      code_lengths[1] = num_symbols - 1;
    }
    ok = __cc_VP8LHuffmanTreeBuildExplicit(tree, code_lengths, codes, symbols,
                                      alphabet_size, num_symbols);
  } else {
    int i;
    int code_length_code_lengths[19] = { 0 };
    const int num_codes = __cc_VP8LReadBits(br, 4) + 4;
    if (num_codes > 19) {
      dec->status_ = VP8_STATUS_BITSTREAM_ERROR;
      return 0;
    }

    memset(code_lengths, 0, alphabet_size * sizeof(*code_lengths));

    for (i = 0; i < num_codes; ++i) {
      code_length_code_lengths[kCodeLengthCodeOrder[i]] = __cc_VP8LReadBits(br, 3);
    }
    ok = __cc_ReadHuffmanCodeLengths(dec, code_length_code_lengths, alphabet_size,
                                code_lengths);
    ok = ok && __cc_VP8LHuffmanTreeBuildImplicit(tree, code_lengths, huff_codes,
                                            alphabet_size);
  }
  ok = ok && !br->error_;
  if (!ok) {
    dec->status_ = VP8_STATUS_BITSTREAM_ERROR;
    return 0;
  }
  return 1;
}

__cc_HTreeGroup* __cc_VP8LHtreeGroupsNew(int num_htree_groups);

static int __cc_ReadHuffmanCodes(__cc_VP8LDecoder* const dec, int xsize, int ysize,
                            int color_cache_bits, int allow_recursion) {
  int i, j;
  __cc_VP8LBitReader* const br = &dec->br_;
  __cc_VP8LMetadata* const hdr = &dec->hdr_;
  uint32_t* huffman_image = ((void *)0);
  __cc_HTreeGroup* htree_groups = ((void *)0);
  int num_htree_groups = 1;
  int max_alphabet_size = 0;
  int* code_lengths = ((void *)0);
  int* huff_codes = ((void *)0);

  if (allow_recursion && __cc_VP8LReadBits(br, 1)) {

    const int huffman_precision = __cc_VP8LReadBits(br, 3) + 2;
    const int huffman_xsize = __cc_VP8LSubSampleSize(xsize, huffman_precision);
    const int huffman_ysize = __cc_VP8LSubSampleSize(ysize, huffman_precision);
    const int huffman_pixs = huffman_xsize * huffman_ysize;
    if (!__cc_DecodeImageStream(huffman_xsize, huffman_ysize, 0, dec,
                           &huffman_image)) {
      dec->status_ = VP8_STATUS_BITSTREAM_ERROR;
      goto Error;
    }
    hdr->huffman_subsample_bits_ = huffman_precision;
    for (i = 0; i < huffman_pixs; ++i) {

      const int group = (huffman_image[i] >> 8) & 0xffff;
      huffman_image[i] = group;
      if (group >= num_htree_groups) {
        num_htree_groups = group + 1;
      }
    }
  }

  if (br->error_) goto Error;


  for (j = 0; j < 5; ++j) {
    int alphabet_size = kAlphabetSize[j];
    if (j == 0 && color_cache_bits > 0) {
      alphabet_size += 1 << color_cache_bits;
    }
    if (max_alphabet_size < alphabet_size) {
      max_alphabet_size = alphabet_size;
    }
  }

  htree_groups = __cc_VP8LHtreeGroupsNew(num_htree_groups);
  code_lengths =
      (int*)__cc_WebPSafeCalloc((uint64_t)max_alphabet_size, sizeof(*code_lengths));
  huff_codes =
      (int*)__cc_WebPSafeMalloc((uint64_t)max_alphabet_size, sizeof(*huff_codes));

  if (htree_groups == ((void *)0) || code_lengths == ((void *)0) || huff_codes == ((void *)0)) {
    dec->status_ = VP8_STATUS_OUT_OF_MEMORY;
    goto Error;
  }

  for (i = 0; i < num_htree_groups; ++i) {
    __cc_HuffmanTree* const htrees = htree_groups[i].htrees_;
    for (j = 0; j < 5; ++j) {
      int alphabet_size = kAlphabetSize[j];
      __cc_HuffmanTree* const htree = htrees + j;
      if (j == 0 && color_cache_bits > 0) {
        alphabet_size += 1 << color_cache_bits;
      }
      if (!__cc_ReadHuffmanCode(alphabet_size, dec, code_lengths, huff_codes,
                           htree)) {
        goto Error;
      }
    }
  }
  __cc_WebPSafeFree(huff_codes);
  __cc_WebPSafeFree(code_lengths);


  hdr->huffman_image_ = huffman_image;
  hdr->num_htree_groups_ = num_htree_groups;
  hdr->htree_groups_ = htree_groups;
  return 1;

 Error:
  __cc_WebPSafeFree(huff_codes);
  __cc_WebPSafeFree(code_lengths);
  __cc_WebPSafeFree(huffman_image);
  __cc_VP8LHtreeGroupsFree(htree_groups, num_htree_groups);
  return 0;
}



static int __cc_DecodeImageStream(int xsize, int ysize,
                             int is_level0,
                             __cc_VP8LDecoder* const dec,
                             uint32_t** const decoded_data) {
  int ok = 1;
  int transform_xsize = xsize;
  int transform_ysize = ysize;
  __cc_VP8LBitReader* const br = &dec->br_;
  __cc_VP8LMetadata* const hdr = &dec->hdr_;
  uint32_t* data = ((void *)0);
  int color_cache_bits = 0;


  if (is_level0) {
    while (ok && __cc_VP8LReadBits(br, 1)) {
      ok = __cc_ReadTransform(&transform_xsize, &transform_ysize, dec);
    }
  }


  if (ok && __cc_VP8LReadBits(br, 1)) {
    color_cache_bits = __cc_VP8LReadBits(br, 4);
    ok = (color_cache_bits >= 1 && color_cache_bits <= 11);
    if (!ok) {
      dec->status_ = VP8_STATUS_BITSTREAM_ERROR;
      goto End;
    }
  }


  ok = ok && __cc_ReadHuffmanCodes(dec, transform_xsize, transform_ysize,
                              color_cache_bits, is_level0);
  if (!ok) {
    dec->status_ = VP8_STATUS_BITSTREAM_ERROR;
    goto End;
  }


  if (color_cache_bits > 0) {
    hdr->color_cache_size_ = 1 << color_cache_bits;
    if (!__cc_VP8LColorCacheInit(&hdr->color_cache_, color_cache_bits)) {
      dec->status_ = VP8_STATUS_OUT_OF_MEMORY;
      ok = 0;
      goto End;
    }
  } else {
    hdr->color_cache_size_ = 0;
  }
  __cc_UpdateDecoder(dec, transform_xsize, transform_ysize);

  if (is_level0) {
    dec->state_ = READ_HDR;
    goto End;
  }

  {
    const uint64_t total_size = (uint64_t)transform_xsize * transform_ysize;
    data = (uint32_t*)__cc_WebPSafeMalloc(total_size, sizeof(*data));
    if (data == ((void *)0)) {
      dec->status_ = VP8_STATUS_OUT_OF_MEMORY;
      ok = 0;
      goto End;
    }
  }


  ok = __cc_DecodeImageData(dec, data, transform_xsize, transform_ysize,
                       transform_ysize, ((void *)0));
  ok = ok && !br->error_;

 End:

  if (!ok) {
    __cc_WebPSafeFree(data);
    __cc_ClearMetadata(hdr);


    if (dec->status_ == VP8_STATUS_BITSTREAM_ERROR && dec->br_.eos_) {
      dec->status_ = VP8_STATUS_SUSPENDED;
    }
  } else {
    if (decoded_data != ((void *)0)) {
      *decoded_data = data;
    } else {


      ((data == ((void *)0)) ? (void) (0) : __assert_fail ("data == ((void *)0)", "extraction.1.c", 7737, __PRETTY_FUNCTION__));
      ((is_level0) ? (void) (0) : __assert_fail ("is_level0", "extraction.1.c", 7738, __PRETTY_FUNCTION__));
    }
    dec->last_pixel_ = 0;
    if (!is_level0) __cc_ClearMetadata(hdr);
  }
  return ok;
}

int __cc_VP8LDecodeAlphaHeader(__cc_ALPHDecoder* const alph_dec,
                          const uint8_t* const data, size_t data_size,
                          uint8_t* const output) {
  int ok = 0;
  __cc_VP8LDecoder* dec;
  __cc_VP8Io* io;
  ((alph_dec != ((void *)0)) ? (void) (0) : __assert_fail ("alph_dec != ((void *)0)", "extraction.1.c", 7752, __PRETTY_FUNCTION__));
  alph_dec->vp8l_dec_ = __cc_VP8LNew();
  if (alph_dec->vp8l_dec_ == ((void *)0)) return 0;
  dec = alph_dec->vp8l_dec_;

  dec->width_ = alph_dec->width_;
  dec->height_ = alph_dec->height_;
  dec->io_ = &alph_dec->io_;
  io = dec->io_;

  __cc_VP8InitIo(io);
  __cc_WebPInitCustomIo(((void *)0), io);
  io->opaque = output;
  io->width = alph_dec->width_;
  io->height = alph_dec->height_;

  dec->status_ = VP8_STATUS_OK;
  __cc_VP8LInitBitReader(&dec->br_, data, data_size);

  dec->action_ = READ_HDR;
  if (!__cc_DecodeImageStream(alph_dec->width_, alph_dec->height_, 1, dec, ((void *)0))) {
    goto Err;
  }




  if (dec->next_transform_ == 1 &&
      dec->transforms_[0].type_ == COLOR_INDEXING_TRANSFORM &&
      __cc_Is8bOptimizable(&dec->hdr_)) {
    alph_dec->use_8b_decode = 1;
    ok = __cc_AllocateInternalBuffers8b(dec);
  } else {

    alph_dec->use_8b_decode = 0;
    ok = __cc_AllocateInternalBuffers32b(dec, alph_dec->width_);
  }

  if (!ok) goto Err;

  dec->action_ = READ_DATA;
  return 1;

 Err:
  __cc_VP8LDelete(alph_dec->vp8l_dec_);
  alph_dec->vp8l_dec_ = ((void *)0);
  return 0;
}

static void __cc_DC16(uint8_t *dst) {
  int DC = 16;
  int j;
  for (j = 0; j < 16; ++j) {
    DC += dst[-1 + j * 32] + dst[j - 32];
  }
  __cc_Put16(DC >> 5, dst);
}

static inline int __cc_clip(int v, int M) {
  return v < 0 ? 0 : v > M ? M : v;
}

static const uint8_t kDcTable[128] = {
  4, 5, 6, 7, 8, 9, 10, 10,
  11, 12, 13, 14, 15, 16, 17, 17,
  18, 19, 20, 20, 21, 21, 22, 22,
  23, 23, 24, 25, 25, 26, 27, 28,
  29, 30, 31, 32, 33, 34, 35, 36,
  37, 37, 38, 39, 40, 41, 42, 43,
  44, 45, 46, 46, 47, 48, 49, 50,
  51, 52, 53, 54, 55, 56, 57, 58,
  59, 60, 61, 62, 63, 64, 65, 66,
  67, 68, 69, 70, 71, 72, 73, 74,
  75, 76, 76, 77, 78, 79, 80, 81,
  82, 83, 84, 85, 86, 87, 88, 89,
  91, 93, 95, 96, 98, 100, 101, 102,
  104, 106, 108, 110, 112, 114, 116, 118,
  122, 124, 126, 128, 130, 132, 134, 136,
  138, 140, 143, 145, 148, 151, 154, 157
};

static const uint16_t kAcTable[128] = {
  4, 5, 6, 7, 8, 9, 10, 11,
  12, 13, 14, 15, 16, 17, 18, 19,
  20, 21, 22, 23, 24, 25, 26, 27,
  28, 29, 30, 31, 32, 33, 34, 35,
  36, 37, 38, 39, 40, 41, 42, 43,
  44, 45, 46, 47, 48, 49, 50, 51,
  52, 53, 54, 55, 56, 57, 58, 60,
  62, 64, 66, 68, 70, 72, 74, 76,
  78, 80, 82, 84, 86, 88, 90, 92,
  94, 96, 98, 100, 102, 104, 106, 108,
  110, 112, 114, 116, 119, 122, 125, 128,
  131, 134, 137, 140, 143, 146, 149, 152,
  155, 158, 161, 164, 167, 170, 173, 177,
  181, 185, 189, 193, 197, 201, 205, 209,
  213, 217, 221, 225, 229, 234, 239, 245,
  249, 254, 259, 264, 269, 274, 279, 284
};

void __cc_VP8ParseQuant(__cc_VP8Decoder* const dec) {
  __cc_VP8BitReader* const br = &dec->br_;
  const int base_q0 = __cc_VP8GetValue(br, 7);
  const int dqy1_dc = __cc_VP8Get(br) ? __cc_VP8GetSignedValue(br, 4) : 0;
  const int dqy2_dc = __cc_VP8Get(br) ? __cc_VP8GetSignedValue(br, 4) : 0;
  const int dqy2_ac = __cc_VP8Get(br) ? __cc_VP8GetSignedValue(br, 4) : 0;
  const int dquv_dc = __cc_VP8Get(br) ? __cc_VP8GetSignedValue(br, 4) : 0;
  const int dquv_ac = __cc_VP8Get(br) ? __cc_VP8GetSignedValue(br, 4) : 0;

  const __cc_VP8SegmentHeader* const hdr = &dec->segment_hdr_;
  int i;

  for (i = 0; i < NUM_MB_SEGMENTS; ++i) {
    int q;
    if (hdr->use_segment_) {
      q = hdr->quantizer_[i];
      if (!hdr->absolute_delta_) {
        q += base_q0;
      }
    } else {
      if (i > 0) {
        dec->dqm_[i] = dec->dqm_[0];
        continue;
      } else {
        q = base_q0;
      }
    }
    {
      __cc_VP8QuantMatrix* const m = &dec->dqm_[i];
      m->y1_mat_[0] = kDcTable[__cc_clip(q + dqy1_dc, 127)];
      m->y1_mat_[1] = kAcTable[__cc_clip(q + 0, 127)];

      m->y2_mat_[0] = kDcTable[__cc_clip(q + dqy2_dc, 127)] * 2;



      m->y2_mat_[1] = (kAcTable[__cc_clip(q + dqy2_ac, 127)] * 101581) >> 16;
      if (m->y2_mat_[1] < 8) m->y2_mat_[1] = 8;

      m->uv_mat_[0] = kDcTable[__cc_clip(q + dquv_dc, 117)];
      m->uv_mat_[1] = kAcTable[__cc_clip(q + dquv_ac, 127)];

      m->uv_quant_ = q + dquv_ac;
    }
  }
}

static inline
int __cc_WebPRescalerInputDone(const __cc_WebPRescaler* const rescaler) {
  return (rescaler->src_y >= rescaler->src_height);
}

static void __cc_ImportRowExpandC(__cc_WebPRescaler* const wrk, const uint8_t* src) {
  const int x_stride = wrk->num_channels;
  const int x_out_max = wrk->dst_width * wrk->num_channels;
  int channel;
  ((!__cc_WebPRescalerInputDone(wrk)) ? (void) (0) : __assert_fail ("!WebPRescalerInputDone(wrk)", "extraction.1.c", 7908, __PRETTY_FUNCTION__));
  ((wrk->x_expand) ? (void) (0) : __assert_fail ("wrk->x_expand", "extraction.1.c", 7909, __PRETTY_FUNCTION__));
  for (channel = 0; channel < x_stride; ++channel) {
    int x_in = channel;
    int x_out = channel;

    int accum = wrk->x_add;
    int left = src[x_in];
    int right = (wrk->src_width > 1) ? src[x_in + x_stride] : left;
    x_in += x_stride;
    while (1) {
      wrk->frow[x_out] = right * wrk->x_add + (left - right) * accum;
      x_out += x_stride;
      if (x_out >= x_out_max) break;
      accum -= wrk->x_sub;
      if (accum < 0) {
        left = right;
        x_in += x_stride;
        ((x_in < wrk->src_width * x_stride) ? (void) (0) : __assert_fail ("x_in < wrk->src_width * x_stride", "extraction.1.c", 7926, __PRETTY_FUNCTION__));
        right = src[x_in];
        accum += wrk->x_add;
      }
    }
    ((wrk->x_sub == 0 || accum == 0) ? (void) (0) : __assert_fail ("wrk->x_sub == 0 || accum == 0", "extraction.1.c", 7931, __PRETTY_FUNCTION__));
  }
}

typedef void (*__cc_WebPRescalerImportRowFunc)(__cc_WebPRescaler* const wrk,
                                          const uint8_t* src);

static __cc_WebPRescalerImportRowFunc WebPRescalerImportRowExpand;

static __cc_WebPRescalerImportRowFunc WebPRescalerImportRowShrink;

void __cc_WebPRescalerImportRow(__cc_WebPRescaler* const wrk, const uint8_t* src) {
  ((!__cc_WebPRescalerInputDone(wrk)) ? (void) (0) : __assert_fail ("!WebPRescalerInputDone(wrk)", "extraction.1.c", 7943, __PRETTY_FUNCTION__));
  if (!wrk->x_expand) {
    WebPRescalerImportRowShrink(wrk, src);
  } else {
    WebPRescalerImportRowExpand(wrk, src);
  }
}

int __cc_WebPRescalerImport(__cc_WebPRescaler* const wrk, int num_lines,
                       const uint8_t* src, int src_stride) {
  int total_imported = 0;
  while (total_imported < num_lines && !__cc_WebPRescalerHasPendingOutput(wrk)) {
    if (wrk->y_expand) {
      __cc_rescaler_t* const tmp = wrk->irow;
      wrk->irow = wrk->frow;
      wrk->frow = tmp;
    }
    __cc_WebPRescalerImportRow(wrk, src);
    if (!wrk->y_expand) {
      int x;
      for (x = 0; x < wrk->num_channels * wrk->dst_width; ++x) {
        wrk->irow[x] += wrk->frow[x];
      }
    }
    ++wrk->src_y;
    src += src_stride;
    ++total_imported;
    wrk->y_accum -= wrk->y_sub;
  }
  return total_imported;
}

typedef void (*__cc_WebPRescalerExportRowFunc)(__cc_WebPRescaler* const wrk);
static void __cc_ImportRowShrinkC(__cc_WebPRescaler* const wrk, const uint8_t* src) {
  const int x_stride = wrk->num_channels;
  const int x_out_max = wrk->dst_width * wrk->num_channels;
  int channel;
  ((!__cc_WebPRescalerInputDone(wrk)) ? (void) (0) : __assert_fail ("!WebPRescalerInputDone(wrk)", "extraction.1.c", 7989, __PRETTY_FUNCTION__));
  ((!wrk->x_expand) ? (void) (0) : __assert_fail ("!wrk->x_expand", "extraction.1.c", 7990, __PRETTY_FUNCTION__));
  for (channel = 0; channel < x_stride; ++channel) {
    int x_in = channel;
    int x_out = channel;
    uint32_t sum = 0;
    int accum = 0;
    while (x_out < x_out_max) {
      uint32_t base = 0;
      accum += wrk->x_add;
      while (accum > 0) {
        accum -= wrk->x_sub;
        ((x_in < wrk->src_width * x_stride) ? (void) (0) : __assert_fail ("x_in < wrk->src_width * x_stride", "extraction.1.c", 8001, __PRETTY_FUNCTION__));
        base = src[x_in];
        sum += base;
        x_in += x_stride;
      }
      {
        const __cc_rescaler_t frac = base * (-accum);
        wrk->frow[x_out] = sum * wrk->x_sub - frac;

        sum = (int)(((uint64_t)(frac) * (wrk->fx_scale) + ((1ull << 32) >> 1)) >> 32);
      }
      x_out += x_stride;
    }
    ((accum == 0) ? (void) (0) : __assert_fail ("accum == 0", "extraction.1.c", 8014, __PRETTY_FUNCTION__));
  }
}

static __cc_WebPRescalerExportRowFunc WebPRescalerExportRowExpand;

__cc_VP8StatusCode __cc_WebPDecode(const uint8_t* data, size_t data_size,
                         __cc_WebPDecoderConfig* config) {
  __cc_WebPDecParams params;
  __cc_VP8StatusCode status;

  if (config == ((void *)0)) {
    return VP8_STATUS_INVALID_PARAM;
  }

  status = __cc_GetFeatures(data, data_size, &config->input);
  if (status != VP8_STATUS_OK) {
    if (status == VP8_STATUS_NOT_ENOUGH_DATA) {
      return VP8_STATUS_BITSTREAM_ERROR;
    }
    return status;
  }

  __cc_WebPResetDecParams(&params);
  params.output = &config->output;
  params.options = &config->options;
  status = __cc_DecodeInto(data, data_size, &params);

  return status;
}

int __cc_WebPEncodingSetError(const __cc_WebPPicture* const pic,
                         __cc_WebPEncodingError error) {
  (((int)error < VP8_ENC_ERROR_LAST) ? (void) (0) : __assert_fail ("(int)error < VP8_ENC_ERROR_LAST", "extraction.1.c", 8047, __PRETTY_FUNCTION__));
  (((int)error >= VP8_ENC_OK) ? (void) (0) : __assert_fail ("(int)error >= VP8_ENC_OK", "extraction.1.c", 8048, __PRETTY_FUNCTION__));
  ((__cc_WebPPicture*)pic)->error_code = error;
  return 0;
}

static void __cc_ShiftBytes(__cc_VP8LBitReader* const br) {
  while (br->bit_pos_ >= 8 && br->pos_ < br->len_) {
    br->val_ >>= 8;
    br->val_ |= ((__cc_vp8l_val_t)br->buf_[br->pos_]) << (64 - 8);
    ++br->pos_;
    br->bit_pos_ -= 8;
  }
  br->eos_ = __cc_VP8LIsEndOfStream(br);
}

__cc_VP8LProcessBlueAndRedFunc VP8LAddGreenToBlueAndRed;

void __cc_VP8LConvertBGRAToRGB565_C(const uint32_t* src,
                               int num_pixels, uint8_t* dst) {
  const uint32_t* const src_end = src + num_pixels;
  while (src < src_end) {
    const uint32_t argb = *src++;
    const uint8_t rg = ((argb >> 16) & 0xf8) | ((argb >> 13) & 0x7);
    const uint8_t gb = ((argb >> 5) & 0xe0) | ((argb >> 3) & 0x1f);




    *dst++ = rg;
    *dst++ = gb;

  }
}

void __cc_VP8LConvertBGRAToRGBA4444_C(const uint32_t* src,
                                 int num_pixels, uint8_t* dst) {
  const uint32_t* const src_end = src + num_pixels;
  while (src < src_end) {
    const uint32_t argb = *src++;
    const uint8_t rg = ((argb >> 16) & 0xf0) | ((argb >> 12) & 0xf);
    const uint8_t ba = ((argb >> 0) & 0xf0) | ((argb >> 28) & 0xf);




    *dst++ = rg;
    *dst++ = ba;

  }
}

__cc_VP8LConvertFunc VP8LConvertBGRAToRGBA;

static double __cc_ExtraCost(const uint32_t* population, int length) {
  int i;
  double cost = 0.;
  for (i = 2; i < length - 2; ++i) cost += (i >> 1) * population[i + 2];
  return cost;
}

static int __cc_is_big_endian(void) {
  static const union {
    uint16_t w;
    uint8_t b[2];
  } tmp = { 1 };
  return (tmp.b[0] != 1);
}

static void __cc_CopyOrSwap(const uint32_t* src, int num_pixels, uint8_t* dst,
                       int swap_on_big_endian) {
  if (__cc_is_big_endian() == swap_on_big_endian) {
    const uint32_t* const src_end = src + num_pixels;
    while (src < src_end) {
      const uint32_t argb = *src++;



      *(uint32_t*)dst = __cc_BSwap32(argb);
      dst += sizeof(argb);
    }
  } else {
    memcpy(dst, src, num_pixels * sizeof(*src));
  }
}

extern __cc_VP8LConvertFunc VP8LConvertBGRAToRGB565;

extern __cc_VP8LConvertFunc VP8LConvertBGRAToBGR;

extern __cc_VP8LConvertFunc VP8LConvertBGRAToRGB;

extern __cc_VP8LConvertFunc VP8LConvertBGRAToRGBA;

void __cc_VP8LConvertFromBGRA(const uint32_t* const in_data, int num_pixels,
                         __cc_WEBP_CSP_MODE out_colorspace, uint8_t* const rgba) {
  switch (out_colorspace) {
    case MODE_RGB:
      VP8LConvertBGRAToRGB(in_data, num_pixels, rgba);
      break;
    case MODE_RGBA:
      VP8LConvertBGRAToRGBA(in_data, num_pixels, rgba);
      break;
    case MODE_rgbA:
      VP8LConvertBGRAToRGBA(in_data, num_pixels, rgba);
      WebPApplyAlphaMultiply(rgba, 0, num_pixels, 1, 0);
      break;
    case MODE_BGR:
      VP8LConvertBGRAToBGR(in_data, num_pixels, rgba);
      break;
    case MODE_BGRA:
      __cc_CopyOrSwap(in_data, num_pixels, rgba, 1);
      break;
    case MODE_bgrA:
      __cc_CopyOrSwap(in_data, num_pixels, rgba, 1);
      WebPApplyAlphaMultiply(rgba, 0, num_pixels, 1, 0);
      break;
    case MODE_ARGB:
      __cc_CopyOrSwap(in_data, num_pixels, rgba, 0);
      break;
    case MODE_Argb:
      __cc_CopyOrSwap(in_data, num_pixels, rgba, 0);
      WebPApplyAlphaMultiply(rgba, 1, num_pixels, 1, 0);
      break;
    case MODE_RGBA_4444:
      VP8LConvertBGRAToRGBA4444(in_data, num_pixels, rgba);
      break;
    case MODE_rgbA_4444:
      VP8LConvertBGRAToRGBA4444(in_data, num_pixels, rgba);
      WebPApplyAlphaMultiply4444(rgba, num_pixels, 1, 0);
      break;
    case MODE_RGB_565:
      VP8LConvertBGRAToRGB565(in_data, num_pixels, rgba);
      break;
    default:
      ((0) ? (void) (0) : __assert_fail ("0", "extraction.1.c", 8194, __PRETTY_FUNCTION__));
  }
}

__cc_VP8LFastLog2SlowFunc VP8LFastSLog2Slow;

void __cc_VP8LConvertBGRAToBGR_C(const uint32_t* src,
                            int num_pixels, uint8_t* dst) {
  const uint32_t* const src_end = src + num_pixels;
  while (src < src_end) {
    const uint32_t argb = *src++;
    *dst++ = (argb >> 0) & 0xff;
    *dst++ = (argb >> 8) & 0xff;
    *dst++ = (argb >> 16) & 0xff;
  }
}

static inline int __cc_AddSubtractComponentFull(int a, int b, int c) {
  return __cc_Clip255(a + b - c);
}

static inline uint32_t __cc_ClampedAddSubtractFull(uint32_t c0, uint32_t c1,
                                                   uint32_t c2) {
  const int a = __cc_AddSubtractComponentFull(c0 >> 24, c1 >> 24, c2 >> 24);
  const int r = __cc_AddSubtractComponentFull((c0 >> 16) & 0xff,
                                         (c1 >> 16) & 0xff,
                                         (c2 >> 16) & 0xff);
  const int g = __cc_AddSubtractComponentFull((c0 >> 8) & 0xff,
                                         (c1 >> 8) & 0xff,
                                         (c2 >> 8) & 0xff);
  const int b = __cc_AddSubtractComponentFull(c0 & 0xff, c1 & 0xff, c2 & 0xff);
  return ((uint32_t)a << 24) | (r << 16) | (g << 8) | b;
}

static uint32_t __cc_Predictor12(uint32_t left, const uint32_t* const top) {
  const uint32_t pred = __cc_ClampedAddSubtractFull(left, top[0], top[-1]);
  return pred;
}

__cc_VP8LCostCombinedCountFunc VP8LHuffmanCostCombinedCount;

void __cc_VP8LSubtractGreenFromBlueAndRed_C(uint32_t* argb_data, int num_pixels) {
  int i;
  for (i = 0; i < num_pixels; ++i) {
    const uint32_t argb = argb_data[i];
    const uint32_t green = (argb >> 8) & 0xff;
    const uint32_t new_r = (((argb >> 16) & 0xff) - green) & 0xff;
    const uint32_t new_b = ((argb & 0xff) - green) & 0xff;
    argb_data[i] = (argb & 0xff00ff00) | (new_r << 16) | new_b;
  }
}



static inline int __cc_Sub3(int a, int b, int c) {
  const int pb = b - c;
  const int pa = a - c;
  return abs(pb) - abs(pa);
}

static inline uint32_t __cc_Select(uint32_t a, uint32_t b, uint32_t c) {
  const int pa_minus_pb =
      __cc_Sub3((a >> 24) , (b >> 24) , (c >> 24) ) +
      __cc_Sub3((a >> 16) & 0xff, (b >> 16) & 0xff, (c >> 16) & 0xff) +
      __cc_Sub3((a >> 8) & 0xff, (b >> 8) & 0xff, (c >> 8) & 0xff) +
      __cc_Sub3((a ) & 0xff, (b ) & 0xff, (c ) & 0xff);
  return (pa_minus_pb <= 0) ? a : b;
}

static inline uint32_t __cc_ColorTransformDelta(int8_t color_pred,
                                                int8_t color) {
  return (uint32_t)((int)(color_pred) * color) >> 5;
}

static inline uint32_t __cc_Average2(uint32_t a0, uint32_t a1) {
  return (((a0 ^ a1) & 0xfefefefeL) >> 1) + (a0 & a1);
}

static uint32_t __cc_Predictor9(uint32_t left, const uint32_t* const top) {
  const uint32_t pred = __cc_Average2(top[0], top[1]);
  (void)left;
  return pred;
}

static inline int __cc_needs_filter2(const uint8_t* p,
                                     int step, int t, int it) {
  const int p3 = p[-4 * step], p2 = p[-3 * step], p1 = p[-2 * step];
  const int p0 = p[-step], q0 = p[0];
  const int q1 = p[step], q2 = p[2 * step], q3 = p[3 * step];
  if ((4 * VP8kabs0[p0 - q0] + VP8kabs0[p1 - q1]) > t) return 0;
  return VP8kabs0[p3 - p2] <= it && VP8kabs0[p2 - p1] <= it &&
         VP8kabs0[p1 - p0] <= it && VP8kabs0[q3 - q2] <= it &&
         VP8kabs0[q2 - q1] <= it && VP8kabs0[q1 - q0] <= it;
}



static __cc_WebPRescalerExportRowFunc WebPRescalerExportRowShrink;

void __cc_WebPRescalerExportRow(__cc_WebPRescaler* const wrk) {
  if (wrk->y_accum <= 0) {
    ((!__cc_WebPRescalerOutputDone(wrk)) ? (void) (0) : __assert_fail ("!WebPRescalerOutputDone(wrk)", "extraction.1.c", 8295, __PRETTY_FUNCTION__));
    if (wrk->y_expand) {
      WebPRescalerExportRowExpand(wrk);
    } else if (wrk->fxy_scale) {
      WebPRescalerExportRowShrink(wrk);
    } else {
      int i;
      ((wrk->src_width == 1 && wrk->dst_width <= 2) ? (void) (0) : __assert_fail ("wrk->src_width == 1 && wrk->dst_width <= 2", "extraction.1.c", 8302, __PRETTY_FUNCTION__));
      ((wrk->src_height == 1 && wrk->dst_height == 1) ? (void) (0) : __assert_fail ("wrk->src_height == 1 && wrk->dst_height == 1", "extraction.1.c", 8303, __PRETTY_FUNCTION__));
      for (i = 0; i < wrk->num_channels * wrk->dst_width; ++i) {
        wrk->dst[i] = wrk->irow[i];
        wrk->irow[i] = 0;
      }
    }
    wrk->y_accum += wrk->y_add;
    wrk->dst += wrk->dst_stride;
    ++wrk->dst_y;
  }
}

static inline uint32_t __cc_Average3(uint32_t a0, uint32_t a1, uint32_t a2) {
  return __cc_Average2(__cc_Average2(a0, a2), a1);
}

static uint32_t __cc_Predictor5(uint32_t left, const uint32_t* const top) {
  const uint32_t pred = __cc_Average3(left, top[0], top[1]);
  return pred;
}

__cc_VP8LConvertFunc VP8LConvertBGRAToBGR;

static uint32_t __cc_Predictor4(uint32_t left, const uint32_t* const top) {
  (void)left;
  return top[-1];
}

static inline uint8_t __cc_GetAlphaValue(uint32_t val) {
  return (val >> 8) & 0xff;
}

static inline uint8_t __cc_GetAlphaIndex(uint8_t idx) {
  return idx;
}

void __cc_VP8LColorIndexInverseTransformAlpha(const __cc_VP8LTransform* const transform, int y_start, int y_end, const uint8_t* src, uint8_t* dst) { int y; const int bits_per_pixel = 8 >> transform->bits_; const int width = transform->xsize_; const uint32_t* const color_map = transform->data_; if (bits_per_pixel < 8) { const int pixels_per_byte = 1 << transform->bits_; const int count_mask = pixels_per_byte - 1; const uint32_t bit_mask = (1 << bits_per_pixel) - 1; for (y = y_start; y < y_end; ++y) { uint32_t packed_pixels = 0; int x; for (x = 0; x < width; ++x) { if ((x & count_mask) == 0) packed_pixels = __cc_GetAlphaIndex(*src++); *dst++ = __cc_GetAlphaValue(color_map[packed_pixels & bit_mask]); packed_pixels >>= bits_per_pixel; } } } else { for (y = y_start; y < y_end; ++y) { int x; for (x = 0; x < width; ++x) { *dst++ = __cc_GetAlphaValue(color_map[__cc_GetAlphaIndex(*src++)]); } } } }


__cc_VP8LConvertFunc VP8LConvertBGRAToRGB;

static void __cc_HistogramAdd(const __cc_VP8LHistogram* const a,
                         const __cc_VP8LHistogram* const b,
                         __cc_VP8LHistogram* const out) {
  int i;
  const int literal_size = __cc_VP8LHistogramNumCodes(a->palette_code_bits_);
  ((a->palette_code_bits_ == b->palette_code_bits_) ? (void) (0) : __assert_fail ("a->palette_code_bits_ == b->palette_code_bits_", "extraction.1.c", 8349, __PRETTY_FUNCTION__));
  if (b != out) {
    for (i = 0; i < literal_size; ++i) {
      out->literal_[i] = a->literal_[i] + b->literal_[i];
    }
    for (i = 0; i < 40; ++i) {
      out->distance_[i] = a->distance_[i] + b->distance_[i];
    }
    for (i = 0; i < 256; ++i) {
      out->red_[i] = a->red_[i] + b->red_[i];
      out->blue_[i] = a->blue_[i] + b->blue_[i];
      out->alpha_[i] = a->alpha_[i] + b->alpha_[i];
    }
  } else {
    for (i = 0; i < literal_size; ++i) {
      out->literal_[i] += a->literal_[i];
    }
    for (i = 0; i < 40; ++i) {
      out->distance_[i] += a->distance_[i];
    }
    for (i = 0; i < 256; ++i) {
      out->red_[i] += a->red_[i];
      out->blue_[i] += a->blue_[i];
      out->alpha_[i] += a->alpha_[i];
    }
  }
}

__cc_VP8LHistogramAddFunc VP8LHistogramAdd;

static __cc_VP8LStreaks __cc_HuffmanCostCombinedCount(const uint32_t* X,
                                            const uint32_t* Y, int length) {
  int i;
  int streak = 0;
  __cc_VP8LStreaks stats;
  memset(&stats, 0, sizeof(stats));
  for (i = 0; i < length - 1; ++i) {
    const int xy = X[i] + Y[i];
    const int xy_next = X[i + 1] + Y[i + 1];
    ++streak;
    if (xy == xy_next) {
      continue;
    }
    stats.counts[xy != 0] += (streak > 3);
    stats.streaks[xy != 0][(streak > 3)] += streak;
    streak = 0;
  }
  {
    const int xy = X[i] + Y[i];
    ++streak;
    stats.counts[xy != 0] += (streak > 3);
    stats.streaks[xy != 0][(streak > 3)] += streak;
  }
  return stats;
}

static __cc_VP8LStreaks __cc_HuffmanCostCount(const uint32_t* population, int length) {
  int i;
  int streak = 0;
  __cc_VP8LStreaks stats;
  memset(&stats, 0, sizeof(stats));
  for (i = 0; i < length - 1; ++i) {
    ++streak;
    if (population[i] == population[i + 1]) {
      continue;
    }
    stats.counts[population[i] != 0] += (streak > 3);
    stats.streaks[population[i] != 0][(streak > 3)] += streak;
    streak = 0;
  }
  ++streak;
  stats.counts[population[i] != 0] += (streak > 3);
  stats.streaks[population[i] != 0][(streak > 3)] += streak;
  return stats;
}

__cc_VP8LCostCountFunc VP8LHuffmanCostCount;

__cc_VP8LCostFunc VP8LExtraCost;



const float kLog2Table[256] = {
  0.0000000000000000f, 0.0000000000000000f,
  1.0000000000000000f, 1.5849625007211560f,
  2.0000000000000000f, 2.3219280948873621f,
  2.5849625007211560f, 2.8073549220576041f,
  3.0000000000000000f, 3.1699250014423121f,
  3.3219280948873621f, 3.4594316186372973f,
  3.5849625007211560f, 3.7004397181410921f,
  3.8073549220576041f, 3.9068905956085187f,
  4.0000000000000000f, 4.0874628412503390f,
  4.1699250014423121f, 4.2479275134435852f,
  4.3219280948873626f, 4.3923174227787606f,
  4.4594316186372973f, 4.5235619560570130f,
  4.5849625007211560f, 4.6438561897747243f,
  4.7004397181410917f, 4.7548875021634682f,
  4.8073549220576037f, 4.8579809951275718f,
  4.9068905956085187f, 4.9541963103868749f,
  5.0000000000000000f, 5.0443941193584533f,
  5.0874628412503390f, 5.1292830169449663f,
  5.1699250014423121f, 5.2094533656289501f,
  5.2479275134435852f, 5.2854022188622487f,
  5.3219280948873626f, 5.3575520046180837f,
  5.3923174227787606f, 5.4262647547020979f,
  5.4594316186372973f, 5.4918530963296747f,
  5.5235619560570130f, 5.5545888516776376f,
  5.5849625007211560f, 5.6147098441152083f,
  5.6438561897747243f, 5.6724253419714951f,
  5.7004397181410917f, 5.7279204545631987f,
  5.7548875021634682f, 5.7813597135246599f,
  5.8073549220576037f, 5.8328900141647412f,
  5.8579809951275718f, 5.8826430493618415f,
  5.9068905956085187f, 5.9307373375628866f,
  5.9541963103868749f, 5.9772799234999167f,
  6.0000000000000000f, 6.0223678130284543f,
  6.0443941193584533f, 6.0660891904577720f,
  6.0874628412503390f, 6.1085244567781691f,
  6.1292830169449663f, 6.1497471195046822f,
  6.1699250014423121f, 6.1898245588800175f,
  6.2094533656289501f, 6.2288186904958804f,
  6.2479275134435852f, 6.2667865406949010f,
  6.2854022188622487f, 6.3037807481771030f,
  6.3219280948873626f, 6.3398500028846243f,
  6.3575520046180837f, 6.3750394313469245f,
  6.3923174227787606f, 6.4093909361377017f,
  6.4262647547020979f, 6.4429434958487279f,
  6.4594316186372973f, 6.4757334309663976f,
  6.4918530963296747f, 6.5077946401986963f,
  6.5235619560570130f, 6.5391588111080309f,
  6.5545888516776376f, 6.5698556083309478f,
  6.5849625007211560f, 6.5999128421871278f,
  6.6147098441152083f, 6.6293566200796094f,
  6.6438561897747243f, 6.6582114827517946f,
  6.6724253419714951f, 6.6865005271832185f,
  6.7004397181410917f, 6.7142455176661224f,
  6.7279204545631987f, 6.7414669864011464f,
  6.7548875021634682f, 6.7681843247769259f,
  6.7813597135246599f, 6.7944158663501061f,
  6.8073549220576037f, 6.8201789624151878f,
  6.8328900141647412f, 6.8454900509443747f,
  6.8579809951275718f, 6.8703647195834047f,
  6.8826430493618415f, 6.8948177633079437f,
  6.9068905956085187f, 6.9188632372745946f,
  6.9307373375628866f, 6.9425145053392398f,
  6.9541963103868749f, 6.9657842846620869f,
  6.9772799234999167f, 6.9886846867721654f,
  7.0000000000000000f, 7.0112272554232539f,
  7.0223678130284543f, 7.0334230015374501f,
  7.0443941193584533f, 7.0552824355011898f,
  7.0660891904577720f, 7.0768155970508308f,
  7.0874628412503390f, 7.0980320829605263f,
  7.1085244567781691f, 7.1189410727235076f,
  7.1292830169449663f, 7.1395513523987936f,
  7.1497471195046822f, 7.1598713367783890f,
  7.1699250014423121f, 7.1799090900149344f,
  7.1898245588800175f, 7.1996723448363644f,
  7.2094533656289501f, 7.2191685204621611f,
  7.2288186904958804f, 7.2384047393250785f,
  7.2479275134435852f, 7.2573878426926521f,
  7.2667865406949010f, 7.2761244052742375f,
  7.2854022188622487f, 7.2946207488916270f,
  7.3037807481771030f, 7.3128829552843557f,
  7.3219280948873626f, 7.3309168781146167f,
  7.3398500028846243f, 7.3487281542310771f,
  7.3575520046180837f, 7.3663222142458160f,
  7.3750394313469245f, 7.3837042924740519f,
  7.3923174227787606f, 7.4008794362821843f,
  7.4093909361377017f, 7.4178525148858982f,
  7.4262647547020979f, 7.4346282276367245f,
  7.4429434958487279f, 7.4512111118323289f,
  7.4594316186372973f, 7.4676055500829976f,
  7.4757334309663976f, 7.4838157772642563f,
  7.4918530963296747f, 7.4998458870832056f,
  7.5077946401986963f, 7.5156998382840427f,
  7.5235619560570130f, 7.5313814605163118f,
  7.5391588111080309f, 7.5468944598876364f,
  7.5545888516776376f, 7.5622424242210728f,
  7.5698556083309478f, 7.5774288280357486f,
  7.5849625007211560f, 7.5924570372680806f,
  7.5999128421871278f, 7.6073303137496104f,
  7.6147098441152083f, 7.6220518194563764f,
  7.6293566200796094f, 7.6366246205436487f,
  7.6438561897747243f, 7.6510516911789281f,
  7.6582114827517946f, 7.6653359171851764f,
  7.6724253419714951f, 7.6794800995054464f,
  7.6865005271832185f, 7.6934869574993252f,
  7.7004397181410917f, 7.7073591320808825f,
  7.7142455176661224f, 7.7210991887071855f,
  7.7279204545631987f, 7.7347096202258383f,
  7.7414669864011464f, 7.7481928495894605f,
  7.7548875021634682f, 7.7615512324444795f,
  7.7681843247769259f, 7.7747870596011736f,
  7.7813597135246599f, 7.7879025593914317f,
  7.7944158663501061f, 7.8008998999203047f,
  7.8073549220576037f, 7.8137811912170374f,
  7.8201789624151878f, 7.8265484872909150f,
  7.8328900141647412f, 7.8392037880969436f,
  7.8454900509443747f, 7.8517490414160571f,
  7.8579809951275718f, 7.8641861446542797f,
  7.8703647195834047f, 7.8765169465649993f,
  7.8826430493618415f, 7.8887432488982591f,
  7.8948177633079437f, 7.9008668079807486f,
  7.9068905956085187f, 7.9128893362299619f,
  7.9188632372745946f, 7.9248125036057812f,
  7.9307373375628866f, 7.9366379390025709f,
  7.9425145053392398f, 7.9483672315846778f,
  7.9541963103868749f, 7.9600019320680805f,
  7.9657842846620869f, 7.9715435539507719f,
  7.9772799234999167f, 7.9829935746943103f,
  7.9886846867721654f, 7.9943534368588577f
};

static float __cc_FastSLog2Slow(uint32_t v) {
  ((v >= 256) ? (void) (0) : __assert_fail ("v >= 256", "extraction.1.c", 8563, __PRETTY_FUNCTION__));
  if (v < 65536) {
    int log_cnt = 0;
    uint32_t y = 1;
    int correction = 0;
    const float v_f = (float)v;
    const uint32_t orig_v = v;
    do {
      ++log_cnt;
      v = v >> 1;
      y = y << 1;
    } while (v >= 256);






    correction = (23 * (orig_v & (y - 1))) >> 4;
    return v_f * (kLog2Table[v] + log_cnt) + correction;
  } else {
    return (float)(1.44269504088896338700465094007086 * v * log((double)v));
  }
}



static float __cc_FastLog2Slow(uint32_t v) {
  ((v >= 256) ? (void) (0) : __assert_fail ("v >= 256", "extraction.1.c", 8591, __PRETTY_FUNCTION__));
  if (v < 65536) {
    int log_cnt = 0;
    uint32_t y = 1;
    const uint32_t orig_v = v;
    double log_2;
    do {
      ++log_cnt;
      v = v >> 1;
      y = y << 1;
    } while (v >= 256);
    log_2 = kLog2Table[v] + log_cnt;
    if (orig_v >= 4096) {


      const int correction = (23 * (orig_v & (y - 1))) >> 4;
      log_2 += (double)correction / orig_v;
    }
    return (float)log_2;
  } else {
    return (float)(1.44269504088896338700465094007086 * log((double)v));
  }
}

__cc_VP8LCostCombinedFunc VP8LExtraCostCombined;

void __cc_VP8LConvertBGRAToRGBA_C(const uint32_t* src,
                             int num_pixels, uint8_t* dst) {
  const uint32_t* const src_end = src + num_pixels;
  while (src < src_end) {
    const uint32_t argb = *src++;
    *dst++ = (argb >> 16) & 0xff;
    *dst++ = (argb >> 8) & 0xff;
    *dst++ = (argb >> 0) & 0xff;
    *dst++ = (argb >> 24) & 0xff;
  }
}

__cc_VP8LFastLog2SlowFunc VP8LFastLog2Slow;

__cc_VP8LProcessBlueAndRedFunc VP8LSubtractGreenFromBlueAndRed;

void __cc_VP8LConvertBGRAToRGB_C(const uint32_t* src,
                            int num_pixels, uint8_t* dst) {
  const uint32_t* const src_end = src + num_pixels;
  while (src < src_end) {
    const uint32_t argb = *src++;
    *dst++ = (argb >> 16) & 0xff;
    *dst++ = (argb >> 8) & 0xff;
    *dst++ = (argb >> 0) & 0xff;
  }
}

__cc_VP8LConvertFunc VP8LConvertBGRAToRGB565;

void __cc_VP8LTransformColor_C(const __cc_VP8LMultipliers* const m, uint32_t* data,
                          int num_pixels) {
  int i;
  for (i = 0; i < num_pixels; ++i) {
    const uint32_t argb = data[i];
    const uint32_t green = argb >> 8;
    const uint32_t red = argb >> 16;
    uint32_t new_red = red;
    uint32_t new_blue = argb;
    new_red -= __cc_ColorTransformDelta(m->green_to_red_, green);
    new_red &= 0xff;
    new_blue -= __cc_ColorTransformDelta(m->green_to_blue_, green);
    new_blue -= __cc_ColorTransformDelta(m->red_to_blue_, red);
    new_blue &= 0xff;
    data[i] = (argb & 0xff00ff00u) | (new_red << 16) | (new_blue);
  }
}

__cc_VP8LTransformColorFunc VP8LTransformColor;

__cc_VP8LPredictorFunc VP8LPredictors[16];

void __cc_VP8LTransformColorInverse_C(const __cc_VP8LMultipliers* const m, uint32_t* data,
                                 int num_pixels) {
  int i;
  for (i = 0; i < num_pixels; ++i) {
    const uint32_t argb = data[i];
    const uint32_t green = argb >> 8;
    const uint32_t red = argb >> 16;
    uint32_t new_red = red;
    uint32_t new_blue = argb;
    new_red += __cc_ColorTransformDelta(m->green_to_red_, green);
    new_red &= 0xff;
    new_blue += __cc_ColorTransformDelta(m->green_to_blue_, green);
    new_blue += __cc_ColorTransformDelta(m->red_to_blue_, new_red);
    new_blue &= 0xff;
    data[i] = (argb & 0xff00ff00u) | (new_red << 16) | (new_blue);
  }
}

void __cc_VP8LAddGreenToBlueAndRed_C(uint32_t* data, int num_pixels) {
  int i;
  for (i = 0; i < num_pixels; ++i) {
    const uint32_t argb = data[i];
    const uint32_t green = ((argb >> 8) & 0xff);
    uint32_t red_blue = (argb & 0x00ff00ffu);
    red_blue += (green << 16) | green;
    red_blue &= 0x00ff00ffu;
    data[i] = (argb & 0xff00ff00u) | red_blue;
  }
}

static inline uint32_t __cc_ClampedAddSubtractHalf(uint32_t c0, uint32_t c1,
                                                   uint32_t c2) {
  const uint32_t ave = __cc_Average2(c0, c1);
  const int a = __cc_AddSubtractComponentHalf(ave >> 24, c2 >> 24);
  const int r = __cc_AddSubtractComponentHalf((ave >> 16) & 0xff, (c2 >> 16) & 0xff);
  const int g = __cc_AddSubtractComponentHalf((ave >> 8) & 0xff, (c2 >> 8) & 0xff);
  const int b = __cc_AddSubtractComponentHalf((ave >> 0) & 0xff, (c2 >> 0) & 0xff);
  return ((uint32_t)a << 24) | (r << 16) | (g << 8) | b;
}

static uint32_t __cc_Predictor13(uint32_t left, const uint32_t* const top) {
  const uint32_t pred = __cc_ClampedAddSubtractHalf(left, top[0], top[-1]);
  return pred;
}

static uint32_t __cc_Predictor11(uint32_t left, const uint32_t* const top) {
  const uint32_t pred = __cc_Select(top[0], left, top[-1]);
  return pred;
}

static inline uint32_t __cc_Average4(uint32_t a0, uint32_t a1,
                                     uint32_t a2, uint32_t a3) {
  return __cc_Average2(__cc_Average2(a0, a1), __cc_Average2(a2, a3));
}

static uint32_t __cc_Predictor10(uint32_t left, const uint32_t* const top) {
  const uint32_t pred = __cc_Average4(left, top[-1], top[0], top[1]);
  return pred;
}

static uint32_t __cc_Predictor7(uint32_t left, const uint32_t* const top) {
  const uint32_t pred = __cc_Average2(left, top[0]);
  return pred;
}

static uint32_t __cc_Predictor6(uint32_t left, const uint32_t* const top) {
  const uint32_t pred = __cc_Average2(left, top[-1]);
  return pred;
}

static uint32_t __cc_Predictor3(uint32_t left, const uint32_t* const top) {
  (void)left;
  return top[1];
}

static uint32_t __cc_Predictor8(uint32_t left, const uint32_t* const top) {
  const uint32_t pred = __cc_Average2(top[-1], top[0]);
  (void)left;
  return pred;
}

static const __cc_VP8LPredictorFunc kPredictorsC[16] = {
  __cc_Predictor0, __cc_Predictor1, __cc_Predictor2, __cc_Predictor3,
  __cc_Predictor4, __cc_Predictor5, __cc_Predictor6, __cc_Predictor7,
  __cc_Predictor8, __cc_Predictor9, __cc_Predictor10, __cc_Predictor11,
  __cc_Predictor12, __cc_Predictor13,
  __cc_Predictor0, __cc_Predictor0
};

__cc_VP8LConvertFunc VP8LConvertBGRAToRGBA4444;

__cc_VP8LTransformColorFunc VP8LTransformColorInverse;

static double __cc_ExtraCostCombined(const uint32_t* X, const uint32_t* Y,
                                int length) {
  int i;
  double cost = 0.;
  for (i = 2; i < length - 2; ++i) {
    const int xy = X[i + 2] + Y[i + 2];
    cost += (i >> 1) * xy;
  }
  return cost;
}

static volatile __cc_VP8CPUInfo lossless_last_cpuinfo_used =
    (__cc_VP8CPUInfo)&lossless_last_cpuinfo_used;

void __cc_VP8LDspInit(void) {
  if (lossless_last_cpuinfo_used == VP8GetCPUInfo) return;

  memcpy(VP8LPredictors, kPredictorsC, sizeof(VP8LPredictors));

  VP8LSubtractGreenFromBlueAndRed = __cc_VP8LSubtractGreenFromBlueAndRed_C;
  VP8LAddGreenToBlueAndRed = __cc_VP8LAddGreenToBlueAndRed_C;

  VP8LTransformColor = __cc_VP8LTransformColor_C;
  VP8LTransformColorInverse = __cc_VP8LTransformColorInverse_C;

  VP8LConvertBGRAToRGB = __cc_VP8LConvertBGRAToRGB_C;
  VP8LConvertBGRAToRGBA = __cc_VP8LConvertBGRAToRGBA_C;
  VP8LConvertBGRAToRGBA4444 = __cc_VP8LConvertBGRAToRGBA4444_C;
  VP8LConvertBGRAToRGB565 = __cc_VP8LConvertBGRAToRGB565_C;
  VP8LConvertBGRAToBGR = __cc_VP8LConvertBGRAToBGR_C;

  VP8LFastLog2Slow = __cc_FastLog2Slow;
  VP8LFastSLog2Slow = __cc_FastSLog2Slow;

  VP8LExtraCost = __cc_ExtraCost;
  VP8LExtraCostCombined = __cc_ExtraCostCombined;

  VP8LHuffmanCostCount = __cc_HuffmanCostCount;
  VP8LHuffmanCostCombinedCount = __cc_HuffmanCostCombinedCount;

  VP8LHistogramAdd = __cc_HistogramAdd;


  if (VP8GetCPUInfo != ((void *)0)) {
  }
  lossless_last_cpuinfo_used = VP8GetCPUInfo;
}

static int __cc_ReadImageInfo(__cc_VP8LBitReader* const br,
                         int* const width, int* const height,
                         int* const has_alpha) {
  if (__cc_VP8LReadBits(br, 8) != 0x2f) return 0;
  *width = __cc_VP8LReadBits(br, 14) + 1;
  *height = __cc_VP8LReadBits(br, 14) + 1;
  *has_alpha = __cc_VP8LReadBits(br, 1);
  if (__cc_VP8LReadBits(br, 3) != 0) return 0;
  return 1;
}

int __cc_VP8LDecodeHeader(__cc_VP8LDecoder* const dec, __cc_VP8Io* const io) {
  int width, height, has_alpha;

  if (dec == ((void *)0)) return 0;
  if (io == ((void *)0)) {
    dec->status_ = VP8_STATUS_INVALID_PARAM;
    return 0;
  }

  dec->io_ = io;
  dec->status_ = VP8_STATUS_OK;
  __cc_VP8LInitBitReader(&dec->br_, io->data, io->data_size);
  if (!__cc_ReadImageInfo(&dec->br_, &width, &height, &has_alpha)) {
    dec->status_ = VP8_STATUS_BITSTREAM_ERROR;
    goto Error;
  }
  dec->state_ = READ_DIM;
  io->width = width;
  io->height = height;

  dec->action_ = READ_HDR;
  if (!__cc_DecodeImageStream(width, height, 1, dec, ((void *)0))) goto Error;
  return 1;

 Error:
  __cc_VP8LClear(dec);
  ((dec->status_ != VP8_STATUS_OK) ? (void) (0) : __assert_fail ("dec->status_ != VP8_STATUS_OK", "extraction.1.c", 8861, __PRETTY_FUNCTION__));
  return 0;
}

int __cc_VP8LCheckSignature(const uint8_t* const data, size_t size) {
  return (size >= 5 &&
          data[0] == 0x2f &&
          (data[4] >> 5) == 0);
}

int __cc_VP8LGetInfo(const uint8_t* data, size_t data_size,
                int* const width, int* const height, int* const has_alpha) {
  if (data == ((void *)0) || data_size < 5) {
    return 0;
  } else if (!__cc_VP8LCheckSignature(data, data_size)) {
    return 0;
  } else {
    int w, h, a;
    __cc_VP8LBitReader br;
    __cc_VP8LInitBitReader(&br, data, data_size);
    if (!__cc_ReadImageInfo(&br, &w, &h, &a)) {
      return 0;
    }
    if (width != ((void *)0)) *width = w;
    if (height != ((void *)0)) *height = h;
    if (has_alpha != ((void *)0)) *has_alpha = a;
    return 1;
  }
}

static __cc_VP8StatusCode __cc_AllocateBuffer(__cc_WebPDecBuffer* const buffer) {
  const int w = buffer->width;
  const int h = buffer->height;
  const __cc_WEBP_CSP_MODE mode = buffer->colorspace;

  if (w <= 0 || h <= 0 || !__cc_IsValidColorspace(mode)) {
    return VP8_STATUS_INVALID_PARAM;
  }

  if (!buffer->is_external_memory && buffer->private_memory == ((void *)0)) {
    uint8_t* output;
    int uv_stride = 0, a_stride = 0;
    uint64_t uv_size = 0, a_size = 0, total_size;


    const int stride = w * kModeBpp[mode];
    const uint64_t size = (uint64_t)stride * h;

    if (!__cc_WebPIsRGBMode(mode)) {
      uv_stride = (w + 1) / 2;
      uv_size = (uint64_t)uv_stride * ((h + 1) / 2);
      if (mode == MODE_YUVA) {
        a_stride = w;
        a_size = (uint64_t)a_stride * h;
      }
    }
    total_size = size + 2 * uv_size + a_size;


    output = (uint8_t*)__cc_WebPSafeMalloc(total_size, sizeof(*output));
    if (output == ((void *)0)) {
      return VP8_STATUS_OUT_OF_MEMORY;
    }
    buffer->private_memory = output;

    if (!__cc_WebPIsRGBMode(mode)) {
      __cc_WebPYUVABuffer* const buf = &buffer->u.YUVA;
      buf->y = output;
      buf->y_stride = stride;
      buf->y_size = (size_t)size;
      buf->u = output + size;
      buf->u_stride = uv_stride;
      buf->u_size = (size_t)uv_size;
      buf->v = output + size + uv_size;
      buf->v_stride = uv_stride;
      buf->v_size = (size_t)uv_size;
      if (mode == MODE_YUVA) {
        buf->a = output + size + 2 * uv_size;
      }
      buf->a_size = (size_t)a_size;
      buf->a_stride = a_stride;
    } else {
      __cc_WebPRGBABuffer* const buf = &buffer->u.RGBA;
      buf->rgba = output;
      buf->stride = stride;
      buf->size = (size_t)size;
    }
  }
  return __cc_CheckDecBuffer(buffer);
}

__cc_VP8StatusCode __cc_WebPAllocateDecBuffer(int w, int h,
                                    const __cc_WebPDecoderOptions* const options,
                                    __cc_WebPDecBuffer* const out) {
  __cc_VP8StatusCode status;
  if (out == ((void *)0) || w <= 0 || h <= 0) {
    return VP8_STATUS_INVALID_PARAM;
  }
  if (options != ((void *)0)) {
    if (options->use_cropping) {
      const int cw = options->crop_width;
      const int ch = options->crop_height;
      const int x = options->crop_left & ~1;
      const int y = options->crop_top & ~1;
      if (x < 0 || y < 0 || cw <= 0 || ch <= 0 || x + cw > w || y + ch > h) {
        return VP8_STATUS_INVALID_PARAM;
      }
      w = cw;
      h = ch;
    }
    if (options->use_scaling) {
      if (options->scaled_width <= 0 || options->scaled_height <= 0) {
        return VP8_STATUS_INVALID_PARAM;
      }
      w = options->scaled_width;
      h = options->scaled_height;
    }
  }
  out->width = w;
  out->height = h;


  status = __cc_AllocateBuffer(out);
  if (status != VP8_STATUS_OK) return status;







  return status;
}

int __cc_VP8GetThreadMethod(const __cc_WebPDecoderOptions* const options,
                       const __cc_WebPHeaderStructure* const headers,
                       int width, int height) {
  if (options == ((void *)0) || options->use_threads == 0) {
    return 0;
  }
  (void)headers;
  (void)width;
  (void)height;
  ((headers == ((void *)0) || !headers->is_lossless) ? (void) (0) : __assert_fail ("headers == ((void *)0) || !headers->is_lossless", "extraction.1.c", 9004, __PRETTY_FUNCTION__));
  return 0;

}

static void __cc_Copy32b(uint8_t* dst, uint8_t* src) {
  memcpy(dst, src, 4);
}

int __cc_WebPPictureAlloc(__cc_WebPPicture* picture) {
  if (picture != ((void *)0)) {
    const int width = picture->width;
    const int height = picture->height;

    __cc_WebPPictureFree(picture);

    if (!picture->use_argb) {
      return __cc_WebPPictureAllocYUVA(picture, width, height);
    } else {
      return __cc_WebPPictureAllocARGB(picture, width, height);
    }
  }
  return 1;
}



void __cc_VP8LDoFillBitWindow(__cc_VP8LBitReader* const br) {
  ((br->bit_pos_ >= 32) ? (void) (0) : __assert_fail ("br->bit_pos_ >= 32", "extraction.1.c", 9040, __PRETTY_FUNCTION__));



  if (br->pos_ + sizeof(br->val_) < br->len_) {
    br->val_ >>= 32;
    br->bit_pos_ -= 32;


    br->val_ |= (__cc_vp8l_val_t)*(const uint32_t*)(br->buf_ + br->pos_) <<
                (64 - 32);
    br->pos_ += 4;
    return;
  }

  __cc_ShiftBytes(br);
}

static void __cc_DC8uvNoLeft(uint8_t *dst) {
  int dc0 = 4;
  int i;
  for (i = 0; i < 8; ++i) {
    dc0 += dst[i - 32];
  }
  __cc_Put8x8uv(dc0 >> 3, dst);
}

extern int __cc_WebPPictureImportRGBA(
    __cc_WebPPicture* picture, const uint8_t* rgba, int rgba_stride);

extern const uint8_t* const VP8kclip1;

extern const int8_t* const VP8ksclip2;

extern const int8_t* const VP8ksclip1;

static inline void __cc_do_filter2(uint8_t* p, int step) {
  const int p1 = p[-2*step], p0 = p[-step], q0 = p[0], q1 = p[step];
  const int a = 3 * (q0 - p0) + VP8ksclip1[p1 - q1];
  const int a1 = VP8ksclip2[(a + 4) >> 3];
  const int a2 = VP8ksclip2[(a + 3) >> 3];
  p[-step] = VP8kclip1[p0 + a2];
  p[ 0] = VP8kclip1[q0 - a1];
}

static inline int __cc_needs_filter(const uint8_t* p, int step, int t) {
  const int p1 = p[-2 * step], p0 = p[-step], q0 = p[0], q1 = p[step];
  return ((4 * VP8kabs0[p0 - q0] + VP8kabs0[p1 - q1]) <= t);
}

static void __cc_SimpleHFilter16(uint8_t* p, int stride, int thresh) {
  int i;
  const int thresh2 = 2 * thresh + 1;
  for (i = 0; i < 16; ++i) {
    if (__cc_needs_filter(p + i * stride, 1, thresh2)) {
      __cc_do_filter2(p + i * stride, 1);
    }
  }
}

static void __cc_SimpleHFilter16i(uint8_t* p, int stride, int thresh) {
  int k;
  for (k = 3; k > 0; --k) {
    p += 4;
    __cc_SimpleHFilter16(p, stride, thresh);
  }
}

static void __cc_ExportRowShrinkC(__cc_WebPRescaler* const wrk) {
  int x_out;
  uint8_t* const dst = wrk->dst;
  __cc_rescaler_t* const irow = wrk->irow;
  const int x_out_max = wrk->dst_width * wrk->num_channels;
  const __cc_rescaler_t* const frow = wrk->frow;
  const uint32_t yscale = wrk->fy_scale * (-wrk->y_accum);
  ((!__cc_WebPRescalerOutputDone(wrk)) ? (void) (0) : __assert_fail ("!WebPRescalerOutputDone(wrk)", "extraction.1.c", 9115, __PRETTY_FUNCTION__));
  ((wrk->y_accum <= 0) ? (void) (0) : __assert_fail ("wrk->y_accum <= 0", "extraction.1.c", 9116, __PRETTY_FUNCTION__));
  ((!wrk->y_expand) ? (void) (0) : __assert_fail ("!wrk->y_expand", "extraction.1.c", 9117, __PRETTY_FUNCTION__));
  if (yscale) {
    for (x_out = 0; x_out < x_out_max; ++x_out) {
      const uint32_t frac = (uint32_t)(((uint64_t)(frow[x_out]) * (yscale) + ((1ull << 32) >> 1)) >> 32);
      const int v = (int)(((uint64_t)(irow[x_out] - frac) * (wrk->fxy_scale) + ((1ull << 32) >> 1)) >> 32);
      ((v >= 0 && v <= 255) ? (void) (0) : __assert_fail ("v >= 0 && v <= 255", "extraction.1.c", 9122, __PRETTY_FUNCTION__));
      dst[x_out] = v;
      irow[x_out] = frac;
    }
  } else {
    for (x_out = 0; x_out < x_out_max; ++x_out) {
      const int v = (int)(((uint64_t)(irow[x_out]) * (wrk->fxy_scale) + ((1ull << 32) >> 1)) >> 32);
      ((v >= 0 && v <= 255) ? (void) (0) : __assert_fail ("v >= 0 && v <= 255", "extraction.1.c", 9129, __PRETTY_FUNCTION__));
      dst[x_out] = v;
      irow[x_out] = 0;
    }
  }
}



static inline int __cc_VP8YUVToR(int y, int v) {
  return __cc_VP8Clip8(19077 * y + 26149 * v + (-19077 * 16 - 26149 * 128 + YUV_HALF2));
}





static inline int __cc_VP8YUVToB(int y, int u) {
  return __cc_VP8Clip8(19077 * y + 33050 * u + (-19077 * 16 - 33050 * 128 + YUV_HALF2));
}

static inline void __cc_VP8YuvToBgr(int y, int u, int v,
                                    uint8_t* const bgr) {
  bgr[0] = __cc_VP8YUVToB(y, u);
  bgr[1] = __cc_VP8YUVToG(y, u, v);
  bgr[2] = __cc_VP8YUVToR(y, v);
}

static inline void __cc_VP8YuvToBgra(uint8_t y, uint8_t u, uint8_t v,
                                     uint8_t* const bgra) {
  __cc_VP8YuvToBgr(y, u, v, bgra);
  bgra[3] = 0xff;
}
static void __cc_YuvToBgraRow(const uint8_t* y, const uint8_t* u, const uint8_t* v, uint8_t* dst, int len) { const uint8_t* const end = dst + (len & ~1) * 4; while (dst != end) { __cc_VP8YuvToBgra(y[0], u[0], v[0], dst); __cc_VP8YuvToBgra(y[1], u[0], v[0], dst + 4); y += 2; ++u; ++v; dst += 2 * 4; } if (len & 1) { __cc_VP8YuvToBgra(y[0], u[0], v[0], dst); } }

static inline void __cc_VP8YuvToRgb(int y, int u, int v,
                                    uint8_t* const rgb) {
  rgb[0] = __cc_VP8YUVToR(y, v);
  rgb[1] = __cc_VP8YUVToG(y, u, v);
  rgb[2] = __cc_VP8YUVToB(y, u);
}

static inline void __cc_VP8YuvToRgba(uint8_t y, uint8_t u, uint8_t v,
                                     uint8_t* const rgba) {
  __cc_VP8YuvToRgb(y, u, v, rgba);
  rgba[3] = 0xff;
}

static void __cc_YuvToRgbaRow(const uint8_t* y, const uint8_t* u, const uint8_t* v, uint8_t* dst, int len) { const uint8_t* const end = dst + (len & ~1) * 4; while (dst != end) { __cc_VP8YuvToRgba(y[0], u[0], v[0], dst); __cc_VP8YuvToRgba(y[1], u[0], v[0], dst + 4); y += 2; ++u; ++v; dst += 2 * 4; } if (len & 1) { __cc_VP8YuvToRgba(y[0], u[0], v[0], dst); } }

static void __cc_YuvToBgrRow(const uint8_t* y, const uint8_t* u, const uint8_t* v, uint8_t* dst, int len) { const uint8_t* const end = dst + (len & ~1) * 3; while (dst != end) { __cc_VP8YuvToBgr(y[0], u[0], v[0], dst); __cc_VP8YuvToBgr(y[1], u[0], v[0], dst + 3); y += 2; ++u; ++v; dst += 2 * 3; } if (len & 1) { __cc_VP8YuvToBgr(y[0], u[0], v[0], dst); } }

static inline void __cc_VP8YuvToRgb565(int y, int u, int v,
                                       uint8_t* const rgb) {
  const int r = __cc_VP8YUVToR(y, v);
  const int g = __cc_VP8YUVToG(y, u, v);
  const int b = __cc_VP8YUVToB(y, u);
  const int rg = (r & 0xf8) | (g >> 5);
  const int gb = ((g << 3) & 0xe0) | (b >> 3);




  rgb[0] = rg;
  rgb[1] = gb;

}

static void __cc_YuvToRgb565Row(const uint8_t* y, const uint8_t* u, const uint8_t* v, uint8_t* dst, int len) { const uint8_t* const end = dst + (len & ~1) * 2; while (dst != end) { __cc_VP8YuvToRgb565(y[0], u[0], v[0], dst); __cc_VP8YuvToRgb565(y[1], u[0], v[0], dst + 2); y += 2; ++u; ++v; dst += 2 * 2; } if (len & 1) { __cc_VP8YuvToRgb565(y[0], u[0], v[0], dst); } }

static void __cc_YuvToRgbRow(const uint8_t* y, const uint8_t* u, const uint8_t* v, uint8_t* dst, int len) { const uint8_t* const end = dst + (len & ~1) * 3; while (dst != end) { __cc_VP8YuvToRgb(y[0], u[0], v[0], dst); __cc_VP8YuvToRgb(y[1], u[0], v[0], dst + 3); y += 2; ++u; ++v; dst += 2 * 3; } if (len & 1) { __cc_VP8YuvToRgb(y[0], u[0], v[0], dst); } }

static inline void __cc_do_filter4(uint8_t* p, int step) {
  const int p1 = p[-2*step], p0 = p[-step], q0 = p[0], q1 = p[step];
  const int a = 3 * (q0 - p0);
  const int a1 = VP8ksclip2[(a + 4) >> 3];
  const int a2 = VP8ksclip2[(a + 3) >> 3];
  const int a3 = (a1 + 1) >> 1;
  p[-2*step] = VP8kclip1[p1 + a3];
  p[- step] = VP8kclip1[p0 + a2];
  p[ 0] = VP8kclip1[q0 - a1];
  p[ step] = VP8kclip1[q1 - a3];
}

static inline int __cc_hev(const uint8_t* p, int step, int thresh) {
  const int p1 = p[-2*step], p0 = p[-step], q0 = p[0], q1 = p[step];
  return (VP8kabs0[p1 - p0] > thresh) || (VP8kabs0[q1 - q0] > thresh);
}

static inline void __cc_FilterLoop24(uint8_t* p,
                                     int hstride, int vstride, int size,
                                     int thresh, int ithresh, int hev_thresh) {
  const int thresh2 = 2 * thresh + 1;
  while (size-- > 0) {
    if (__cc_needs_filter2(p, hstride, thresh2, ithresh)) {
      if (__cc_hev(p, hstride, hev_thresh)) {
        __cc_do_filter2(p, hstride);
      } else {
        __cc_do_filter4(p, hstride);
      }
    }
    p += vstride;
  }
}

static void __cc_HFilter8i(uint8_t* u, uint8_t* v, int stride,
                      int thresh, int ithresh, int hev_thresh) {
  __cc_FilterLoop24(u + 4, 1, stride, 8, thresh, ithresh, hev_thresh);
  __cc_FilterLoop24(v + 4, 1, stride, 8, thresh, ithresh, hev_thresh);
}

static volatile __cc_VP8CPUInfo yuv_last_cpuinfo_used =
    (__cc_VP8CPUInfo)&yuv_last_cpuinfo_used;

void* __cc_WebPSafeMalloc(uint64_t nmemb, size_t size) {
  void* ptr;
  do {} while (0);
  if (!__cc_CheckSizeArgumentsOverflow(nmemb, size)) return ((void *)0);
  ((nmemb * size > 0) ? (void) (0) : __assert_fail ("nmemb * size > 0", "extraction.1.c", 9266, __PRETTY_FUNCTION__));
  ptr = malloc((size_t)(nmemb * size));
  do {} while (0);
  return ptr;
}

__cc_WebPSamplerRowFunc WebPSamplers[MODE_LAST];

static inline void __cc_VP8YuvToArgb(uint8_t y, uint8_t u, uint8_t v,
                                     uint8_t* const argb) {
  argb[0] = 0xff;
  __cc_VP8YuvToRgb(y, u, v, argb + 1);
}

static void __cc_YuvToArgbRow(const uint8_t* y, const uint8_t* u, const uint8_t* v, uint8_t* dst, int len) { const uint8_t* const end = dst + (len & ~1) * 4; while (dst != end) { __cc_VP8YuvToArgb(y[0], u[0], v[0], dst); __cc_VP8YuvToArgb(y[1], u[0], v[0], dst + 4); y += 2; ++u; ++v; dst += 2 * 4; } if (len & 1) { __cc_VP8YuvToArgb(y[0], u[0], v[0], dst); } }

static inline void __cc_VP8YuvToRgba4444(int y, int u, int v,
                                         uint8_t* const argb) {
  const int r = __cc_VP8YUVToR(y, v);
  const int g = __cc_VP8YUVToG(y, u, v);
  const int b = __cc_VP8YUVToB(y, u);
  const int rg = (r & 0xf0) | (g >> 4);
  const int ba = (b & 0xf0) | 0x0f;




  argb[0] = rg;
  argb[1] = ba;

}

static void __cc_YuvToRgba4444Row(const uint8_t* y, const uint8_t* u, const uint8_t* v, uint8_t* dst, int len) { const uint8_t* const end = dst + (len & ~1) * 2; while (dst != end) { __cc_VP8YuvToRgba4444(y[0], u[0], v[0], dst); __cc_VP8YuvToRgba4444(y[1], u[0], v[0], dst + 2); y += 2; ++u; ++v; dst += 2 * 2; } if (len & 1) { __cc_VP8YuvToRgba4444(y[0], u[0], v[0], dst); } }

void __cc_WebPInitSamplers(void) {
  if (yuv_last_cpuinfo_used == VP8GetCPUInfo) return;

  WebPSamplers[MODE_RGB] = __cc_YuvToRgbRow;
  WebPSamplers[MODE_RGBA] = __cc_YuvToRgbaRow;
  WebPSamplers[MODE_BGR] = __cc_YuvToBgrRow;
  WebPSamplers[MODE_BGRA] = __cc_YuvToBgraRow;
  WebPSamplers[MODE_ARGB] = __cc_YuvToArgbRow;
  WebPSamplers[MODE_RGBA_4444] = __cc_YuvToRgba4444Row;
  WebPSamplers[MODE_RGB_565] = __cc_YuvToRgb565Row;
  WebPSamplers[MODE_rgbA] = __cc_YuvToRgbaRow;
  WebPSamplers[MODE_bgrA] = __cc_YuvToBgraRow;
  WebPSamplers[MODE_Argb] = __cc_YuvToArgbRow;
  WebPSamplers[MODE_rgbA_4444] = __cc_YuvToRgba4444Row;


  if (VP8GetCPUInfo != ((void *)0)) {
  }
  yuv_last_cpuinfo_used = VP8GetCPUInfo;
}
static void __cc_UpsampleBgrLinePair(const uint8_t* top_y, const uint8_t* bottom_y, const uint8_t* top_u, const uint8_t* top_v, const uint8_t* cur_u, const uint8_t* cur_v, uint8_t* top_dst, uint8_t* bottom_dst, int len) { int x; const int last_pixel_pair = (len - 1) >> 1; uint32_t tl_uv = ((top_u[0]) | ((top_v[0]) << 16)); uint32_t l_uv = ((cur_u[0]) | ((cur_v[0]) << 16)); ((top_y != ((void *)0)) ? (void) (0) : __assert_fail ("top_y != ((void *)0)", "extraction.1.c", 9391, __PRETTY_FUNCTION__)); { const uint32_t uv0 = (3 * tl_uv + l_uv + 0x00020002u) >> 2; __cc_VP8YuvToBgr(top_y[0], uv0 & 0xff, (uv0 >> 16), top_dst); } if (bottom_y != ((void *)0)) { const uint32_t uv0 = (3 * l_uv + tl_uv + 0x00020002u) >> 2; __cc_VP8YuvToBgr(bottom_y[0], uv0 & 0xff, (uv0 >> 16), bottom_dst); } for (x = 1; x <= last_pixel_pair; ++x) { const uint32_t t_uv = ((top_u[x]) | ((top_v[x]) << 16)); const uint32_t uv = ((cur_u[x]) | ((cur_v[x]) << 16)); const uint32_t avg = tl_uv + t_uv + l_uv + uv + 0x00080008u; const uint32_t diag_12 = (avg + 2 * (t_uv + l_uv)) >> 3; const uint32_t diag_03 = (avg + 2 * (tl_uv + uv)) >> 3; { const uint32_t uv0 = (diag_12 + tl_uv) >> 1; const uint32_t uv1 = (diag_03 + t_uv) >> 1; __cc_VP8YuvToBgr(top_y[2 * x - 1], uv0 & 0xff, (uv0 >> 16), top_dst + (2 * x - 1) * 3); __cc_VP8YuvToBgr(top_y[2 * x - 0], uv1 & 0xff, (uv1 >> 16), top_dst + (2 * x - 0) * 3); } if (bottom_y != ((void *)0)) { const uint32_t uv0 = (diag_03 + l_uv) >> 1; const uint32_t uv1 = (diag_12 + uv) >> 1; __cc_VP8YuvToBgr(bottom_y[2 * x - 1], uv0 & 0xff, (uv0 >> 16), bottom_dst + (2 * x - 1) * 3); __cc_VP8YuvToBgr(bottom_y[2 * x + 0], uv1 & 0xff, (uv1 >> 16), bottom_dst + (2 * x + 0) * 3); } tl_uv = t_uv; l_uv = uv; } if (!(len & 1)) { { const uint32_t uv0 = (3 * tl_uv + l_uv + 0x00020002u) >> 2; __cc_VP8YuvToBgr(top_y[len - 1], uv0 & 0xff, (uv0 >> 16), top_dst + (len - 1) * 3); } if (bottom_y != ((void *)0)) { const uint32_t uv0 = (3 * l_uv + tl_uv + 0x00020002u) >> 2; __cc_VP8YuvToBgr(bottom_y[len - 1], uv0 & 0xff, (uv0 >> 16), bottom_dst + (len - 1) * 3); } } }

static volatile __cc_VP8CPUInfo upsampling_last_cpuinfo_used2 =
    (__cc_VP8CPUInfo)&upsampling_last_cpuinfo_used2;
static void __cc_Yuv444ToRgba(const uint8_t* y, const uint8_t* u, const uint8_t* v, uint8_t* dst, int len) { int i; for (i = 0; i < len; ++i) __cc_VP8YuvToRgba(y[i], u[i], v[i], &dst[i * 4]); }

static void __cc_UpsampleBgraLinePair(const uint8_t* top_y, const uint8_t* bottom_y, const uint8_t* top_u, const uint8_t* top_v, const uint8_t* cur_u, const uint8_t* cur_v, uint8_t* top_dst, uint8_t* bottom_dst, int len) { int x; const int last_pixel_pair = (len - 1) >> 1; uint32_t tl_uv = ((top_u[0]) | ((top_v[0]) << 16)); uint32_t l_uv = ((cur_u[0]) | ((cur_v[0]) << 16)); ((top_y != ((void *)0)) ? (void) (0) : __assert_fail ("top_y != ((void *)0)", "extraction.1.c", 9405, __PRETTY_FUNCTION__)); { const uint32_t uv0 = (3 * tl_uv + l_uv + 0x00020002u) >> 2; __cc_VP8YuvToBgra(top_y[0], uv0 & 0xff, (uv0 >> 16), top_dst); } if (bottom_y != ((void *)0)) { const uint32_t uv0 = (3 * l_uv + tl_uv + 0x00020002u) >> 2; __cc_VP8YuvToBgra(bottom_y[0], uv0 & 0xff, (uv0 >> 16), bottom_dst); } for (x = 1; x <= last_pixel_pair; ++x) { const uint32_t t_uv = ((top_u[x]) | ((top_v[x]) << 16)); const uint32_t uv = ((cur_u[x]) | ((cur_v[x]) << 16)); const uint32_t avg = tl_uv + t_uv + l_uv + uv + 0x00080008u; const uint32_t diag_12 = (avg + 2 * (t_uv + l_uv)) >> 3; const uint32_t diag_03 = (avg + 2 * (tl_uv + uv)) >> 3; { const uint32_t uv0 = (diag_12 + tl_uv) >> 1; const uint32_t uv1 = (diag_03 + t_uv) >> 1; __cc_VP8YuvToBgra(top_y[2 * x - 1], uv0 & 0xff, (uv0 >> 16), top_dst + (2 * x - 1) * 4); __cc_VP8YuvToBgra(top_y[2 * x - 0], uv1 & 0xff, (uv1 >> 16), top_dst + (2 * x - 0) * 4); } if (bottom_y != ((void *)0)) { const uint32_t uv0 = (diag_03 + l_uv) >> 1; const uint32_t uv1 = (diag_12 + uv) >> 1; __cc_VP8YuvToBgra(bottom_y[2 * x - 1], uv0 & 0xff, (uv0 >> 16), bottom_dst + (2 * x - 1) * 4); __cc_VP8YuvToBgra(bottom_y[2 * x + 0], uv1 & 0xff, (uv1 >> 16), bottom_dst + (2 * x + 0) * 4); } tl_uv = t_uv; l_uv = uv; } if (!(len & 1)) { { const uint32_t uv0 = (3 * tl_uv + l_uv + 0x00020002u) >> 2; __cc_VP8YuvToBgra(top_y[len - 1], uv0 & 0xff, (uv0 >> 16), top_dst + (len - 1) * 4); } if (bottom_y != ((void *)0)) { const uint32_t uv0 = (3 * l_uv + tl_uv + 0x00020002u) >> 2; __cc_VP8YuvToBgra(bottom_y[len - 1], uv0 & 0xff, (uv0 >> 16), bottom_dst + (len - 1) * 4); } } }

extern int __cc_WebPPictureImportRGB(
    __cc_WebPPicture* picture, const uint8_t* rgb, int rgb_stride);

static void __cc_UpsampleRgbaLinePair(const uint8_t* top_y, const uint8_t* bottom_y, const uint8_t* top_u, const uint8_t* top_v, const uint8_t* cur_u, const uint8_t* cur_v, uint8_t* top_dst, uint8_t* bottom_dst, int len) { int x; const int last_pixel_pair = (len - 1) >> 1; uint32_t tl_uv = ((top_u[0]) | ((top_v[0]) << 16)); uint32_t l_uv = ((cur_u[0]) | ((cur_v[0]) << 16)); ((top_y != ((void *)0)) ? (void) (0) : __assert_fail ("top_y != ((void *)0)", "extraction.1.c", 9410, __PRETTY_FUNCTION__)); { const uint32_t uv0 = (3 * tl_uv + l_uv + 0x00020002u) >> 2; __cc_VP8YuvToRgba(top_y[0], uv0 & 0xff, (uv0 >> 16), top_dst); } if (bottom_y != ((void *)0)) { const uint32_t uv0 = (3 * l_uv + tl_uv + 0x00020002u) >> 2; __cc_VP8YuvToRgba(bottom_y[0], uv0 & 0xff, (uv0 >> 16), bottom_dst); } for (x = 1; x <= last_pixel_pair; ++x) { const uint32_t t_uv = ((top_u[x]) | ((top_v[x]) << 16)); const uint32_t uv = ((cur_u[x]) | ((cur_v[x]) << 16)); const uint32_t avg = tl_uv + t_uv + l_uv + uv + 0x00080008u; const uint32_t diag_12 = (avg + 2 * (t_uv + l_uv)) >> 3; const uint32_t diag_03 = (avg + 2 * (tl_uv + uv)) >> 3; { const uint32_t uv0 = (diag_12 + tl_uv) >> 1; const uint32_t uv1 = (diag_03 + t_uv) >> 1; __cc_VP8YuvToRgba(top_y[2 * x - 1], uv0 & 0xff, (uv0 >> 16), top_dst + (2 * x - 1) * 4); __cc_VP8YuvToRgba(top_y[2 * x - 0], uv1 & 0xff, (uv1 >> 16), top_dst + (2 * x - 0) * 4); } if (bottom_y != ((void *)0)) { const uint32_t uv0 = (diag_03 + l_uv) >> 1; const uint32_t uv1 = (diag_12 + uv) >> 1; __cc_VP8YuvToRgba(bottom_y[2 * x - 1], uv0 & 0xff, (uv0 >> 16), bottom_dst + (2 * x - 1) * 4); __cc_VP8YuvToRgba(bottom_y[2 * x + 0], uv1 & 0xff, (uv1 >> 16), bottom_dst + (2 * x + 0) * 4); } tl_uv = t_uv; l_uv = uv; } if (!(len & 1)) { { const uint32_t uv0 = (3 * tl_uv + l_uv + 0x00020002u) >> 2; __cc_VP8YuvToRgba(top_y[len - 1], uv0 & 0xff, (uv0 >> 16), top_dst + (len - 1) * 4); } if (bottom_y != ((void *)0)) { const uint32_t uv0 = (3 * l_uv + tl_uv + 0x00020002u) >> 2; __cc_VP8YuvToRgba(bottom_y[len - 1], uv0 & 0xff, (uv0 >> 16), bottom_dst + (len - 1) * 4); } } }

static void __cc_Yuv444ToRgb565(const uint8_t* y, const uint8_t* u, const uint8_t* v, uint8_t* dst, int len) { int i; for (i = 0; i < len; ++i) __cc_VP8YuvToRgb565(y[i], u[i], v[i], &dst[i * 2]); }

static void __cc_UpsampleArgbLinePair(const uint8_t* top_y, const uint8_t* bottom_y, const uint8_t* top_u, const uint8_t* top_v, const uint8_t* cur_u, const uint8_t* cur_v, uint8_t* top_dst, uint8_t* bottom_dst, int len) { int x; const int last_pixel_pair = (len - 1) >> 1; uint32_t tl_uv = ((top_u[0]) | ((top_v[0]) << 16)); uint32_t l_uv = ((cur_u[0]) | ((cur_v[0]) << 16)); ((top_y != ((void *)0)) ? (void) (0) : __assert_fail ("top_y != ((void *)0)", "extraction.1.c", 9414, __PRETTY_FUNCTION__)); { const uint32_t uv0 = (3 * tl_uv + l_uv + 0x00020002u) >> 2; __cc_VP8YuvToArgb(top_y[0], uv0 & 0xff, (uv0 >> 16), top_dst); } if (bottom_y != ((void *)0)) { const uint32_t uv0 = (3 * l_uv + tl_uv + 0x00020002u) >> 2; __cc_VP8YuvToArgb(bottom_y[0], uv0 & 0xff, (uv0 >> 16), bottom_dst); } for (x = 1; x <= last_pixel_pair; ++x) { const uint32_t t_uv = ((top_u[x]) | ((top_v[x]) << 16)); const uint32_t uv = ((cur_u[x]) | ((cur_v[x]) << 16)); const uint32_t avg = tl_uv + t_uv + l_uv + uv + 0x00080008u; const uint32_t diag_12 = (avg + 2 * (t_uv + l_uv)) >> 3; const uint32_t diag_03 = (avg + 2 * (tl_uv + uv)) >> 3; { const uint32_t uv0 = (diag_12 + tl_uv) >> 1; const uint32_t uv1 = (diag_03 + t_uv) >> 1; __cc_VP8YuvToArgb(top_y[2 * x - 1], uv0 & 0xff, (uv0 >> 16), top_dst + (2 * x - 1) * 4); __cc_VP8YuvToArgb(top_y[2 * x - 0], uv1 & 0xff, (uv1 >> 16), top_dst + (2 * x - 0) * 4); } if (bottom_y != ((void *)0)) { const uint32_t uv0 = (diag_03 + l_uv) >> 1; const uint32_t uv1 = (diag_12 + uv) >> 1; __cc_VP8YuvToArgb(bottom_y[2 * x - 1], uv0 & 0xff, (uv0 >> 16), bottom_dst + (2 * x - 1) * 4); __cc_VP8YuvToArgb(bottom_y[2 * x + 0], uv1 & 0xff, (uv1 >> 16), bottom_dst + (2 * x + 0) * 4); } tl_uv = t_uv; l_uv = uv; } if (!(len & 1)) { { const uint32_t uv0 = (3 * tl_uv + l_uv + 0x00020002u) >> 2; __cc_VP8YuvToArgb(top_y[len - 1], uv0 & 0xff, (uv0 >> 16), top_dst + (len - 1) * 4); } if (bottom_y != ((void *)0)) { const uint32_t uv0 = (3 * l_uv + tl_uv + 0x00020002u) >> 2; __cc_VP8YuvToArgb(bottom_y[len - 1], uv0 & 0xff, (uv0 >> 16), bottom_dst + (len - 1) * 4); } } }

static void __cc_UpsampleRgba4444LinePair(const uint8_t* top_y, const uint8_t* bottom_y, const uint8_t* top_u, const uint8_t* top_v, const uint8_t* cur_u, const uint8_t* cur_v, uint8_t* top_dst, uint8_t* bottom_dst, int len) { int x; const int last_pixel_pair = (len - 1) >> 1; uint32_t tl_uv = ((top_u[0]) | ((top_v[0]) << 16)); uint32_t l_uv = ((cur_u[0]) | ((cur_v[0]) << 16)); ((top_y != ((void *)0)) ? (void) (0) : __assert_fail ("top_y != ((void *)0)", "extraction.1.c", 9416, __PRETTY_FUNCTION__)); { const uint32_t uv0 = (3 * tl_uv + l_uv + 0x00020002u) >> 2; __cc_VP8YuvToRgba4444(top_y[0], uv0 & 0xff, (uv0 >> 16), top_dst); } if (bottom_y != ((void *)0)) { const uint32_t uv0 = (3 * l_uv + tl_uv + 0x00020002u) >> 2; __cc_VP8YuvToRgba4444(bottom_y[0], uv0 & 0xff, (uv0 >> 16), bottom_dst); } for (x = 1; x <= last_pixel_pair; ++x) { const uint32_t t_uv = ((top_u[x]) | ((top_v[x]) << 16)); const uint32_t uv = ((cur_u[x]) | ((cur_v[x]) << 16)); const uint32_t avg = tl_uv + t_uv + l_uv + uv + 0x00080008u; const uint32_t diag_12 = (avg + 2 * (t_uv + l_uv)) >> 3; const uint32_t diag_03 = (avg + 2 * (tl_uv + uv)) >> 3; { const uint32_t uv0 = (diag_12 + tl_uv) >> 1; const uint32_t uv1 = (diag_03 + t_uv) >> 1; __cc_VP8YuvToRgba4444(top_y[2 * x - 1], uv0 & 0xff, (uv0 >> 16), top_dst + (2 * x - 1) * 2); __cc_VP8YuvToRgba4444(top_y[2 * x - 0], uv1 & 0xff, (uv1 >> 16), top_dst + (2 * x - 0) * 2); } if (bottom_y != ((void *)0)) { const uint32_t uv0 = (diag_03 + l_uv) >> 1; const uint32_t uv1 = (diag_12 + uv) >> 1; __cc_VP8YuvToRgba4444(bottom_y[2 * x - 1], uv0 & 0xff, (uv0 >> 16), bottom_dst + (2 * x - 1) * 2); __cc_VP8YuvToRgba4444(bottom_y[2 * x + 0], uv1 & 0xff, (uv1 >> 16), bottom_dst + (2 * x + 0) * 2); } tl_uv = t_uv; l_uv = uv; } if (!(len & 1)) { { const uint32_t uv0 = (3 * tl_uv + l_uv + 0x00020002u) >> 2; __cc_VP8YuvToRgba4444(top_y[len - 1], uv0 & 0xff, (uv0 >> 16), top_dst + (len - 1) * 2); } if (bottom_y != ((void *)0)) { const uint32_t uv0 = (3 * l_uv + tl_uv + 0x00020002u) >> 2; __cc_VP8YuvToRgba4444(bottom_y[len - 1], uv0 & 0xff, (uv0 >> 16), bottom_dst + (len - 1) * 2); } } }

static void __cc_Yuv444ToRgb(const uint8_t* y, const uint8_t* u, const uint8_t* v, uint8_t* dst, int len) { int i; for (i = 0; i < len; ++i) __cc_VP8YuvToRgb(y[i], u[i], v[i], &dst[i * 3]); }

__cc_WebPUpsampleLinePairFunc WebPUpsamplers[MODE_LAST];

static const int8_t sclip1[1020 + 1020 + 1] = {
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
  0x80, 0x80, 0x80, 0x80, 0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87,
  0x88, 0x89, 0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f, 0x90, 0x91, 0x92, 0x93,
  0x94, 0x95, 0x96, 0x97, 0x98, 0x99, 0x9a, 0x9b, 0x9c, 0x9d, 0x9e, 0x9f,
  0xa0, 0xa1, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7, 0xa8, 0xa9, 0xaa, 0xab,
  0xac, 0xad, 0xae, 0xaf, 0xb0, 0xb1, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6, 0xb7,
  0xb8, 0xb9, 0xba, 0xbb, 0xbc, 0xbd, 0xbe, 0xbf, 0xc0, 0xc1, 0xc2, 0xc3,
  0xc4, 0xc5, 0xc6, 0xc7, 0xc8, 0xc9, 0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf,
  0xd0, 0xd1, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda, 0xdb,
  0xdc, 0xdd, 0xde, 0xdf, 0xe0, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7,
  0xe8, 0xe9, 0xea, 0xeb, 0xec, 0xed, 0xee, 0xef, 0xf0, 0xf1, 0xf2, 0xf3,
  0xf4, 0xf5, 0xf6, 0xf7, 0xf8, 0xf9, 0xfa, 0xfb, 0xfc, 0xfd, 0xfe, 0xff,
  0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0a, 0x0b,
  0x0c, 0x0d, 0x0e, 0x0f, 0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17,
  0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f, 0x20, 0x21, 0x22, 0x23,
  0x24, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f,
  0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3a, 0x3b,
  0x3c, 0x3d, 0x3e, 0x3f, 0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47,
  0x48, 0x49, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f, 0x50, 0x51, 0x52, 0x53,
  0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5a, 0x5b, 0x5c, 0x5d, 0x5e, 0x5f,
  0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69, 0x6a, 0x6b,
  0x6c, 0x6d, 0x6e, 0x6f, 0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77,
  0x78, 0x79, 0x7a, 0x7b, 0x7c, 0x7d, 0x7e, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f,
  0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f
};



void __cc_VP8InitClipTables(void) {
}

static const uint8_t clip1[255 + 511 + 1] = {
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08,
  0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f, 0x10, 0x11, 0x12, 0x13, 0x14,
  0x15, 0x16, 0x17, 0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f, 0x20,
  0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2a, 0x2b, 0x2c,
  0x2d, 0x2e, 0x2f, 0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38,
  0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f, 0x40, 0x41, 0x42, 0x43, 0x44,
  0x45, 0x46, 0x47, 0x48, 0x49, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f, 0x50,
  0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5a, 0x5b, 0x5c,
  0x5d, 0x5e, 0x5f, 0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68,
  0x69, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f, 0x70, 0x71, 0x72, 0x73, 0x74,
  0x75, 0x76, 0x77, 0x78, 0x79, 0x7a, 0x7b, 0x7c, 0x7d, 0x7e, 0x7f, 0x80,
  0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x8a, 0x8b, 0x8c,
  0x8d, 0x8e, 0x8f, 0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98,
  0x99, 0x9a, 0x9b, 0x9c, 0x9d, 0x9e, 0x9f, 0xa0, 0xa1, 0xa2, 0xa3, 0xa4,
  0xa5, 0xa6, 0xa7, 0xa8, 0xa9, 0xaa, 0xab, 0xac, 0xad, 0xae, 0xaf, 0xb0,
  0xb1, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6, 0xb7, 0xb8, 0xb9, 0xba, 0xbb, 0xbc,
  0xbd, 0xbe, 0xbf, 0xc0, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7, 0xc8,
  0xc9, 0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf, 0xd0, 0xd1, 0xd2, 0xd3, 0xd4,
  0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda, 0xdb, 0xdc, 0xdd, 0xde, 0xdf, 0xe0,
  0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9, 0xea, 0xeb, 0xec,
  0xed, 0xee, 0xef, 0xf0, 0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8,
  0xf9, 0xfa, 0xfb, 0xfc, 0xfd, 0xfe, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
};

const int8_t* const VP8ksclip1 = &sclip1[1020];

static const int8_t sclip2[112 + 112 + 1] = {
  0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0,
  0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0,
  0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0,
  0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0,
  0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0,
  0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0,
  0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0,
  0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0,
  0xf0, 0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8, 0xf9, 0xfa, 0xfb,
  0xfc, 0xfd, 0xfe, 0xff, 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
  0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f,
  0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f,
  0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f,
  0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f,
  0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f,
  0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f,
  0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f,
  0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f,
  0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f
};

static const uint8_t abs0[255 + 255 + 1] = {
  0xff, 0xfe, 0xfd, 0xfc, 0xfb, 0xfa, 0xf9, 0xf8, 0xf7, 0xf6, 0xf5, 0xf4,
  0xf3, 0xf2, 0xf1, 0xf0, 0xef, 0xee, 0xed, 0xec, 0xeb, 0xea, 0xe9, 0xe8,
  0xe7, 0xe6, 0xe5, 0xe4, 0xe3, 0xe2, 0xe1, 0xe0, 0xdf, 0xde, 0xdd, 0xdc,
  0xdb, 0xda, 0xd9, 0xd8, 0xd7, 0xd6, 0xd5, 0xd4, 0xd3, 0xd2, 0xd1, 0xd0,
  0xcf, 0xce, 0xcd, 0xcc, 0xcb, 0xca, 0xc9, 0xc8, 0xc7, 0xc6, 0xc5, 0xc4,
  0xc3, 0xc2, 0xc1, 0xc0, 0xbf, 0xbe, 0xbd, 0xbc, 0xbb, 0xba, 0xb9, 0xb8,
  0xb7, 0xb6, 0xb5, 0xb4, 0xb3, 0xb2, 0xb1, 0xb0, 0xaf, 0xae, 0xad, 0xac,
  0xab, 0xaa, 0xa9, 0xa8, 0xa7, 0xa6, 0xa5, 0xa4, 0xa3, 0xa2, 0xa1, 0xa0,
  0x9f, 0x9e, 0x9d, 0x9c, 0x9b, 0x9a, 0x99, 0x98, 0x97, 0x96, 0x95, 0x94,
  0x93, 0x92, 0x91, 0x90, 0x8f, 0x8e, 0x8d, 0x8c, 0x8b, 0x8a, 0x89, 0x88,
  0x87, 0x86, 0x85, 0x84, 0x83, 0x82, 0x81, 0x80, 0x7f, 0x7e, 0x7d, 0x7c,
  0x7b, 0x7a, 0x79, 0x78, 0x77, 0x76, 0x75, 0x74, 0x73, 0x72, 0x71, 0x70,
  0x6f, 0x6e, 0x6d, 0x6c, 0x6b, 0x6a, 0x69, 0x68, 0x67, 0x66, 0x65, 0x64,
  0x63, 0x62, 0x61, 0x60, 0x5f, 0x5e, 0x5d, 0x5c, 0x5b, 0x5a, 0x59, 0x58,
  0x57, 0x56, 0x55, 0x54, 0x53, 0x52, 0x51, 0x50, 0x4f, 0x4e, 0x4d, 0x4c,
  0x4b, 0x4a, 0x49, 0x48, 0x47, 0x46, 0x45, 0x44, 0x43, 0x42, 0x41, 0x40,
  0x3f, 0x3e, 0x3d, 0x3c, 0x3b, 0x3a, 0x39, 0x38, 0x37, 0x36, 0x35, 0x34,
  0x33, 0x32, 0x31, 0x30, 0x2f, 0x2e, 0x2d, 0x2c, 0x2b, 0x2a, 0x29, 0x28,
  0x27, 0x26, 0x25, 0x24, 0x23, 0x22, 0x21, 0x20, 0x1f, 0x1e, 0x1d, 0x1c,
  0x1b, 0x1a, 0x19, 0x18, 0x17, 0x16, 0x15, 0x14, 0x13, 0x12, 0x11, 0x10,
  0x0f, 0x0e, 0x0d, 0x0c, 0x0b, 0x0a, 0x09, 0x08, 0x07, 0x06, 0x05, 0x04,
  0x03, 0x02, 0x01, 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08,
  0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f, 0x10, 0x11, 0x12, 0x13, 0x14,
  0x15, 0x16, 0x17, 0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f, 0x20,
  0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2a, 0x2b, 0x2c,
  0x2d, 0x2e, 0x2f, 0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38,
  0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f, 0x40, 0x41, 0x42, 0x43, 0x44,
  0x45, 0x46, 0x47, 0x48, 0x49, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f, 0x50,
  0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5a, 0x5b, 0x5c,
  0x5d, 0x5e, 0x5f, 0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68,
  0x69, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f, 0x70, 0x71, 0x72, 0x73, 0x74,
  0x75, 0x76, 0x77, 0x78, 0x79, 0x7a, 0x7b, 0x7c, 0x7d, 0x7e, 0x7f, 0x80,
  0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x8a, 0x8b, 0x8c,
  0x8d, 0x8e, 0x8f, 0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98,
  0x99, 0x9a, 0x9b, 0x9c, 0x9d, 0x9e, 0x9f, 0xa0, 0xa1, 0xa2, 0xa3, 0xa4,
  0xa5, 0xa6, 0xa7, 0xa8, 0xa9, 0xaa, 0xab, 0xac, 0xad, 0xae, 0xaf, 0xb0,
  0xb1, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6, 0xb7, 0xb8, 0xb9, 0xba, 0xbb, 0xbc,
  0xbd, 0xbe, 0xbf, 0xc0, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7, 0xc8,
  0xc9, 0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf, 0xd0, 0xd1, 0xd2, 0xd3, 0xd4,
  0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda, 0xdb, 0xdc, 0xdd, 0xde, 0xdf, 0xe0,
  0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9, 0xea, 0xeb, 0xec,
  0xed, 0xee, 0xef, 0xf0, 0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8,
  0xf9, 0xfa, 0xfb, 0xfc, 0xfd, 0xfe, 0xff
};

const uint8_t* const VP8kabs0 = &abs0[255];

static void __cc_DC16NoTopLeft(uint8_t *dst) {
  __cc_Put16(0x80, dst);
}

static void __cc_HE16(uint8_t *dst) {
  int j;
  for (j = 16; j > 0; --j) {
    memset(dst, dst[-1], 16);
    dst += 32;
  }
}

static void __cc_DC16NoTop(uint8_t *dst) {
  int DC = 8;
  int j;
  for (j = 0; j < 16; ++j) {
    DC += dst[-1 + j * 32];
  }
  __cc_Put16(DC >> 4, dst);
}

static inline void __cc_TrueMotion(uint8_t *dst, int size) {
  const uint8_t* top = dst - 32;
  const uint8_t* const clip0 = VP8kclip1 - top[-1];
  int y;
  for (y = 0; y < size; ++y) {
    const uint8_t* const clip = clip0 + dst[-1];
    int x;
    for (x = 0; x < size; ++x) {
      dst[x] = clip[top[x]];
    }
    dst += 32;
  }
}

static void __cc_TM16(uint8_t *dst) { __cc_TrueMotion(dst, 16); }

static void __cc_VE16(uint8_t *dst) {
  int j;
  for (j = 0; j < 16; ++j) {
    memcpy(dst + j * 32, dst - 32, 16);
  }
}

const __cc_VP8PredFunc VP8PredLuma16[NUM_B_DC_MODES] = {
  __cc_DC16, __cc_TM16, __cc_VE16, __cc_HE16,
  __cc_DC16NoTop, __cc_DC16NoLeft, __cc_DC16NoTopLeft
};

static void __cc_UpsampleRgb565LinePair(const uint8_t* top_y, const uint8_t* bottom_y, const uint8_t* top_u, const uint8_t* top_v, const uint8_t* cur_u, const uint8_t* cur_v, uint8_t* top_dst, uint8_t* bottom_dst, int len) { int x; const int last_pixel_pair = (len - 1) >> 1; uint32_t tl_uv = ((top_u[0]) | ((top_v[0]) << 16)); uint32_t l_uv = ((cur_u[0]) | ((cur_v[0]) << 16)); ((top_y != ((void *)0)) ? (void) (0) : __assert_fail ("top_y != ((void *)0)", "extraction.1.c", 9806, __PRETTY_FUNCTION__)); { const uint32_t uv0 = (3 * tl_uv + l_uv + 0x00020002u) >> 2; __cc_VP8YuvToRgb565(top_y[0], uv0 & 0xff, (uv0 >> 16), top_dst); } if (bottom_y != ((void *)0)) { const uint32_t uv0 = (3 * l_uv + tl_uv + 0x00020002u) >> 2; __cc_VP8YuvToRgb565(bottom_y[0], uv0 & 0xff, (uv0 >> 16), bottom_dst); } for (x = 1; x <= last_pixel_pair; ++x) { const uint32_t t_uv = ((top_u[x]) | ((top_v[x]) << 16)); const uint32_t uv = ((cur_u[x]) | ((cur_v[x]) << 16)); const uint32_t avg = tl_uv + t_uv + l_uv + uv + 0x00080008u; const uint32_t diag_12 = (avg + 2 * (t_uv + l_uv)) >> 3; const uint32_t diag_03 = (avg + 2 * (tl_uv + uv)) >> 3; { const uint32_t uv0 = (diag_12 + tl_uv) >> 1; const uint32_t uv1 = (diag_03 + t_uv) >> 1; __cc_VP8YuvToRgb565(top_y[2 * x - 1], uv0 & 0xff, (uv0 >> 16), top_dst + (2 * x - 1) * 2); __cc_VP8YuvToRgb565(top_y[2 * x - 0], uv1 & 0xff, (uv1 >> 16), top_dst + (2 * x - 0) * 2); } if (bottom_y != ((void *)0)) { const uint32_t uv0 = (diag_03 + l_uv) >> 1; const uint32_t uv1 = (diag_12 + uv) >> 1; __cc_VP8YuvToRgb565(bottom_y[2 * x - 1], uv0 & 0xff, (uv0 >> 16), bottom_dst + (2 * x - 1) * 2); __cc_VP8YuvToRgb565(bottom_y[2 * x + 0], uv1 & 0xff, (uv1 >> 16), bottom_dst + (2 * x + 0) * 2); } tl_uv = t_uv; l_uv = uv; } if (!(len & 1)) { { const uint32_t uv0 = (3 * tl_uv + l_uv + 0x00020002u) >> 2; __cc_VP8YuvToRgb565(top_y[len - 1], uv0 & 0xff, (uv0 >> 16), top_dst + (len - 1) * 2); } if (bottom_y != ((void *)0)) { const uint32_t uv0 = (3 * l_uv + tl_uv + 0x00020002u) >> 2; __cc_VP8YuvToRgb565(bottom_y[len - 1], uv0 & 0xff, (uv0 >> 16), bottom_dst + (len - 1) * 2); } } }

static void __cc_DC4(uint8_t *dst) {
  uint32_t dc = 4;
  int i;
  for (i = 0; i < 4; ++i) dc += dst[i - 32] + dst[-1 + i * 32];
  dc >>= 3;
  for (i = 0; i < 4; ++i) memset(dst + i * 32, dc, 4);
}

int __cc_WebPIoInitFromOptions(const __cc_WebPDecoderOptions* const options,
                          __cc_VP8Io* const io, __cc_WEBP_CSP_MODE src_colorspace) {
  const int W = io->width;
  const int H = io->height;
  int x = 0, y = 0, w = W, h = H;


  io->use_cropping = (options != ((void *)0)) && (options->use_cropping > 0);
  if (io->use_cropping) {
    w = options->crop_width;
    h = options->crop_height;
    x = options->crop_left;
    y = options->crop_top;
    if (!__cc_WebPIsRGBMode(src_colorspace)) {
      x &= ~1;
      y &= ~1;
    }
    if (x < 0 || y < 0 || w <= 0 || h <= 0 || x + w > W || y + h > H) {
      return 0;
    }
  }
  io->crop_left = x;
  io->crop_top = y;
  io->crop_right = x + w;
  io->crop_bottom = y + h;
  io->mb_w = w;
  io->mb_h = h;


  io->use_scaling = (options != ((void *)0)) && (options->use_scaling > 0);
  if (io->use_scaling) {
    if (options->scaled_width <= 0 || options->scaled_height <= 0) {
      return 0;
    }
    io->scaled_width = options->scaled_width;
    io->scaled_height = options->scaled_height;
  }


  io->bypass_filtering = options && options->bypass_filtering;



  io->fancy_upsampling = (options == ((void *)0)) || (!options->no_fancy_upsampling);


  if (io->use_scaling) {

    io->bypass_filtering = (io->scaled_width < W * 3 / 4) &&
                           (io->scaled_height < H * 3 / 4);
    io->fancy_upsampling = 0;
  }
  return 1;
}




static void __cc_TransformDC(const int16_t *in, uint8_t* dst) {
  const int DC = in[0] + 4;
  int i, j;
  for (j = 0; j < 4; ++j) {
    for (i = 0; i < 4; ++i) {
      dst[i + j * 32] = __cc_clip_8b(dst[i + j * 32] + ((DC) >> 3));
    }
  }
}

static const int kC1 = 20091 + (1 << 16);



static void __cc_TransformOne(const int16_t* in, uint8_t* dst) {
  int C[4 * 4], *tmp;
  int i;
  tmp = C;
  for (i = 0; i < 4; ++i) {
    const int a = in[0] + in[8];
    const int b = in[0] - in[8];
    const int c = (((in[4]) * (kC2)) >> 16) - (((in[12]) * (kC1)) >> 16);
    const int d = (((in[4]) * (kC1)) >> 16) + (((in[12]) * (kC2)) >> 16);
    tmp[0] = a + d;
    tmp[1] = b + c;
    tmp[2] = b - c;
    tmp[3] = a - d;
    tmp += 4;
    in++;
  }







  tmp = C;
  for (i = 0; i < 4; ++i) {
    const int dc = tmp[0] + 4;
    const int a = dc + tmp[8];
    const int b = dc - tmp[8];
    const int c = (((tmp[4]) * (kC2)) >> 16) - (((tmp[12]) * (kC1)) >> 16);
    const int d = (((tmp[4]) * (kC1)) >> 16) + (((tmp[12]) * (kC2)) >> 16);
    dst[0 + 0 * 32] = __cc_clip_8b(dst[0 + 0 * 32] + ((a + d) >> 3));
    dst[1 + 0 * 32] = __cc_clip_8b(dst[1 + 0 * 32] + ((b + c) >> 3));
    dst[2 + 0 * 32] = __cc_clip_8b(dst[2 + 0 * 32] + ((b - c) >> 3));
    dst[3 + 0 * 32] = __cc_clip_8b(dst[3 + 0 * 32] + ((a - d) >> 3));
    tmp++;
    dst += 32;
  }
}





static void __cc_RD4(uint8_t *dst) {
  const int I = dst[-1 + 0 * 32];
  const int J = dst[-1 + 1 * 32];
  const int K = dst[-1 + 2 * 32];
  const int L = dst[-1 + 3 * 32];
  const int X = dst[-1 - 32];
  const int A = dst[0 - 32];
  const int B = dst[1 - 32];
  const int C = dst[2 - 32];
  const int D = dst[3 - 32];
  dst[(0) + (3) * 32] = (((J) + 2 * (K) + (L) + 2) >> 2);
  dst[(0) + (2) * 32] = dst[(1) + (3) * 32] = (((I) + 2 * (J) + (K) + 2) >> 2);
  dst[(0) + (1) * 32] = dst[(1) + (2) * 32] = dst[(2) + (3) * 32] = (((X) + 2 * (I) + (J) + 2) >> 2);
  dst[(0) + (0) * 32] = dst[(1) + (1) * 32] = dst[(2) + (2) * 32] = dst[(3) + (3) * 32] = (((A) + 2 * (X) + (I) + 2) >> 2);
  dst[(1) + (0) * 32] = dst[(2) + (1) * 32] = dst[(3) + (2) * 32] = (((B) + 2 * (A) + (X) + 2) >> 2);
  dst[(2) + (0) * 32] = dst[(3) + (1) * 32] = (((C) + 2 * (B) + (A) + 2) >> 2);
  dst[(3) + (0) * 32] = (((D) + 2 * (C) + (B) + 2) >> 2);
}

static void __cc_DC8uvNoTop(uint8_t *dst) {
  int dc0 = 4;
  int i;
  for (i = 0; i < 8; ++i) {
    dc0 += dst[-1 + i * 32];
  }
  __cc_Put8x8uv(dc0 >> 3, dst);
}

static void __cc_HE8uv(uint8_t *dst) {
  int j;
  for (j = 0; j < 8; ++j) {
    memset(dst, dst[-1], 8);
    dst += 32;
  }
}

static void __cc_VE8uv(uint8_t *dst) {
  int j;
  for (j = 0; j < 8; ++j) {
    memcpy(dst + j * 32, dst - 32, 8);
  }
}

static void __cc_TM8uv(uint8_t *dst) { __cc_TrueMotion(dst, 8); }

static void __cc_DC8uvNoTopLeft(uint8_t *dst) {
  __cc_Put8x8uv(0x80, dst);
}

static void __cc_DC8uv(uint8_t *dst) {
  int dc0 = 8;
  int i;
  for (i = 0; i < 8; ++i) {
    dc0 += dst[i - 32] + dst[-1 + i * 32];
  }
  __cc_Put8x8uv(dc0 >> 4, dst);
}

const __cc_VP8PredFunc VP8PredChroma8[NUM_B_DC_MODES] = {
  __cc_DC8uv, __cc_TM8uv, __cc_VE8uv, __cc_HE8uv,
  __cc_DC8uvNoTop, __cc_DC8uvNoLeft, __cc_DC8uvNoTopLeft
};



static void __cc_HD4(uint8_t *dst) {
  const int I = dst[-1 + 0 * 32];
  const int J = dst[-1 + 1 * 32];
  const int K = dst[-1 + 2 * 32];
  const int L = dst[-1 + 3 * 32];
  const int X = dst[-1 - 32];
  const int A = dst[0 - 32];
  const int B = dst[1 - 32];
  const int C = dst[2 - 32];

  dst[(0) + (0) * 32] = dst[(2) + (1) * 32] = (((I) + (X) + 1) >> 1);
  dst[(0) + (1) * 32] = dst[(2) + (2) * 32] = (((J) + (I) + 1) >> 1);
  dst[(0) + (2) * 32] = dst[(2) + (3) * 32] = (((K) + (J) + 1) >> 1);
  dst[(0) + (3) * 32] = (((L) + (K) + 1) >> 1);

  dst[(3) + (0) * 32] = (((A) + 2 * (B) + (C) + 2) >> 2);
  dst[(2) + (0) * 32] = (((X) + 2 * (A) + (B) + 2) >> 2);
  dst[(1) + (0) * 32] = dst[(3) + (1) * 32] = (((I) + 2 * (X) + (A) + 2) >> 2);
  dst[(1) + (1) * 32] = dst[(3) + (2) * 32] = (((J) + 2 * (I) + (X) + 2) >> 2);
  dst[(1) + (2) * 32] = dst[(3) + (3) * 32] = (((K) + 2 * (J) + (I) + 2) >> 2);
  dst[(1) + (3) * 32] = (((L) + 2 * (K) + (J) + 2) >> 2);
}

static void __cc_VE4(uint8_t *dst) {
  const uint8_t* top = dst - 32;
  const uint8_t vals[4] = {
    (((top[-1]) + 2 * (top[0]) + (top[1]) + 2) >> 2),
    (((top[ 0]) + 2 * (top[1]) + (top[2]) + 2) >> 2),
    (((top[ 1]) + 2 * (top[2]) + (top[3]) + 2) >> 2),
    (((top[ 2]) + 2 * (top[3]) + (top[4]) + 2) >> 2)
  };
  int i;
  for (i = 0; i < 4; ++i) {
    memcpy(dst + i * 32, vals, sizeof(vals));
  }
}



int __cc_VP8LDecodeAlphaHeader(struct ALPHDecoder* const alph_dec,
                          const uint8_t* const data, size_t data_size,
                          uint8_t* const output);

static int __cc_ALPHInit(__cc_ALPHDecoder* const dec, const uint8_t* data,
                    size_t data_size, int width, int height, uint8_t* output) {
  int ok = 0;
  const uint8_t* const alpha_data = data + 1;
  const size_t alpha_data_size = data_size - 1;
  int rsrv;

  ((width > 0 && height > 0) ? (void) (0) : __assert_fail ("width > 0 && height > 0", "extraction.1.c", 10046, __PRETTY_FUNCTION__));
  ((data != ((void *)0) && output != ((void *)0)) ? (void) (0) : __assert_fail ("data != ((void *)0) && output != ((void *)0)", "extraction.1.c", 10047, __PRETTY_FUNCTION__));

  dec->width_ = width;
  dec->height_ = height;

  if (data_size <= 1) {
    return 0;
  }

  dec->method_ = (data[0] >> 0) & 0x03;
  dec->filter_ = (data[0] >> 2) & 0x03;
  dec->pre_processing_ = (data[0] >> 4) & 0x03;
  rsrv = (data[0] >> 6) & 0x03;
  if (dec->method_ < 0 ||
      dec->method_ > 1 ||
      dec->filter_ >= WEBP_FILTER_LAST ||
      dec->pre_processing_ > 1 ||
      rsrv != 0) {
    return 0;
  }

  if (dec->method_ == 0) {
    const size_t alpha_decoded_size = dec->width_ * dec->height_;
    ok = (alpha_data_size >= alpha_decoded_size);
  } else {
    ((dec->method_ == 1) ? (void) (0) : __assert_fail ("dec->method_ == 1", "extraction.1.c", 10072, __PRETTY_FUNCTION__));
    ok = __cc_VP8LDecodeAlphaHeader(dec, alpha_data, alpha_data_size, output);
  }
  return ok;
}

static inline void __cc_do_filter6(uint8_t* p, int step) {
  const int p2 = p[-3*step], p1 = p[-2*step], p0 = p[-step];
  const int q0 = p[0], q1 = p[step], q2 = p[2*step];
  const int a = VP8ksclip1[3 * (q0 - p0) + VP8ksclip1[p1 - q1]];

  const int a1 = (27 * a + 63) >> 7;
  const int a2 = (18 * a + 63) >> 7;
  const int a3 = (9 * a + 63) >> 7;
  p[-3*step] = VP8kclip1[p2 + a3];
  p[-2*step] = VP8kclip1[p1 + a2];
  p[- step] = VP8kclip1[p0 + a1];
  p[ 0] = VP8kclip1[q0 - a1];
  p[ step] = VP8kclip1[q1 - a2];
  p[ 2*step] = VP8kclip1[q2 - a3];
}

static inline void __cc_FilterLoop26(uint8_t* p,
                                     int hstride, int vstride, int size,
                                     int thresh, int ithresh, int hev_thresh) {
  const int thresh2 = 2 * thresh + 1;
  while (size-- > 0) {
    if (__cc_needs_filter2(p, hstride, thresh2, ithresh)) {
      if (__cc_hev(p, hstride, hev_thresh)) {
        __cc_do_filter2(p, hstride);
      } else {
        __cc_do_filter6(p, hstride);
      }
    }
    p += vstride;
  }
}

__cc_VP8DecIdct VP8TransformAC3;

static void __cc_UpsampleRgbLinePair(const uint8_t* top_y, const uint8_t* bottom_y, const uint8_t* top_u, const uint8_t* top_v, const uint8_t* cur_u, const uint8_t* cur_v, uint8_t* top_dst, uint8_t* bottom_dst, int len) { int x; const int last_pixel_pair = (len - 1) >> 1; uint32_t tl_uv = ((top_u[0]) | ((top_v[0]) << 16)); uint32_t l_uv = ((cur_u[0]) | ((cur_v[0]) << 16)); ((top_y != ((void *)0)) ? (void) (0) : __assert_fail ("top_y != ((void *)0)", "extraction.1.c", 10112, __PRETTY_FUNCTION__)); { const uint32_t uv0 = (3 * tl_uv + l_uv + 0x00020002u) >> 2; __cc_VP8YuvToRgb(top_y[0], uv0 & 0xff, (uv0 >> 16), top_dst); } if (bottom_y != ((void *)0)) { const uint32_t uv0 = (3 * l_uv + tl_uv + 0x00020002u) >> 2; __cc_VP8YuvToRgb(bottom_y[0], uv0 & 0xff, (uv0 >> 16), bottom_dst); } for (x = 1; x <= last_pixel_pair; ++x) { const uint32_t t_uv = ((top_u[x]) | ((top_v[x]) << 16)); const uint32_t uv = ((cur_u[x]) | ((cur_v[x]) << 16)); const uint32_t avg = tl_uv + t_uv + l_uv + uv + 0x00080008u; const uint32_t diag_12 = (avg + 2 * (t_uv + l_uv)) >> 3; const uint32_t diag_03 = (avg + 2 * (tl_uv + uv)) >> 3; { const uint32_t uv0 = (diag_12 + tl_uv) >> 1; const uint32_t uv1 = (diag_03 + t_uv) >> 1; __cc_VP8YuvToRgb(top_y[2 * x - 1], uv0 & 0xff, (uv0 >> 16), top_dst + (2 * x - 1) * 3); __cc_VP8YuvToRgb(top_y[2 * x - 0], uv1 & 0xff, (uv1 >> 16), top_dst + (2 * x - 0) * 3); } if (bottom_y != ((void *)0)) { const uint32_t uv0 = (diag_03 + l_uv) >> 1; const uint32_t uv1 = (diag_12 + uv) >> 1; __cc_VP8YuvToRgb(bottom_y[2 * x - 1], uv0 & 0xff, (uv0 >> 16), bottom_dst + (2 * x - 1) * 3); __cc_VP8YuvToRgb(bottom_y[2 * x + 0], uv1 & 0xff, (uv1 >> 16), bottom_dst + (2 * x + 0) * 3); } tl_uv = t_uv; l_uv = uv; } if (!(len & 1)) { { const uint32_t uv0 = (3 * tl_uv + l_uv + 0x00020002u) >> 2; __cc_VP8YuvToRgb(top_y[len - 1], uv0 & 0xff, (uv0 >> 16), top_dst + (len - 1) * 3); } if (bottom_y != ((void *)0)) { const uint32_t uv0 = (3 * l_uv + tl_uv + 0x00020002u) >> 2; __cc_VP8YuvToRgb(bottom_y[len - 1], uv0 & 0xff, (uv0 >> 16), bottom_dst + (len - 1) * 3); } } }

static void __cc_HU4(uint8_t *dst) {
  const int I = dst[-1 + 0 * 32];
  const int J = dst[-1 + 1 * 32];
  const int K = dst[-1 + 2 * 32];
  const int L = dst[-1 + 3 * 32];
  dst[(0) + (0) * 32] = (((I) + (J) + 1) >> 1);
  dst[(2) + (0) * 32] = dst[(0) + (1) * 32] = (((J) + (K) + 1) >> 1);
  dst[(2) + (1) * 32] = dst[(0) + (2) * 32] = (((K) + (L) + 1) >> 1);
  dst[(1) + (0) * 32] = (((I) + 2 * (J) + (K) + 2) >> 2);
  dst[(3) + (0) * 32] = dst[(1) + (1) * 32] = (((J) + 2 * (K) + (L) + 2) >> 2);
  dst[(3) + (1) * 32] = dst[(1) + (2) * 32] = (((K) + 2 * (L) + (L) + 2) >> 2);
  dst[(3) + (2) * 32] = dst[(2) + (2) * 32] =
    dst[(0) + (3) * 32] = dst[(1) + (3) * 32] = dst[(2) + (3) * 32] = dst[(3) + (3) * 32] = L;
}

void __cc_WebPFreeDecBuffer(__cc_WebPDecBuffer* buffer) {
  if (buffer != ((void *)0)) {
    if (!buffer->is_external_memory) {
      __cc_WebPSafeFree(buffer->private_memory);
    }
    buffer->private_memory = ((void *)0);
  }
}

void __cc_VP8InitRandom(__cc_VP8Random* const rg, float dithering) {
  memcpy(rg->tab_, kRandomTable, sizeof(rg->tab_));
  rg->index1_ = 0;
  rg->index2_ = 31;
  rg->amp_ = (dithering < 0.0) ? 0
           : (dithering > 1.0) ? (1 << 8)
           : (uint32_t)((1 << 8) * dithering);
}

static void __cc_VFilter16(uint8_t* p, int stride,
                      int thresh, int ithresh, int hev_thresh) {
  __cc_FilterLoop26(p, stride, 1, 16, thresh, ithresh, hev_thresh);
}

__cc_VP8LumaFilterFunc VP8HFilter16i;

__cc_VP8ChromaFilterFunc VP8VFilter8;

static void __cc_VFilter8i(uint8_t* u, uint8_t* v, int stride,
                      int thresh, int ithresh, int hev_thresh) {
  __cc_FilterLoop24(u + 4 * stride, stride, 1, 8, thresh, ithresh, hev_thresh);
  __cc_FilterLoop24(v + 4 * stride, stride, 1, 8, thresh, ithresh, hev_thresh);
}

__cc_VP8DecIdct VP8TransformDC;

static void __cc_HFilter16i(uint8_t* p, int stride,
                       int thresh, int ithresh, int hev_thresh) {
  int k;
  for (k = 3; k > 0; --k) {
    p += 4;
    __cc_FilterLoop24(p, 1, stride, 16, thresh, ithresh, hev_thresh);
  }
}

static void __cc_HE4(uint8_t *dst) {
  const int A = dst[-1 - 32];
  const int B = dst[-1];
  const int C = dst[-1 + 32];
  const int D = dst[-1 + 2 * 32];
  const int E = dst[-1 + 3 * 32];
  *(uint32_t*)(dst + 0 * 32) = 0x01010101U * (((A) + 2 * (B) + (C) + 2) >> 2);
  *(uint32_t*)(dst + 1 * 32) = 0x01010101U * (((B) + 2 * (C) + (D) + 2) >> 2);
  *(uint32_t*)(dst + 2 * 32) = 0x01010101U * (((C) + 2 * (D) + (E) + 2) >> 2);
  *(uint32_t*)(dst + 3 * 32) = 0x01010101U * (((D) + 2 * (E) + (E) + 2) >> 2);
}

__cc_VP8SimpleFilterFunc VP8SimpleVFilter16i;

int __cc_VP8LDecodeAlphaImageStream(struct ALPHDecoder* const alph_dec,
                               int last_row);

__cc_ALPHDecoder* __cc_ALPHNew(void) {
  __cc_ALPHDecoder* const dec = (__cc_ALPHDecoder*)__cc_WebPSafeCalloc(1ULL, sizeof(*dec));
  return dec;
}

__cc_VP8DecIdct VP8TransformDCUV;



__cc_VP8SimpleFilterFunc VP8SimpleHFilter16i;

static void __cc_TransformAC3(const int16_t* in, uint8_t* dst) {
  const int a = in[0] + 4;
  const int c4 = (((in[4]) * (kC2)) >> 16);
  const int d4 = (((in[4]) * (kC1)) >> 16);
  const int c1 = (((in[1]) * (kC2)) >> 16);
  const int d1 = (((in[1]) * (kC1)) >> 16);
  do { const int DC = (a + d4); dst[0 + 0 * 32] = __cc_clip_8b(dst[0 + 0 * 32] + ((DC + (d1)) >> 3)); dst[1 + 0 * 32] = __cc_clip_8b(dst[1 + 0 * 32] + ((DC + (c1)) >> 3)); dst[2 + 0 * 32] = __cc_clip_8b(dst[2 + 0 * 32] + ((DC - (c1)) >> 3)); dst[3 + 0 * 32] = __cc_clip_8b(dst[3 + 0 * 32] + ((DC - (d1)) >> 3)); } while (0);
  do { const int DC = (a + c4); dst[0 + 1 * 32] = __cc_clip_8b(dst[0 + 1 * 32] + ((DC + (d1)) >> 3)); dst[1 + 1 * 32] = __cc_clip_8b(dst[1 + 1 * 32] + ((DC + (c1)) >> 3)); dst[2 + 1 * 32] = __cc_clip_8b(dst[2 + 1 * 32] + ((DC - (c1)) >> 3)); dst[3 + 1 * 32] = __cc_clip_8b(dst[3 + 1 * 32] + ((DC - (d1)) >> 3)); } while (0);
  do { const int DC = (a - c4); dst[0 + 2 * 32] = __cc_clip_8b(dst[0 + 2 * 32] + ((DC + (d1)) >> 3)); dst[1 + 2 * 32] = __cc_clip_8b(dst[1 + 2 * 32] + ((DC + (c1)) >> 3)); dst[2 + 2 * 32] = __cc_clip_8b(dst[2 + 2 * 32] + ((DC - (c1)) >> 3)); dst[3 + 2 * 32] = __cc_clip_8b(dst[3 + 2 * 32] + ((DC - (d1)) >> 3)); } while (0);
  do { const int DC = (a - d4); dst[0 + 3 * 32] = __cc_clip_8b(dst[0 + 3 * 32] + ((DC + (d1)) >> 3)); dst[1 + 3 * 32] = __cc_clip_8b(dst[1 + 3 * 32] + ((DC + (c1)) >> 3)); dst[2 + 3 * 32] = __cc_clip_8b(dst[2 + 3 * 32] + ((DC - (c1)) >> 3)); dst[3 + 3 * 32] = __cc_clip_8b(dst[3 + 3 * 32] + ((DC - (d1)) >> 3)); } while (0);
}

__cc_VP8LumaFilterFunc VP8VFilter16i;

static void __cc_SimpleVFilter16(uint8_t* p, int stride, int thresh) {
  int i;
  const int thresh2 = 2 * thresh + 1;
  for (i = 0; i < 16; ++i) {
    if (__cc_needs_filter(p + i, stride, thresh2)) {
      __cc_do_filter2(p + i, stride);
    }
  }
}

static void __cc_SimpleVFilter16i(uint8_t* p, int stride, int thresh) {
  int k;
  for (k = 3; k > 0; --k) {
    p += 4 * stride;
    __cc_SimpleVFilter16(p, stride, thresh);
  }
}

static void __cc_VL4(uint8_t *dst) {
  const int A = dst[0 - 32];
  const int B = dst[1 - 32];
  const int C = dst[2 - 32];
  const int D = dst[3 - 32];
  const int E = dst[4 - 32];
  const int F = dst[5 - 32];
  const int G = dst[6 - 32];
  const int H = dst[7 - 32];
  dst[(0) + (0) * 32] = (((A) + (B) + 1) >> 1);
  dst[(1) + (0) * 32] = dst[(0) + (2) * 32] = (((B) + (C) + 1) >> 1);
  dst[(2) + (0) * 32] = dst[(1) + (2) * 32] = (((C) + (D) + 1) >> 1);
  dst[(3) + (0) * 32] = dst[(2) + (2) * 32] = (((D) + (E) + 1) >> 1);

  dst[(0) + (1) * 32] = (((A) + 2 * (B) + (C) + 2) >> 2);
  dst[(1) + (1) * 32] = dst[(0) + (3) * 32] = (((B) + 2 * (C) + (D) + 2) >> 2);
  dst[(2) + (1) * 32] = dst[(1) + (3) * 32] = (((C) + 2 * (D) + (E) + 2) >> 2);
  dst[(3) + (1) * 32] = dst[(2) + (3) * 32] = (((D) + 2 * (E) + (F) + 2) >> 2);
              dst[(3) + (2) * 32] = (((E) + 2 * (F) + (G) + 2) >> 2);
              dst[(3) + (3) * 32] = (((F) + 2 * (G) + (H) + 2) >> 2);
}

__cc_VP8SimpleFilterFunc VP8SimpleVFilter16;

__cc_VP8LumaFilterFunc VP8VFilter16;

extern const __cc_VP8PredFunc VP8PredLuma16[ ];

extern __cc_VP8DecIdct2 VP8Transform;

uint32_t __cc_VP8GetValue(__cc_VP8BitReader* const br, int bits) {
  uint32_t v = 0;
  while (bits-- > 0) {
    v |= __cc_VP8GetBit(br, 0x80) << bits;
  }
  return v;
}

int32_t __cc_VP8GetSignedValue(__cc_VP8BitReader* const br, int bits) {
  const int value = __cc_VP8GetValue(br, bits);
  return __cc_VP8Get(br) ? -value : value;
}

int __cc_VP8LHuffmanCodeLengthsToCodes(
    const int* const code_lengths, int code_lengths_size,
    int* const huff_codes) {
  int symbol;
  int code_len;
  int code_length_hist[15 + 1] = { 0 };
  int curr_code;
  int next_codes[15 + 1] = { 0 };
  int max_code_length = 0;

  ((code_lengths != ((void *)0)) ? (void) (0) : __assert_fail ("code_lengths != ((void *)0)", "extraction.1.c", 10286, __PRETTY_FUNCTION__));
  ((code_lengths_size > 0) ? (void) (0) : __assert_fail ("code_lengths_size > 0", "extraction.1.c", 10287, __PRETTY_FUNCTION__));
  ((huff_codes != ((void *)0)) ? (void) (0) : __assert_fail ("huff_codes != ((void *)0)", "extraction.1.c", 10288, __PRETTY_FUNCTION__));


  for (symbol = 0; symbol < code_lengths_size; ++symbol) {
    if (code_lengths[symbol] > max_code_length) {
      max_code_length = code_lengths[symbol];
    }
  }
  if (max_code_length > 15) return 0;


  for (symbol = 0; symbol < code_lengths_size; ++symbol) {
    ++code_length_hist[code_lengths[symbol]];
  }
  code_length_hist[0] = 0;




  curr_code = 0;
  next_codes[0] = -1;
  for (code_len = 1; code_len <= max_code_length; ++code_len) {
    curr_code = (curr_code + code_length_hist[code_len - 1]) << 1;
    next_codes[code_len] = curr_code;
  }


  for (symbol = 0; symbol < code_lengths_size; ++symbol) {
    if (code_lengths[symbol] > 0) {
      huff_codes[symbol] = next_codes[code_lengths[symbol]]++;
    } else {
      huff_codes[symbol] = (-1);
    }
  }
  return 1;
}

static void __cc_Yuv444ToRgba4444(const uint8_t* y, const uint8_t* u, const uint8_t* v, uint8_t* dst, int len) { int i; for (i = 0; i < len; ++i) __cc_VP8YuvToRgba4444(y[i], u[i], v[i], &dst[i * 2]); }

static void __cc_VFilter8(uint8_t* u, uint8_t* v, int stride,
                     int thresh, int ithresh, int hev_thresh) {
  __cc_FilterLoop26(u, stride, 1, 8, thresh, ithresh, hev_thresh);
  __cc_FilterLoop26(v, stride, 1, 8, thresh, ithresh, hev_thresh);
}

static void __cc_HFilter8(uint8_t* u, uint8_t* v, int stride,
                     int thresh, int ithresh, int hev_thresh) {
  __cc_FilterLoop26(u, 1, stride, 8, thresh, ithresh, hev_thresh);
  __cc_FilterLoop26(v, 1, stride, 8, thresh, ithresh, hev_thresh);
}

__cc_VP8StatusCode __cc_WebPGetFeaturesInternal(const uint8_t* data, size_t data_size,
                                      __cc_WebPBitstreamFeatures* features,
                                      int version) {
  if ((((version) >> 8) != ((0x0203) >> 8))) {
    return VP8_STATUS_INVALID_PARAM;
  }
  if (features == ((void *)0)) {
    return VP8_STATUS_INVALID_PARAM;
  }
  return __cc_GetFeatures(data, data_size, features);
}

__cc_VP8LumaFilterFunc VP8HFilter16;

int __cc_ExUtilLoadWebP(const char* const in_file,
                   const uint8_t** data, size_t* data_size,
                   __cc_WebPBitstreamFeatures* bitstream);

static inline int __cc_WebPInitDecoderConfig(__cc_WebPDecoderConfig* config) {
  return __cc_WebPInitDecoderConfigInternal(config, 0x0203);
}

__cc_VP8StatusCode __cc_ExUtilDecodeWebP(const uint8_t* const data, size_t data_size,
                               int verbose, __cc_WebPDecoderConfig* const config);

const int8_t* const VP8ksclip2 = &sclip2[112];

typedef void (*__cc_WebPUnfilterFunc)(int width, int height, int stride,
                                 int row, int num_rows, uint8_t* data);

extern const __cc_WebPUnfilterFunc WebPUnfilters[WEBP_FILTER_LAST];

static int __cc_ALPHDecode(__cc_VP8Decoder* const dec, int row, int num_rows) {
  __cc_ALPHDecoder* const alph_dec = dec->alph_dec_;
  const int width = alph_dec->width_;
  const int height = alph_dec->height_;
  __cc_WebPUnfilterFunc unfilter_func = WebPUnfilters[alph_dec->filter_];
  uint8_t* const output = dec->alpha_plane_;
  if (alph_dec->method_ == 0) {
    const size_t offset = row * width;
    const size_t num_pixels = num_rows * width;
    ((dec->alpha_data_size_ >= 1 + offset + num_pixels) ? (void) (0) : __assert_fail ("dec->alpha_data_size_ >= 1 + offset + num_pixels", "extraction.1.c", 10380, __PRETTY_FUNCTION__));
    memcpy(dec->alpha_plane_ + offset,
           dec->alpha_data_ + 1 + offset, num_pixels);
  } else {
    ((alph_dec->vp8l_dec_ != ((void *)0)) ? (void) (0) : __assert_fail ("alph_dec->vp8l_dec_ != ((void *)0)", "extraction.1.c", 10384, __PRETTY_FUNCTION__));
    if (!__cc_VP8LDecodeAlphaImageStream(alph_dec, row + num_rows)) {
      return 0;
    }
  }

  if (unfilter_func != ((void *)0)) {
    unfilter_func(width, height, width, row, num_rows, output);
  }

  if (row + num_rows == dec->pic_hdr_.height_) {
    dec->is_alpha_decoded_ = 1;
  }
  return 1;
}

__cc_VP8StatusCode __cc_VP8EnterCritical(__cc_VP8Decoder* const dec, __cc_VP8Io* const io) {


  if (io->setup != ((void *)0) && !io->setup(io)) {
    __cc_VP8SetError(dec, VP8_STATUS_USER_ABORT, "Frame setup failed");
    return dec->status_;
  }


  if (io->bypass_filtering) {
    dec->filter_type_ = 0;
  }
  {
    const int extra_pixels = kFilterExtraRows[dec->filter_type_];
    if (dec->filter_type_ == 2) {

      dec->tl_mb_x_ = 0;
      dec->tl_mb_y_ = 0;
    } else {




      dec->tl_mb_x_ = (io->crop_left - extra_pixels) >> 4;
      dec->tl_mb_y_ = (io->crop_top - extra_pixels) >> 4;
      if (dec->tl_mb_x_ < 0) dec->tl_mb_x_ = 0;
      if (dec->tl_mb_y_ < 0) dec->tl_mb_y_ = 0;
    }

    dec->br_mb_y_ = (io->crop_bottom + 15 + extra_pixels) >> 4;
    dec->br_mb_x_ = (io->crop_right + 15 + extra_pixels) >> 4;
    if (dec->br_mb_x_ > dec->mb_w_) {
      dec->br_mb_x_ = dec->mb_w_;
    }
    if (dec->br_mb_y_ > dec->mb_h_) {
      dec->br_mb_y_ = dec->mb_h_;
    }
  }
  __cc_PrecomputeFilterStrengths(dec);
  return VP8_STATUS_OK;
}

uint32_t __cc_VP8LReadBits(__cc_VP8LBitReader* const br, int n_bits) {
  ((n_bits >= 0) ? (void) (0) : __assert_fail ("n_bits >= 0", "extraction.1.c", 10456, __PRETTY_FUNCTION__));

  if (!br->eos_ && n_bits <= 24) {
    const uint32_t val =
        (uint32_t)(br->val_ >> br->bit_pos_) & kBitMask[n_bits];
    const int new_bits = br->bit_pos_ + n_bits;
    br->bit_pos_ = new_bits;
    __cc_ShiftBytes(br);
    return val;
  } else {
    br->error_ = 1;
    return 0;
  }
}

extern __cc_VP8DecIdct VP8TransformDC;

static void __cc_Yuv444ToArgb(const uint8_t* y, const uint8_t* u, const uint8_t* v, uint8_t* dst, int len) { int i; for (i = 0; i < len; ++i) __cc_VP8YuvToArgb(y[i], u[i], v[i], &dst[i * 4]); }

static void __cc_Yuv444ToBgra(const uint8_t* y, const uint8_t* u, const uint8_t* v, uint8_t* dst, int len) { int i; for (i = 0; i < len; ++i) __cc_VP8YuvToBgra(y[i], u[i], v[i], &dst[i * 4]); }

static void __cc_Yuv444ToBgr(const uint8_t* y, const uint8_t* u, const uint8_t* v, uint8_t* dst, int len) { int i; for (i = 0; i < len; ++i) __cc_VP8YuvToBgr(y[i], u[i], v[i], &dst[i * 3]); }

const __cc_WebPYUV444Converter WebPYUV444Converters[MODE_LAST] = {
  __cc_Yuv444ToRgb,
  __cc_Yuv444ToRgba,
  __cc_Yuv444ToBgr,
  __cc_Yuv444ToBgra,
  __cc_Yuv444ToArgb,
  __cc_Yuv444ToRgba4444,
  __cc_Yuv444ToRgb565,
  __cc_Yuv444ToRgba,
  __cc_Yuv444ToBgra,
  __cc_Yuv444ToArgb,
  __cc_Yuv444ToRgba4444
};

int __cc_VP8LColorCacheInit(__cc_VP8LColorCache* const cc, int hash_bits) {
  const int hash_size = 1 << hash_bits;
  ((cc != ((void *)0)) ? (void) (0) : __assert_fail ("cc != ((void *)0)", "extraction.1.c", 10495, __PRETTY_FUNCTION__));
  ((hash_bits > 0) ? (void) (0) : __assert_fail ("hash_bits > 0", "extraction.1.c", 10496, __PRETTY_FUNCTION__));
  cc->colors_ = (uint32_t*)__cc_WebPSafeCalloc((uint64_t)hash_size,
                                          sizeof(*cc->colors_));
  if (cc->colors_ == ((void *)0)) return 0;
  cc->hash_shift_ = 32 - hash_bits;
  return 1;
}



int __cc_ExUtilLoadWebP(const char* const in_file,
                   const uint8_t** data, size_t* data_size,
                   __cc_WebPBitstreamFeatures* bitstream) {
  __cc_VP8StatusCode status;
  __cc_WebPBitstreamFeatures local_features;
  if (!__cc_ExUtilReadFile(in_file, data, data_size)) return 0;

  if (bitstream == ((void *)0)) {
    bitstream = &local_features;
  }

  status = __cc_WebPGetFeatures(*data, *data_size, bitstream);
  if (status != VP8_STATUS_OK) {
    free((void*)*data);
    *data = ((void *)0);
    *data_size = 0;
    __cc_ExUtilPrintWebPError(in_file, status);
    return 0;
  }
  return 1;
}

static void __cc_HorizontalUnfilter(int width, int height, int stride, int row,
                               int num_rows, uint8_t* data) {
  __cc_DoHorizontalFilter(data, width, height, stride, row, num_rows, 1, data);
}

const __cc_WebPUnfilterFunc WebPUnfilters[WEBP_FILTER_LAST] = {
  ((void *)0),
  __cc_HorizontalUnfilter,
  __cc_VerticalUnfilter,
  __cc_GradientUnfilter
};

extern const __cc_VP8PredFunc VP8PredLuma4[ ];

const uint8_t kVP8Log2Range[128] = {
     7, 6, 6, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 4, 4,
  3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
  2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
  2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
  0
};

__cc_VP8ChromaFilterFunc VP8HFilter8i;

static void __cc_HFilter16(uint8_t* p, int stride,
                      int thresh, int ithresh, int hev_thresh) {
  __cc_FilterLoop26(p, 1, stride, 16, thresh, ithresh, hev_thresh);
}

static void __cc_VFilter16i(uint8_t* p, int stride,
                       int thresh, int ithresh, int hev_thresh) {
  int k;
  for (k = 3; k > 0; --k) {
    p += 4 * stride;
    __cc_FilterLoop24(p, stride, 1, 16, thresh, ithresh, hev_thresh);
  }
}

static volatile __cc_VP8CPUInfo dec_last_cpuinfo_used =
    (__cc_VP8CPUInfo)&dec_last_cpuinfo_used;

__cc_VP8ChromaFilterFunc VP8HFilter8;

__cc_VP8ChromaFilterFunc VP8VFilter8i;

static void __cc_TransformWHT(const int16_t* in, int16_t* out) {
  int tmp[16];
  int i;
  for (i = 0; i < 4; ++i) {
    const int a0 = in[0 + i] + in[12 + i];
    const int a1 = in[4 + i] + in[ 8 + i];
    const int a2 = in[4 + i] - in[ 8 + i];
    const int a3 = in[0 + i] - in[12 + i];
    tmp[0 + i] = a0 + a1;
    tmp[8 + i] = a0 - a1;
    tmp[4 + i] = a3 + a2;
    tmp[12 + i] = a3 - a2;
  }
  for (i = 0; i < 4; ++i) {
    const int dc = tmp[0 + i * 4] + 3;
    const int a0 = dc + tmp[3 + i * 4];
    const int a1 = tmp[1 + i * 4] + tmp[2 + i * 4];
    const int a2 = tmp[1 + i * 4] - tmp[2 + i * 4];
    const int a3 = dc - tmp[3 + i * 4];
    out[ 0] = (a0 + a1) >> 3;
    out[16] = (a3 + a2) >> 3;
    out[32] = (a0 - a1) >> 3;
    out[48] = (a3 - a2) >> 3;
    out += 64;
  }
}

static void __cc_TransformDCUV(const int16_t* in, uint8_t* dst) {
  if (in[0 * 16]) VP8TransformDC(in + 0 * 16, dst);
  if (in[1 * 16]) VP8TransformDC(in + 1 * 16, dst + 4);
  if (in[2 * 16]) VP8TransformDC(in + 2 * 16, dst + 4 * 32);
  if (in[3 * 16]) VP8TransformDC(in + 3 * 16, dst + 4 * 32 + 4);
}

static void __cc_TransformUV(const int16_t* in, uint8_t* dst) {
  VP8Transform(in + 0 * 16, dst, 1);
  VP8Transform(in + 2 * 16, dst + 4 * 32, 1);
}

__cc_VP8DecIdct VP8TransformUV;

void (*VP8TransformWHT)(const int16_t* in, int16_t* out);

static void __cc_TransformTwo(const int16_t* in, uint8_t* dst, int do_two) {
  __cc_TransformOne(in, dst);
  if (do_two) {
    __cc_TransformOne(in + 16, dst + 4);
  }
}

void __cc_VP8InitClipTables(void);

void __cc_VP8DspInit(void) {
  if (dec_last_cpuinfo_used == VP8GetCPUInfo) return;

  __cc_VP8InitClipTables();

  VP8TransformWHT = __cc_TransformWHT;
  VP8Transform = __cc_TransformTwo;
  VP8TransformUV = __cc_TransformUV;
  VP8TransformDC = __cc_TransformDC;
  VP8TransformDCUV = __cc_TransformDCUV;
  VP8TransformAC3 = __cc_TransformAC3;

  VP8VFilter16 = __cc_VFilter16;
  VP8HFilter16 = __cc_HFilter16;
  VP8VFilter8 = __cc_VFilter8;
  VP8HFilter8 = __cc_HFilter8;
  VP8VFilter16i = __cc_VFilter16i;
  VP8HFilter16i = __cc_HFilter16i;
  VP8VFilter8i = __cc_VFilter8i;
  VP8HFilter8i = __cc_HFilter8i;
  VP8SimpleVFilter16 = __cc_SimpleVFilter16;
  VP8SimpleHFilter16 = __cc_SimpleHFilter16;
  VP8SimpleVFilter16i = __cc_SimpleVFilter16i;
  VP8SimpleHFilter16i = __cc_SimpleHFilter16i;


  if (VP8GetCPUInfo != ((void *)0)) {
  }
  dec_last_cpuinfo_used = VP8GetCPUInfo;
}



void __cc_WebPSamplerProcessPlane(const uint8_t* y, int y_stride,
                             const uint8_t* u, const uint8_t* v, int uv_stride,
                             uint8_t* dst, int dst_stride,
                             int width, int height, __cc_WebPSamplerRowFunc func) {
  int j;
  for (j = 0; j < height; ++j) {
    func(y, u, v, dst, width);
    y += y_stride;
    if (j & 1) {
      u += uv_stride;
      v += uv_stride;
    }
    dst += dst_stride;
  }
}

int __cc_WebPRescaleNeededLines(const __cc_WebPRescaler* const wrk, int max_num_lines) {
  const int num_lines = (wrk->y_accum + wrk->y_sub - 1) / wrk->y_sub;
  return (num_lines > max_num_lines) ? max_num_lines : num_lines;
}




static void __cc_ExportRowExpandC(__cc_WebPRescaler* const wrk) {
  int x_out;
  uint8_t* const dst = wrk->dst;
  __cc_rescaler_t* const irow = wrk->irow;
  const int x_out_max = wrk->dst_width * wrk->num_channels;
  const __cc_rescaler_t* const frow = wrk->frow;
  ((!__cc_WebPRescalerOutputDone(wrk)) ? (void) (0) : __assert_fail ("!WebPRescalerOutputDone(wrk)", "extraction.1.c", 10705, __PRETTY_FUNCTION__));
  ((wrk->y_accum <= 0) ? (void) (0) : __assert_fail ("wrk->y_accum <= 0", "extraction.1.c", 10706, __PRETTY_FUNCTION__));
  ((wrk->y_expand) ? (void) (0) : __assert_fail ("wrk->y_expand", "extraction.1.c", 10707, __PRETTY_FUNCTION__));
  ((wrk->y_sub != 0) ? (void) (0) : __assert_fail ("wrk->y_sub != 0", "extraction.1.c", 10708, __PRETTY_FUNCTION__));
  if (wrk->y_accum == 0) {
    for (x_out = 0; x_out < x_out_max; ++x_out) {
      const uint32_t J = frow[x_out];
      const int v = (int)(((uint64_t)(J) * (wrk->fy_scale) + ((1ull << 32) >> 1)) >> 32);
      ((v >= 0 && v <= 255) ? (void) (0) : __assert_fail ("v >= 0 && v <= 255", "extraction.1.c", 10713, __PRETTY_FUNCTION__));
      dst[x_out] = v;
    }
  } else {
    const uint32_t B = ((uint32_t)(((uint64_t)(-wrk->y_accum) << 32) / (wrk->y_sub)));
    const uint32_t A = (uint32_t)((1ull << 32) - B);
    for (x_out = 0; x_out < x_out_max; ++x_out) {
      const uint64_t I = (uint64_t)A * frow[x_out]
                       + (uint64_t)B * irow[x_out];
      const uint32_t J = (uint32_t)((I + ((1ull << 32) >> 1)) >> 32);
      const int v = (int)(((uint64_t)(J) * (wrk->fy_scale) + ((1ull << 32) >> 1)) >> 32);
      ((v >= 0 && v <= 255) ? (void) (0) : __assert_fail ("v >= 0 && v <= 255", "extraction.1.c", 10724, __PRETTY_FUNCTION__));
      dst[x_out] = v;
    }
  }
}

int __cc_WebPPictureImportRGB(__cc_WebPPicture* picture,
                         const uint8_t* rgb, int rgb_stride) {
  return (picture != ((void *)0)) ? __cc_Import(picture, rgb, rgb_stride, 3, 0, 0) : 0;
}

int __cc_VP8LDecodeImage(__cc_VP8LDecoder* const dec) {
  __cc_VP8Io* io = ((void *)0);
  __cc_WebPDecParams* params = ((void *)0);


  if (dec == ((void *)0)) return 0;

  dec->status_ = VP8_STATUS_BITSTREAM_ERROR;
  ((dec->hdr_.htree_groups_ != ((void *)0)) ? (void) (0) : __assert_fail ("dec->hdr_.htree_groups_ != ((void *)0)", "extraction.1.c", 10743, __PRETTY_FUNCTION__));
  ((dec->hdr_.num_htree_groups_ > 0) ? (void) (0) : __assert_fail ("dec->hdr_.num_htree_groups_ > 0", "extraction.1.c", 10744, __PRETTY_FUNCTION__));

  io = dec->io_;
  ((io != ((void *)0)) ? (void) (0) : __assert_fail ("io != ((void *)0)", "extraction.1.c", 10747, __PRETTY_FUNCTION__));
  params = (__cc_WebPDecParams*)io->opaque;
  ((params != ((void *)0)) ? (void) (0) : __assert_fail ("params != ((void *)0)", "extraction.1.c", 10749, __PRETTY_FUNCTION__));
  dec->output_ = params->output;
  ((dec->output_ != ((void *)0)) ? (void) (0) : __assert_fail ("dec->output_ != ((void *)0)", "extraction.1.c", 10751, __PRETTY_FUNCTION__));


  if (!__cc_WebPIoInitFromOptions(params->options, io, MODE_BGRA)) {
    dec->status_ = VP8_STATUS_INVALID_PARAM;
    goto Err;
  }

  if (!__cc_AllocateInternalBuffers32b(dec, io->width)) goto Err;

  if (io->use_scaling && !__cc_AllocateAndInitRescaler(dec, io)) goto Err;

  if (io->use_scaling || __cc_WebPIsPremultipliedMode(dec->output_->colorspace)) {

    __cc_WebPInitAlphaProcessing();
  }


  dec->action_ = READ_DATA;
  if (!__cc_DecodeImageData(dec, dec->pixels_, dec->width_, dec->height_,
                       dec->height_, __cc_ProcessRows)) {
    goto Err;
  }


  params->last_y = dec->last_out_row_;
  __cc_VP8LClear(dec);
  return 1;

 Err:
  __cc_VP8LClear(dec);
  ((dec->status_ != VP8_STATUS_OK) ? (void) (0) : __assert_fail ("dec->status_ != VP8_STATUS_OK", "extraction.1.c", 10782, __PRETTY_FUNCTION__));
  return 0;
}

int __cc_VP8LDecodeAlphaImageStream(__cc_ALPHDecoder* const alph_dec, int last_row) {
  __cc_VP8LDecoder* const dec = alph_dec->vp8l_dec_;
  ((dec != ((void *)0)) ? (void) (0) : __assert_fail ("dec != ((void *)0)", "extraction.1.c", 10788, __PRETTY_FUNCTION__));
  ((dec->action_ == READ_DATA) ? (void) (0) : __assert_fail ("dec->action_ == READ_DATA", "extraction.1.c", 10789, __PRETTY_FUNCTION__));
  ((last_row <= dec->height_) ? (void) (0) : __assert_fail ("last_row <= dec->height_", "extraction.1.c", 10790, __PRETTY_FUNCTION__));

  if (dec->last_pixel_ == dec->width_ * dec->height_) {
    return 1;
  }


  return alph_dec->use_8b_decode ?
      __cc_DecodeAlphaData(dec, (uint8_t*)dec->pixels_, dec->width_, dec->height_,
                      last_row) :
      __cc_DecodeImageData(dec, dec->pixels_, dec->width_, dec->height_,
                      last_row, __cc_ExtractAlphaRows);
}

static inline void __cc_DoTransform(uint32_t bits, const int16_t* const src,
                                    uint8_t* const dst) {
  switch (bits >> 30) {
    case 3:
      VP8Transform(src, dst, 0);
      break;
    case 2:
      VP8TransformAC3(src, dst);
      break;
    case 1:
      VP8TransformDC(src, dst);
      break;
    default:
      break;
  }
}

static void __cc_ReconstructRow(const __cc_VP8Decoder* const dec,
                           const __cc_VP8ThreadContext* ctx) {
  int j;
  int mb_x;
  const int mb_y = ctx->mb_y_;
  const int cache_id = ctx->id_;
  uint8_t* const y_dst = dec->yuv_b_ + (32 * 1 + 8);
  uint8_t* const u_dst = dec->yuv_b_ + ((32 * 1 + 8) + 32 * 16 + 32);
  uint8_t* const v_dst = dec->yuv_b_ + (((32 * 1 + 8) + 32 * 16 + 32) + 16);
  for (mb_x = 0; mb_x < dec->mb_w_; ++mb_x) {
    const __cc_VP8MBData* const block = ctx->mb_data_ + mb_x;



    if (mb_x > 0) {
      for (j = -1; j < 16; ++j) {
        __cc_Copy32b(&y_dst[j * 32 - 4], &y_dst[j * 32 + 12]);
      }
      for (j = -1; j < 8; ++j) {
        __cc_Copy32b(&u_dst[j * 32 - 4], &u_dst[j * 32 + 4]);
        __cc_Copy32b(&v_dst[j * 32 - 4], &v_dst[j * 32 + 4]);
      }
    } else {
      for (j = 0; j < 16; ++j) {
        y_dst[j * 32 - 1] = 129;
      }
      for (j = 0; j < 8; ++j) {
        u_dst[j * 32 - 1] = 129;
        v_dst[j * 32 - 1] = 129;
      }

      if (mb_y > 0) {
        y_dst[-1 - 32] = u_dst[-1 - 32] = v_dst[-1 - 32] = 129;
      }
    }
    {

      __cc_VP8TopSamples* const top_yuv = dec->yuv_t_ + mb_x;
      const int16_t* const coeffs = block->coeffs_;
      uint32_t bits = block->non_zero_y_;
      int n;

      if (mb_y > 0) {
        memcpy(y_dst - 32, top_yuv[0].y, 16);
        memcpy(u_dst - 32, top_yuv[0].u, 8);
        memcpy(v_dst - 32, top_yuv[0].v, 8);
      } else if (mb_x == 0) {


        memset(y_dst - 32 - 1, 127, 16 + 4 + 1);
        memset(u_dst - 32 - 1, 127, 8 + 1);
        memset(v_dst - 32 - 1, 127, 8 + 1);
      }


      if (block->is_i4x4_) {
        uint32_t* const top_right = (uint32_t*)(y_dst - 32 + 16);

        if (mb_y > 0) {
          if (mb_x >= dec->mb_w_ - 1) {
            memset(top_right, top_yuv[0].y[15], sizeof(*top_right));
          } else {
            memcpy(top_right, top_yuv[1].y, sizeof(*top_right));
          }
        }

        top_right[32] = top_right[2 * 32] = top_right[3 * 32] = top_right[0];


        for (n = 0; n < 16; ++n, bits <<= 2) {
          uint8_t* const dst = y_dst + kScan[n];
          VP8PredLuma4[block->imodes_[n]](dst);
          __cc_DoTransform(bits, coeffs + n * 16, dst);
        }
      } else {
        const int pred_func = __cc_CheckMode(mb_x, mb_y,
                                        block->imodes_[0]);
        VP8PredLuma16[pred_func](y_dst);
        if (bits != 0) {
          for (n = 0; n < 16; ++n, bits <<= 2) {
            __cc_DoTransform(bits, coeffs + n * 16, y_dst + kScan[n]);
          }
        }
      }
      {

        const uint32_t bits_uv = block->non_zero_uv_;
        const int pred_func = __cc_CheckMode(mb_x, mb_y, block->uvmode_);
        VP8PredChroma8[pred_func](u_dst);
        VP8PredChroma8[pred_func](v_dst);
        __cc_DoUVTransform(bits_uv >> 0, coeffs + 16 * 16, u_dst);
        __cc_DoUVTransform(bits_uv >> 8, coeffs + 20 * 16, v_dst);
      }


      if (mb_y < dec->mb_h_ - 1) {
        memcpy(top_yuv[0].y, y_dst + 15 * 32, 16);
        memcpy(top_yuv[0].u, u_dst + 7 * 32, 8);
        memcpy(top_yuv[0].v, v_dst + 7 * 32, 8);
      }
    }

    {
      const int y_offset = cache_id * 16 * dec->cache_y_stride_;
      const int uv_offset = cache_id * 8 * dec->cache_uv_stride_;
      uint8_t* const y_out = dec->cache_y_ + mb_x * 16 + y_offset;
      uint8_t* const u_out = dec->cache_u_ + mb_x * 8 + uv_offset;
      uint8_t* const v_out = dec->cache_v_ + mb_x * 8 + uv_offset;
      for (j = 0; j < 16; ++j) {
        memcpy(y_out + j * dec->cache_y_stride_, y_dst + j * 32, 16);
      }
      for (j = 0; j < 8; ++j) {
        memcpy(u_out + j * dec->cache_uv_stride_, u_dst + j * 32, 8);
        memcpy(v_out + j * dec->cache_uv_stride_, v_dst + j * 32, 8);
      }
    }
  }
}

const uint8_t* const VP8kclip1 = &clip1[255];

static void __cc_LD4(uint8_t *dst) {
  const int A = dst[0 - 32];
  const int B = dst[1 - 32];
  const int C = dst[2 - 32];
  const int D = dst[3 - 32];
  const int E = dst[4 - 32];
  const int F = dst[5 - 32];
  const int G = dst[6 - 32];
  const int H = dst[7 - 32];
  dst[(0) + (0) * 32] = (((A) + 2 * (B) + (C) + 2) >> 2);
  dst[(1) + (0) * 32] = dst[(0) + (1) * 32] = (((B) + 2 * (C) + (D) + 2) >> 2);
  dst[(2) + (0) * 32] = dst[(1) + (1) * 32] = dst[(0) + (2) * 32] = (((C) + 2 * (D) + (E) + 2) >> 2);
  dst[(3) + (0) * 32] = dst[(2) + (1) * 32] = dst[(1) + (2) * 32] = dst[(0) + (3) * 32] = (((D) + 2 * (E) + (F) + 2) >> 2);
  dst[(3) + (1) * 32] = dst[(2) + (2) * 32] = dst[(1) + (3) * 32] = (((E) + 2 * (F) + (G) + 2) >> 2);
  dst[(3) + (2) * 32] = dst[(2) + (3) * 32] = (((F) + 2 * (G) + (H) + 2) >> 2);
  dst[(3) + (3) * 32] = (((G) + 2 * (H) + (H) + 2) >> 2);
}

const uint8_t* __cc_VP8DecompressAlphaRows(__cc_VP8Decoder* const dec,
                                      int row, int num_rows) {
  const int width = dec->pic_hdr_.width_;
  const int height = dec->pic_hdr_.height_;

  if (row < 0 || num_rows <= 0 || row + num_rows > height) {
    return ((void *)0);
  }

  if (row == 0) {

    ((dec->alpha_plane_ != ((void *)0)) ? (void) (0) : __assert_fail ("dec->alpha_plane_ != ((void *)0)", "extraction.1.c", 10971, __PRETTY_FUNCTION__));
    dec->alph_dec_ = __cc_ALPHNew();
    if (dec->alph_dec_ == ((void *)0)) return ((void *)0);
    if (!__cc_ALPHInit(dec->alph_dec_, dec->alpha_data_, dec->alpha_data_size_,
                  width, height, dec->alpha_plane_)) {
      __cc_ALPHDelete(dec->alph_dec_);
      dec->alph_dec_ = ((void *)0);
      return ((void *)0);
    }

    if (dec->alph_dec_->pre_processing_ != 1) {
      dec->alpha_dithering_ = 0;
    } else {
      num_rows = height;
    }
  }

  if (!dec->is_alpha_decoded_) {
    int ok = 0;
    ((dec->alph_dec_ != ((void *)0)) ? (void) (0) : __assert_fail ("dec->alph_dec_ != ((void *)0)", "extraction.1.c", 10990, __PRETTY_FUNCTION__));
    ok = __cc_ALPHDecode(dec, row, num_rows);
    if (ok && dec->alpha_dithering_ > 0) {
      ok = __cc_WebPDequantizeLevels(dec->alpha_plane_, width, height,
                                dec->alpha_dithering_);
    }
    if (!ok || dec->is_alpha_decoded_) {
      __cc_ALPHDelete(dec->alph_dec_);
      dec->alph_dec_ = ((void *)0);
    }
    if (!ok) return ((void *)0);
  }


  return dec->alpha_plane_ + row * width;
}

static void __cc_VR4(uint8_t *dst) {
  const int I = dst[-1 + 0 * 32];
  const int J = dst[-1 + 1 * 32];
  const int K = dst[-1 + 2 * 32];
  const int X = dst[-1 - 32];
  const int A = dst[0 - 32];
  const int B = dst[1 - 32];
  const int C = dst[2 - 32];
  const int D = dst[3 - 32];
  dst[(0) + (0) * 32] = dst[(1) + (2) * 32] = (((X) + (A) + 1) >> 1);
  dst[(1) + (0) * 32] = dst[(2) + (2) * 32] = (((A) + (B) + 1) >> 1);
  dst[(2) + (0) * 32] = dst[(3) + (2) * 32] = (((B) + (C) + 1) >> 1);
  dst[(3) + (0) * 32] = (((C) + (D) + 1) >> 1);

  dst[(0) + (3) * 32] = (((K) + 2 * (J) + (I) + 2) >> 2);
  dst[(0) + (2) * 32] = (((J) + 2 * (I) + (X) + 2) >> 2);
  dst[(0) + (1) * 32] = dst[(1) + (3) * 32] = (((I) + 2 * (X) + (A) + 2) >> 2);
  dst[(1) + (1) * 32] = dst[(2) + (3) * 32] = (((X) + 2 * (A) + (B) + 2) >> 2);
  dst[(2) + (1) * 32] = dst[(3) + (3) * 32] = (((A) + 2 * (B) + (C) + 2) >> 2);
  dst[(3) + (1) * 32] = (((B) + 2 * (C) + (D) + 2) >> 2);
}

int __cc_VP8LHuffmanTreeBuildImplicit(__cc_HuffmanTree* const tree,
                                 const int* const code_lengths,
                                 int* const codes,
                                 int code_lengths_size) {
  int symbol;
  int num_symbols = 0;
  int root_symbol = 0;

  ((tree != ((void *)0)) ? (void) (0) : __assert_fail ("tree != ((void *)0)", "extraction.1.c", 11037, __PRETTY_FUNCTION__));
  ((code_lengths != ((void *)0)) ? (void) (0) : __assert_fail ("code_lengths != ((void *)0)", "extraction.1.c", 11038, __PRETTY_FUNCTION__));


  for (symbol = 0; symbol < code_lengths_size; ++symbol) {
    if (code_lengths[symbol] > 0) {

      ++num_symbols;
      root_symbol = symbol;
    }
  }


  if (!__cc_TreeInit(tree, num_symbols)) return 0;


  if (num_symbols == 1) {
    const int max_symbol = code_lengths_size;
    if (root_symbol < 0 || root_symbol >= max_symbol) {
      __cc_VP8LHuffmanTreeFree(tree);
      return 0;
    }
    return __cc_TreeAddSymbol(tree, root_symbol, 0, 0);
  } else {
    int ok = 0;
    memset(codes, 0, code_lengths_size * sizeof(*codes));

    if (!__cc_VP8LHuffmanCodeLengthsToCodes(code_lengths, code_lengths_size,
                                       codes)) {
      goto End;
    }


    for (symbol = 0; symbol < code_lengths_size; ++symbol) {
      if (code_lengths[symbol] > 0) {
        if (!__cc_TreeAddSymbol(tree, symbol, codes[symbol],
                           code_lengths[symbol])) {
          goto End;
        }
      }
    }
    ok = 1;
 End:
    ok = ok && __cc_IsFull(tree);
    if (!ok) __cc_VP8LHuffmanTreeFree(tree);
    return ok;
  }
}

static void __cc_TM4(uint8_t *dst) { __cc_TrueMotion(dst, 4); }

void __cc_WebPRescalerInit(__cc_WebPRescaler* const wrk, int src_width, int src_height,
                      uint8_t* const dst,
                      int dst_width, int dst_height, int dst_stride,
                      int num_channels, __cc_rescaler_t* const work) {
  const int x_add = src_width, x_sub = dst_width;
  const int y_add = src_height, y_sub = dst_height;
  wrk->x_expand = (src_width < dst_width);
  wrk->y_expand = (src_height < dst_height);
  wrk->src_width = src_width;
  wrk->src_height = src_height;
  wrk->dst_width = dst_width;
  wrk->dst_height = dst_height;
  wrk->src_y = 0;
  wrk->dst_y = 0;
  wrk->dst = dst;
  wrk->dst_stride = dst_stride;
  wrk->num_channels = num_channels;


  wrk->x_add = wrk->x_expand ? (x_sub - 1) : x_add;
  wrk->x_sub = wrk->x_expand ? (x_add - 1) : x_sub;
  if (!wrk->x_expand) {
    wrk->fx_scale = ((uint32_t)(((uint64_t)(1) << 32) / (wrk->x_sub)));
  }

  wrk->y_add = wrk->y_expand ? y_add - 1 : y_add;
  wrk->y_sub = wrk->y_expand ? y_sub - 1 : y_sub;
  wrk->y_accum = wrk->y_expand ? wrk->y_sub : wrk->y_add;
  if (!wrk->y_expand) {

    const uint64_t ratio =
        (uint64_t)dst_height * (1ull << 32) / (wrk->x_add * wrk->y_add);
    if (ratio != (uint32_t)ratio) {


      wrk->fxy_scale = 0;
    } else {
      wrk->fxy_scale = (uint32_t)ratio;
    }
    wrk->fy_scale = ((uint32_t)(((uint64_t)(1) << 32) / (wrk->y_sub)));
  } else {
    wrk->fy_scale = ((uint32_t)(((uint64_t)(1) << 32) / (wrk->x_add)));

  }
  wrk->irow = work;
  wrk->frow = work + num_channels * dst_width;
  memset(work, 0, 2 * dst_width * num_channels * sizeof(*work));

  if (WebPRescalerImportRowExpand == ((void *)0)) {
    WebPRescalerImportRowExpand = __cc_ImportRowExpandC;
    WebPRescalerImportRowShrink = __cc_ImportRowShrinkC;
    WebPRescalerExportRowExpand = __cc_ExportRowExpandC;
    WebPRescalerExportRowShrink = __cc_ExportRowShrinkC;
    if (VP8GetCPUInfo != ((void *)0)) {
    }
  }
}

int __cc_WebPInitDecoderConfigInternal(__cc_WebPDecoderConfig* config,
                                  int version) {
  if ((((version) >> 8) != ((0x0203) >> 8))) {
    return 0;
  }
  if (config == ((void *)0)) {
    return 0;
  }
  memset(config, 0, sizeof(*config));
  __cc_DefaultFeatures(&config->input);
  __cc_WebPInitDecBuffer(&config->output);
  return 1;
}

int ReadWebP(const char* const in_file, __cc_WebPPicture* const pic,
             int keep_alpha, __cc_Metadata* const metadata) {
  int ok = 0;
  size_t data_size = 0;
  const uint8_t* data = ((void *)0);
  __cc_VP8StatusCode status = VP8_STATUS_OK;
  __cc_WebPDecoderConfig config;
  __cc_WebPDecBuffer* const output_buffer = &config.output;
  __cc_WebPBitstreamFeatures* const bitstream = &config.input;


  if (metadata != ((void *)0)) {
    fprintf(stderr, "Warning: metadata extraction from WebP is unsupported.\n");
  }

  if (!__cc_WebPInitDecoderConfig(&config)) {
    fprintf(stderr, "Library version mismatch!\n");
    return 0;
  }

  if (__cc_ExUtilLoadWebP(in_file, &data, &data_size, bitstream)) {
    const int has_alpha = keep_alpha && bitstream->has_alpha;
    output_buffer->colorspace = has_alpha ? MODE_RGBA : MODE_RGB;

    status = __cc_ExUtilDecodeWebP(data, data_size, 0, &config);
    if (status == VP8_STATUS_OK) {
      const uint8_t* const rgba = output_buffer->u.RGBA.rgba;
      const int stride = output_buffer->u.RGBA.stride;
      pic->width = output_buffer->width;
      pic->height = output_buffer->height;
      pic->use_argb = 1;
      ok = has_alpha ? __cc_WebPPictureImportRGBA(pic, rgba, stride)
                     : __cc_WebPPictureImportRGB(pic, rgba, stride);
    }
  }

  if (status != VP8_STATUS_OK) {
    __cc_ExUtilPrintWebPError(in_file, status);
  }

  free((void*)data);
  __cc_WebPFreeDecBuffer(output_buffer);
  return ok;
}

void __cc_WebPMultARGBRows(uint8_t* ptr, int stride, int width, int num_rows,
                      int inverse) {
  int n;
  for (n = 0; n < num_rows; ++n) {
    WebPMultARGBRow((uint32_t*)ptr, width, inverse);
    ptr += stride;
  }
}

int __cc_WebPRescalerExport(__cc_WebPRescaler* const rescaler) {
  int total_exported = 0;
  while (__cc_WebPRescalerHasPendingOutput(rescaler)) {
    __cc_WebPRescalerExportRow(rescaler);
    ++total_exported;
  }
  return total_exported;
}

void __cc_WebPInitUpsamplers(void) {
  if (upsampling_last_cpuinfo_used2 == VP8GetCPUInfo) return;


  WebPUpsamplers[MODE_RGB] = __cc_UpsampleRgbLinePair;
  WebPUpsamplers[MODE_RGBA] = __cc_UpsampleRgbaLinePair;
  WebPUpsamplers[MODE_BGR] = __cc_UpsampleBgrLinePair;
  WebPUpsamplers[MODE_BGRA] = __cc_UpsampleBgraLinePair;
  WebPUpsamplers[MODE_ARGB] = __cc_UpsampleArgbLinePair;
  WebPUpsamplers[MODE_RGBA_4444] = __cc_UpsampleRgba4444LinePair;
  WebPUpsamplers[MODE_RGB_565] = __cc_UpsampleRgb565LinePair;
  WebPUpsamplers[MODE_rgbA] = __cc_UpsampleRgbaLinePair;
  WebPUpsamplers[MODE_bgrA] = __cc_UpsampleBgraLinePair;
  WebPUpsamplers[MODE_Argb] = __cc_UpsampleArgbLinePair;
  WebPUpsamplers[MODE_rgbA_4444] = __cc_UpsampleRgba4444LinePair;


  if (VP8GetCPUInfo != ((void *)0)) {
  }

  upsampling_last_cpuinfo_used2 = VP8GetCPUInfo;
}

const __cc_VP8PredFunc VP8PredLuma4[NUM_BMODES] = {
  __cc_DC4, __cc_TM4, __cc_VE4, __cc_HE4, __cc_RD4, __cc_VR4, __cc_LD4, __cc_VL4, __cc_HD4, __cc_HU4
};

int __cc_VP8InitIoInternal(__cc_VP8Io* const io, int version) {
  if ((((version) >> 8) != ((0x0203) >> 8))) {
    return 0;
  }
  if (io != ((void *)0)) {
    memset(io, 0, sizeof(*io));
  }
  return 1;
}

void __cc_WebPSafeFree(void* const ptr) {
  if (ptr != ((void *)0)) {
    do {} while (0);
    do {} while (0);
  }
  free(ptr);
}

void __cc_VP8ResetProba(__cc_VP8Proba* const proba) {
  memset(proba->segments_, 255u, sizeof(proba->segments_));

}

void __cc_VP8InitDithering(const __cc_WebPDecoderOptions* const options,
                      __cc_VP8Decoder* const dec) {
  ((dec != ((void *)0)) ? (void) (0) : __assert_fail ("dec != ((void *)0)", "extraction.1.c", 11293, __PRETTY_FUNCTION__));
  if (options != ((void *)0)) {
    const int d = options->dithering_strength;
    const int max_amp = (1 << 8) - 1;
    const int f = (d < 0) ? 0 : (d > 100) ? max_amp : (d * max_amp / 100);
    if (f > 0) {
      int s;
      int all_amp = 0;
      for (s = 0; s < NUM_MB_SEGMENTS; ++s) {
        __cc_VP8QuantMatrix* const dqm = &dec->dqm_[s];
        if (dqm->uv_quant_ < 12) {

          const int idx = (dqm->uv_quant_ < 0) ? 0 : dqm->uv_quant_;
          dqm->dither_ = (f * kQuantToDitherAmp[idx]) >> 3;
        }
        all_amp |= dqm->dither_;
      }
      if (all_amp != 0) {
        __cc_VP8InitRandom(&dec->dithering_rg_, 1.0f);
        dec->dither_ = 1;
      }
    }
  }
}
